# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: hangups/googlechat.proto

from google.protobuf.internal import enum_type_wrapper
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()




DESCRIPTOR = _descriptor.FileDescriptor(
  name='hangups/googlechat.proto',
  package='',
  syntax='proto2',
  serialized_options=None,
  create_key=_descriptor._internal_create_key,
  serialized_pb=b'\n\x18hangups/googlechat.proto\"-\n\x06UserId\x12\n\n\x02id\x18\x01 \x01(\t\x12\x17\n\x04type\x18\x02 \x01(\x0e\x32\t.UserType\"\xe3\x01\n\x04User\x12\x18\n\x07user_id\x18\x01 \x01(\x0b\x32\x07.UserId\x12\x0c\n\x04name\x18\x02 \x01(\t\x12\x12\n\navatar_url\x18\x03 \x01(\t\x12\r\n\x05\x65mail\x18\x04 \x01(\t\x12\x12\n\nfirst_name\x18\x05 \x01(\t\x12\x11\n\tlast_name\x18\x06 \x01(\t\x12\x0f\n\x07\x64\x65leted\x18\x07 \x01(\x08\x12\x14\n\x0cis_anonymous\x18\x08 \x01(\x08\x12\x0e\n\x06gender\x18\n \x01(\t\x12\x32\n\x12\x62lock_relationship\x18\r \x01(\x0b\x32\x16.UserBlockRelationship\"W\n\x15UserBlockRelationship\x12\x1f\n\x17is_blocked_by_requester\x18\x01 \x01(\x08\x12\x1d\n\x15has_blocked_requester\x18\x02 \x01(\x08\"\xb9\x01\n\x0b\x44ndSettings\x12.\n\tdnd_state\x18\x01 \x01(\x0e\x32\x1b.DndSettings.DndState_State\x12\x1c\n\x14\x64nd_expiry_time_usec\x18\x02 \x01(\x03\x12%\n\x1dstate_remaining_duration_usec\x18\x03 \x01(\x03\"5\n\x0e\x44ndState_State\x12\x0b\n\x07UNKNOWN\x10\x00\x12\r\n\tAVAILABLE\x10\x01\x12\x07\n\x03\x44ND\x10\x02\"\x18\n\x05\x45moji\x12\x0f\n\x07unicode\x18\x01 \x01(\t\"u\n\x0c\x43ustomStatus\x12\x13\n\x0bstatus_text\x18\x01 \x01(\t\x12\x14\n\x0cstatus_emoji\x18\x02 \x01(\t\x12#\n\x1bstate_expiry_timestamp_usec\x18\x03 \x01(\x03\x12\x15\n\x05\x65moji\x18\x04 \x01(\x0b\x32\x06.Emoji\"\xa0\x02\n\nUserStatus\x12\x18\n\x07user_id\x18\x01 \x01(\x0b\x32\x07.UserId\x12\"\n\x0c\x64nd_settings\x18\x02 \x01(\x0b\x32\x0c.DndSettings\x12\x1d\n\x15update_timestamp_usec\x18\x03 \x01(\x03\x12*\n\"dnd_settings_update_timestamp_usec\x18\x07 \x01(\x03\x12\x1d\n\x15status_timestamp_usec\x18\x04 \x01(\x03\x12\x17\n\x0fpresence_shared\x18\x05 \x01(\x08\x12$\n\rcustom_status\x18\x06 \x01(\x0b\x32\r.CustomStatus\x12+\n#custom_status_update_timestamp_usec\x18\x08 \x01(\x03\":\n\x16UserStatusUpdatedEvent\x12 \n\x0buser_status\x18\x01 \x01(\x0b\x32\x0b.UserStatus\"Y\n\x14GetUserStatusRequest\x12\x19\n\x08user_ids\x18\x01 \x03(\x0b\x32\x07.UserId\x12&\n\x0erequest_header\x18\x64 \x01(\x0b\x32\x0e.RequestHeader\";\n\x15GetUserStatusResponse\x12\"\n\ruser_statuses\x18\x01 \x03(\x0b\x32\x0b.UserStatus\"B\n\x18GetSelfUserStatusRequest\x12&\n\x0erequest_header\x18\x64 \x01(\x0b\x32\x0e.RequestHeader\"c\n\x19GetSelfUserStatusResponse\x12 \n\x0buser_status\x18\x01 \x01(\x0b\x32\x0b.UserStatus\x12$\n\ruser_revision\x18\x02 \x01(\x0b\x32\r.ReadRevision\"\xc3\x04\n\x19\x43lientFeatureCapabilities\x12\x1b\n\x13\x65nable_all_features\x18\x01 \x01(\x08\x12L\n\x18spaces_level_for_testing\x18\x02 \x01(\x0e\x32*.ClientFeatureCapabilities.CapabilityLevel\x12I\n\x15\x64ms_level_for_testing\x18\x03 \x01(\x0e\x32*.ClientFeatureCapabilities.CapabilityLevel\x12\x44\n\x10post_rooms_level\x18\x04 \x01(\x0e\x32*.ClientFeatureCapabilities.CapabilityLevel\x12K\n\x17spam_room_invites_level\x18\x05 \x01(\x0e\x32*.ClientFeatureCapabilities.CapabilityLevel\x12\x43\n\x0ftombstone_level\x18\x06 \x01(\x0e\x32*.ClientFeatureCapabilities.CapabilityLevel\x12K\n\x17rich_text_viewing_level\x18\x07 \x01(\x0e\x32*.ClientFeatureCapabilities.CapabilityLevel\"K\n\x0f\x43\x61pabilityLevel\x12\x0f\n\x0bUNSUPPORTED\x10\x00\x12\x12\n\x0e\x44\x41TA_SUPPORTED\x10\x01\x12\x13\n\x0f\x46ULLY_SUPPORTED\x10\x02\"\xc0\x02\n\rRequestHeader\x12\x10\n\x08trace_id\x18\x01 \x01(\x03\x12.\n\x0b\x63lient_type\x18\x02 \x01(\x0e\x32\x19.RequestHeader.ClientType\x12\x16\n\x0e\x63lient_version\x18\x03 \x01(\x03\x12\x0e\n\x06locale\x18\x04 \x01(\t\x12?\n\x1b\x63lient_feature_capabilities\x18\x05 \x01(\x0b\x32\x1a.ClientFeatureCapabilities\"\x83\x01\n\nClientType\x12\x0b\n\x07UNKNOWN\x10\x00\x12\x0b\n\x07\x41NDROID\x10\x01\x12\x07\n\x03IOS\x10\x02\x12\x07\n\x03WEB\x10\x03\x12\x13\n\x0f\x43LASSIC_INTEROP\x10\x04\x12\x14\n\x10INTERNAL_TESTING\x10\x05\x12\x0f\n\x0bWEB_DESKTOP\x10\x06\x12\r\n\tWEB_GMAIL\x10\x07\"\x1d\n\x06Member\x12\x13\n\x04user\x18\x01 \x01(\x0b\x32\x05.User\"$\n\x08MemberId\x12\x18\n\x07user_id\x18\x01 \x01(\x0b\x32\x07.UserId\"d\n\x0cMembershipId\x12\x1c\n\tmember_id\x18\x01 \x01(\x0b\x32\t.MemberId\x12\x1a\n\x08space_id\x18\x02 \x01(\x0b\x32\x08.SpaceId\x12\x1a\n\x08group_id\x18\x03 \x01(\x0b\x32\x08.GroupId\"\xbc\x01\n\nMembership\x12\x19\n\x02id\x18\x01 \x01(\x0b\x32\r.MembershipId\x12\x13\n\x0b\x63reate_time\x18\x02 \x01(\x03\x12*\n\x10membership_state\x18\x03 \x01(\x0e\x32\x10.MembershipState\x12(\n\x0finvite_category\x18\x04 \x01(\x0e\x32\x0f.InviteCategory\x12(\n\x0fmembership_role\x18\x05 \x01(\x0e\x32\x0f.MembershipRole\"C\n\rMemberProfile\x12\x19\n\x02id\x18\x01 \x01(\x0b\x32\r.MembershipId\x12\x17\n\x06member\x18\x02 \x01(\x0b\x32\x07.Member\"\x81\x01\n\x11GetMembersRequest\x12&\n\x0erequest_header\x18\x64 \x01(\x0b\x32\x0e.RequestHeader\x12\x1d\n\nmember_ids\x18\x01 \x03(\x0b\x32\t.MemberId\x12%\n\x0emembership_ids\x18\x02 \x03(\x0b\x32\r.MembershipId\"W\n\x12GetMembersResponse\x12\x18\n\x07members\x18\x01 \x03(\x0b\x32\x07.Member\x12\'\n\x0fmember_profiles\x18\x02 \x03(\x0b\x32\x0e.MemberProfile\"\xa1\x01\n\x0cUserPresence\x12\x18\n\x07user_id\x18\x01 \x01(\x0b\x32\x07.UserId\x12\x1b\n\x08presence\x18\x02 \x01(\x0e\x32\t.Presence\x12\x14\n\x0c\x61\x63tive_until\x18\x04 \x01(\x03\x12\"\n\tdnd_state\x18\x03 \x01(\x0e\x32\x0f.DndState_State\x12 \n\x0buser_status\x18\x05 \x01(\x0b\x32\x0b.UserStatus\"\x96\x01\n\x16GetUserPresenceRequest\x12&\n\x0erequest_header\x18\x64 \x01(\x0b\x32\x0e.RequestHeader\x12\x19\n\x08user_ids\x18\x01 \x03(\x0b\x32\x07.UserId\x12\x1c\n\x14include_active_until\x18\x02 \x01(\x08\x12\x1b\n\x13include_user_status\x18\x03 \x01(\x08\"@\n\x17GetUserPresenceResponse\x12%\n\x0euser_presences\x18\x01 \x03(\x0b\x32\r.UserPresence\"\x1f\n\x11JAddOnsIdentifier\x12\n\n\x02id\x18\x01 \x01(\t\"\xc3\x08\n\x14JAddOnsFormattedText\x12\x1e\n\x02id\x18\x04 \x01(\x0b\x32\x12.JAddOnsIdentifier\x12\x15\n\roriginal_text\x18\x01 \x01(\t\x12K\n\x17\x66ormatted_text_elements\x18\x02 \x03(\x0b\x32*.JAddOnsFormattedText.FormattedTextElement\x12\x33\n\ntext_align\x18\x03 \x01(\x0e\x32\x1f.JAddOnsFormattedText.TextAlign\x1a\xc3\x06\n\x14\x46ormattedTextElement\x12J\n\x0bstyled_text\x18\x01 \x01(\x0b\x32\x35.JAddOnsFormattedText.FormattedTextElement.StyledText\x12G\n\thyperlink\x18\x02 \x01(\x0b\x32\x34.JAddOnsFormattedText.FormattedTextElement.HyperLink\x1ak\n\x08\x44\x61teTime\x12\x13\n\x0btime_millis\x18\x01 \x01(\x03\x12 \n\x18time_zone_offset_minutes\x18\x02 \x01(\x05\x12\x11\n\tdate_only\x18\x03 \x01(\x08\x12\x15\n\rfloating_time\x18\x04 \x01(\x08\x1a\xe8\x03\n\nStyledText\x12\x0c\n\x04text\x18\x01 \x01(\t\x12\x45\n\x08\x64\x61tetime\x18\x04 \x01(\x0b\x32\x33.JAddOnsFormattedText.FormattedTextElement.DateTime\x12K\n\x06styles\x18\x02 \x03(\x0e\x32;.JAddOnsFormattedText.FormattedTextElement.StyledText.Style\x12U\n\x0b\x66ont_weight\x18\x05 \x01(\x0e\x32@.JAddOnsFormattedText.FormattedTextElement.StyledText.FontWeight\x12\r\n\x05\x63olor\x18\x03 \x01(\x05\x12)\n\x0ctheme_colors\x18\x07 \x01(\x0b\x32\x13.JAddOnsThemeColors\"k\n\x05Style\x12\x08\n\x04NONE\x10\x00\x12\x13\n\x0f\x42OLD_DEPRECATED\x10\x01\x12\n\n\x06ITALIC\x10\x02\x12\r\n\tUNDERLINE\x10\x03\x12\x11\n\rSTRIKETHROUGH\x10\x04\x12\x06\n\x02\x42R\x10\x05\x12\r\n\tUPPERCASE\x10\x06\":\n\nFontWeight\x12\x0b\n\x07REGULAR\x10\x00\x12\t\n\x05LIGHT\x10\x01\x12\n\n\x06MEDIUM\x10\x02\x12\x08\n\x04\x42OLD\x10\x03\x1a>\n\tHyperLink\x12\x0c\n\x04link\x18\x01 \x01(\t\x12\x15\n\roriginal_link\x18\x03 \x01(\t\x12\x0c\n\x04text\x18\x02 \x01(\t\",\n\tTextAlign\x12\x08\n\x04LEFT\x10\x00\x12\n\n\x06\x43\x45NTER\x10\x01\x12\t\n\x05RIGHT\x10\x02\"\xca\x01\n\x15JAddOnsImageCropStyle\x12\x32\n\x04type\x18\x01 \x01(\x0e\x32$.JAddOnsImageCropStyle.ImageCropType\x12\x14\n\x0c\x61spect_ratio\x18\x02 \x01(\x01\"g\n\rImageCropType\x12\x15\n\x11\x43ROP_TYPE_NOT_SET\x10\x00\x12\n\n\x06SQUARE\x10\x01\x12\n\n\x06\x43IRCLE\x10\x02\x12\x14\n\x10RECTANGLE_CUSTOM\x10\x03\x12\x11\n\rRECTANGLE_4_3\x10\x04\"I\n\x12JAddOnsThemeColors\x12\x19\n\x11light_theme_color\x18\x01 \x01(\x05\x12\x18\n\x10\x64\x61rk_theme_color\x18\x02 \x01(\x05\"\xbc\x02\n\x0fJAddOnsOpenLink\x12\x0b\n\x03url\x18\x01 \x01(\t\x12\x15\n\roriginal_link\x18\x05 \x01(\t\x12(\n\x07open_as\x18\x02 \x01(\x0e\x32\x17.JAddOnsOpenLink.OpenAs\x12*\n\x08on_close\x18\x03 \x01(\x0e\x32\x18.JAddOnsOpenLink.OnClose\x12\x36\n\x0eload_indicator\x18\x04 \x01(\x0e\x32\x1e.JAddOnsOpenLink.LoadIndicator\"$\n\x06OpenAs\x12\r\n\tFULL_SIZE\x10\x00\x12\x0b\n\x07OVERLAY\x10\x01\")\n\x07OnClose\x12\x0b\n\x07NOTHING\x10\x00\x12\x11\n\rRELOAD_ADD_ON\x10\x01\"&\n\rLoadIndicator\x12\x08\n\x04NONE\x10\x00\x12\x0b\n\x07SPINNER\x10\x01\"\xe5\t\n\x0fJAddOnsCardItem\x12/\n\x06header\x18\x01 \x01(\x0b\x32\x1f.JAddOnsCardItem.CardItemHeader\x12\x32\n\x08sections\x18\x02 \x03(\x0b\x32 .JAddOnsCardItem.CardItemSection\x12\x35\n\x0c\x63\x61rd_actions\x18\x03 \x03(\x0b\x32\x1f.JAddOnsCardItem.CardItemAction\x12\x0c\n\x04name\x18\x04 \x01(\t\x12:\n\x0c\x66ixed_footer\x18\x05 \x01(\x0b\x32$.JAddOnsCardItem.CardItemFixedFooter\x12>\n\x0erefresh_action\x18\x06 \x01(\x0b\x32&.JAddOnsCardItem.CardItemRefreshAction\x12\x34\n\rdisplay_style\x18\x07 \x01(\x0e\x32\x1d.JAddOnsCardItem.DisplayStyle\x12\x39\n\x10peek_card_header\x18\x08 \x01(\x0b\x32\x1f.JAddOnsCardItem.CardItemHeader\x12\x34\n\x17\x62\x61\x63kground_theme_colors\x18\t \x01(\x0b\x32\x13.JAddOnsThemeColors\x1a\xc5\x01\n\x0e\x43\x61rdItemHeader\x12$\n\x05title\x18\x01 \x01(\x0b\x32\x15.JAddOnsFormattedText\x12\'\n\x08subtitle\x18\x02 \x01(\x0b\x32\x15.JAddOnsFormattedText\x12\x39\n\x0bimage_style\x18\x03 \x01(\x0e\x32$.JAddOnsImageCropStyle.ImageCropType\x12\x11\n\timage_url\x18\x04 \x01(\t\x12\x16\n\x0eimage_alt_text\x18\x05 \x01(\t\x1a\xc6\x01\n\x0f\x43\x61rdItemSection\x12\x1e\n\x02id\x18\x06 \x01(\x0b\x32\x12.JAddOnsIdentifier\x12\x13\n\x0b\x64\x65scription\x18\x01 \x01(\t\x12%\n\x06header\x18\x05 \x01(\x0b\x32\x15.JAddOnsFormattedText\x12\x1f\n\x07widgets\x18\x02 \x03(\x0b\x32\x0e.JAddOnsWidget\x12\x13\n\x0b\x63ollapsable\x18\x03 \x01(\x08\x12!\n\x19num_uncollapsable_widgets\x18\x04 \x01(\x05\x1aI\n\x0e\x43\x61rdItemAction\x12\x14\n\x0c\x61\x63tion_label\x18\x01 \x01(\t\x12!\n\x08on_click\x18\x02 \x01(\x0b\x32\x0f.JAddOnsOnClick\x1a\xa5\x01\n\x13\x43\x61rdItemFixedFooter\x12&\n\x07\x62uttons\x18\x01 \x03(\x0b\x32\x15.JAddOnsWidget.Button\x12\x31\n\x0eprimary_button\x18\x02 \x01(\x0b\x32\x19.JAddOnsWidget.TextButton\x12\x33\n\x10secondary_button\x18\x03 \x01(\x0b\x32\x19.JAddOnsWidget.TextButton\x1a;\n\x15\x43\x61rdItemRefreshAction\x12\"\n\x06method\x18\x01 \x01(\x0b\x32\x12.JAddOnsFormAction\"D\n\x0c\x44isplayStyle\x12\x1d\n\x19\x44ISPLAY_STYLE_UNSPECIFIED\x10\x00\x12\x08\n\x04PEEK\x10\x01\x12\x0b\n\x07REPLACE\x10\x02\"1\n\x0fJAddOnsPushCard\x12\x1e\n\x04\x63\x61rd\x18\x01 \x01(\x0b\x32\x10.JAddOnsCardItem\"\xba\x01\n\x0eJAddOnsOnClick\x12\x0c\n\x04link\x18\x01 \x01(\t\x12\"\n\x06\x61\x63tion\x18\x02 \x01(\x0b\x32\x12.JAddOnsFormAction\x12#\n\topen_link\x18\x05 \x01(\x0b\x32\x10.JAddOnsOpenLink\x12,\n\x10open_link_action\x18\x07 \x01(\x0b\x32\x12.JAddOnsFormAction\x12#\n\tpush_card\x18\x08 \x01(\x0b\x32\x10.JAddOnsPushCard\"!\n\x11JAddOnsTextWidget\x12\x0c\n\x04line\x18\x01 \x03(\t\"\x93\x01\n\x15JAddOnsImageComponent\x12\x11\n\timage_url\x18\x01 \x01(\t\x12\x10\n\x08\x61lt_text\x18\x02 \x01(\t\x12*\n\ncrop_style\x18\x03 \x01(\x0b\x32\x16.JAddOnsImageCropStyle\x12)\n\x0c\x62order_style\x18\x04 \x01(\x0b\x32\x13.JAddOnsBorderStyle\"\xfe\x03\n\x0bJAddOnsGrid\x12\x1e\n\x02id\x18\x01 \x01(\x0b\x32\x12.JAddOnsIdentifier\x12\r\n\x05title\x18\x02 \x01(\t\x12$\n\x05items\x18\x03 \x03(\x0b\x32\x15.JAddOnsGrid.GridItem\x12)\n\x0c\x62order_style\x18\x04 \x01(\x0b\x32\x13.JAddOnsBorderStyle\x12\x13\n\x0bnum_columns\x18\x05 \x01(\x05\x12!\n\x08on_click\x18\x06 \x01(\x0b\x32\x0f.JAddOnsOnClick\x1a\xb6\x02\n\x08GridItem\x12\x12\n\nidentifier\x18\x01 \x01(\t\x12%\n\x05image\x18\x02 \x01(\x0b\x32\x16.JAddOnsImageComponent\x12\r\n\x05title\x18\x03 \x01(\t\x12\x10\n\x08subtitle\x18\x04 \x01(\t\x12\x36\n\x0etext_alignment\x18\x05 \x01(\x0e\x32\x1e.JAddOnsWidget.HorizontalAlign\x12\x34\n\x06layout\x18\t \x01(\x0e\x32$.JAddOnsGrid.GridItem.GridItemLayout\x12!\n\x08on_click\x18\n \x01(\x0b\x32\x0f.JAddOnsOnClick\"=\n\x0eGridItemLayout\x12\x0b\n\x07NOT_SET\x10\x00\x12\x0e\n\nTEXT_BELOW\x10\x01\x12\x0e\n\nTEXT_ABOVE\x10\x02\"\xc6\x01\n\x12JAddOnsBorderStyle\x12,\n\x04type\x18\x01 \x01(\x0e\x32\x1e.JAddOnsBorderStyle.BorderType\x12)\n\x0cstroke_color\x18\x02 \x01(\x0b\x32\x13.JAddOnsThemeColors\x12\x15\n\rcorner_radius\x18\x03 \x01(\x05\"@\n\nBorderType\x12\x17\n\x13\x42ORDER_TYPE_NOT_SET\x10\x00\x12\r\n\tNO_BORDER\x10\x01\x12\n\n\x06STROKE\x10\x02\"9\n\x17JAddOnsLabelContentPair\x12\r\n\x05label\x18\x01 \x01(\t\x12\x0f\n\x07\x63ontent\x18\x02 \x01(\t\"U\n\x1dJAddOnsLabelContentPairWidget\x12\x34\n\x12label_content_pair\x18\x01 \x03(\x0b\x32\x18.JAddOnsLabelContentPair\"\xeb \n\rJAddOnsWidget\x12\'\n\x0btext_widget\x18\x01 \x01(\x0b\x32\x12.JAddOnsTextWidget\x12\x41\n\x19label_content_pair_widget\x18\x02 \x01(\x0b\x32\x1e.JAddOnsLabelContentPairWidget\x12\x34\n\x0etext_paragraph\x18\x03 \x01(\x0b\x32\x1c.JAddOnsWidget.TextParagraph\x12\x33\n\x0etext_key_value\x18\x04 \x01(\x0b\x32\x1b.JAddOnsWidget.TextKeyValue\x12\x35\n\x0fimage_key_value\x18\x05 \x01(\x0b\x32\x1c.JAddOnsWidget.ImageKeyValue\x12#\n\x05image\x18\t \x01(\x0b\x32\x14.JAddOnsWidget.Image\x12*\n\tkey_value\x18\r \x01(\x0b\x32\x17.JAddOnsWidget.KeyValue\x12\'\n\x07\x64ivider\x18\x10 \x01(\x0b\x32\x16.JAddOnsWidget.Divider\x12\x1a\n\x04grid\x18\x11 \x01(\x0b\x32\x0c.JAddOnsGrid\x12!\n\x04menu\x18\n \x01(\x0b\x32\x13.JAddOnsWidget.Menu\x12,\n\ntext_field\x18\x0b \x01(\x0b\x32\x18.JAddOnsWidget.TextField\x12:\n\x11selection_control\x18\x0c \x01(\x0b\x32\x1f.JAddOnsWidget.SelectionControl\x12\x37\n\x10\x64\x61te_time_picker\x18\x0e \x01(\x0b\x32\x1d.JAddOnsWidget.DateTimePicker\x12&\n\x07\x62uttons\x18\x08 \x03(\x0b\x32\x15.JAddOnsWidget.Button\x12\x38\n\x10horizontal_align\x18\x0f \x01(\x0e\x32\x1e.JAddOnsWidget.HorizontalAlign\x1a\x34\n\rTextParagraph\x12#\n\x04text\x18\x01 \x01(\x0b\x32\x15.JAddOnsFormattedText\x1az\n\x0cTextKeyValue\x12\"\n\x03key\x18\x01 \x01(\x0b\x32\x15.JAddOnsFormattedText\x12#\n\x04text\x18\x02 \x01(\x0b\x32\x15.JAddOnsFormattedText\x12!\n\x08on_click\x18\x03 \x01(\x0b\x32\x0f.JAddOnsOnClick\x1ai\n\rImageKeyValue\x12\x10\n\x08icon_url\x18\x01 \x01(\t\x12#\n\x04text\x18\x02 \x01(\x0b\x32\x15.JAddOnsFormattedText\x12!\n\x08on_click\x18\x03 \x01(\x0b\x32\x0f.JAddOnsOnClick\x1a\x8a\x01\n\x05Image\x12\x1e\n\x02id\x18\x05 \x01(\x0b\x32\x12.JAddOnsIdentifier\x12\x16\n\x0e\x66ife_image_url\x18\x01 \x01(\t\x12!\n\x08on_click\x18\x02 \x01(\x0b\x32\x0f.JAddOnsOnClick\x12\x14\n\x0c\x61spect_ratio\x18\x03 \x01(\x01\x12\x10\n\x08\x61lt_text\x18\x04 \x01(\t\x1a\x65\n\x04Icon\x12\x10\n\x08icon_url\x18\x01 \x01(\t\x12\x10\n\x08\x61lt_text\x18\x02 \x01(\t\x12\x39\n\x0bimage_style\x18\x03 \x01(\x0e\x32$.JAddOnsImageCropStyle.ImageCropType\x1a\xe8\x05\n\x08KeyValue\x12\x10\n\x08icon_url\x18\x01 \x01(\t\x12\x15\n\ricon_alt_text\x18\t \x01(\t\x12\x39\n\x0bimage_style\x18\n \x01(\x0e\x32$.JAddOnsImageCropStyle.ImageCropType\x12\'\n\nstart_icon\x18\x0b \x01(\x0b\x32\x13.JAddOnsWidget.Icon\x12(\n\ttop_label\x18\x02 \x01(\x0b\x32\x15.JAddOnsFormattedText\x12&\n\x07\x63ontent\x18\x03 \x01(\x0b\x32\x15.JAddOnsFormattedText\x12\x19\n\x11\x63ontent_multiline\x18\x08 \x01(\x08\x12+\n\x0c\x62ottom_label\x18\x04 \x01(\x0b\x32\x15.JAddOnsFormattedText\x12!\n\x08on_click\x18\x05 \x01(\x0b\x32\x0f.JAddOnsOnClick\x12%\n\x06\x62utton\x18\x06 \x01(\x0b\x32\x15.JAddOnsWidget.Button\x12;\n\rswitch_widget\x18\x07 \x01(\x0b\x32$.JAddOnsWidget.KeyValue.SwitchWidget\x12%\n\x08\x65nd_icon\x18\x0c \x01(\x0b\x32\x13.JAddOnsWidget.Icon\x1a\x86\x02\n\x0cSwitchWidget\x12\x1e\n\x02id\x18\x05 \x01(\x0b\x32\x12.JAddOnsIdentifier\x12\x0c\n\x04name\x18\x01 \x01(\t\x12\r\n\x05value\x18\x02 \x01(\t\x12\x10\n\x08selected\x18\x03 \x01(\x08\x12%\n\ton_change\x18\x04 \x01(\x0b\x32\x12.JAddOnsFormAction\x12\x46\n\x0c\x63ontrol_type\x18\x06 \x01(\x0e\x32\x30.JAddOnsWidget.KeyValue.SwitchWidget.ControlType\"8\n\x0b\x43ontrolType\x12\x0f\n\x0bUNSPECIFIED\x10\x00\x12\n\n\x06SWITCH\x10\x01\x12\x0c\n\x08\x43HECKBOX\x10\x02\x1a\t\n\x07\x44ivider\x1a\xb2\x01\n\x04Menu\x12\x0c\n\x04name\x18\x01 \x01(\t\x12+\n\x05items\x18\x02 \x03(\x0b\x32\x1c.JAddOnsWidget.Menu.MenuItem\x12%\n\ton_change\x18\x04 \x01(\x0b\x32\x12.JAddOnsFormAction\x12\r\n\x05label\x18\x05 \x01(\t\x1a\x39\n\x08MenuItem\x12\x0c\n\x04text\x18\x01 \x01(\t\x12\r\n\x05value\x18\x04 \x01(\t\x12\x10\n\x08selected\x18\x03 \x01(\x08\x1am\n\x0c\x41utoComplete\x12;\n\x05items\x18\x01 \x03(\x0b\x32,.JAddOnsWidget.AutoComplete.AutoCompleteItem\x1a \n\x10\x41utoCompleteItem\x12\x0c\n\x04text\x18\x01 \x01(\t\x1a\xf5\x03\n\tTextField\x12\x1e\n\x02id\x18\x0b \x01(\x0b\x32\x12.JAddOnsIdentifier\x12\x0c\n\x04name\x18\x01 \x01(\t\x12\x11\n\tmax_lines\x18\x02 \x01(\x05\x12\x34\n\x04type\x18\x05 \x01(\x0e\x32&.JAddOnsWidget.TextField.TextFieldType\x12\r\n\x05label\x18\x03 \x01(\t\x12\x11\n\thint_text\x18\x04 \x01(\t\x12\r\n\x05value\x18\x06 \x01(\t\x12\x34\n\tline_type\x18\x07 \x01(\x0e\x32!.JAddOnsWidget.TextField.LineType\x12%\n\ton_change\x18\x08 \x01(\x0b\x32\x12.JAddOnsFormAction\x12\x32\n\rauto_complete\x18\t \x01(\x0b\x32\x1b.JAddOnsWidget.AutoComplete\x12\x32\n\x16\x61uto_complete_callback\x18\n \x01(\x0b\x32\x12.JAddOnsFormAction\x12)\n!auto_complete_multiple_selections\x18\x0c \x01(\x08\"*\n\rTextFieldType\x12\x0b\n\x07REGULAR\x10\x00\x12\x0c\n\x08\x46LOATING\x10\x01\"$\n\x08LineType\x12\n\n\x06SINGLE\x10\x00\x12\x0c\n\x08MULTIPLE\x10\x01\x1a\x9d\x03\n\x10SelectionControl\x12\x1e\n\x02id\x18\x06 \x01(\x0b\x32\x12.JAddOnsIdentifier\x12\x0c\n\x04name\x18\x01 \x01(\t\x12;\n\x04type\x18\x02 \x01(\x0e\x32-.JAddOnsWidget.SelectionControl.SelectionType\x12<\n\x05items\x18\x03 \x03(\x0b\x32-.JAddOnsWidget.SelectionControl.SelectionItem\x12%\n\ton_change\x18\x04 \x01(\x0b\x32\x12.JAddOnsFormAction\x12\r\n\x05label\x18\x05 \x01(\t\x1a^\n\rSelectionItem\x12\x1e\n\x02id\x18\x04 \x01(\x0b\x32\x12.JAddOnsIdentifier\x12\x0c\n\x04text\x18\x01 \x01(\t\x12\x10\n\x08selected\x18\x02 \x01(\x08\x12\r\n\x05value\x18\x03 \x01(\t\"J\n\rSelectionType\x12\r\n\tCHECK_BOX\x10\x00\x12\x10\n\x0cRADIO_BUTTON\x10\x01\x12\n\n\x06SWITCH\x10\x02\x12\x0c\n\x08\x44ROPDOWN\x10\x03\x1a\xe1\x02\n\x0e\x44\x61teTimePicker\x12\x0c\n\x04name\x18\x01 \x01(\t\x12\r\n\x05label\x18\x02 \x01(\t\x12>\n\x04type\x18\x03 \x01(\x0e\x32\x30.JAddOnsWidget.DateTimePicker.DateTimePickerType\x12\x16\n\x0evalue_ms_epoch\x18\x04 \x01(\x03\x12\x1c\n\x14timezone_offset_date\x18\x05 \x01(\x05\x12%\n\ton_change\x18\x06 \x01(\x0b\x32\x12.JAddOnsFormAction\x12\r\n\x05\x63olor\x18\x07 \x01(\x05\x12)\n\x0ctheme_colors\x18\x08 \x01(\x0b\x32\x13.JAddOnsThemeColors\"[\n\x12\x44\x61teTimePickerType\x12\x14\n\x10UNSPECIFIED_TYPE\x10\x00\x12\x11\n\rDATE_AND_TIME\x10\x01\x12\r\n\tDATE_ONLY\x10\x02\x12\r\n\tTIME_ONLY\x10\x03\x1a\xc8\x02\n\nTextButton\x12\x1e\n\x02id\x18\x03 \x01(\x0b\x32\x12.JAddOnsIdentifier\x12#\n\x04text\x18\x01 \x01(\x0b\x32\x15.JAddOnsFormattedText\x12!\n\x08on_click\x18\x02 \x01(\x0b\x32\x0f.JAddOnsOnClick\x12\x10\n\x08\x64isabled\x18\x04 \x01(\x08\x12.\n\x05style\x18\x05 \x01(\x0e\x32\x1f.JAddOnsWidget.TextButton.Style\x12\x18\n\x10\x62\x61\x63kground_color\x18\x06 \x01(\x05\x12\x34\n\x17\x62\x61\x63kground_theme_colors\x18\x07 \x01(\x0b\x32\x13.JAddOnsThemeColors\x12\x10\n\x08\x61lt_text\x18\x08 \x01(\t\".\n\x05Style\x12\x0f\n\x0bUNSPECIFIED\x10\x00\x12\x08\n\x04TEXT\x10\x01\x12\n\n\x06\x46ILLED\x10\x02\x1ap\n\x0bImageButton\x12\x1e\n\x02id\x18\x04 \x01(\x0b\x32\x12.JAddOnsIdentifier\x12\x10\n\x08icon_url\x18\x01 \x01(\t\x12!\n\x08on_click\x18\x02 \x01(\x0b\x32\x0f.JAddOnsOnClick\x12\x0c\n\x04name\x18\x03 \x01(\t\x1aj\n\x06\x42utton\x12.\n\x0btext_button\x18\x01 \x01(\x0b\x32\x19.JAddOnsWidget.TextButton\x12\x30\n\x0cimage_button\x18\x02 \x01(\x0b\x32\x1a.JAddOnsWidget.ImageButton\">\n\x0fHorizontalAlign\x12\x0b\n\x07NOT_SET\x10\x00\x12\t\n\x05START\x10\x01\x12\n\n\x06\x43\x45NTER\x10\x02\x12\x07\n\x03\x45ND\x10\x03\"\xc9\x02\n\x11JAddOnsFormAction\x12\x1a\n\x12\x61\x63tion_method_name\x18\x01 \x01(\t\x12\x36\n\nparameters\x18\x04 \x03(\x0b\x32\".JAddOnsFormAction.ActionParameter\x12\x1d\n\x15parameters_deprecated\x18\x03 \x03(\t\x12\x18\n\x10\x65ncrypted_action\x18\x07 \x01(\t\x12\x38\n\x0eload_indicator\x18\x05 \x01(\x0e\x32 .JAddOnsFormAction.LoadIndicator\x12\x16\n\x0epersist_values\x18\x06 \x01(\x08\x1a-\n\x0f\x41\x63tionParameter\x12\x0b\n\x03key\x18\x01 \x01(\t\x12\r\n\x05value\x18\x02 \x01(\t\"&\n\rLoadIndicator\x12\x0b\n\x07SPINNER\x10\x00\x12\x08\n\x04NONE\x10\x01\"\x95\n\n\x16JAddOnsContextualAddOn\x12\x30\n\x07toolbar\x18\x01 \x01(\x0b\x32\x1f.JAddOnsContextualAddOn.Toolbar\x12+\n\x05\x63\x61rds\x18\x02 \x03(\x0b\x32\x1c.JAddOnsContextualAddOn.Card\x1az\n\x07Toolbar\x12#\n\x04name\x18\x01 \x01(\x0b\x32\x15.JAddOnsFormattedText\x12\x10\n\x08icon_url\x18\x02 \x01(\t\x12\r\n\x05\x63olor\x18\x03 \x01(\x05\x12)\n\x0ctheme_colors\x18\x04 \x01(\x0b\x32\x13.JAddOnsThemeColors\x1a\x9f\x08\n\x04\x43\x61rd\x12\x37\n\x06header\x18\x01 \x01(\x0b\x32\'.JAddOnsContextualAddOn.Card.CardHeader\x12\x36\n\x08sections\x18\x02 \x03(\x0b\x32$.JAddOnsContextualAddOn.Card.Section\x12=\n\x0c\x63\x61rd_actions\x18\x03 \x03(\x0b\x32\'.JAddOnsContextualAddOn.Card.CardAction\x12\x0c\n\x04name\x18\x04 \x01(\t\x12>\n\x0c\x66ixed_footer\x18\x05 \x01(\x0b\x32(.JAddOnsContextualAddOn.Card.FixedFooter\x12\x42\n\x0erefresh_action\x18\x06 \x01(\x0b\x32*.JAddOnsContextualAddOn.Card.RefreshAction\x12\x34\n\x17\x62\x61\x63kground_theme_colors\x18\x07 \x01(\x0b\x32\x13.JAddOnsThemeColors\x1a\xc1\x01\n\nCardHeader\x12$\n\x05title\x18\x01 \x01(\x0b\x32\x15.JAddOnsFormattedText\x12\'\n\x08subtitle\x18\x02 \x01(\x0b\x32\x15.JAddOnsFormattedText\x12\x39\n\x0bimage_style\x18\x03 \x01(\x0e\x32$.JAddOnsImageCropStyle.ImageCropType\x12\x11\n\timage_url\x18\x04 \x01(\t\x12\x16\n\x0eimage_alt_text\x18\x05 \x01(\t\x1a\xbe\x01\n\x07Section\x12\x1e\n\x02id\x18\x06 \x01(\x0b\x32\x12.JAddOnsIdentifier\x12\x13\n\x0b\x64\x65scription\x18\x01 \x01(\t\x12%\n\x06header\x18\x05 \x01(\x0b\x32\x15.JAddOnsFormattedText\x12\x1f\n\x07widgets\x18\x02 \x03(\x0b\x32\x0e.JAddOnsWidget\x12\x13\n\x0b\x63ollapsable\x18\x03 \x01(\x08\x12!\n\x19num_uncollapsable_widgets\x18\x04 \x01(\x05\x1a\x45\n\nCardAction\x12\x14\n\x0c\x61\x63tion_label\x18\x01 \x01(\t\x12!\n\x08on_click\x18\x02 \x01(\x0b\x32\x0f.JAddOnsOnClick\x1a\x9d\x01\n\x0b\x46ixedFooter\x12&\n\x07\x62uttons\x18\x01 \x03(\x0b\x32\x15.JAddOnsWidget.Button\x12\x31\n\x0eprimary_button\x18\x02 \x01(\x0b\x32\x19.JAddOnsWidget.TextButton\x12\x33\n\x10secondary_button\x18\x03 \x01(\x0b\x32\x19.JAddOnsWidget.TextButton\x1a\x33\n\rRefreshAction\x12\"\n\x06method\x18\x01 \x01(\x0b\x32\x12.JAddOnsFormAction\"\x14\n\x04Html\x12\x0c\n\x04html\x18\x02 \x01(\t\"%\n\x0eHtmlAttachment\x12\x13\n\x04html\x18\x02 \x01(\x0b\x32\x05.Html\"\x8b\x01\n\nAttachment\x12\x1d\n\x04html\x18\x01 \x01(\x0b\x32\x0f.HtmlAttachment\x12,\n\x0b\x61\x64\x64_on_data\x18\x02 \x01(\x0b\x32\x17.JAddOnsContextualAddOn\x12\x17\n\x06\x61pp_id\x18\x05 \x01(\x0b\x32\x07.UserId\x12\x17\n\x08\x61pp_user\x18\x06 \x01(\x0b\x32\x05.User\"m\n\x08Reaction\x12\x15\n\x05\x65moji\x18\x01 \x01(\x0b\x32\x06.Emoji\x12\r\n\x05\x63ount\x18\x02 \x01(\x05\x12!\n\x19\x63urrent_user_participated\x18\x03 \x01(\x08\x12\x18\n\x10\x63reate_timestamp\x18\x04 \x01(\x03\"-\n\x0fMessageParentId\x12\x1a\n\x08topic_id\x18\x04 \x01(\x0b\x32\x08.TopicId\"D\n\tMessageId\x12#\n\tparent_id\x18\x01 \x01(\x0b\x32\x10.MessageParentId\x12\x12\n\nmessage_id\x18\x02 \x01(\t\"\xf8\x05\n\x07Message\x12\x16\n\x02id\x18\x01 \x01(\x0b\x32\n.MessageId\x12,\n\rmessage_state\x18\x14 \x01(\x0e\x32\x15.Message.MessageState\x12\x16\n\x07\x63reator\x18\x02 \x01(\x0b\x32\x05.User\x12\x13\n\x0b\x63reate_time\x18\x03 \x01(\x03\x12\x18\n\x10last_update_time\x18\x04 \x01(\x03\x12\x13\n\x0b\x64\x65lete_time\x18\x08 \x01(\x03\x12\x16\n\x0elast_edit_time\x18\x11 \x01(\x03\x12\x11\n\ttext_body\x18\n \x01(\t\x12 \n\x0b\x61nnotations\x18\x0b \x03(\x0b\x32\x0b.Annotation\x12\x10\n\x08local_id\x18\x0e \x01(\t\x12 \n\x0b\x61ttachments\x18\x0f \x03(\x0b\x32\x0b.Attachment\x12.\n\x12retention_settings\x18\x12 \x01(\x0b\x32\x12.RetentionSettings\x12\x1c\n\treactions\x18\x15 \x03(\x0b\x32\t.Reaction\x12\x17\n\x0flast_reply_time\x18\x05 \x01(\x03\x12\x16\n\x0elast_view_time\x18\x06 \x01(\x03\x12\x19\n\x11last_mention_time\x18\x07 \x01(\x03\x12\x1a\n\x12num_unread_replies\x18\t \x01(\x03\x12\x1c\n\nlast_reply\x18\x0c \x01(\x0b\x32\x08.Message\x12\x11\n\tsort_time\x18\r \x01(\x03\x12*\n\x0cmessage_type\x18\x1c \x01(\x0e\x32\x14.Message.MessageType\x12\'\n\x12\x63reator_membership\x18\x1e \x01(\x0b\x32\x0b.Membership\";\n\x0cMessageState\x12\x0b\n\x07UNKNOWN\x10\x00\x12\x11\n\rGROUP_VISIBLE\x10\x01\x12\x0b\n\x07ON_HOLD\x10\x02\"Q\n\x0bMessageType\x12\x1c\n\x18MESSAGE_TYPE_UNSPECIFIED\x10\x00\x12\x10\n\x0cUSER_MESSAGE\x10\x01\x12\x12\n\x0eSYSTEM_MESSAGE\x10\x02\"\xa0\x03\n\rDriveMetadata\x12\n\n\x02id\x18\x01 \x01(\t\x12\r\n\x05title\x18\x02 \x01(\t\x12\x15\n\rthumbnail_url\x18\x03 \x01(\t\x12\x17\n\x0fthumbnail_width\x18\x07 \x01(\x05\x12\x18\n\x10thumbnail_height\x18\x08 \x01(\x05\x12\x10\n\x08mimetype\x18\x04 \x01(\t\x12\x19\n\x11\x65xternal_mimetype\x18\x0e \x01(\t\x12!\n\x19organization_display_name\x18\r \x01(\t\x12\x18\n\x10\x65ncrypted_doc_id\x18\n \x01(\x08\x12\x14\n\x0curl_fragment\x18\x10 \x01(\t\x12&\n\tembed_url\x18\x12 \x01(\x0b\x32\x13.TrustedResourceUrl\x12\x10\n\x08is_owner\x18\x13 \x01(\x08\x12\x11\n\tcan_share\x18\x05 \x01(\x08\x12\x10\n\x08\x63\x61n_edit\x18\x0f \x01(\x08\x12\x10\n\x08\x63\x61n_view\x18\x06 \x01(\x08\x12\x19\n\x11should_not_render\x18\t \x01(\x08\x12\x1e\n\x16is_download_restricted\x18\x11 \x01(\x08\"\x12\n\x03Url\x12\x0b\n\x03url\x18\x03 \x01(\t\"*\n\x12TrustedResourceUrl\x12\x14\n\x0cresource_url\x18\x04 \x01(\t\"\xc8\x02\n\x0bUrlMetadata\x12\r\n\x05title\x18\x01 \x01(\t\x12\x0f\n\x07snippet\x18\x02 \x01(\t\x12\x11\n\timage_url\x18\x03 \x01(\t\x12\x14\n\x0cimage_height\x18\x04 \x01(\t\x12\x13\n\x0bimage_width\x18\x05 \x01(\t\x12\x11\n\x03url\x18\x07 \x01(\x0b\x32\x04.Url\x12\x15\n\x07gws_url\x18\r \x01(\x0b\x32\x04.Url\x12\x1a\n\x0credirect_url\x18\x0e \x01(\x0b\x32\x04.Url\x12$\n\x1cgws_url_expiration_timestamp\x18\x0f \x01(\x03\x12\x0e\n\x06\x64omain\x18\x08 \x01(\t\x12\x19\n\x11should_not_render\x18\t \x01(\x08\x12\x18\n\x10int_image_height\x18\n \x01(\x05\x12\x17\n\x0fint_image_width\x18\x0b \x01(\x05\x12\x11\n\tmime_type\x18\x0c \x01(\t\"\x81\x01\n\x0eUploadMetadata\x12\x18\n\x10\x61ttachment_token\x18\x01 \x01(\t\x12\x14\n\x0c\x63ontent_name\x18\x03 \x01(\t\x12\x14\n\x0c\x63ontent_type\x18\x04 \x01(\t\x12\x10\n\x08local_id\x18\x06 \x01(\t\x12\x17\n\x0f\x63loned_drive_id\x18\t \x01(\t\"\xba\x02\n\x0e\x46ormatMetadata\x12/\n\x0b\x66ormat_type\x18\x01 \x01(\x0e\x32\x1a.FormatMetadata.FormatType\x12\x12\n\nfont_color\x18\x02 \x01(\x05\"\xe2\x01\n\nFormatType\x12\x14\n\x10TYPE_UNSPECIFIED\x10\x00\x12\x08\n\x04\x42OLD\x10\x01\x12\n\n\x06ITALIC\x10\x02\x12\n\n\x06STRIKE\x10\x03\x12\x0f\n\x0bSOURCE_CODE\x10\x04\x12\r\n\tMONOSPACE\x10\x05\x12\n\n\x06HIDDEN\x10\x06\x12\x13\n\x0fMONOSPACE_BLOCK\x10\x07\x12\r\n\tUNDERLINE\x10\x08\x12\x0e\n\nFONT_COLOR\x10\t\x12\x11\n\rBULLETED_LIST\x10\n\x12\x16\n\x12\x42ULLETED_LIST_ITEM\x10\x0b\x12\x11\n\rCLIENT_HIDDEN\x10\x0c\"\x8d\x05\n\nAnnotation\x12\x1d\n\x04type\x18\x01 \x01(\x0e\x32\x0f.AnnotationType\x12\x13\n\x0bstart_index\x18\x02 \x01(\x05\x12\x0e\n\x06length\x18\x03 \x01(\x05\x12\x10\n\x08local_id\x18\t \x01(\t\x12\x11\n\tunique_id\x18\x13 \x01(\t\x12\x34\n\x10\x63hip_render_type\x18\x14 \x01(\x0e\x32\x1a.Annotation.ChipRenderType\x12\x1a\n\x12server_invalidated\x18\r \x01(\x08\x12\x33\n\x15user_mention_metadata\x18\x05 \x01(\x0b\x32\x14.UserMentionMetadata\x12(\n\x0f\x66ormat_metadata\x18\x08 \x01(\x0b\x32\x0f.FormatMetadata\x12\x35\n\x16slash_command_metadata\x18\x0f \x01(\x0b\x32\x15.SlashCommandMetadata\x12&\n\x0e\x64rive_metadata\x18\x04 \x01(\x0b\x32\x0e.DriveMetadata\x12\"\n\x0curl_metadata\x18\x07 \x01(\x0b\x32\x0c.UrlMetadata\x12(\n\x0fupload_metadata\x18\n \x01(\x0b\x32\x0f.UploadMetadata\x12\x36\n\x12membership_changed\x18\x0b \x01(\x0b\x32\x1a.MembershipChangedMetadata\x12*\n\x0croom_updated\x18\x0e \x01(\x0b\x32\x14.RoomUpdatedMetadata\"T\n\x0e\x43hipRenderType\x12\x0b\n\x07UNKNOWN\x10\x00\x12\n\n\x06RENDER\x10\x01\x12\x16\n\x12RENDER_IF_POSSIBLE\x10\x02\x12\x11\n\rDO_NOT_RENDER\x10\x03\"G\n\rTypingContext\x12\x1a\n\x08group_id\x18\x01 \x01(\x0b\x32\x08.GroupId\x12\x1a\n\x08topic_id\x18\x02 \x01(\x0b\x32\x08.TopicId\"}\n\x15SetTypingStateRequest\x12&\n\x0erequest_header\x18\x64 \x01(\x0b\x32\x0e.RequestHeader\x12\x1b\n\x05state\x18\x01 \x01(\x0e\x32\x0c.TypingState\x12\x1f\n\x07\x63ontext\x18\x02 \x01(\x0b\x32\x0e.TypingContext\"6\n\x16SetTypingStateResponse\x12\x1c\n\x14start_timestamp_usec\x18\x01 \x01(\x03\"\x15\n\x04\x44mId\x12\r\n\x05\x64m_id\x18\x01 \x01(\t\"\x1b\n\x07SpaceId\x12\x10\n\x08space_id\x18\x01 \x01(\t\";\n\x07GroupId\x12\x1a\n\x08space_id\x18\x01 \x01(\x0b\x32\x08.SpaceId\x12\x14\n\x05\x64m_id\x18\x03 \x01(\x0b\x32\x05.DmId\"\x99\x07\n\x05Group\x12\x1a\n\x08group_id\x18\x01 \x01(\x0b\x32\x08.GroupId\x12\x0c\n\x04name\x18\x02 \x01(\t\x12\x16\n\x07\x63reator\x18\x04 \x01(\x0b\x32\x05.User\x12\x13\n\x0b\x63reate_time\x18\x05 \x01(\x03\x12\x1a\n\x12last_modified_time\x18\x06 \x01(\x03\x12\x11\n\tsort_time\x18\x07 \x01(\x03\x12\x15\n\rrevision_time\x18\r \x01(\x03\x12)\n\x10group_read_state\x18\x0c \x01(\x0b\x32\x0f.GroupReadState\x12\x1e\n\x16retention_horizon_time\x18\x0e \x01(\x03\x12.\n\x12retention_settings\x18\x10 \x01(\x0b\x32\x12.RetentionSettings\x12\x0f\n\x07is_flat\x18\x11 \x01(\x08\x12\x17\n\x0finterop_enabled\x18\x12 \x01(\x08\x12\"\n\x1aretention_duration_seconds\x18\x13 \x01(\x03\x12$\n\ngroup_type\x18\x16 \x01(\x0e\x32\x10.Group.GroupType\x12$\n\nvisibility\x18\x17 \x01(\x0b\x32\x10.GroupVisibility\x12\x14\n\x0croster_email\x18\x19 \x01(\t\x12$\n\nflat_group\x18\x1a \x01(\x0b\x32\x10.Group.FlatGroup\x12,\n\x0ethreaded_group\x18\x1b \x01(\x0b\x32\x14.Group.ThreadedGroup\x12\x12\n\navatar_url\x18\x1c \x01(\t\x12\x46\n\x1c\x61ttribute_checker_group_type\x18! \x01(\x0e\x32 .SharedAttributeCheckerGroupType\x12\x1e\n\nname_users\x18\" \x01(\x0b\x32\n.NameUsers\x12/\n\x13group_support_level\x18& \x01(\x0e\x32\x12.GroupSupportLevel\x12\x39\n\x18group_unsupported_reason\x18\' \x01(\x0e\x32\x17.GroupUnsupportedReason\x12!\n\x19typing_indicators_enabled\x18+ \x01(\x08\x1a\x0b\n\tFlatGroup\x1a\x0f\n\rThreadedGroup\"K\n\tGroupType\x12\x1a\n\x16GROUP_TYPE_UNSPECIFIED\x10\x00\x12\x08\n\x04ROOM\x10\x01\x12\x0c\n\x08HUMAN_DM\x10\x02\x12\n\n\x06\x42OT_DM\x10\x03\"H\n\x10GroupReadStateId\x12\x18\n\x07user_id\x18\x01 \x01(\x0b\x32\x07.UserId\x12\x1a\n\x08group_id\x18\x02 \x01(\x0b\x32\x08.GroupId\"\x96\x01\n\x0bInviteState\x12 \n\x0finviter_user_id\x18\x01 \x01(\x0b\x32\x07.UserId\x12)\n\x15invitation_message_id\x18\x02 \x01(\x0b\x32\n.MessageId\x12\x18\n\x10show_welcome_mat\x18\x03 \x01(\x08\x12 \n\x0binvite_type\x18\x05 \x01(\x0b\x32\x0b.InviteType\"\x88\x06\n\x0eGroupReadState\x12\x1d\n\x02id\x18\x01 \x01(\x0b\x32\x11.GroupReadStateId\x12\x16\n\x0elast_read_time\x18\x02 \x01(\x03\x12\x1c\n\x14unread_message_count\x18\x04 \x01(\x03\x12\x0f\n\x07starred\x18\x05 \x01(\x08\x12\x18\n\x10update_timestamp\x18\x06 \x01(\x03\x12%\n\x1dunread_subscribed_topic_count\x18\x07 \x01(\x03\x12;\n3unread_subscribed_topic_count_consistency_timestamp\x18\x0b \x01(\x03\x12*\n\x18unread_subscribed_topics\x18\x08 \x03(\x0b\x32\x08.TopicId\x12\x16\n\x0ehide_timestamp\x18\t \x01(\x03\x12\x1f\n\x17\x63lear_history_timestamp\x18\x0e \x01(\x03\x12\"\n\x0cinvite_state\x18\n \x01(\x0b\x32\x0c.InviteState\x12\x39\n\x15notification_settings\x18\x0c \x01(\x0b\x32\x1a.GroupNotificationSettings\x12.\n\x12retention_settings\x18\r \x01(\x0b\x32\x12.RetentionSettings\x12\x0f\n\x07\x62locked\x18\x0f \x01(\x08\x12(\n\x0fmembership_role\x18\x16 \x01(\x0e\x32\x0f.MembershipRole\x12*\n\x10membership_state\x18\x10 \x01(\x0e\x32\x10.MembershipState\x12(\n\x0finvite_category\x18\x11 \x01(\x0e\x32\x0f.InviteCategory\x12%\n\x1dmark_as_unread_timestamp_usec\x18\x12 \x01(\x03\x12\x1d\n\x15visible_in_world_view\x18\x13 \x01(\x08\x12(\n show_notification_card_in_stream\x18\x14 \x01(\x08\x12\x1d\n\x0cjoined_users\x18\x17 \x03(\x0b\x32\x07.UserId\"\xe6\x02\n\x19GroupNotificationSettings\x12@\n\x05state\x18\x01 \x01(\x0e\x32\x31.GroupNotificationSettings.GroupNotificationState\x12\x44\n\nroom_state\x18\x02 \x01(\x0e\x32\x30.GroupNotificationSettings.RoomNotificationState\"P\n\x16GroupNotificationState\x12\x1e\n\x1aUNKNOWN_NOTIFICATION_STATE\x10\x00\x12\t\n\x05MUTED\x10\x01\x12\x0b\n\x07UNMUTED\x10\x02\"o\n\x15RoomNotificationState\x12\x11\n\rNOTIFY_ALWAYS\x10\x00\x12 \n\x1cNOTIFY_LESS_WITH_NEW_THREADS\x10\x01\x12\x0f\n\x0bNOTIFY_LESS\x10\x02\x12\x10\n\x0cNOTIFY_NEVER\x10\x03\"\xb4\x01\n\x11RetentionSettings\x12\x30\n\x05state\x18\x01 \x01(\x0e\x32!.RetentionSettings.RetentionState\x12\x18\n\x10\x65xpiry_timestamp\x18\x02 \x01(\x03\"S\n\x0eRetentionState\x12\x1b\n\x17UNKNOWN_RETENTION_STATE\x10\x00\x12\r\n\tPERMANENT\x10\x01\x12\x15\n\x11\x45PHEMERAL_ONE_DAY\x10\x02\"7\n\x07TopicId\x12\x1a\n\x08group_id\x18\x03 \x01(\x0b\x32\x08.GroupId\x12\x10\n\x08topic_id\x18\x02 \x01(\t\"\xd6\x01\n\x05Topic\x12\x14\n\x02id\x18\x01 \x01(\x0b\x32\x08.TopicId\x12\x11\n\tsort_time\x18\x02 \x01(\x03\x12\x18\n\x10\x63reate_time_usec\x18\x0f \x01(\x03\x12\x19\n\x07replies\x18\x07 \x03(\x0b\x32\x08.Message\x12\x19\n\x11is_system_message\x18\x0c \x01(\x08\x12.\n\x12retention_settings\x18\r \x01(\x0b\x32\x12.RetentionSettings\x12$\n\x1c\x63ontains_more_unread_replies\x18\x0e \x01(\x08\"0\n\x0bMessageInfo\x12!\n\x19\x61\x63\x63\x65pt_format_annotations\x18\x01 \x01(\x08\"\xa5\x02\n\x12\x43reateTopicRequest\x12&\n\x0erequest_header\x18\x64 \x01(\x0b\x32\x0e.RequestHeader\x12\x1a\n\x08group_id\x18\x05 \x01(\x0b\x32\x08.GroupId\x12\x11\n\ttext_body\x18\x02 \x01(\t\x12 \n\x0b\x61nnotations\x18\x03 \x03(\x0b\x32\x0b.Annotation\x12.\n\x12retention_settings\x18\x06 \x01(\x0b\x32\x12.RetentionSettings\x12\x10\n\x08local_id\x18\x04 \x01(\t\x12\x1c\n\x14topic_and_message_id\x18\x07 \x01(\t\x12\x12\n\nhistory_v2\x18\x08 \x01(\x08\x12\"\n\x0cmessage_info\x18\t \x01(\x0b\x32\x0c.MessageInfo\"\x83\x01\n\x13\x43reateTopicResponse\x12\x15\n\x05topic\x18\x01 \x01(\x0b\x32\x06.Topic\x12&\n\x0egroup_revision\x18\x02 \x01(\x0b\x32\x0e.WriteRevision\x12-\n\x16\x63urrent_group_revision\x18\x03 \x01(\x0b\x32\r.ReadRevision\"6\n\x0bInviteeInfo\x12\x18\n\x07user_id\x18\x01 \x01(\x0b\x32\x07.UserId\x12\r\n\x05\x65mail\x18\x02 \x01(\t\"7\n\x11InviteeMemberInfo\x12\"\n\x0cinvitee_info\x18\x01 \x01(\x0b\x32\x0c.InviteeInfo\"\xc8\x03\n\x11SpaceCreationInfo\x12\x0c\n\x04name\x18\x01 \x01(\t\x12$\n\nvisibility\x18\x04 \x01(\x0b\x32\x10.GroupVisibility\x12\x30\n\nflat_group\x18\x05 \x01(\x0b\x32\x1c.SpaceCreationInfo.FlatGroup\x12\x38\n\x0ethreaded_group\x18\x06 \x01(\x0b\x32 .SpaceCreationInfo.ThreadedGroup\x12!\n\x19has_server_generated_name\x18\x07 \x01(\x08\x12\x30\n\x14invitee_member_infos\x18\x08 \x03(\x0b\x32\x12.InviteeMemberInfo\x12\x30\n\nspace_type\x18\x0b \x01(\x0b\x32\x1c.SpaceCreationInfo.SpaceType\x12\x46\n\x1c\x61ttribute_checker_group_type\x18\r \x01(\x0e\x32 .SharedAttributeCheckerGroupType\x12\x19\n\x11shared_drive_name\x18\x0e \x01(\t\x1a\x0b\n\tFlatGroup\x1a\x0f\n\rThreadedGroup\x1a\x0b\n\tSpaceType\"\x95\x01\n\x12\x43reateGroupRequest\x12&\n\x0erequest_header\x18\x64 \x01(\x0b\x32\x0e.RequestHeader\x12!\n\x05space\x18\x01 \x01(\x0b\x32\x12.SpaceCreationInfo\x12\x10\n\x08local_id\x18\x03 \x01(\t\x12\"\n\x1ashould_find_existing_space\x18\x04 \x01(\x08\"9\n\x16\x43reateMembershipResult\x12\x1f\n\nmembership\x18\x01 \x01(\x0b\x32\x0b.Membership\"\xa5\x01\n\x13\x43reateGroupResponse\x12\x15\n\x05group\x18\x01 \x01(\x0b\x32\x06.Group\x12&\n\x0egroup_revision\x18\x02 \x01(\x0b\x32\x0e.WriteRevision\x12%\n\ruser_revision\x18\x03 \x01(\x0b\x32\x0e.WriteRevision\x12(\n\x07results\x18\x04 \x03(\x0b\x32\x17.CreateMembershipResult\"\xcf\x02\n\x0f\x43reateDmRequest\x12&\n\x0erequest_header\x18\x64 \x01(\x0b\x32\x0e.RequestHeader\x12\x34\n\rfetch_options\x18\x08 \x03(\x0e\x32\x1d.CreateDmRequest.FetchOptions\x12\x18\n\x07members\x18\x01 \x03(\x0b\x32\x07.UserId\x12\x1e\n\x08invitees\x18\x07 \x03(\x0b\x32\x0c.InviteeInfo\x12.\n\x12retention_settings\x18\x03 \x01(\x0b\x32\x12.RetentionSettings\x12\x10\n\x08local_id\x18\x04 \x01(\t\x12\x1c\n\x14topic_and_message_id\x18\x05 \x01(\t\"D\n\x0c\x46\x65tchOptions\x12\x0b\n\x07UNKNOWN\x10\x00\x12\'\n#INCLUDE_BOT_MEMBERSHIPS_IN_RESPONSE\x10\x01\"\x87\x01\n\x10\x43reateDmResponse\x12\x12\n\x02\x64m\x18\x01 \x01(\x0b\x32\x06.Group\x12\x15\n\x05topic\x18\x02 \x01(\x0b\x32\x06.Topic\x12&\n\x0egroup_revision\x18\x03 \x01(\x0b\x32\x0e.WriteRevision\x12 \n\x0bmemberships\x18\x04 \x03(\x0b\x32\x0b.Membership\"\xdf\x01\n\x12ListTopicsResponse\x12\x16\n\x06topics\x18\x01 \x03(\x0b\x32\x06.Topic\x12$\n\ruser_revision\x18\x02 \x01(\x0b\x32\r.ReadRevision\x12%\n\x0egroup_revision\x18\x03 \x01(\x0b\x32\r.ReadRevision\x12\x1c\n\x14\x63ontains_first_topic\x18\x04 \x01(\x08\x12\x1b\n\x13\x63ontains_last_topic\x18\x05 \x01(\x08\x12)\n\x10read_receipt_set\x18\x06 \x01(\x0b\x32\x0f.ReadReceiptSet\"\xc0\x03\n\x11ListTopicsRequest\x12&\n\x0erequest_header\x18\x64 \x01(\x0b\x32\x0e.RequestHeader\x12\x1a\n\x08group_id\x18\x08 \x01(\x0b\x32\x08.GroupId\x12\x1c\n\x14page_size_for_topics\x18\x02 \x01(\x05\x12$\n\x1cpage_size_for_unread_replies\x18\x06 \x01(\x05\x12\"\n\x1apage_size_for_read_replies\x18\x07 \x01(\x05\x12\x1d\n\x15page_size_for_replies\x18\x03 \x01(\x05\x12\x36\n\rfetch_options\x18\x05 \x03(\x0e\x32\x1f.ListTopicsRequest.FetchOptions\x12/\n\x13user_not_older_than\x18\t \x01(\x0b\x32\x12.ReferenceRevision\x12\x30\n\x14group_not_older_than\x18\n \x01(\x0b\x32\x12.ReferenceRevision\"E\n\x0c\x46\x65tchOptions\x12\x08\n\x04USER\x10\x01\x12\x18\n\x14TOTAL_MESSAGE_COUNTS\x10\x02\x12\x11\n\rREAD_RECEIPTS\x10\x03\"\xc7\x01\n\x12ListMembersRequest\x12&\n\x0erequest_header\x18\x64 \x01(\x0b\x32\x0e.RequestHeader\x12\x1a\n\x08space_id\x18\x01 \x01(\x0b\x32\x08.SpaceId\x12\x1a\n\x08group_id\x18\x05 \x01(\x0b\x32\x08.GroupId\x12\x11\n\tpage_size\x18\x03 \x01(\x05\x12\x12\n\npage_token\x18\x04 \x01(\t\x12*\n\x0enot_older_than\x18\x06 \x01(\x0b\x32\x12.ReferenceRevision\"\xd7\x01\n\x13ListMembersResponse\x12 \n\x0bmemberships\x18\x01 \x03(\x0b\x32\x0b.Membership\x12\x18\n\x07members\x18\x02 \x03(\x0b\x32\x07.Member\x12\x1d\n\nmember_ids\x18\x05 \x03(\x0b\x32\t.MemberId\x12\x17\n\x0fnext_page_token\x18\x03 \x01(\t\x12%\n\x0egroup_revision\x18\x04 \x01(\x0b\x32\r.ReadRevision\x12%\n\x12invited_member_ids\x18\x06 \x03(\x0b\x32\t.MemberId\"<\n\x0bReadReceipt\x12\x18\n\x10read_time_micros\x18\x02 \x01(\x03\x12\x13\n\x04user\x18\x03 \x01(\x0b\x32\x05.User\"F\n\x0eReadReceiptSet\x12\x0f\n\x07\x65nabled\x18\x01 \x01(\x08\x12#\n\rread_receipts\x18\x02 \x03(\x0b\x32\x0c.ReadReceipt\"\x97\x01\n\x13WebPushNotification\x12\x19\n\x07message\x18\x01 \x01(\x0b\x32\x08.Message\x12\x13\n\x0bsender_name\x18\x02 \x01(\t\x12\x19\n\x11sender_avatar_url\x18\x03 \x01(\t\x12\x12\n\ngroup_name\x18\x04 \x01(\t\x12!\n\x05\x63\x61use\x18\x05 \x01(\x0e\x32\x12.NotificationCause\"\xaa\x01\n\x18\x41ndroidLocalNotification\x12\x11\n\tthread_id\x18\x01 \x01(\t\x12\x0f\n\x07type_id\x18\x02 \x01(\t\x12\x12\n\ngroup_name\x18\x03 \x01(\t\x12\x11\n\ttext_body\x18\x06 \x01(\t\x12\x10\n\x08icon_url\x18\x04 \x01(\t\x12\r\n\x05title\x18\x07 \x01(\t\x12\x10\n\x08group_id\x18\x08 \x01(\t\x12\x10\n\x08sub_text\x18\t \x01(\t\"\x88\x02\n\x14IosLocalNotification\x12\x18\n\x10\x61pns_collapse_id\x18\x01 \x01(\t\x12\x11\n\tthread_id\x18\x02 \x01(\t\x12\r\n\x05title\x18\x03 \x01(\t\x12\x0c\n\x04\x62ody\x18\x04 \x01(\t\x12\x10\n\x08\x63\x61tegory\x18\x05 \x01(\t\x12\x15\n\rgroup_is_flat\x18\x06 \x01(\x08\x12!\n\x19message_is_off_the_record\x18\x07 \x01(\x08\x12\x46\n\x1c\x61ttribute_checker_group_type\x18\t \x01(\x0e\x32 .SharedAttributeCheckerGroupType\x12\x12\n\nnavigation\x18\x08 \x01(\t\"\x8f\x01\n\x17MobileLocalNotification\x12=\n\x1a\x61ndroid_local_notification\x18\x01 \x01(\x0b\x32\x19.AndroidLocalNotification\x12\x35\n\x16ios_local_notification\x18\x02 \x01(\x0b\x32\x15.IosLocalNotification\"\x82\x01\n\x0cMessageEvent\x12\x19\n\x07message\x18\x01 \x01(\x0b\x32\x08.Message\x12\"\n\x1alast_message_in_topic_time\x18\x04 \x01(\x03\x12\x1a\n\x12prev_revision_time\x18\x05 \x01(\x03\x12\x17\n\x0fis_head_message\x18\x06 \x01(\x08\"\x8f\x01\n\x17TypingStateChangedEvent\x12\x1b\n\x05state\x18\x01 \x01(\x0e\x32\x0c.TypingState\x12\x18\n\x07user_id\x18\x02 \x01(\x0b\x32\x07.UserId\x12\x1f\n\x07\x63ontext\x18\x03 \x01(\x0b\x32\x0e.TypingContext\x12\x1c\n\x14start_timestamp_usec\x18\x04 \x01(\x03\"\x9f\x01\n\x16MembershipChangedEvent\x12#\n\x0enew_membership\x18\x01 \x01(\x0b\x32\x0b.Membership\x12\x30\n\x16prior_membership_state\x18\x02 \x01(\x0e\x32\x10.MembershipState\x12.\n\x15prior_membership_role\x18\x03 \x01(\x0e\x32\x0f.MembershipRole\"`\n\x17ReadReceiptChangedEvent\x12\x1a\n\x08group_id\x18\x01 \x01(\x0b\x32\x08.GroupId\x12)\n\x10read_receipt_set\x18\x02 \x01(\x0b\x32\x0f.ReadReceiptSet\"A\n\x10GroupViewedEvent\x12\x1a\n\x08group_id\x18\x01 \x01(\x0b\x32\x08.GroupId\x12\x11\n\tview_time\x18\x02 \x01(\x03\"\xa2\x04\n\x18WebPushNotificationEvent\x12*\n\x0cnotification\x18\x01 \x01(\x0b\x32\x14.WebPushNotification\x12N\n\x16\x64ispatch_approach_type\x18\x02 \x01(\x0e\x32..WebPushNotificationEvent.DispatchApproachType\x12;\n\x19mobile_local_notification\x18\x03 \x01(\x0b\x32\x18.MobileLocalNotification\x12>\n\x0e\x65ndpoint_types\x18\x04 \x03(\x0e\x32&.WebPushNotificationEvent.EndpointType\"\xae\x01\n\x14\x44ispatchApproachType\x12&\n\"DISPATCH_APPROACH_TYPE_UNSPECIFIED\x10\x00\x12\x17\n\x13INTERACTIVE_SESSION\x10\x01\x12\x1d\n\x19\x42ROADCAST_TO_WEB_SESSIONS\x10\x02\x12\x36\n2INTERACTIVE_SESSIONS_INCLUDE_VISIBLE_WEB_UNDER_TTL\x10\x03\"\\\n\x0c\x45ndpointType\x12\x1d\n\x19\x45NDPOINT_TYPE_UNSPECIFIED\x10\x00\x12\x15\n\x11MOBILE_WEBCHANNEL\x10\x01\x12\x16\n\x12\x44\x45SKTOP_WEBCHANNEL\x10\x02\"\xbb\x02\n\x13StreamEventsRequest\x12\x1b\n\x08platform\x18\x04 \x01(\x0e\x32\t.Platform\x12 \n\x0b\x63lient_info\x18\x05 \x01(\x0b\x32\x0b.ClientInfo\x12\x19\n\x11\x63lient_session_id\x18\x06 \x01(\x03\x12\x11\n\tsample_id\x18\x01 \x01(\t\x12\x12\n\nsample_ids\x18\x07 \x03(\t\x12\x1e\n\nping_event\x18\x02 \x01(\x0b\x32\n.PingEvent\x12-\n\x12\x63lock_sync_request\x18\x03 \x01(\x0b\x32\x11.ClockSyncRequest\x12\x39\n\x18group_subscription_event\x18\x08 \x01(\x0b\x32\x17.GroupSubscriptionEvent\x12\x19\n\x11test_user_gaia_id\x18\x64 \x01(\x03\"_\n\nClientInfo\x12\x1b\n\x08platform\x18\x01 \x01(\x0e\x32\t.Platform\x12\x1c\n\x06origin\x18\x02 \x01(\x0e\x32\x0c.EventOrigin\x12\x16\n\x0esource_machine\x18\x03 \x01(\t\"\x89\x02\n\x18\x43lientNotificationsState\x12V\n\x14\x64\x65vice_setting_state\x18\x01 \x01(\x0e\x32\x38.ClientNotificationsState.DeviceNotificationSettingState\"\x94\x01\n\x1e\x44\x65viceNotificationSettingState\x12-\n)DEVICE_NOTIFICATION_SETTING_STATE_UNKNOWN\x10\x00\x12 \n\x1c\x44\x45VICE_NOTIFICATIONS_ENABLED\x10\x01\x12!\n\x1d\x44\x45VICE_NOTIFICATIONS_DISABLED\x10\x02\"\xfb\x05\n\tPingEvent\x12\x1f\n\x05state\x18\x01 \x01(\x0e\x32\x10.PingEvent.State\x12\x41\n\x17\x61pplication_focus_state\x18\x03 \x01(\x0e\x32 .PingEvent.ApplicationFocusState\x12 \n\x18last_interactive_time_ms\x18\x04 \x01(\x03\x12\x43\n\x18\x63lient_interactive_state\x18\x05 \x01(\x0e\x32!.PingEvent.ClientInteractiveState\x12$\n\x1c\x63lient_notifications_enabled\x18\x06 \x01(\x08\x12\x36\n\x13notifications_state\x18\x07 \x01(\x0b\x32\x19.ClientNotificationsState\x12\x1a\n\x12pwa_dedupe_enabled\x18\x08 \x01(\x08\x12\x39\n\x13\x64\x65vice_active_state\x18\t \x01(\x0e\x32\x1c.PingEvent.DeviceActiveState\"8\n\x05State\x12\x15\n\x11STATE_UNSPECIFIED\x10\x00\x12\n\n\x06\x41\x43TIVE\x10\x01\x12\x0c\n\x08INACTIVE\x10\x02\"l\n\x15\x41pplicationFocusState\x12\x1b\n\x17\x46OCUS_STATE_UNSPECIFIED\x10\x00\x12\x1a\n\x16\x46OCUS_STATE_FOREGROUND\x10\x01\x12\x1a\n\x16\x46OCUS_STATE_BACKGROUND\x10\x02\"\\\n\x16\x43lientInteractiveState\x12\x0b\n\x07UNKNOWN\x10\x00\x12\x0f\n\x0bINTERACTIVE\x10\x01\x12\x0b\n\x07\x46OCUSED\x10\x02\x12\x0b\n\x07VISIBLE\x10\x03\x12\n\n\x06HIDDEN\x10\x04\"h\n\x11\x44\x65viceActiveState\x12\x11\n\rSTATE_UNKNOWN\x10\x00\x12\x15\n\x11STATE_IDLE_LOCKED\x10\x01\x12\x17\n\x13STATE_IDLE_UNLOCKED\x10\x02\x12\x10\n\x0cSTATE_ACTIVE\x10\x03\",\n\x10\x43lockSyncRequest\x12\x18\n\x10origin_time_msec\x18\x01 \x01(\x03\"d\n\x11\x43lockSyncResponse\x12\x18\n\x10origin_time_msec\x18\x01 \x01(\x03\x12\x19\n\x11receive_time_msec\x18\x02 \x01(\x03\x12\x1a\n\x12transmit_time_msec\x18\x03 \x01(\x03\"5\n\x16GroupSubscriptionEvent\x12\x1b\n\tgroup_ids\x18\x01 \x03(\x0b\x32\x08.GroupId\"q\n\x14StreamEventsResponse\x12\x15\n\x05\x65vent\x18\x01 \x01(\x0b\x32\x06.Event\x12\x11\n\tsample_id\x18\x02 \x01(\t\x12/\n\x13\x63lock_sync_response\x18\x03 \x01(\x0b\x32\x12.ClockSyncResponse\"\x93\x10\n\x05\x45vent\x12\x1a\n\x08group_id\x18\x01 \x01(\x0b\x32\x08.GroupId\x12\x1e\n\x04type\x18\x03 \x01(\x0e\x32\x10.Event.EventType\x12\x1e\n\x04\x62ody\x18\x04 \x01(\x0b\x32\x10.Event.EventBody\x12\x18\n\x07user_id\x18\x05 \x01(\x0b\x32\x07.UserId\x12%\n\ruser_revision\x18\x06 \x01(\x0b\x32\x0e.WriteRevision\x12&\n\x0egroup_revision\x18\x07 \x01(\x0b\x32\x0e.WriteRevision\x12 \n\x06\x62odies\x18\x08 \x03(\x0b\x32\x10.Event.EventBody\x1a\xb4\x03\n\tEventBody\x12\'\n\x0cgroup_viewed\x18\x03 \x01(\x0b\x32\x11.GroupViewedEvent\x12%\n\x0emessage_posted\x18\x06 \x01(\x0b\x32\r.MessageEvent\x12\x38\n\x15web_push_notification\x18\n \x01(\x0b\x32\x19.WebPushNotificationEvent\x12\x33\n\x12membership_changed\x18\x0e \x01(\x0b\x32\x17.MembershipChangedEvent\x12:\n\x19user_status_updated_event\x18\x17 \x01(\x0b\x32\x17.UserStatusUpdatedEvent\x12<\n\x1atyping_state_changed_event\x18\x1a \x01(\x0b\x32\x18.TypingStateChangedEvent\x12\x36\n\x14read_receipt_changed\x18! \x01(\x0b\x32\x18.ReadReceiptChangedEvent\x12$\n\nevent_type\x18\x0c \x01(\x0e\x32\x10.Event.EventType\x12\x10\n\x08trace_id\x18\x14 \x01(\x03\"\xeb\n\n\tEventType\x12\x0b\n\x07UNKNOWN\x10\x00\x12\x17\n\x13USER_ADDED_TO_GROUP\x10\x01\x12\x1b\n\x17USER_REMOVED_FROM_GROUP\x10\x02\x12\x10\n\x0cGROUP_VIEWED\x10\x03\x12\x10\n\x0cTOPIC_VIEWED\x10\x04\x12\x11\n\rGROUP_UPDATED\x10\x05\x12\x12\n\x0eMESSAGE_POSTED\x10\x06\x12\x13\n\x0fMESSAGE_UPDATED\x10\x07\x12\x13\n\x0fMESSAGE_DELETED\x10\x08\x12\x16\n\x12TOPIC_MUTE_CHANGED\x10\t\x12\x19\n\x15USER_SETTINGS_CHANGED\x10\n\x12\x11\n\rGROUP_STARRED\x10\x0b\x12\x19\n\x15WEB_PUSH_NOTIFICATION\x10\x0c\x12/\n+GROUP_UNREAD_SUBSCRIBED_TOPIC_COUNT_UPDATED\x10\r\x12\x18\n\x14INVITE_COUNT_UPDATED\x10\x0e\x12\x16\n\x12MEMBERSHIP_CHANGED\x10\x0f\x12\x16\n\x12GROUP_HIDE_CHANGED\x10\x10\x12\x1b\n\x17\x44RIVE_ACL_FIX_PROCESSED\x10\x11\x12\'\n#GROUP_NOTIFICATION_SETTINGS_UPDATED\x10\x12\x12\x1e\n\x1aRETENTION_SETTINGS_UPDATED\x10\x13\x12\x11\n\rTOPIC_CREATED\x10\x14\x12\x1a\n\x16ON_HOLD_MESSAGE_POSTED\x10\x15\x12\x1b\n\x17ON_HOLD_MESSAGE_UPDATED\x10\x16\x12\x1d\n\x19ON_HOLD_MESSAGE_PUBLISHED\x10\x17\x12\x13\n\x0fMESSAGE_REACTED\x10\x18\x12\x1d\n\x19USER_STATUS_UPDATED_EVENT\x10\x19\x12$\n GROUP_RETENTION_SETTINGS_UPDATED\x10\x1a\x12$\n USER_WORKING_HOURS_UPDATED_EVENT\x10\x1b\x12\x19\n\x15MESSAGE_SMART_REPLIES\x10\x1c\x12\x18\n\x14TYPING_STATE_CHANGED\x10\x1d\x12\x11\n\rGROUP_DELETED\x10\x1e\x12\x17\n\x13\x42LOCK_STATE_CHANGED\x10\x1f\x12\x11\n\rCLEAR_HISTORY\x10 \x12\x11\n\rSESSION_READY\x10!\x12 \n\x1cGROUP_SORT_TIMESTAMP_CHANGED\x10\"\x12\x1e\n\x1aGSUITE_INTEGRATION_UPDATED\x10#\x12\x18\n\x14READ_RECEIPT_CHANGED\x10$\x12\x12\n\x0eMARK_AS_UNREAD\x10%\x12\x0f\n\x0bGROUP_NO_OP\x10&\x12\x1a\n\x16INVALIDATE_GROUP_CACHE\x10\'\x12\x0e\n\nUSER_NO_OP\x10(\x12\x19\n\x15INVALIDATE_USER_CACHE\x10)\x12#\n\x1fUSER_DENORMALIZED_GROUP_UPDATED\x10*\x12&\n\"USER_PRESENCE_SHARED_UPDATED_EVENT\x10+\x12\x1e\n\x1aNOTIFICATIONS_CARD_UPDATED\x10,\x12\'\n#USER_HUB_AVAILABILITY_UPDATED_EVENT\x10-\x12\x1a\n\x16USER_OWNERSHIP_UPDATED\x10.\x12!\n\x1dSHARED_DRIVE_CREATE_SCHEDULED\x10/\x12\x18\n\x14SHARED_DRIVE_UPDATED\x10\x30\x12\"\n\x1eMESSAGE_PERSONAL_LABEL_UPDATED\x10\x31\x12\x17\n\x13USER_QUOTA_EXCEEDED\x10\x32\"A\n\nInviteType\"3\n\x04Type\x12\x0b\n\x07UNKNOWN\x10\x00\x12\x0e\n\nAT_MENTION\x10\x01\x12\x0e\n\nDIRECT_ADD\x10\x02\"7\n\x0cGroupDetails\x12\x13\n\x0b\x64\x65scription\x18\x01 \x01(\t\x12\x12\n\nguidelines\x18\x02 \x01(\t\"\xca\x03\n\x13RoomUpdatedMetadata\x12\x0c\n\x04name\x18\x02 \x01(\t\x12$\n\nvisibility\x18\x03 \x01(\x0b\x32\x10.GroupVisibility\x12\"\n\x1agroup_link_sharing_enabled\x18\x05 \x01(\x08\x12@\n\x0frename_metadata\x18\x06 \x01(\x0b\x32\'.RoomUpdatedMetadata.RoomRenameMetadata\x12P\n\x16group_details_metadata\x18\x07 \x01(\x0b\x32\x30.RoomUpdatedMetadata.GroupDetailsUpdatedMetadata\x12\x18\n\tinitiator\x18\x04 \x01(\x0b\x32\x05.User\x1a\x39\n\x12RoomRenameMetadata\x12\x10\n\x08new_name\x18\x01 \x01(\t\x12\x11\n\tprev_name\x18\x02 \x01(\t\x1ar\n\x1bGroupDetailsUpdatedMetadata\x12(\n\x11new_group_details\x18\x01 \x01(\x0b\x32\r.GroupDetails\x12)\n\x12prev_group_details\x18\x02 \x01(\x0b\x32\r.GroupDetails\"\xbb\x19\n\x0cMeetingSpace\x12\x18\n\x10meeting_space_id\x18\x01 \x01(\t\x12\x14\n\x0cmeeting_code\x18\x02 \x01(\t\x12\x13\n\x0bmeeting_url\x18\x03 \x01(\t\x12/\n\x0cphone_access\x18\x05 \x03(\x0b\x32\x19.MeetingSpace.PhoneAccess\x12\x42\n\x16universal_phone_access\x18\x08 \x01(\x0b\x32\".MeetingSpace.UniversalPhoneAccess\x12)\n\tcall_info\x18\x06 \x01(\x0b\x32\x16.MeetingSpace.CallInfo\x12\x15\n\rmeeting_alias\x18\x07 \x01(\t\x12\x33\n\x0egateway_access\x18\t \x01(\x0b\x32\x1b.MeetingSpace.GatewayAccess\x12\x15\n\rmore_join_url\x18\n \x01(\t\x12\x31\n\x15\x61\x63\x63\x65pted_number_class\x18\x0b \x03(\x0e\x32\x12.DialInNumberClass\x12:\n\x12gateway_sip_access\x18\r \x03(\x0b\x32\x1e.MeetingSpace.GatewaySipAccess\x12\x37\n\x10\x62roadcast_access\x18\x0e \x01(\x0b\x32\x1d.MeetingSpace.BroadcastAccess\x12(\n\x08settings\x18\x0f \x01(\x0b\x32\x16.MeetingSpace.Settings\x1a|\n\x0bPhoneAccess\x12\x14\n\x0cphone_number\x18\x01 \x01(\t\x12\x1e\n\x16\x66ormatted_phone_number\x18\x05 \x01(\t\x12\x0b\n\x03pin\x18\x02 \x01(\t\x12\x13\n\x0bregion_code\x18\x03 \x01(\t\x12\x15\n\rlanguage_code\x18\x04 \x01(\t\x1a:\n\x14UniversalPhoneAccess\x12\x0b\n\x03pin\x18\x01 \x01(\t\x12\x15\n\rpstn_info_url\x18\x02 \x01(\t\x1a>\n\tPresenter\x12\x1b\n\x13presenter_device_id\x18\x01 \x01(\t\x12\x14\n\x0c\x62y_device_id\x18\x02 \x01(\t\x1a\xcd\x04\n\rRecordingInfo\x12\x45\n\x10recording_status\x18\x01 \x01(\x0e\x32+.MeetingSpace.RecordingInfo.RecordingStatus\x12\x14\n\x0crecording_id\x18\x02 \x01(\t\x12\x1a\n\x12producer_device_id\x18\x03 \x01(\t\x12J\n\x16latest_recording_event\x18\x04 \x01(\x0b\x32*.MeetingSpace.RecordingInfo.RecordingEvent\x12\x1a\n\x12owner_display_name\x18\x05 \x01(\t\x12=\n\x1arecording_application_type\x18\x06 \x01(\x0e\x32\x19.RecordingApplicationType\x1a\xa6\x01\n\x0eRecordingEvent\x12\x11\n\tdevice_id\x18\x01 \x01(\t\x12\x42\n\x04type\x18\x02 \x01(\x0e\x32\x34.MeetingSpace.RecordingInfo.RecordingEvent.EventType\"=\n\tEventType\x12\x1f\n\x1bRECORDING_EVENT_UNSPECIFIED\x10\x00\x12\x0f\n\x0bUSER_ACTION\x10\x01\"s\n\x0fRecordingStatus\x12\x19\n\x15RECORDING_UNSPECIFIED\x10\x00\x12\x16\n\x12RECORDING_INACTIVE\x10\x01\x12\x16\n\x12RECORDING_STARTING\x10\x02\x12\x15\n\x11RECORDING_STARTED\x10\x03\x1a\xaf\x05\n\x14StreamingSessionInfo\x12\x39\n\x06status\x18\x01 \x01(\x0e\x32).MeetingSpace.StreamingSessionInfo.Status\x12\x12\n\nsession_id\x18\x02 \x01(\t\x12\x33\n\x10\x61pplication_type\x18\x03 \x01(\x0e\x32\x19.RecordingApplicationType\x12M\n\x14latest_session_event\x18\x04 \x01(\x0b\x32/.MeetingSpace.StreamingSessionInfo.SessionEvent\x12\x1a\n\x12owner_display_name\x18\x05 \x01(\t\x12\x34\n\x14viewer_access_policy\x18\x06 \x01(\x0e\x32\x16.BroadcastAccessPolicy\x12\x18\n\x10training_enabled\x18\x07 \x01(\x08\x12J\n\x0cviewer_stats\x18\x08 \x01(\x0b\x32\x34.MeetingSpace.StreamingSessionInfo.StreamViewerStats\x1a\x8e\x01\n\x0cSessionEvent\x12\x11\n\tdevice_id\x18\x01 \x01(\t\x12G\n\x04type\x18\x02 \x01(\x0e\x32\x39.MeetingSpace.StreamingSessionInfo.SessionEvent.EventType\"\"\n\tEventType\x12\x15\n\x11\x45VENT_UNSPECIFIED\x10\x00\x1a\x33\n\x11StreamViewerStats\x12\x1e\n\x16\x65stimated_viewer_count\x18\x01 \x01(\x03\"F\n\x06Status\x12\x16\n\x12STATUS_UNSPECIFIED\x10\x00\x12\x0c\n\x08INACTIVE\x10\x01\x12\x0c\n\x08STARTING\x10\x02\x12\x08\n\x04LIVE\x10\x03\x1a\xf4\x06\n\x08\x43\x61llInfo\x12*\n\tpresenter\x18\x01 \x01(\x0b\x32\x17.MeetingSpace.Presenter\x12\x33\n\x0erecording_info\x18\x02 \x01(\x0b\x32\x1b.MeetingSpace.RecordingInfo\x12\x19\n\x11\x63\x61lendar_event_id\x18\x03 \x01(\t\x12\x19\n\x11organization_name\x18\x04 \x01(\t\x12\x1a\n\x12max_joined_devices\x18\x06 \x01(\x05\x12\x1a\n\x12media_backend_info\x18\x08 \x01(\t\x12>\n\x12streaming_sessions\x18\t \x03(\x0b\x32\".MeetingSpace.StreamingSessionInfo\x12\x35\n\x08settings\x18\x0b \x01(\x0b\x32#.MeetingSpace.CallInfo.CallSettings\x12\x38\n\x0cpaygate_info\x18\r \x01(\x0b\x32\".MeetingSpace.CallInfo.PaygateInfo\x12#\n\x1bsupported_caption_languages\x18\x0e \x03(\t\x12\x30\n\x08\x63se_info\x18\x0f \x01(\x0b\x32\x1e.MeetingSpace.CallInfo.CseInfo\x1a\xc8\x01\n\x0c\x43\x61llSettings\x12\x13\n\x0b\x61\x63\x63\x65ss_lock\x18\x01 \x01(\x08\x12\x11\n\tchat_lock\x18\x02 \x01(\x08\x12\x14\n\x0cpresent_lock\x18\x03 \x01(\x08\x12!\n\x19\x61ttendance_report_enabled\x18\x04 \x01(\x08\x12\x12\n\naudio_lock\x18\x05 \x01(\x08\x12\x12\n\nvideo_lock\x18\x06 \x01(\x08\x12\x1a\n\x12moderation_enabled\x18\x07 \x01(\x08\x12\x13\n\x0b\x63se_enabled\x18\x08 \x01(\x08\x1a\xa5\x01\n\x0bPaygateInfo\x12\x1b\n\x13show_upgrade_promos\x18\x01 \x01(\x08\x12\x42\n\x1d\x63\x61ll_ending_soon_warning_time\x18\x02 \x01(\x0b\x32\x1b.ComGoogleProtobufTimestamp\x12\x35\n\x10\x63\x61ll_ending_time\x18\x03 \x01(\x0b\x32\x1b.ComGoogleProtobufTimestamp\x1a\x1e\n\x07\x43seInfo\x12\x13\n\x0bwrapped_key\x18\x01 \x01(\x0c\x1a \n\rGatewayAccess\x12\x0f\n\x07\x65nabled\x18\x01 \x01(\x08\x1a\x38\n\x10GatewaySipAccess\x12\x0b\n\x03uri\x18\x01 \x01(\t\x12\x17\n\x0fsip_access_code\x18\x02 \x01(\t\x1a#\n\x0f\x42roadcastAccess\x12\x10\n\x08view_url\x18\x01 \x01(\t\x1a\x9c\x01\n\x08Settings\x12\x13\n\x0b\x61\x63\x63\x65ss_lock\x18\x01 \x01(\x08\x12!\n\x19\x61ttendance_report_enabled\x18\x02 \x01(\x08\x12\x11\n\tchat_lock\x18\x03 \x01(\x08\x12\x14\n\x0cpresent_lock\x18\x04 \x01(\x08\x12\x1a\n\x12moderation_enabled\x18\x05 \x01(\x08\x12\x13\n\x0b\x63se_enabled\x18\x06 \x01(\x08\"z\n\x11VideoCallMetadata\x12$\n\rmeeting_space\x18\x01 \x01(\x0b\x32\r.MeetingSpace\x12$\n\x1cwas_created_in_current_group\x18\x02 \x01(\x08\x12\x19\n\x11should_not_render\x18\x03 \x01(\x08\"\x9f\x05\n\x19MembershipChangedMetadata\x12-\n\x04type\x18\x01 \x01(\x0e\x32\x1f.MembershipChangedMetadata.Type\x12K\n\x14\x61\x66\x66\x65\x63ted_memberships\x18\x06 \x03(\x0b\x32-.MembershipChangedMetadata.AffectedMembership\x12\x1a\n\tinitiator\x18\x02 \x01(\x0b\x32\x07.UserId\x12#\n\x10\x61\x66\x66\x65\x63ted_members\x18\x03 \x03(\x0b\x32\t.MemberId\x12 \n\x11initiator_profile\x18\x04 \x01(\x0b\x32\x05.User\x12)\n\x18\x61\x66\x66\x65\x63ted_member_profiles\x18\x05 \x03(\x0b\x32\x07.Member\x1a\xcb\x01\n\x12\x41\x66\x66\x65\x63tedMembership\x12\"\n\x0f\x61\x66\x66\x65\x63ted_member\x18\x01 \x01(\x0b\x32\t.MemberId\x12\x30\n\x16prior_membership_state\x18\x02 \x01(\x0e\x32\x10.MembershipState\x12.\n\x15prior_membership_role\x18\x03 \x01(\x0e\x32\x0f.MembershipRole\x12/\n\x16target_membership_role\x18\x04 \x01(\x0e\x32\x0f.MembershipRole\"\xa9\x01\n\x04Type\x12\x14\n\x10TYPE_UNSPECIFIED\x10\x00\x12\x0b\n\x07INVITED\x10\x01\x12\n\n\x06JOINED\x10\x02\x12\t\n\x05\x41\x44\x44\x45\x44\x10\x03\x12\x0b\n\x07REMOVED\x10\x04\x12\x08\n\x04LEFT\x10\x05\x12\r\n\tBOT_ADDED\x10\x06\x12\x0f\n\x0b\x42OT_REMOVED\x10\x07\x12\x1e\n\x1aKICKED_DUE_TO_OTR_CONFLICT\x10\x08\x12\x10\n\x0cROLE_UPDATED\x10\t\"\x86\x02\n\x13UserMentionMetadata\x12\x13\n\x02id\x18\x01 \x01(\x0b\x32\x07.UserId\x12\"\n\x0cinvitee_info\x18\x03 \x01(\x0b\x32\x0c.InviteeInfo\x12\'\n\x04type\x18\x02 \x01(\x0e\x32\x19.UserMentionMetadata.Type\x12\x14\n\x0c\x64isplay_name\x18\x04 \x01(\t\x12\x0e\n\x06gender\x18\x05 \x01(\t\"g\n\x04Type\x12\x14\n\x10TYPE_UNSPECIFIED\x10\x00\x12\n\n\x06INVITE\x10\x01\x12\x0c\n\x08UNINVITE\x10\x02\x12\x0b\n\x07MENTION\x10\x03\x12\x0f\n\x0bMENTION_ALL\x10\x04\x12\x11\n\rFAILED_TO_ADD\x10\x05\"\xf6\x01\n\x14SlashCommandMetadata\x12\x13\n\x02id\x18\x01 \x01(\x0b\x32\x07.UserId\x12(\n\x04type\x18\x02 \x01(\x0e\x32\x1a.SlashCommandMetadata.Type\x12\x14\n\x0c\x63ommand_name\x18\x03 \x01(\t\x12\x12\n\ncommand_id\x18\x04 \x01(\x03\x12\x16\n\x0e\x61rguments_hint\x18\x05 \x01(\t\x12\x17\n\x0ftriggers_dialog\x18\x06 \x01(\x08\"D\n\x04Type\x12\x14\n\x10TYPE_UNSPECIFIED\x10\x00\x12\x07\n\x03\x41\x44\x44\x10\x01\x12\n\n\x06INVOKE\x10\x02\x12\x11\n\rFAILED_TO_ADD\x10\x03\"J\n\x0fGroupVisibility\"7\n\x0fVisibilityState\x12\x0b\n\x07UNKNOWN\x10\x00\x12\x0b\n\x07PRIVATE\x10\x01\x12\n\n\x06PUBLIC\x10\x02\"<\n\x1a\x43omGoogleProtobufTimestamp\x12\x0f\n\x07seconds\x18\x01 \x01(\x03\x12\r\n\x05nanos\x18\x02 \x01(\x05\">\n\x14GetServerTimeRequest\x12&\n\x0erequest_header\x18\x64 \x01(\x0b\x32\x0e.RequestHeader\"G\n\x15GetServerTimeResponse\x12.\n\ttimestamp\x18\x01 \x01(\x0b\x32\x1b.ComGoogleProtobufTimestamp\"N\n\x0c\x43\x61tchUpRange\x12\x1f\n\x17\x66rom_revision_timestamp\x18\x01 \x01(\x03\x12\x1d\n\x15to_revision_timestamp\x18\x02 \x01(\x03\"\x9f\x01\n\x13\x43\x61tchUpGroupRequest\x12&\n\x0erequest_header\x18\x64 \x01(\x0b\x32\x0e.RequestHeader\x12\x1a\n\x08group_id\x18\x01 \x01(\x0b\x32\x08.GroupId\x12\x1c\n\x05range\x18\x02 \x01(\x0b\x32\r.CatchUpRange\x12\x11\n\tpage_size\x18\x03 \x01(\x05\x12\x13\n\x0b\x63utoff_size\x18\x04 \x01(\x05\"\x82\x01\n\x12\x43\x61tchUpUserRequest\x12&\n\x0erequest_header\x18\x64 \x01(\x0b\x32\x0e.RequestHeader\x12\x1c\n\x05range\x18\x01 \x01(\x0b\x32\r.CatchUpRange\x12\x11\n\tpage_size\x18\x02 \x01(\x05\x12\x13\n\x0b\x63utoff_size\x18\x03 \x01(\x05\"\xf8\x01\n\x0f\x43\x61tchUpResponse\x12\x16\n\x06\x65vents\x18\x01 \x03(\x0b\x32\x06.Event\x12/\n\x06status\x18\x02 \x01(\x0e\x32\x1f.CatchUpResponse.ResponseStatus\"\x9b\x01\n\x0eResponseStatus\x12\x0b\n\x07UNKNOWN\x10\x00\x12\r\n\tCOMPLETED\x10\x01\x12\r\n\tPAGINATED\x10\x02\x12\x1b\n\x17\x41\x42ORTED_CUTOFF_EXCEEDED\x10\x03\x12\x1e\n\x1a\x41\x42ORTED_CACHE_INVALIDATION\x10\x04\x12!\n\x1d\x41\x42ORTED_FROM_REVISION_TOO_OLD\x10\x05\"\xc5\x03\n\x0fGetGroupRequest\x12&\n\x0erequest_header\x18\x64 \x01(\x0b\x32\x0e.RequestHeader\x12\x1a\n\x08group_id\x18\x01 \x01(\x0b\x32\x08.GroupId\x12\x34\n\rfetch_options\x18\x04 \x03(\x0e\x32\x1d.GetGroupRequest.FetchOptions\x12/\n\x13user_not_older_than\x18\x02 \x01(\x0b\x32\x12.ReferenceRevision\x12\x30\n\x14group_not_older_than\x18\x03 \x01(\x0b\x32\x12.ReferenceRevision\x12\x1a\n\x12include_invite_dms\x18\x05 \x01(\x08\"\xb8\x01\n\x0c\x46\x65tchOptions\x12\x0b\n\x07MEMBERS\x10\x00\x12\x0c\n\x08INVITEES\x10\x01\x12\x13\n\x0fMEMBER_IDS_ONLY\x10\x02\x12\x1c\n\x18PROFILE_IN_READ_RECEIPTS\x10\x03\x12\x13\n\x0fINCLUDE_SNIPPET\x10\x04\x12\x1e\n\x1aINCLUDE_DYNAMIC_GROUP_NAME\x10\x05\x12%\n!INCLUDE_ACCOUNT_USER_CAPABILITIES\x10\x06\"\xd7\x02\n\x10GetGroupResponse\x12\x15\n\x05group\x18\x01 \x01(\x0b\x32\x06.Group\x12 \n\x0bmemberships\x18\x04 \x03(\x0b\x32\x0b.Membership\x12$\n\ruser_revision\x18\x02 \x01(\x0b\x32\r.ReadRevision\x12%\n\x0egroup_revision\x18\x03 \x01(\x0b\x32\r.ReadRevision\x12*\n\x10membership_state\x18\x06 \x01(\x0e\x32\x10.MembershipState\x12$\n\x11joined_member_ids\x18\x07 \x03(\x0b\x32\t.MemberId\x12%\n\x12invited_member_ids\x18\x08 \x03(\x0b\x32\t.MemberId\x12)\n\x10read_receipt_set\x18\t \x01(\x0b\x32\x0f.ReadReceiptSet\x12\x19\n\x07snippet\x18\n \x01(\x0b\x32\x08.Message\"\x88\x05\n\x0cWorldSection\x12:\n\x12world_section_type\x18\x01 \x01(\x0e\x32\x1e.WorldSection.WorldSectionType\"\xbb\x04\n\x10WorldSectionType\x12\"\n\x1eWORLD_SECTION_TYPE_UNSPECIFIED\x10\x00\x12!\n\x1dSTARRED_DIRECT_MESSAGE_PEOPLE\x10\x01\x12\x11\n\rSTARRED_ROOMS\x10\x02\x12\x1f\n\x1bSTARRED_DIRECT_MESSAGE_BOTS\x10\x03\x12%\n!NON_STARRED_DIRECT_MESSAGE_PEOPLE\x10\x04\x12\x15\n\x11NON_STARRED_ROOMS\x10\x05\x12#\n\x1fNON_STARRED_DIRECT_MESSAGE_BOTS\x10\x06\x12\x1d\n\x19\x41LL_DIRECT_MESSAGE_PEOPLE\x10\x07\x12\r\n\tALL_ROOMS\x10\x08\x12\x1b\n\x17\x41LL_DIRECT_MESSAGE_BOTS\x10\t\x12\x15\n\x11INVITED_DM_PEOPLE\x10\n\x12\x1a\n\x16SPAM_INVITED_DM_PEOPLE\x10\x0b\x12#\n\x1fSTARRED_DIRECT_MESSAGE_EVERYONE\x10\x0c\x12\'\n#NON_STARRED_DIRECT_MESSAGE_EVERYONE\x10\r\x12\x1f\n\x1b\x41LL_DIRECT_MESSAGE_EVERYONE\x10\x0e\x12)\n%STARRED_DMS_AND_STARRED_UNNAMED_ROOMS\x10\x0f\x12\x31\n-NON_STARRED_DMS_AND_NON_STARRED_UNNAMED_ROOMS\x10\x10\"\xaa\x07\n\x0bWorldFilter\x12\x30\n\rstarred_state\x18\x01 \x01(\x0e\x32\x19.WorldFilter.StarredState\x12\x36\n\x10visibility_state\x18\x02 \x01(\x0e\x32\x1c.WorldFilter.VisibilityState\x12*\n\nread_state\x18\x07 \x01(\x0e\x32\x16.WorldFilter.ReadState\x12,\n\x0b\x62lock_state\x18\x08 \x01(\x0e\x32\x17.WorldFilter.BlockState\x12,\n\x0bnamed_state\x18\t \x01(\x0e\x32\x17.WorldFilter.NamedState\x12*\n\x10membership_state\x18\x03 \x01(\x0e\x32\x10.MembershipState\x12(\n\x0finvite_category\x18\x04 \x01(\x0e\x32\x0f.InviteCategory\x12,\n\x0bmember_type\x18\x05 \x01(\x0e\x32\x17.WorldFilter.MemberType\x12*\n\ngroup_type\x18\x06 \x01(\x0e\x32\x16.WorldFilter.GroupType\"K\n\x0cStarredState\x12\x1d\n\x19STARRED_STATE_UNSPECIFIED\x10\x00\x12\x0b\n\x07STARRED\x10\x01\x12\x0f\n\x0bNON_STARRED\x10\x02\"L\n\x0fVisibilityState\x12 \n\x1cVISIBILITY_STATE_UNSPECIFIED\x10\x00\x12\x0b\n\x07VISIBLE\x10\x01\x12\n\n\x06HIDDEN\x10\x02\"Z\n\tReadState\x12\x1a\n\x16READ_STATE_UNSPECIFIED\x10\x00\x12\x08\n\x04READ\x10\x01\x12\n\n\x06UNREAD\x10\x02\x12\x1b\n\x17UNREAD_SUBSCRIBED_TOPIC\x10\x03\"E\n\nBlockState\x12\x1b\n\x17\x42LOCK_STATE_UNSPECIFIED\x10\x00\x12\x0b\n\x07\x42LOCKED\x10\x01\x12\r\n\tUNBLOCKED\x10\x02\"A\n\nNamedState\x12\x1b\n\x17NAMED_STATE_UNSPECIFIED\x10\x00\x12\t\n\x05NAMED\x10\x01\x12\x0b\n\x07UNNAMED\x10\x02\"=\n\nMemberType\x12\x1b\n\x17MEMBER_TYPE_UNSPECIFIED\x10\x00\x12\t\n\x05HUMAN\x10\x01\x12\x07\n\x03\x42OT\x10\x02\"9\n\tGroupType\x12\x1a\n\x16GROUP_TYPE_UNSPECIFIED\x10\x00\x12\x06\n\x02\x44M\x10\x01\x12\x08\n\x04ROOM\x10\x02\"\\\n\tNameUsers\x12\x1e\n\rname_user_ids\x18\x01 \x03(\x0b\x32\x07.UserId\x12\x1b\n\x13has_more_name_users\x18\x02 \x01(\x08\x12\x12\n\ngroup_name\x18\x03 \x01(\t\"\xea\x08\n\rWorldItemLite\x12\x1a\n\x08group_id\x18\x01 \x01(\x0b\x32\x08.GroupId\x12%\n\x0egroup_revision\x18\x02 \x01(\x0b\x32\r.ReadRevision\x12\x16\n\x0esort_timestamp\x18\x03 \x01(\x03\x12#\n\nread_state\x18\x04 \x01(\x0b\x32\x0f.GroupReadState\x12\x11\n\troom_name\x18\x05 \x01(\t\x12,\n\ndm_members\x18\x06 \x01(\x0b\x32\x18.WorldItemLite.DmMembers\x12\x1e\n\nname_users\x18\x14 \x01(\x0b\x32\n.NameUsers\x12%\n\x1dretention_horizon_time_micros\x18\x08 \x01(\x03\x12\"\n\x1aretention_duration_seconds\x18\t \x01(\x03\x12,\n\ngroup_lite\x18\x07 \x01(\x0b\x32\x18.WorldItemLite.GroupLite\x12\x19\n\x07message\x18\r \x01(\x0b\x32\x08.Message\x12\x1a\n\x12is_message_blocked\x18\x19 \x01(\x08\x12,\n\nflat_group\x18\x0e \x01(\x0b\x32\x18.WorldItemLite.FlatGroup\x12\x34\n\x0ethreaded_group\x18\x0f \x01(\x0b\x32\x1c.WorldItemLite.ThreadedGroup\x12\x12\n\navatar_url\x18\x10 \x01(\t\x12\x46\n\x1c\x61ttribute_checker_group_type\x18\x13 \x01(\x0e\x32 .SharedAttributeCheckerGroupType\x12/\n\x13group_support_level\x18\x17 \x01(\x0e\x32\x12.GroupSupportLevel\x12\x39\n\x18group_unsupported_reason\x18\x18 \x01(\x0e\x32\x17.GroupUnsupportedReason\x1aV\n\x0eMembershipLite\x12\x18\n\x07user_id\x18\x01 \x01(\x0b\x32\x07.UserId\x12*\n\x10membership_state\x18\x02 \x01(\x0e\x32\x10.MembershipState\x1av\n\tDmMembers\x12\x18\n\x07members\x18\x01 \x03(\x0b\x32\x07.UserId\x12\x32\n\x0bmemberships\x18\x02 \x03(\x0b\x32\x1d.WorldItemLite.MembershipLite\x12\x1b\n\x13has_all_memberships\x18\x03 \x01(\x08\x1a\xad\x01\n\tGroupLite\x12\x1b\n\ncreator_id\x18\x01 \x01(\x0b\x32\x07.UserId\x12\x13\n\x0b\x63reate_time\x18\x02 \x01(\x03\x12\x0f\n\x07is_flat\x18\x03 \x01(\x08\x12.\n\x12retention_settings\x18\x04 \x01(\x0b\x32\x12.RetentionSettings\x12\x17\n\x0finterop_enabled\x18\x05 \x01(\x08\x12\x14\n\x0croster_email\x18\x07 \x01(\t\x1a\x0b\n\tFlatGroup\x1a\x0f\n\rThreadedGroup\"\xd8\x01\n\x13WorldSectionRequest\x12\x11\n\tpage_size\x18\x01 \x01(\x05\x12$\n\rworld_section\x18\x02 \x01(\x0b\x32\r.WorldSection\x12\"\n\x0cworld_filter\x18\x04 \x01(\x0b\x32\x0c.WorldFilter\x12$\n\x1cnum_world_items_with_snippet\x18\x05 \x01(\x05\x12$\n\x1c\x61nchor_sort_timestamp_micros\x18\x03 \x01(\x03\x12\x18\n\x10pagination_token\x18\x06 \x01(\t\"\xe7\x01\n\x14WorldSectionResponse\x12$\n\rworld_section\x18\x01 \x01(\x0b\x32\r.WorldSection\x12\"\n\x0cworld_filter\x18\x04 \x01(\x0b\x32\x0c.WorldFilter\x12#\n\x0bworld_items\x18\x02 \x03(\x0b\x32\x0e.WorldItemLite\x12.\n&next_page_anchor_sort_timestamp_micros\x18\x03 \x01(\x03\x12\x16\n\x0ehas_more_items\x18\x05 \x01(\x08\x12\x18\n\x10pagination_token\x18\x06 \x01(\t\"\xde\x03\n\x15PaginatedWorldRequest\x12&\n\x0erequest_header\x18\x01 \x01(\x0b\x32\x0e.RequestHeader\x12\x34\n\x16world_section_requests\x18\x02 \x03(\x0b\x32\x14.WorldSectionRequest\x12\x1f\n\x17world_consistency_token\x18\x03 \x01(\t\x12:\n\rfetch_options\x18\x04 \x03(\x0e\x32#.PaginatedWorldRequest.FetchOptions\x12\x1e\n\x16\x66\x65tch_from_user_spaces\x18\x05 \x01(\x08\x12*\n\"receive_world_update_notifications\x18\x06 \x01(\x08\x12(\n fetch_snippets_for_unnamed_rooms\x18\x07 \x01(\x08\"\x93\x01\n\x0c\x46\x65tchOptions\x12\x0b\n\x07UNKNOWN\x10\x00\x12\x16\n\x12\x45XCLUDE_GROUP_LITE\x10\x01\x12\x1a\n\x16\x46\x45TCH_BOTS_IN_HUMAN_DM\x10\x02\x12$\n FETCH_SPACE_INTEGRATION_PAYLOADS\x10\x03\x12\x1c\n\x18\x46\x45TCH_GROUPS_D3_POLICIES\x10\x04\"\xbc\x01\n\x16PaginatedWorldResponse\x12\x36\n\x17world_section_responses\x18\x01 \x03(\x0b\x32\x15.WorldSectionResponse\x12\x1f\n\x17world_consistency_token\x18\x02 \x01(\t\x12$\n\ruser_revision\x18\x03 \x01(\x0b\x32\r.ReadRevision\x12#\n\x0bworld_items\x18\x04 \x03(\x0b\x32\x0e.WorldItemLite\"\xa9\x01\n\x18RemoveMembershipsRequest\x12&\n\x0erequest_header\x18\x64 \x01(\x0b\x32\x0e.RequestHeader\x12\x1d\n\nmember_ids\x18\x01 \x03(\x0b\x32\t.MemberId\x12\x1a\n\x08group_id\x18\x02 \x01(\x0b\x32\x08.GroupId\x12*\n\x10membership_state\x18\x03 \x01(\x0e\x32\x10.MembershipState\"}\n\x16RemoveMembershipResult\x12&\n\x0erequest_header\x18\x64 \x01(\x0b\x32\x0e.RequestHeader\x12\x1c\n\tmember_id\x18\x01 \x01(\x0b\x32\t.MemberId\x12\x1d\n\x15\x61ttached_roster_names\x18\x04 \x03(\t\"\xb0\x01\n\x19RemoveMembershipsResponse\x12(\n\x07results\x18\x01 \x03(\x0b\x32\x17.RemoveMembershipResult\x12.\n\x12retention_settings\x18\x03 \x01(\x0b\x32\x12.RetentionSettings\x12\x39\n!retention_settings_group_revision\x18\x04 \x01(\x0b\x32\x0e.WriteRevision\"^\n\x10HideGroupRequest\x12&\n\x0erequest_header\x18\x64 \x01(\x0b\x32\x0e.RequestHeader\x12\x14\n\x02id\x18\x01 \x01(\x0b\x32\x08.GroupId\x12\x0c\n\x04hide\x18\x02 \x01(\x08\"_\n\x11HideGroupResponse\x12#\n\nread_state\x18\x01 \x01(\x0b\x32\x0f.GroupReadState\x12%\n\ruser_revision\x18\x03 \x01(\x0b\x32\x0e.WriteRevision\"\x9b\x01\n\x1aInviteNotificationSettings\x12>\n\x06option\x18\x01 \x03(\x0e\x32..InviteNotificationSettings.NotificationOption\"=\n\x12NotificationOption\x12\x0f\n\x0bUNSPECIFIED\x10\x00\x12\x16\n\x12\x45MAIL_NOTIFICATION\x10\x01\"\x96\x02\n\x17\x43reateMembershipRequest\x12&\n\x0erequest_header\x18\x64 \x01(\x0b\x32\x0e.RequestHeader\x12\x1d\n\nmember_ids\x18\x02 \x03(\x0b\x32\t.MemberId\x12\x30\n\x14invitee_member_infos\x18\x08 \x03(\x0b\x32\x12.InviteeMemberInfo\x12*\n\x10membership_state\x18\x03 \x01(\x0e\x32\x10.MembershipState\x12\x1a\n\x08group_id\x18\x04 \x01(\x0b\x32\x08.GroupId\x12:\n\x15notification_settings\x18\x05 \x01(\x0b\x32\x1b.InviteNotificationSettings\"\xd7\x01\n\x18\x43reateMembershipResponse\x12(\n\x07results\x18\x02 \x03(\x0b\x32\x17.CreateMembershipResult\x12&\n\x0egroup_revision\x18\x04 \x03(\x0b\x32\x0e.WriteRevision\x12.\n\x12retention_settings\x18\x05 \x01(\x0b\x32\x12.RetentionSettings\x12\x39\n!retention_settings_group_revision\x18\x06 \x01(\x0b\x32\x0e.WriteRevision\"q\n\x19MarkGroupReadstateRequest\x12&\n\x0erequest_header\x18\x64 \x01(\x0b\x32\x0e.RequestHeader\x12\x14\n\x02id\x18\x01 \x01(\x0b\x32\x08.GroupId\x12\x16\n\x0elast_read_time\x18\x02 \x01(\x03\"h\n\x1aMarkGroupReadstateResponse\x12#\n\nread_state\x18\x01 \x01(\x0b\x32\x0f.GroupReadState\x12%\n\ruser_revision\x18\x02 \x01(\x0b\x32\x0e.WriteRevision\"[\n\x18SetPresenceSharedRequest\x12&\n\x0erequest_header\x18\x64 \x01(\x0b\x32\x0e.RequestHeader\x12\x17\n\x0fpresence_shared\x18\x01 \x01(\x08\"d\n\x19SetPresenceSharedResponse\x12 \n\x0buser_status\x18\x01 \x01(\x0b\x32\x0b.UserStatus\x12%\n\ruser_revision\x18\x02 \x01(\x0b\x32\x0e.WriteRevision\"\xe8\x01\n\x15SetDndDurationRequest\x12&\n\x0erequest_header\x18\x64 \x01(\x0b\x32\x0e.RequestHeader\x12\x37\n\x11\x63urrent_dnd_state\x18\x02 \x01(\x0e\x32\x1c.SetDndDurationRequest.State\x12\x1d\n\x15new_dnd_duration_usec\x18\x01 \x01(\x03\x12!\n\x19\x64nd_expiry_timestamp_usec\x18\x03 \x01(\x03\",\n\x05State\x12\x0b\n\x07UNKNOWN\x10\x00\x12\r\n\tAVAILABLE\x10\x01\x12\x07\n\x03\x44ND\x10\x02\"a\n\x16SetDndDurationResponse\x12 \n\x0buser_status\x18\x01 \x01(\x0b\x32\x0b.UserStatus\x12%\n\ruser_revision\x18\x02 \x01(\x0b\x32\x0e.WriteRevision\"\xbd\x02\n\x12UpdateGroupRequest\x12&\n\x0erequest_header\x18\x64 \x01(\x0b\x32\x0e.RequestHeader\x12\x1a\n\x08space_id\x18\x01 \x01(\x0b\x32\x08.SpaceId\x12\x34\n\x0cupdate_masks\x18\x04 \x03(\x0e\x32\x1e.UpdateGroupRequest.UpdateMask\x12\x0c\n\x04name\x18\x02 \x01(\t\x12$\n\nvisibility\x18\x05 \x01(\x0b\x32\x10.GroupVisibility\"y\n\nUpdateMask\x12\x1b\n\x17UPDATE_MASK_UNSPECIFIED\x10\x00\x12\x08\n\x04NAME\x10\x01\x12\x0e\n\nVISIBILITY\x10\x02\x12\x0f\n\x0b\x41VATAR_INFO\x10\x03\x12\x11\n\rSPACE_DETAILS\x10\x04\x12\x10\n\x0cSHARED_DRIVE\x10\x05\"T\n\x13UpdateGroupResponse\x12\x15\n\x05group\x18\x01 \x01(\x0b\x32\x06.Group\x12&\n\x0egroup_revision\x18\x02 \x01(\x0b\x32\x0e.WriteRevision\"\x95\x01\n\x12\x42lockEntityRequest\x12&\n\x0erequest_header\x18\x64 \x01(\x0b\x32\x0e.RequestHeader\x12\x18\n\x07user_id\x18\x01 \x01(\x0b\x32\x07.UserId\x12\x1a\n\x08group_id\x18\x02 \x01(\x0b\x32\x08.GroupId\x12\x0f\n\x07\x62locked\x18\x03 \x01(\x08\x12\x10\n\x08reported\x18\x04 \x01(\x08\"a\n\x13\x42lockEntityResponse\x12#\n\nread_state\x18\x01 \x01(\x0b\x32\x0f.GroupReadState\x12%\n\ruser_revision\x18\x02 \x01(\x0b\x32\x0e.WriteRevision\"\xc2\x01\n\x16SetCustomStatusRequest\x12&\n\x0erequest_header\x18\x64 \x01(\x0b\x32\x0e.RequestHeader\x12$\n\rcustom_status\x18\x01 \x01(\x0b\x32\r.CustomStatus\x12+\n#custom_status_expiry_timestamp_usec\x18\x02 \x01(\x03\x12-\n%custom_status_remaining_duration_usec\x18\x03 \x01(\x03\"b\n\x17SetCustomStatusResponse\x12 \n\x0buser_status\x18\x01 \x01(\x0b\x32\x0b.UserStatus\x12%\n\ruser_revision\x18\x02 \x01(\x0b\x32\x0e.WriteRevision\":\n\rWriteRevision\x12\x11\n\ttimestamp\x18\x01 \x01(\x03\x12\x16\n\x0eprev_timestamp\x18\x02 \x01(\x03\"!\n\x0cReadRevision\x12\x11\n\ttimestamp\x18\x01 \x01(\x03\"&\n\x11ReferenceRevision\x12\x11\n\ttimestamp\x18\x01 \x01(\x03*\x1e\n\x08UserType\x12\t\n\x05HUMAN\x10\x00\x12\x07\n\x03\x42OT\x10\x01*I\n\x0eInviteCategory\x12\x12\n\x0eUNKNOWN_INVITE\x10\x00\x12\x12\n\x0eREGULAR_INVITE\x10\x01\x12\x0f\n\x0bSPAM_INVITE\x10\x02*_\n\x08Presence\x12\x16\n\x12UNDEFINED_PRESENCE\x10\x00\x12\n\n\x06\x41\x43TIVE\x10\x01\x12\x0c\n\x08INACTIVE\x10\x02\x12\x0b\n\x07UNKNOWN\x10\x03\x12\x14\n\x10SHARING_DISABLED\x10\x04*E\n\x0e\x44ndState_State\x12\x1b\n\x17\x44ND_STATE_STATE_UNKNOWN\x10\x00\x12\r\n\tAVAILABLE\x10\x01\x12\x07\n\x03\x44ND\x10\x02*9\n\x0bTypingState\x12\x11\n\rUNKNOWN_STATE\x10\x00\x12\n\n\x06TYPING\x10\x01\x12\x0b\n\x07STOPPED\x10\x02*~\n\x16GroupUnsupportedReason\x12(\n$GROUP_UNSUPPORTED_REASON_UNSPECIFIED\x10\x00\x12\x1c\n\x18GROUP_DISABLED_ON_CLIENT\x10\x01\x12\x1c\n\x18GROUP_DISABLED_ON_SERVER\x10\x02*M\n\x11GroupSupportLevel\x12\x0f\n\x0bUNSUPPORTED\x10\x00\x12\x12\n\x0e\x44\x41TA_SUPPORTED\x10\x01\x12\x13\n\x0f\x46ULLY_SUPPORTED\x10\x02*\xa0\x01\n\x11NotificationCause\x12\x1e\n\x1aUNKNOWN_NOTIFICATION_CAUSE\x10\x00\x12\x12\n\x0e\x44IRECT_MESSAGE\x10\x01\x12\x0e\n\nAT_MENTION\x10\x02\x12\r\n\tNEW_TOPIC\x10\x03\x12\x1a\n\x16SUBSCRIBED_TOPIC_REPLY\x10\x04\x12\x1c\n\x18UNSUBSCRIBED_TOPIC_REPLY\x10\x05*\xc2\x05\n\x0b\x45ventOrigin\x12\x16\n\x12ORIGIN_UNSPECIFIED\x10\x00\x12\x0b\n\x07\x41NDROID\x10\x64\x12\x0f\n\x0b\x41NDROID_DEV\x10\x65\x12\x14\n\x10\x41NDROID_PRODTEST\x10\x66\x12\x13\n\x0f\x41NDROID_STAGING\x10g\x12\x10\n\x0c\x41NDROID_PROD\x10h\x12\x07\n\x03IOS\x10n\x12\x0b\n\x07IOS_DEV\x10o\x12\x10\n\x0cIOS_PRODTEST\x10p\x12\x0f\n\x0bIOS_STAGING\x10q\x12\x0c\n\x08IOS_PROD\x10r\x12\x0e\n\nWEB_ORIGIN\x10x\x12\x0b\n\x07WEB_DEV\x10y\x12\x10\n\x0cWEB_PRODTEST\x10z\x12\x0c\n\x08WEB_PROD\x10{\x12\x0f\n\x0bWEB_STAGING\x10|\x12\x12\n\x0e\x44\x45SKTOP_ORIGIN\x10}\x12\x0f\n\x0b\x44\x45SKTOP_DEV\x10~\x12\x14\n\x10\x44\x45SKTOP_PRODTEST\x10\x7f\x12\x14\n\x0f\x44\x45SKTOP_STAGING\x10\x80\x01\x12\x11\n\x0c\x44\x45SKTOP_PROD\x10\x81\x01\x12\x15\n\x10\x43OMPANION_ORIGIN\x10\x82\x01\x12\x12\n\rCOMPANION_DEV\x10\x83\x01\x12\x17\n\x12\x43OMPANION_PRODTEST\x10\x84\x01\x12\x16\n\x11\x43OMPANION_STAGING\x10\x85\x01\x12\x13\n\x0e\x43OMPANION_PROD\x10\x86\x01\x12\x15\n\x10WEB_GMAIL_ORIGIN\x10\x87\x01\x12\x12\n\rWEB_GMAIL_DEV\x10\x88\x01\x12\x17\n\x12WEB_GMAIL_PRODTEST\x10\x89\x01\x12\x16\n\x11WEB_GMAIL_STAGING\x10\x8a\x01\x12\x13\n\x0eWEB_GMAIL_PROD\x10\x8b\x01\x12\x13\n\x0eWEB_PWA_ORIGIN\x10\x8c\x01\x12\x10\n\x0bWEB_PWA_DEV\x10\x8d\x01\x12\x15\n\x10WEB_PWA_PRODTEST\x10\x8e\x01\x12\x14\n\x0fWEB_PWA_STAGING\x10\x8f\x01\x12\x11\n\x0cWEB_PWA_PROD\x10\x90\x01*\xa4\x01\n\x08Platform\x12\x16\n\x12UNDEFINED_PLATFORM\x10\x00\x12\x07\n\x03WEB\x10\x01\x12\n\n\x06MOBILE\x10\x02\x12\t\n\x05\x44RONE\x10\x03\x12\x14\n\x10INTEROP_PRESENCE\x10\x04\x12\x13\n\x0f\x43LASSIC_DESKTOP\x10\x05\x12\x11\n\rCLASSIC_PHONE\x10\x06\x12\x13\n\x0f\x43LASSIC_UNKNOWN\x10\x07\x12\r\n\tWEB_GMAIL\x10\x08*\x8c\x04\n\x0e\x41nnotationType\x12\x1b\n\x17\x41NNOTATION_TYPE_UNKNOWN\x10\x00\x12\x07\n\x03URL\x10\x01\x12\x0e\n\nDRIVE_FILE\x10\x02\x12\r\n\tDRIVE_DOC\x10\x03\x12\x0f\n\x0b\x44RIVE_SHEET\x10\x04\x12\x0f\n\x0b\x44RIVE_SLIDE\x10\x05\x12\x10\n\x0cUSER_MENTION\x10\x06\x12\t\n\x05VIDEO\x10\x07\x12\x0f\n\x0b\x46ORMAT_DATA\x10\x08\x12\t\n\x05IMAGE\x10\t\x12\x07\n\x03PDF\x10\n\x12\x0e\n\nVIDEO_CALL\x10\x0b\x12\x16\n\x12MEMBERSHIP_CHANGED\x10\x0c\x12\x13\n\x0fUPLOAD_METADATA\x10\r\x12\x10\n\x0cROOM_UPDATED\x10\x0e\x12\x0e\n\nINVITATION\x10\x0f\x12\x11\n\rSLASH_COMMAND\x10\x10\x12\x16\n\x12GSUITE_INTEGRATION\x10\x11\x12\x0e\n\nDRIVE_FORM\x10\x12\x12$\n GROUP_RETENTION_SETTINGS_UPDATED\x10\x13\x12\x15\n\x11\x42\x41\x42\x45L_PLACEHOLDER\x10\x14\x12\"\n\x1eREAD_RECEIPTS_SETTINGS_UPDATED\x10\x15\x12\x1c\n\x18INCOMING_WEBHOOK_CHANGED\x10\x16\x12\x1e\n\x1aINTEGRATION_CONFIG_UPDATED\x10\x17\x12\x18\n\x14\x43ONSENTED_APP_UNFURL\x10\x18*Z\n\x11\x44ialInNumberClass\x12\x1c\n\x18NUMBER_CLASS_UNSPECIFIED\x10\x00\x12\x0c\n\x08LOW_COST\x10\x01\x12\r\n\tHIGH_COST\x10\x02\x12\n\n\x06LEGACY\x10\x03*v\n\x18RecordingApplicationType\x12*\n&RECORDING_APPLICATION_TYPE_UNSPECIFIED\x10\x00\x12\r\n\tRECORDING\x10\x01\x12\x10\n\x0cGLIVE_STREAM\x10\x03\x12\r\n\tBROADCAST\x10\x04*a\n\x15\x42roadcastAccessPolicy\x12*\n&BROADCASTING_ACCESS_POLICY_UNSPECIFIED\x10\x00\x12\x10\n\x0cORGANIZATION\x10\x01\x12\n\n\x06PUBLIC\x10\x02*F\n\nCallStatus\x12\x17\n\x13UNKNOWN_CALL_STATUS\x10\x00\x12\x0f\n\x0b\x41\x43TIVE_CALL\x10\x01\x12\x0e\n\nENDED_CALL\x10\x02*J\n\nRingStatus\x12\x17\n\x13UNKNOWN_RING_STATUS\x10\x00\x12\x0b\n\x07RINGING\x10\x01\x12\n\n\x06JOINED\x10\x02\x12\n\n\x06MISSED\x10\x03*u\n\x10UserAccountState\x12\x1e\n\x1aUSER_ACCOUNT_STATE_UNKNOWN\x10\x00\x12\x0b\n\x07\x45NABLED\x10\x01\x12\x0c\n\x08\x44ISABLED\x10\x02\x12\x0b\n\x07\x44\x45LETED\x10\x03\x12\x19\n\x15TEMPORARY_UNAVAILABLE\x10\x04*d\n\x0eMembershipRole\x12\x10\n\x0cROLE_UNKNOWN\x10\x00\x12\r\n\tROLE_NONE\x10\x01\x12\x10\n\x0cROLE_INVITEE\x10\x02\x12\x0f\n\x0bROLE_MEMBER\x10\x03\x12\x0e\n\nROLE_OWNER\x10\x04*R\n\x07\x41ppType\x12\x18\n\x14\x41PP_TYPE_UNSPECIFIED\x10\x00\x12\x07\n\x03\x41PP\x10\x01\x12\x0e\n\nGSUITE_APP\x10\x02\x12\x14\n\x10INCOMING_WEBHOOK\x10\x03*x\n\x0fMembershipState\x12\x12\n\x0eMEMBER_UNKNOWN\x10\x00\x12\x12\n\x0eMEMBER_INVITED\x10\x01\x12\x11\n\rMEMBER_JOINED\x10\x02\x12\x17\n\x13MEMBER_NOT_A_MEMBER\x10\x03\x12\x11\n\rMEMBER_FAILED\x10\x04*\xf6\x01\n\x1fSharedAttributeCheckerGroupType\x12,\n(ATTRIBUTE_CHECKER_GROUP_TYPE_UNSPECIFIED\x10\x00\x12\x17\n\x13ONE_TO_ONE_HUMAN_DM\x10\x01\x12\x15\n\x11ONE_TO_ONE_BOT_DM\x10\x02\x12!\n\x1dIMMUTABLE_MEMBERSHIP_GROUP_DM\x10\x03\x12\r\n\tFLAT_ROOM\x10\x04\x12\x11\n\rTHREADED_ROOM\x10\x05\x12!\n\x1dIMMUTABLE_MEMBERSHIP_HUMAN_DM\x10\x06\x12\r\n\tPOST_ROOM\x10\x07'
)

_USERTYPE = _descriptor.EnumDescriptor(
  name='UserType',
  full_name='UserType',
  filename=None,
  file=DESCRIPTOR,
  create_key=_descriptor._internal_create_key,
  values=[
    _descriptor.EnumValueDescriptor(
      name='HUMAN', index=0, number=0,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='BOT', index=1, number=1,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
  ],
  containing_type=None,
  serialized_options=None,
  serialized_start=42521,
  serialized_end=42551,
)
_sym_db.RegisterEnumDescriptor(_USERTYPE)

UserType = enum_type_wrapper.EnumTypeWrapper(_USERTYPE)
_INVITECATEGORY = _descriptor.EnumDescriptor(
  name='InviteCategory',
  full_name='InviteCategory',
  filename=None,
  file=DESCRIPTOR,
  create_key=_descriptor._internal_create_key,
  values=[
    _descriptor.EnumValueDescriptor(
      name='UNKNOWN_INVITE', index=0, number=0,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='REGULAR_INVITE', index=1, number=1,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='SPAM_INVITE', index=2, number=2,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
  ],
  containing_type=None,
  serialized_options=None,
  serialized_start=42553,
  serialized_end=42626,
)
_sym_db.RegisterEnumDescriptor(_INVITECATEGORY)

InviteCategory = enum_type_wrapper.EnumTypeWrapper(_INVITECATEGORY)
_PRESENCE = _descriptor.EnumDescriptor(
  name='Presence',
  full_name='Presence',
  filename=None,
  file=DESCRIPTOR,
  create_key=_descriptor._internal_create_key,
  values=[
    _descriptor.EnumValueDescriptor(
      name='UNDEFINED_PRESENCE', index=0, number=0,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='ACTIVE', index=1, number=1,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='INACTIVE', index=2, number=2,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='UNKNOWN', index=3, number=3,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='SHARING_DISABLED', index=4, number=4,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
  ],
  containing_type=None,
  serialized_options=None,
  serialized_start=42628,
  serialized_end=42723,
)
_sym_db.RegisterEnumDescriptor(_PRESENCE)

Presence = enum_type_wrapper.EnumTypeWrapper(_PRESENCE)
_DNDSTATE_STATE = _descriptor.EnumDescriptor(
  name='DndState_State',
  full_name='DndState_State',
  filename=None,
  file=DESCRIPTOR,
  create_key=_descriptor._internal_create_key,
  values=[
    _descriptor.EnumValueDescriptor(
      name='DND_STATE_STATE_UNKNOWN', index=0, number=0,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='AVAILABLE', index=1, number=1,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='DND', index=2, number=2,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
  ],
  containing_type=None,
  serialized_options=None,
  serialized_start=42725,
  serialized_end=42794,
)
_sym_db.RegisterEnumDescriptor(_DNDSTATE_STATE)

DndState_State = enum_type_wrapper.EnumTypeWrapper(_DNDSTATE_STATE)
_TYPINGSTATE = _descriptor.EnumDescriptor(
  name='TypingState',
  full_name='TypingState',
  filename=None,
  file=DESCRIPTOR,
  create_key=_descriptor._internal_create_key,
  values=[
    _descriptor.EnumValueDescriptor(
      name='UNKNOWN_STATE', index=0, number=0,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='TYPING', index=1, number=1,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='STOPPED', index=2, number=2,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
  ],
  containing_type=None,
  serialized_options=None,
  serialized_start=42796,
  serialized_end=42853,
)
_sym_db.RegisterEnumDescriptor(_TYPINGSTATE)

TypingState = enum_type_wrapper.EnumTypeWrapper(_TYPINGSTATE)
_GROUPUNSUPPORTEDREASON = _descriptor.EnumDescriptor(
  name='GroupUnsupportedReason',
  full_name='GroupUnsupportedReason',
  filename=None,
  file=DESCRIPTOR,
  create_key=_descriptor._internal_create_key,
  values=[
    _descriptor.EnumValueDescriptor(
      name='GROUP_UNSUPPORTED_REASON_UNSPECIFIED', index=0, number=0,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='GROUP_DISABLED_ON_CLIENT', index=1, number=1,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='GROUP_DISABLED_ON_SERVER', index=2, number=2,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
  ],
  containing_type=None,
  serialized_options=None,
  serialized_start=42855,
  serialized_end=42981,
)
_sym_db.RegisterEnumDescriptor(_GROUPUNSUPPORTEDREASON)

GroupUnsupportedReason = enum_type_wrapper.EnumTypeWrapper(_GROUPUNSUPPORTEDREASON)
_GROUPSUPPORTLEVEL = _descriptor.EnumDescriptor(
  name='GroupSupportLevel',
  full_name='GroupSupportLevel',
  filename=None,
  file=DESCRIPTOR,
  create_key=_descriptor._internal_create_key,
  values=[
    _descriptor.EnumValueDescriptor(
      name='UNSUPPORTED', index=0, number=0,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='DATA_SUPPORTED', index=1, number=1,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='FULLY_SUPPORTED', index=2, number=2,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
  ],
  containing_type=None,
  serialized_options=None,
  serialized_start=42983,
  serialized_end=43060,
)
_sym_db.RegisterEnumDescriptor(_GROUPSUPPORTLEVEL)

GroupSupportLevel = enum_type_wrapper.EnumTypeWrapper(_GROUPSUPPORTLEVEL)
_NOTIFICATIONCAUSE = _descriptor.EnumDescriptor(
  name='NotificationCause',
  full_name='NotificationCause',
  filename=None,
  file=DESCRIPTOR,
  create_key=_descriptor._internal_create_key,
  values=[
    _descriptor.EnumValueDescriptor(
      name='UNKNOWN_NOTIFICATION_CAUSE', index=0, number=0,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='DIRECT_MESSAGE', index=1, number=1,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='AT_MENTION', index=2, number=2,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='NEW_TOPIC', index=3, number=3,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='SUBSCRIBED_TOPIC_REPLY', index=4, number=4,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='UNSUBSCRIBED_TOPIC_REPLY', index=5, number=5,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
  ],
  containing_type=None,
  serialized_options=None,
  serialized_start=43063,
  serialized_end=43223,
)
_sym_db.RegisterEnumDescriptor(_NOTIFICATIONCAUSE)

NotificationCause = enum_type_wrapper.EnumTypeWrapper(_NOTIFICATIONCAUSE)
_EVENTORIGIN = _descriptor.EnumDescriptor(
  name='EventOrigin',
  full_name='EventOrigin',
  filename=None,
  file=DESCRIPTOR,
  create_key=_descriptor._internal_create_key,
  values=[
    _descriptor.EnumValueDescriptor(
      name='ORIGIN_UNSPECIFIED', index=0, number=0,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='ANDROID', index=1, number=100,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='ANDROID_DEV', index=2, number=101,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='ANDROID_PRODTEST', index=3, number=102,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='ANDROID_STAGING', index=4, number=103,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='ANDROID_PROD', index=5, number=104,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='IOS', index=6, number=110,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='IOS_DEV', index=7, number=111,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='IOS_PRODTEST', index=8, number=112,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='IOS_STAGING', index=9, number=113,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='IOS_PROD', index=10, number=114,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='WEB_ORIGIN', index=11, number=120,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='WEB_DEV', index=12, number=121,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='WEB_PRODTEST', index=13, number=122,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='WEB_PROD', index=14, number=123,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='WEB_STAGING', index=15, number=124,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='DESKTOP_ORIGIN', index=16, number=125,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='DESKTOP_DEV', index=17, number=126,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='DESKTOP_PRODTEST', index=18, number=127,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='DESKTOP_STAGING', index=19, number=128,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='DESKTOP_PROD', index=20, number=129,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='COMPANION_ORIGIN', index=21, number=130,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='COMPANION_DEV', index=22, number=131,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='COMPANION_PRODTEST', index=23, number=132,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='COMPANION_STAGING', index=24, number=133,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='COMPANION_PROD', index=25, number=134,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='WEB_GMAIL_ORIGIN', index=26, number=135,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='WEB_GMAIL_DEV', index=27, number=136,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='WEB_GMAIL_PRODTEST', index=28, number=137,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='WEB_GMAIL_STAGING', index=29, number=138,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='WEB_GMAIL_PROD', index=30, number=139,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='WEB_PWA_ORIGIN', index=31, number=140,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='WEB_PWA_DEV', index=32, number=141,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='WEB_PWA_PRODTEST', index=33, number=142,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='WEB_PWA_STAGING', index=34, number=143,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='WEB_PWA_PROD', index=35, number=144,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
  ],
  containing_type=None,
  serialized_options=None,
  serialized_start=43226,
  serialized_end=43932,
)
_sym_db.RegisterEnumDescriptor(_EVENTORIGIN)

EventOrigin = enum_type_wrapper.EnumTypeWrapper(_EVENTORIGIN)
_PLATFORM = _descriptor.EnumDescriptor(
  name='Platform',
  full_name='Platform',
  filename=None,
  file=DESCRIPTOR,
  create_key=_descriptor._internal_create_key,
  values=[
    _descriptor.EnumValueDescriptor(
      name='UNDEFINED_PLATFORM', index=0, number=0,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='WEB', index=1, number=1,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='MOBILE', index=2, number=2,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='DRONE', index=3, number=3,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='INTEROP_PRESENCE', index=4, number=4,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='CLASSIC_DESKTOP', index=5, number=5,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='CLASSIC_PHONE', index=6, number=6,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='CLASSIC_UNKNOWN', index=7, number=7,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='WEB_GMAIL', index=8, number=8,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
  ],
  containing_type=None,
  serialized_options=None,
  serialized_start=43935,
  serialized_end=44099,
)
_sym_db.RegisterEnumDescriptor(_PLATFORM)

Platform = enum_type_wrapper.EnumTypeWrapper(_PLATFORM)
_ANNOTATIONTYPE = _descriptor.EnumDescriptor(
  name='AnnotationType',
  full_name='AnnotationType',
  filename=None,
  file=DESCRIPTOR,
  create_key=_descriptor._internal_create_key,
  values=[
    _descriptor.EnumValueDescriptor(
      name='ANNOTATION_TYPE_UNKNOWN', index=0, number=0,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='URL', index=1, number=1,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='DRIVE_FILE', index=2, number=2,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='DRIVE_DOC', index=3, number=3,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='DRIVE_SHEET', index=4, number=4,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='DRIVE_SLIDE', index=5, number=5,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='USER_MENTION', index=6, number=6,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='VIDEO', index=7, number=7,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='FORMAT_DATA', index=8, number=8,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='IMAGE', index=9, number=9,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='PDF', index=10, number=10,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='VIDEO_CALL', index=11, number=11,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='MEMBERSHIP_CHANGED', index=12, number=12,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='UPLOAD_METADATA', index=13, number=13,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='ROOM_UPDATED', index=14, number=14,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='INVITATION', index=15, number=15,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='SLASH_COMMAND', index=16, number=16,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='GSUITE_INTEGRATION', index=17, number=17,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='DRIVE_FORM', index=18, number=18,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='GROUP_RETENTION_SETTINGS_UPDATED', index=19, number=19,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='BABEL_PLACEHOLDER', index=20, number=20,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='READ_RECEIPTS_SETTINGS_UPDATED', index=21, number=21,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='INCOMING_WEBHOOK_CHANGED', index=22, number=22,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='INTEGRATION_CONFIG_UPDATED', index=23, number=23,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='CONSENTED_APP_UNFURL', index=24, number=24,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
  ],
  containing_type=None,
  serialized_options=None,
  serialized_start=44102,
  serialized_end=44626,
)
_sym_db.RegisterEnumDescriptor(_ANNOTATIONTYPE)

AnnotationType = enum_type_wrapper.EnumTypeWrapper(_ANNOTATIONTYPE)
_DIALINNUMBERCLASS = _descriptor.EnumDescriptor(
  name='DialInNumberClass',
  full_name='DialInNumberClass',
  filename=None,
  file=DESCRIPTOR,
  create_key=_descriptor._internal_create_key,
  values=[
    _descriptor.EnumValueDescriptor(
      name='NUMBER_CLASS_UNSPECIFIED', index=0, number=0,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='LOW_COST', index=1, number=1,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='HIGH_COST', index=2, number=2,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='LEGACY', index=3, number=3,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
  ],
  containing_type=None,
  serialized_options=None,
  serialized_start=44628,
  serialized_end=44718,
)
_sym_db.RegisterEnumDescriptor(_DIALINNUMBERCLASS)

DialInNumberClass = enum_type_wrapper.EnumTypeWrapper(_DIALINNUMBERCLASS)
_RECORDINGAPPLICATIONTYPE = _descriptor.EnumDescriptor(
  name='RecordingApplicationType',
  full_name='RecordingApplicationType',
  filename=None,
  file=DESCRIPTOR,
  create_key=_descriptor._internal_create_key,
  values=[
    _descriptor.EnumValueDescriptor(
      name='RECORDING_APPLICATION_TYPE_UNSPECIFIED', index=0, number=0,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='RECORDING', index=1, number=1,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='GLIVE_STREAM', index=2, number=3,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='BROADCAST', index=3, number=4,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
  ],
  containing_type=None,
  serialized_options=None,
  serialized_start=44720,
  serialized_end=44838,
)
_sym_db.RegisterEnumDescriptor(_RECORDINGAPPLICATIONTYPE)

RecordingApplicationType = enum_type_wrapper.EnumTypeWrapper(_RECORDINGAPPLICATIONTYPE)
_BROADCASTACCESSPOLICY = _descriptor.EnumDescriptor(
  name='BroadcastAccessPolicy',
  full_name='BroadcastAccessPolicy',
  filename=None,
  file=DESCRIPTOR,
  create_key=_descriptor._internal_create_key,
  values=[
    _descriptor.EnumValueDescriptor(
      name='BROADCASTING_ACCESS_POLICY_UNSPECIFIED', index=0, number=0,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='ORGANIZATION', index=1, number=1,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='PUBLIC', index=2, number=2,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
  ],
  containing_type=None,
  serialized_options=None,
  serialized_start=44840,
  serialized_end=44937,
)
_sym_db.RegisterEnumDescriptor(_BROADCASTACCESSPOLICY)

BroadcastAccessPolicy = enum_type_wrapper.EnumTypeWrapper(_BROADCASTACCESSPOLICY)
_CALLSTATUS = _descriptor.EnumDescriptor(
  name='CallStatus',
  full_name='CallStatus',
  filename=None,
  file=DESCRIPTOR,
  create_key=_descriptor._internal_create_key,
  values=[
    _descriptor.EnumValueDescriptor(
      name='UNKNOWN_CALL_STATUS', index=0, number=0,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='ACTIVE_CALL', index=1, number=1,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='ENDED_CALL', index=2, number=2,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
  ],
  containing_type=None,
  serialized_options=None,
  serialized_start=44939,
  serialized_end=45009,
)
_sym_db.RegisterEnumDescriptor(_CALLSTATUS)

CallStatus = enum_type_wrapper.EnumTypeWrapper(_CALLSTATUS)
_RINGSTATUS = _descriptor.EnumDescriptor(
  name='RingStatus',
  full_name='RingStatus',
  filename=None,
  file=DESCRIPTOR,
  create_key=_descriptor._internal_create_key,
  values=[
    _descriptor.EnumValueDescriptor(
      name='UNKNOWN_RING_STATUS', index=0, number=0,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='RINGING', index=1, number=1,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='JOINED', index=2, number=2,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='MISSED', index=3, number=3,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
  ],
  containing_type=None,
  serialized_options=None,
  serialized_start=45011,
  serialized_end=45085,
)
_sym_db.RegisterEnumDescriptor(_RINGSTATUS)

RingStatus = enum_type_wrapper.EnumTypeWrapper(_RINGSTATUS)
_USERACCOUNTSTATE = _descriptor.EnumDescriptor(
  name='UserAccountState',
  full_name='UserAccountState',
  filename=None,
  file=DESCRIPTOR,
  create_key=_descriptor._internal_create_key,
  values=[
    _descriptor.EnumValueDescriptor(
      name='USER_ACCOUNT_STATE_UNKNOWN', index=0, number=0,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='ENABLED', index=1, number=1,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='DISABLED', index=2, number=2,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='DELETED', index=3, number=3,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='TEMPORARY_UNAVAILABLE', index=4, number=4,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
  ],
  containing_type=None,
  serialized_options=None,
  serialized_start=45087,
  serialized_end=45204,
)
_sym_db.RegisterEnumDescriptor(_USERACCOUNTSTATE)

UserAccountState = enum_type_wrapper.EnumTypeWrapper(_USERACCOUNTSTATE)
_MEMBERSHIPROLE = _descriptor.EnumDescriptor(
  name='MembershipRole',
  full_name='MembershipRole',
  filename=None,
  file=DESCRIPTOR,
  create_key=_descriptor._internal_create_key,
  values=[
    _descriptor.EnumValueDescriptor(
      name='ROLE_UNKNOWN', index=0, number=0,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='ROLE_NONE', index=1, number=1,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='ROLE_INVITEE', index=2, number=2,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='ROLE_MEMBER', index=3, number=3,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='ROLE_OWNER', index=4, number=4,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
  ],
  containing_type=None,
  serialized_options=None,
  serialized_start=45206,
  serialized_end=45306,
)
_sym_db.RegisterEnumDescriptor(_MEMBERSHIPROLE)

MembershipRole = enum_type_wrapper.EnumTypeWrapper(_MEMBERSHIPROLE)
_APPTYPE = _descriptor.EnumDescriptor(
  name='AppType',
  full_name='AppType',
  filename=None,
  file=DESCRIPTOR,
  create_key=_descriptor._internal_create_key,
  values=[
    _descriptor.EnumValueDescriptor(
      name='APP_TYPE_UNSPECIFIED', index=0, number=0,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='APP', index=1, number=1,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='GSUITE_APP', index=2, number=2,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='INCOMING_WEBHOOK', index=3, number=3,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
  ],
  containing_type=None,
  serialized_options=None,
  serialized_start=45308,
  serialized_end=45390,
)
_sym_db.RegisterEnumDescriptor(_APPTYPE)

AppType = enum_type_wrapper.EnumTypeWrapper(_APPTYPE)
_MEMBERSHIPSTATE = _descriptor.EnumDescriptor(
  name='MembershipState',
  full_name='MembershipState',
  filename=None,
  file=DESCRIPTOR,
  create_key=_descriptor._internal_create_key,
  values=[
    _descriptor.EnumValueDescriptor(
      name='MEMBER_UNKNOWN', index=0, number=0,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='MEMBER_INVITED', index=1, number=1,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='MEMBER_JOINED', index=2, number=2,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='MEMBER_NOT_A_MEMBER', index=3, number=3,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='MEMBER_FAILED', index=4, number=4,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
  ],
  containing_type=None,
  serialized_options=None,
  serialized_start=45392,
  serialized_end=45512,
)
_sym_db.RegisterEnumDescriptor(_MEMBERSHIPSTATE)

MembershipState = enum_type_wrapper.EnumTypeWrapper(_MEMBERSHIPSTATE)
_SHAREDATTRIBUTECHECKERGROUPTYPE = _descriptor.EnumDescriptor(
  name='SharedAttributeCheckerGroupType',
  full_name='SharedAttributeCheckerGroupType',
  filename=None,
  file=DESCRIPTOR,
  create_key=_descriptor._internal_create_key,
  values=[
    _descriptor.EnumValueDescriptor(
      name='ATTRIBUTE_CHECKER_GROUP_TYPE_UNSPECIFIED', index=0, number=0,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='ONE_TO_ONE_HUMAN_DM', index=1, number=1,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='ONE_TO_ONE_BOT_DM', index=2, number=2,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='IMMUTABLE_MEMBERSHIP_GROUP_DM', index=3, number=3,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='FLAT_ROOM', index=4, number=4,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='THREADED_ROOM', index=5, number=5,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='IMMUTABLE_MEMBERSHIP_HUMAN_DM', index=6, number=6,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='POST_ROOM', index=7, number=7,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
  ],
  containing_type=None,
  serialized_options=None,
  serialized_start=45515,
  serialized_end=45761,
)
_sym_db.RegisterEnumDescriptor(_SHAREDATTRIBUTECHECKERGROUPTYPE)

SharedAttributeCheckerGroupType = enum_type_wrapper.EnumTypeWrapper(_SHAREDATTRIBUTECHECKERGROUPTYPE)
HUMAN = 0
BOT = 1
UNKNOWN_INVITE = 0
REGULAR_INVITE = 1
SPAM_INVITE = 2
UNDEFINED_PRESENCE = 0
ACTIVE = 1
INACTIVE = 2
UNKNOWN = 3
SHARING_DISABLED = 4
DND_STATE_STATE_UNKNOWN = 0
AVAILABLE = 1
DND = 2
UNKNOWN_STATE = 0
TYPING = 1
STOPPED = 2
GROUP_UNSUPPORTED_REASON_UNSPECIFIED = 0
GROUP_DISABLED_ON_CLIENT = 1
GROUP_DISABLED_ON_SERVER = 2
UNSUPPORTED = 0
DATA_SUPPORTED = 1
FULLY_SUPPORTED = 2
UNKNOWN_NOTIFICATION_CAUSE = 0
DIRECT_MESSAGE = 1
AT_MENTION = 2
NEW_TOPIC = 3
SUBSCRIBED_TOPIC_REPLY = 4
UNSUBSCRIBED_TOPIC_REPLY = 5
ORIGIN_UNSPECIFIED = 0
ANDROID = 100
ANDROID_DEV = 101
ANDROID_PRODTEST = 102
ANDROID_STAGING = 103
ANDROID_PROD = 104
IOS = 110
IOS_DEV = 111
IOS_PRODTEST = 112
IOS_STAGING = 113
IOS_PROD = 114
WEB_ORIGIN = 120
WEB_DEV = 121
WEB_PRODTEST = 122
WEB_PROD = 123
WEB_STAGING = 124
DESKTOP_ORIGIN = 125
DESKTOP_DEV = 126
DESKTOP_PRODTEST = 127
DESKTOP_STAGING = 128
DESKTOP_PROD = 129
COMPANION_ORIGIN = 130
COMPANION_DEV = 131
COMPANION_PRODTEST = 132
COMPANION_STAGING = 133
COMPANION_PROD = 134
WEB_GMAIL_ORIGIN = 135
WEB_GMAIL_DEV = 136
WEB_GMAIL_PRODTEST = 137
WEB_GMAIL_STAGING = 138
WEB_GMAIL_PROD = 139
WEB_PWA_ORIGIN = 140
WEB_PWA_DEV = 141
WEB_PWA_PRODTEST = 142
WEB_PWA_STAGING = 143
WEB_PWA_PROD = 144
UNDEFINED_PLATFORM = 0
WEB = 1
MOBILE = 2
DRONE = 3
INTEROP_PRESENCE = 4
CLASSIC_DESKTOP = 5
CLASSIC_PHONE = 6
CLASSIC_UNKNOWN = 7
WEB_GMAIL = 8
ANNOTATION_TYPE_UNKNOWN = 0
URL = 1
DRIVE_FILE = 2
DRIVE_DOC = 3
DRIVE_SHEET = 4
DRIVE_SLIDE = 5
USER_MENTION = 6
VIDEO = 7
FORMAT_DATA = 8
IMAGE = 9
PDF = 10
VIDEO_CALL = 11
MEMBERSHIP_CHANGED = 12
UPLOAD_METADATA = 13
ROOM_UPDATED = 14
INVITATION = 15
SLASH_COMMAND = 16
GSUITE_INTEGRATION = 17
DRIVE_FORM = 18
GROUP_RETENTION_SETTINGS_UPDATED = 19
BABEL_PLACEHOLDER = 20
READ_RECEIPTS_SETTINGS_UPDATED = 21
INCOMING_WEBHOOK_CHANGED = 22
INTEGRATION_CONFIG_UPDATED = 23
CONSENTED_APP_UNFURL = 24
NUMBER_CLASS_UNSPECIFIED = 0
LOW_COST = 1
HIGH_COST = 2
LEGACY = 3
RECORDING_APPLICATION_TYPE_UNSPECIFIED = 0
RECORDING = 1
GLIVE_STREAM = 3
BROADCAST = 4
BROADCASTING_ACCESS_POLICY_UNSPECIFIED = 0
ORGANIZATION = 1
PUBLIC = 2
UNKNOWN_CALL_STATUS = 0
ACTIVE_CALL = 1
ENDED_CALL = 2
UNKNOWN_RING_STATUS = 0
RINGING = 1
JOINED = 2
MISSED = 3
USER_ACCOUNT_STATE_UNKNOWN = 0
ENABLED = 1
DISABLED = 2
DELETED = 3
TEMPORARY_UNAVAILABLE = 4
ROLE_UNKNOWN = 0
ROLE_NONE = 1
ROLE_INVITEE = 2
ROLE_MEMBER = 3
ROLE_OWNER = 4
APP_TYPE_UNSPECIFIED = 0
APP = 1
GSUITE_APP = 2
INCOMING_WEBHOOK = 3
MEMBER_UNKNOWN = 0
MEMBER_INVITED = 1
MEMBER_JOINED = 2
MEMBER_NOT_A_MEMBER = 3
MEMBER_FAILED = 4
ATTRIBUTE_CHECKER_GROUP_TYPE_UNSPECIFIED = 0
ONE_TO_ONE_HUMAN_DM = 1
ONE_TO_ONE_BOT_DM = 2
IMMUTABLE_MEMBERSHIP_GROUP_DM = 3
FLAT_ROOM = 4
THREADED_ROOM = 5
IMMUTABLE_MEMBERSHIP_HUMAN_DM = 6
POST_ROOM = 7


_DNDSETTINGS_DNDSTATE_STATE = _descriptor.EnumDescriptor(
  name='DndState_State',
  full_name='DndSettings.DndState_State',
  filename=None,
  file=DESCRIPTOR,
  create_key=_descriptor._internal_create_key,
  values=[
    _descriptor.EnumValueDescriptor(
      name='UNKNOWN', index=0, number=0,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='AVAILABLE', index=1, number=1,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='DND', index=2, number=2,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
  ],
  containing_type=None,
  serialized_options=None,
  serialized_start=527,
  serialized_end=580,
)
_sym_db.RegisterEnumDescriptor(_DNDSETTINGS_DNDSTATE_STATE)

_CLIENTFEATURECAPABILITIES_CAPABILITYLEVEL = _descriptor.EnumDescriptor(
  name='CapabilityLevel',
  full_name='ClientFeatureCapabilities.CapabilityLevel',
  filename=None,
  file=DESCRIPTOR,
  create_key=_descriptor._internal_create_key,
  values=[
    _descriptor.EnumValueDescriptor(
      name='UNSUPPORTED', index=0, number=0,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='DATA_SUPPORTED', index=1, number=1,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='FULLY_SUPPORTED', index=2, number=2,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
  ],
  containing_type=None,
  serialized_options=None,
  serialized_start=1904,
  serialized_end=1979,
)
_sym_db.RegisterEnumDescriptor(_CLIENTFEATURECAPABILITIES_CAPABILITYLEVEL)

_REQUESTHEADER_CLIENTTYPE = _descriptor.EnumDescriptor(
  name='ClientType',
  full_name='RequestHeader.ClientType',
  filename=None,
  file=DESCRIPTOR,
  create_key=_descriptor._internal_create_key,
  values=[
    _descriptor.EnumValueDescriptor(
      name='UNKNOWN', index=0, number=0,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='ANDROID', index=1, number=1,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='IOS', index=2, number=2,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='WEB', index=3, number=3,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='CLASSIC_INTEROP', index=4, number=4,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='INTERNAL_TESTING', index=5, number=5,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='WEB_DESKTOP', index=6, number=6,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='WEB_GMAIL', index=7, number=7,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
  ],
  containing_type=None,
  serialized_options=None,
  serialized_start=2171,
  serialized_end=2302,
)
_sym_db.RegisterEnumDescriptor(_REQUESTHEADER_CLIENTTYPE)

_JADDONSFORMATTEDTEXT_FORMATTEDTEXTELEMENT_STYLEDTEXT_STYLE = _descriptor.EnumDescriptor(
  name='Style',
  full_name='JAddOnsFormattedText.FormattedTextElement.StyledText.Style',
  filename=None,
  file=DESCRIPTOR,
  create_key=_descriptor._internal_create_key,
  values=[
    _descriptor.EnumValueDescriptor(
      name='NONE', index=0, number=0,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='BOLD_DEPRECATED', index=1, number=1,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='ITALIC', index=2, number=2,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='UNDERLINE', index=3, number=3,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='STRIKETHROUGH', index=4, number=4,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='BR', index=5, number=5,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='UPPERCASE', index=6, number=6,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
  ],
  containing_type=None,
  serialized_options=None,
  serialized_start=4187,
  serialized_end=4294,
)
_sym_db.RegisterEnumDescriptor(_JADDONSFORMATTEDTEXT_FORMATTEDTEXTELEMENT_STYLEDTEXT_STYLE)

_JADDONSFORMATTEDTEXT_FORMATTEDTEXTELEMENT_STYLEDTEXT_FONTWEIGHT = _descriptor.EnumDescriptor(
  name='FontWeight',
  full_name='JAddOnsFormattedText.FormattedTextElement.StyledText.FontWeight',
  filename=None,
  file=DESCRIPTOR,
  create_key=_descriptor._internal_create_key,
  values=[
    _descriptor.EnumValueDescriptor(
      name='REGULAR', index=0, number=0,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='LIGHT', index=1, number=1,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='MEDIUM', index=2, number=2,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='BOLD', index=3, number=3,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
  ],
  containing_type=None,
  serialized_options=None,
  serialized_start=4296,
  serialized_end=4354,
)
_sym_db.RegisterEnumDescriptor(_JADDONSFORMATTEDTEXT_FORMATTEDTEXTELEMENT_STYLEDTEXT_FONTWEIGHT)

_JADDONSFORMATTEDTEXT_TEXTALIGN = _descriptor.EnumDescriptor(
  name='TextAlign',
  full_name='JAddOnsFormattedText.TextAlign',
  filename=None,
  file=DESCRIPTOR,
  create_key=_descriptor._internal_create_key,
  values=[
    _descriptor.EnumValueDescriptor(
      name='LEFT', index=0, number=0,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='CENTER', index=1, number=1,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='RIGHT', index=2, number=2,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
  ],
  containing_type=None,
  serialized_options=None,
  serialized_start=4420,
  serialized_end=4464,
)
_sym_db.RegisterEnumDescriptor(_JADDONSFORMATTEDTEXT_TEXTALIGN)

_JADDONSIMAGECROPSTYLE_IMAGECROPTYPE = _descriptor.EnumDescriptor(
  name='ImageCropType',
  full_name='JAddOnsImageCropStyle.ImageCropType',
  filename=None,
  file=DESCRIPTOR,
  create_key=_descriptor._internal_create_key,
  values=[
    _descriptor.EnumValueDescriptor(
      name='CROP_TYPE_NOT_SET', index=0, number=0,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='SQUARE', index=1, number=1,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='CIRCLE', index=2, number=2,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='RECTANGLE_CUSTOM', index=3, number=3,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='RECTANGLE_4_3', index=4, number=4,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
  ],
  containing_type=None,
  serialized_options=None,
  serialized_start=4566,
  serialized_end=4669,
)
_sym_db.RegisterEnumDescriptor(_JADDONSIMAGECROPSTYLE_IMAGECROPTYPE)

_JADDONSOPENLINK_OPENAS = _descriptor.EnumDescriptor(
  name='OpenAs',
  full_name='JAddOnsOpenLink.OpenAs',
  filename=None,
  file=DESCRIPTOR,
  create_key=_descriptor._internal_create_key,
  values=[
    _descriptor.EnumValueDescriptor(
      name='FULL_SIZE', index=0, number=0,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='OVERLAY', index=1, number=1,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
  ],
  containing_type=None,
  serialized_options=None,
  serialized_start=4944,
  serialized_end=4980,
)
_sym_db.RegisterEnumDescriptor(_JADDONSOPENLINK_OPENAS)

_JADDONSOPENLINK_ONCLOSE = _descriptor.EnumDescriptor(
  name='OnClose',
  full_name='JAddOnsOpenLink.OnClose',
  filename=None,
  file=DESCRIPTOR,
  create_key=_descriptor._internal_create_key,
  values=[
    _descriptor.EnumValueDescriptor(
      name='NOTHING', index=0, number=0,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='RELOAD_ADD_ON', index=1, number=1,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
  ],
  containing_type=None,
  serialized_options=None,
  serialized_start=4982,
  serialized_end=5023,
)
_sym_db.RegisterEnumDescriptor(_JADDONSOPENLINK_ONCLOSE)

_JADDONSOPENLINK_LOADINDICATOR = _descriptor.EnumDescriptor(
  name='LoadIndicator',
  full_name='JAddOnsOpenLink.LoadIndicator',
  filename=None,
  file=DESCRIPTOR,
  create_key=_descriptor._internal_create_key,
  values=[
    _descriptor.EnumValueDescriptor(
      name='NONE', index=0, number=0,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='SPINNER', index=1, number=1,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
  ],
  containing_type=None,
  serialized_options=None,
  serialized_start=5025,
  serialized_end=5063,
)
_sym_db.RegisterEnumDescriptor(_JADDONSOPENLINK_LOADINDICATOR)

_JADDONSCARDITEM_DISPLAYSTYLE = _descriptor.EnumDescriptor(
  name='DisplayStyle',
  full_name='JAddOnsCardItem.DisplayStyle',
  filename=None,
  file=DESCRIPTOR,
  create_key=_descriptor._internal_create_key,
  values=[
    _descriptor.EnumValueDescriptor(
      name='DISPLAY_STYLE_UNSPECIFIED', index=0, number=0,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='PEEK', index=1, number=1,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='REPLACE', index=2, number=2,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
  ],
  containing_type=None,
  serialized_options=None,
  serialized_start=6251,
  serialized_end=6319,
)
_sym_db.RegisterEnumDescriptor(_JADDONSCARDITEM_DISPLAYSTYLE)

_JADDONSGRID_GRIDITEM_GRIDITEMLAYOUT = _descriptor.EnumDescriptor(
  name='GridItemLayout',
  full_name='JAddOnsGrid.GridItem.GridItemLayout',
  filename=None,
  file=DESCRIPTOR,
  create_key=_descriptor._internal_create_key,
  values=[
    _descriptor.EnumValueDescriptor(
      name='NOT_SET', index=0, number=0,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='TEXT_BELOW', index=1, number=1,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='TEXT_ABOVE', index=2, number=2,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
  ],
  containing_type=None,
  serialized_options=None,
  serialized_start=7196,
  serialized_end=7257,
)
_sym_db.RegisterEnumDescriptor(_JADDONSGRID_GRIDITEM_GRIDITEMLAYOUT)

_JADDONSBORDERSTYLE_BORDERTYPE = _descriptor.EnumDescriptor(
  name='BorderType',
  full_name='JAddOnsBorderStyle.BorderType',
  filename=None,
  file=DESCRIPTOR,
  create_key=_descriptor._internal_create_key,
  values=[
    _descriptor.EnumValueDescriptor(
      name='BORDER_TYPE_NOT_SET', index=0, number=0,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='NO_BORDER', index=1, number=1,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='STROKE', index=2, number=2,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
  ],
  containing_type=None,
  serialized_options=None,
  serialized_start=7394,
  serialized_end=7458,
)
_sym_db.RegisterEnumDescriptor(_JADDONSBORDERSTYLE_BORDERTYPE)

_JADDONSWIDGET_KEYVALUE_SWITCHWIDGET_CONTROLTYPE = _descriptor.EnumDescriptor(
  name='ControlType',
  full_name='JAddOnsWidget.KeyValue.SwitchWidget.ControlType',
  filename=None,
  file=DESCRIPTOR,
  create_key=_descriptor._internal_create_key,
  values=[
    _descriptor.EnumValueDescriptor(
      name='UNSPECIFIED', index=0, number=0,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='SWITCH', index=1, number=1,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='CHECKBOX', index=2, number=2,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
  ],
  containing_type=None,
  serialized_options=None,
  serialized_start=9558,
  serialized_end=9614,
)
_sym_db.RegisterEnumDescriptor(_JADDONSWIDGET_KEYVALUE_SWITCHWIDGET_CONTROLTYPE)

_JADDONSWIDGET_TEXTFIELD_TEXTFIELDTYPE = _descriptor.EnumDescriptor(
  name='TextFieldType',
  full_name='JAddOnsWidget.TextField.TextFieldType',
  filename=None,
  file=DESCRIPTOR,
  create_key=_descriptor._internal_create_key,
  values=[
    _descriptor.EnumValueDescriptor(
      name='REGULAR', index=0, number=0,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='FLOATING', index=1, number=1,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
  ],
  containing_type=None,
  serialized_options=None,
  serialized_start=10341,
  serialized_end=10383,
)
_sym_db.RegisterEnumDescriptor(_JADDONSWIDGET_TEXTFIELD_TEXTFIELDTYPE)

_JADDONSWIDGET_TEXTFIELD_LINETYPE = _descriptor.EnumDescriptor(
  name='LineType',
  full_name='JAddOnsWidget.TextField.LineType',
  filename=None,
  file=DESCRIPTOR,
  create_key=_descriptor._internal_create_key,
  values=[
    _descriptor.EnumValueDescriptor(
      name='SINGLE', index=0, number=0,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='MULTIPLE', index=1, number=1,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
  ],
  containing_type=None,
  serialized_options=None,
  serialized_start=10385,
  serialized_end=10421,
)
_sym_db.RegisterEnumDescriptor(_JADDONSWIDGET_TEXTFIELD_LINETYPE)

_JADDONSWIDGET_SELECTIONCONTROL_SELECTIONTYPE = _descriptor.EnumDescriptor(
  name='SelectionType',
  full_name='JAddOnsWidget.SelectionControl.SelectionType',
  filename=None,
  file=DESCRIPTOR,
  create_key=_descriptor._internal_create_key,
  values=[
    _descriptor.EnumValueDescriptor(
      name='CHECK_BOX', index=0, number=0,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='RADIO_BUTTON', index=1, number=1,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='SWITCH', index=2, number=2,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='DROPDOWN', index=3, number=3,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
  ],
  containing_type=None,
  serialized_options=None,
  serialized_start=10763,
  serialized_end=10837,
)
_sym_db.RegisterEnumDescriptor(_JADDONSWIDGET_SELECTIONCONTROL_SELECTIONTYPE)

_JADDONSWIDGET_DATETIMEPICKER_DATETIMEPICKERTYPE = _descriptor.EnumDescriptor(
  name='DateTimePickerType',
  full_name='JAddOnsWidget.DateTimePicker.DateTimePickerType',
  filename=None,
  file=DESCRIPTOR,
  create_key=_descriptor._internal_create_key,
  values=[
    _descriptor.EnumValueDescriptor(
      name='UNSPECIFIED_TYPE', index=0, number=0,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='DATE_AND_TIME', index=1, number=1,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='DATE_ONLY', index=2, number=2,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='TIME_ONLY', index=3, number=3,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
  ],
  containing_type=None,
  serialized_options=None,
  serialized_start=11102,
  serialized_end=11193,
)
_sym_db.RegisterEnumDescriptor(_JADDONSWIDGET_DATETIMEPICKER_DATETIMEPICKERTYPE)

_JADDONSWIDGET_TEXTBUTTON_STYLE = _descriptor.EnumDescriptor(
  name='Style',
  full_name='JAddOnsWidget.TextButton.Style',
  filename=None,
  file=DESCRIPTOR,
  create_key=_descriptor._internal_create_key,
  values=[
    _descriptor.EnumValueDescriptor(
      name='UNSPECIFIED', index=0, number=0,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='TEXT', index=1, number=1,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='FILLED', index=2, number=2,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
  ],
  containing_type=None,
  serialized_options=None,
  serialized_start=11478,
  serialized_end=11524,
)
_sym_db.RegisterEnumDescriptor(_JADDONSWIDGET_TEXTBUTTON_STYLE)

_JADDONSWIDGET_HORIZONTALALIGN = _descriptor.EnumDescriptor(
  name='HorizontalAlign',
  full_name='JAddOnsWidget.HorizontalAlign',
  filename=None,
  file=DESCRIPTOR,
  create_key=_descriptor._internal_create_key,
  values=[
    _descriptor.EnumValueDescriptor(
      name='NOT_SET', index=0, number=0,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='START', index=1, number=1,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='CENTER', index=2, number=2,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='END', index=3, number=3,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
  ],
  containing_type=None,
  serialized_options=None,
  serialized_start=11748,
  serialized_end=11810,
)
_sym_db.RegisterEnumDescriptor(_JADDONSWIDGET_HORIZONTALALIGN)

_JADDONSFORMACTION_LOADINDICATOR = _descriptor.EnumDescriptor(
  name='LoadIndicator',
  full_name='JAddOnsFormAction.LoadIndicator',
  filename=None,
  file=DESCRIPTOR,
  create_key=_descriptor._internal_create_key,
  values=[
    _descriptor.EnumValueDescriptor(
      name='SPINNER', index=0, number=0,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='NONE', index=1, number=1,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
  ],
  containing_type=None,
  serialized_options=None,
  serialized_start=12104,
  serialized_end=12142,
)
_sym_db.RegisterEnumDescriptor(_JADDONSFORMACTION_LOADINDICATOR)

_MESSAGE_MESSAGESTATE = _descriptor.EnumDescriptor(
  name='MessageState',
  full_name='Message.MessageState',
  filename=None,
  file=DESCRIPTOR,
  create_key=_descriptor._internal_create_key,
  values=[
    _descriptor.EnumValueDescriptor(
      name='UNKNOWN', index=0, number=0,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='GROUP_VISIBLE', index=1, number=1,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='ON_HOLD', index=2, number=2,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
  ],
  containing_type=None,
  serialized_options=None,
  serialized_start=14498,
  serialized_end=14557,
)
_sym_db.RegisterEnumDescriptor(_MESSAGE_MESSAGESTATE)

_MESSAGE_MESSAGETYPE = _descriptor.EnumDescriptor(
  name='MessageType',
  full_name='Message.MessageType',
  filename=None,
  file=DESCRIPTOR,
  create_key=_descriptor._internal_create_key,
  values=[
    _descriptor.EnumValueDescriptor(
      name='MESSAGE_TYPE_UNSPECIFIED', index=0, number=0,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='USER_MESSAGE', index=1, number=1,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='SYSTEM_MESSAGE', index=2, number=2,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
  ],
  containing_type=None,
  serialized_options=None,
  serialized_start=14559,
  serialized_end=14640,
)
_sym_db.RegisterEnumDescriptor(_MESSAGE_MESSAGETYPE)

_FORMATMETADATA_FORMATTYPE = _descriptor.EnumDescriptor(
  name='FormatType',
  full_name='FormatMetadata.FormatType',
  filename=None,
  file=DESCRIPTOR,
  create_key=_descriptor._internal_create_key,
  values=[
    _descriptor.EnumValueDescriptor(
      name='TYPE_UNSPECIFIED', index=0, number=0,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='BOLD', index=1, number=1,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='ITALIC', index=2, number=2,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='STRIKE', index=3, number=3,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='SOURCE_CODE', index=4, number=4,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='MONOSPACE', index=5, number=5,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='HIDDEN', index=6, number=6,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='MONOSPACE_BLOCK', index=7, number=7,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='UNDERLINE', index=8, number=8,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='FONT_COLOR', index=9, number=9,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='BULLETED_LIST', index=10, number=10,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='BULLETED_LIST_ITEM', index=11, number=11,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='CLIENT_HIDDEN', index=12, number=12,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
  ],
  containing_type=None,
  serialized_options=None,
  serialized_start=15677,
  serialized_end=15903,
)
_sym_db.RegisterEnumDescriptor(_FORMATMETADATA_FORMATTYPE)

_ANNOTATION_CHIPRENDERTYPE = _descriptor.EnumDescriptor(
  name='ChipRenderType',
  full_name='Annotation.ChipRenderType',
  filename=None,
  file=DESCRIPTOR,
  create_key=_descriptor._internal_create_key,
  values=[
    _descriptor.EnumValueDescriptor(
      name='UNKNOWN', index=0, number=0,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='RENDER', index=1, number=1,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='RENDER_IF_POSSIBLE', index=2, number=2,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='DO_NOT_RENDER', index=3, number=3,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
  ],
  containing_type=None,
  serialized_options=None,
  serialized_start=16475,
  serialized_end=16559,
)
_sym_db.RegisterEnumDescriptor(_ANNOTATION_CHIPRENDERTYPE)

_GROUP_GROUPTYPE = _descriptor.EnumDescriptor(
  name='GroupType',
  full_name='Group.GroupType',
  filename=None,
  file=DESCRIPTOR,
  create_key=_descriptor._internal_create_key,
  values=[
    _descriptor.EnumValueDescriptor(
      name='GROUP_TYPE_UNSPECIFIED', index=0, number=0,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='ROOM', index=1, number=1,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='HUMAN_DM', index=2, number=2,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='BOT_DM', index=3, number=3,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
  ],
  containing_type=None,
  serialized_options=None,
  serialized_start=17777,
  serialized_end=17852,
)
_sym_db.RegisterEnumDescriptor(_GROUP_GROUPTYPE)

_GROUPNOTIFICATIONSETTINGS_GROUPNOTIFICATIONSTATE = _descriptor.EnumDescriptor(
  name='GroupNotificationState',
  full_name='GroupNotificationSettings.GroupNotificationState',
  filename=None,
  file=DESCRIPTOR,
  create_key=_descriptor._internal_create_key,
  values=[
    _descriptor.EnumValueDescriptor(
      name='UNKNOWN_NOTIFICATION_STATE', index=0, number=0,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='MUTED', index=1, number=1,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='UNMUTED', index=2, number=2,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
  ],
  containing_type=None,
  serialized_options=None,
  serialized_start=19026,
  serialized_end=19106,
)
_sym_db.RegisterEnumDescriptor(_GROUPNOTIFICATIONSETTINGS_GROUPNOTIFICATIONSTATE)

_GROUPNOTIFICATIONSETTINGS_ROOMNOTIFICATIONSTATE = _descriptor.EnumDescriptor(
  name='RoomNotificationState',
  full_name='GroupNotificationSettings.RoomNotificationState',
  filename=None,
  file=DESCRIPTOR,
  create_key=_descriptor._internal_create_key,
  values=[
    _descriptor.EnumValueDescriptor(
      name='NOTIFY_ALWAYS', index=0, number=0,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='NOTIFY_LESS_WITH_NEW_THREADS', index=1, number=1,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='NOTIFY_LESS', index=2, number=2,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='NOTIFY_NEVER', index=3, number=3,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
  ],
  containing_type=None,
  serialized_options=None,
  serialized_start=19108,
  serialized_end=19219,
)
_sym_db.RegisterEnumDescriptor(_GROUPNOTIFICATIONSETTINGS_ROOMNOTIFICATIONSTATE)

_RETENTIONSETTINGS_RETENTIONSTATE = _descriptor.EnumDescriptor(
  name='RetentionState',
  full_name='RetentionSettings.RetentionState',
  filename=None,
  file=DESCRIPTOR,
  create_key=_descriptor._internal_create_key,
  values=[
    _descriptor.EnumValueDescriptor(
      name='UNKNOWN_RETENTION_STATE', index=0, number=0,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='PERMANENT', index=1, number=1,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='EPHEMERAL_ONE_DAY', index=2, number=2,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
  ],
  containing_type=None,
  serialized_options=None,
  serialized_start=19319,
  serialized_end=19402,
)
_sym_db.RegisterEnumDescriptor(_RETENTIONSETTINGS_RETENTIONSTATE)

_CREATEDMREQUEST_FETCHOPTIONS = _descriptor.EnumDescriptor(
  name='FetchOptions',
  full_name='CreateDmRequest.FetchOptions',
  filename=None,
  file=DESCRIPTOR,
  create_key=_descriptor._internal_create_key,
  values=[
    _descriptor.EnumValueDescriptor(
      name='UNKNOWN', index=0, number=0,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='INCLUDE_BOT_MEMBERSHIPS_IN_RESPONSE', index=1, number=1,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
  ],
  containing_type=None,
  serialized_options=None,
  serialized_start=21377,
  serialized_end=21445,
)
_sym_db.RegisterEnumDescriptor(_CREATEDMREQUEST_FETCHOPTIONS)

_LISTTOPICSREQUEST_FETCHOPTIONS = _descriptor.EnumDescriptor(
  name='FetchOptions',
  full_name='ListTopicsRequest.FetchOptions',
  filename=None,
  file=DESCRIPTOR,
  create_key=_descriptor._internal_create_key,
  values=[
    _descriptor.EnumValueDescriptor(
      name='USER', index=0, number=1,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='TOTAL_MESSAGE_COUNTS', index=1, number=2,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='READ_RECEIPTS', index=2, number=3,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
  ],
  containing_type=None,
  serialized_options=None,
  serialized_start=22191,
  serialized_end=22260,
)
_sym_db.RegisterEnumDescriptor(_LISTTOPICSREQUEST_FETCHOPTIONS)

_WEBPUSHNOTIFICATIONEVENT_DISPATCHAPPROACHTYPE = _descriptor.EnumDescriptor(
  name='DispatchApproachType',
  full_name='WebPushNotificationEvent.DispatchApproachType',
  filename=None,
  file=DESCRIPTOR,
  create_key=_descriptor._internal_create_key,
  values=[
    _descriptor.EnumValueDescriptor(
      name='DISPATCH_APPROACH_TYPE_UNSPECIFIED', index=0, number=0,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='INTERACTIVE_SESSION', index=1, number=1,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='BROADCAST_TO_WEB_SESSIONS', index=2, number=2,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='INTERACTIVE_SESSIONS_INCLUDE_VISIBLE_WEB_UNDER_TTL', index=3, number=3,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
  ],
  containing_type=None,
  serialized_options=None,
  serialized_start=24441,
  serialized_end=24615,
)
_sym_db.RegisterEnumDescriptor(_WEBPUSHNOTIFICATIONEVENT_DISPATCHAPPROACHTYPE)

_WEBPUSHNOTIFICATIONEVENT_ENDPOINTTYPE = _descriptor.EnumDescriptor(
  name='EndpointType',
  full_name='WebPushNotificationEvent.EndpointType',
  filename=None,
  file=DESCRIPTOR,
  create_key=_descriptor._internal_create_key,
  values=[
    _descriptor.EnumValueDescriptor(
      name='ENDPOINT_TYPE_UNSPECIFIED', index=0, number=0,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='MOBILE_WEBCHANNEL', index=1, number=1,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='DESKTOP_WEBCHANNEL', index=2, number=2,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
  ],
  containing_type=None,
  serialized_options=None,
  serialized_start=24617,
  serialized_end=24709,
)
_sym_db.RegisterEnumDescriptor(_WEBPUSHNOTIFICATIONEVENT_ENDPOINTTYPE)

_CLIENTNOTIFICATIONSSTATE_DEVICENOTIFICATIONSETTINGSTATE = _descriptor.EnumDescriptor(
  name='DeviceNotificationSettingState',
  full_name='ClientNotificationsState.DeviceNotificationSettingState',
  filename=None,
  file=DESCRIPTOR,
  create_key=_descriptor._internal_create_key,
  values=[
    _descriptor.EnumValueDescriptor(
      name='DEVICE_NOTIFICATION_SETTING_STATE_UNKNOWN', index=0, number=0,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='DEVICE_NOTIFICATIONS_ENABLED', index=1, number=1,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='DEVICE_NOTIFICATIONS_DISABLED', index=2, number=2,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
  ],
  containing_type=None,
  serialized_options=None,
  serialized_start=25244,
  serialized_end=25392,
)
_sym_db.RegisterEnumDescriptor(_CLIENTNOTIFICATIONSSTATE_DEVICENOTIFICATIONSETTINGSTATE)

_PINGEVENT_STATE = _descriptor.EnumDescriptor(
  name='State',
  full_name='PingEvent.State',
  filename=None,
  file=DESCRIPTOR,
  create_key=_descriptor._internal_create_key,
  values=[
    _descriptor.EnumValueDescriptor(
      name='STATE_UNSPECIFIED', index=0, number=0,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='ACTIVE', index=1, number=1,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='INACTIVE', index=2, number=2,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
  ],
  containing_type=None,
  serialized_options=None,
  serialized_start=25792,
  serialized_end=25848,
)
_sym_db.RegisterEnumDescriptor(_PINGEVENT_STATE)

_PINGEVENT_APPLICATIONFOCUSSTATE = _descriptor.EnumDescriptor(
  name='ApplicationFocusState',
  full_name='PingEvent.ApplicationFocusState',
  filename=None,
  file=DESCRIPTOR,
  create_key=_descriptor._internal_create_key,
  values=[
    _descriptor.EnumValueDescriptor(
      name='FOCUS_STATE_UNSPECIFIED', index=0, number=0,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='FOCUS_STATE_FOREGROUND', index=1, number=1,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='FOCUS_STATE_BACKGROUND', index=2, number=2,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
  ],
  containing_type=None,
  serialized_options=None,
  serialized_start=25850,
  serialized_end=25958,
)
_sym_db.RegisterEnumDescriptor(_PINGEVENT_APPLICATIONFOCUSSTATE)

_PINGEVENT_CLIENTINTERACTIVESTATE = _descriptor.EnumDescriptor(
  name='ClientInteractiveState',
  full_name='PingEvent.ClientInteractiveState',
  filename=None,
  file=DESCRIPTOR,
  create_key=_descriptor._internal_create_key,
  values=[
    _descriptor.EnumValueDescriptor(
      name='UNKNOWN', index=0, number=0,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='INTERACTIVE', index=1, number=1,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='FOCUSED', index=2, number=2,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='VISIBLE', index=3, number=3,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='HIDDEN', index=4, number=4,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
  ],
  containing_type=None,
  serialized_options=None,
  serialized_start=25960,
  serialized_end=26052,
)
_sym_db.RegisterEnumDescriptor(_PINGEVENT_CLIENTINTERACTIVESTATE)

_PINGEVENT_DEVICEACTIVESTATE = _descriptor.EnumDescriptor(
  name='DeviceActiveState',
  full_name='PingEvent.DeviceActiveState',
  filename=None,
  file=DESCRIPTOR,
  create_key=_descriptor._internal_create_key,
  values=[
    _descriptor.EnumValueDescriptor(
      name='STATE_UNKNOWN', index=0, number=0,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='STATE_IDLE_LOCKED', index=1, number=1,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='STATE_IDLE_UNLOCKED', index=2, number=2,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='STATE_ACTIVE', index=3, number=3,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
  ],
  containing_type=None,
  serialized_options=None,
  serialized_start=26054,
  serialized_end=26158,
)
_sym_db.RegisterEnumDescriptor(_PINGEVENT_DEVICEACTIVESTATE)

_EVENT_EVENTTYPE = _descriptor.EnumDescriptor(
  name='EventType',
  full_name='Event.EventType',
  filename=None,
  file=DESCRIPTOR,
  create_key=_descriptor._internal_create_key,
  values=[
    _descriptor.EnumValueDescriptor(
      name='UNKNOWN', index=0, number=0,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='USER_ADDED_TO_GROUP', index=1, number=1,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='USER_REMOVED_FROM_GROUP', index=2, number=2,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='GROUP_VIEWED', index=3, number=3,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='TOPIC_VIEWED', index=4, number=4,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='GROUP_UPDATED', index=5, number=5,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='MESSAGE_POSTED', index=6, number=6,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='MESSAGE_UPDATED', index=7, number=7,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='MESSAGE_DELETED', index=8, number=8,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='TOPIC_MUTE_CHANGED', index=9, number=9,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='USER_SETTINGS_CHANGED', index=10, number=10,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='GROUP_STARRED', index=11, number=11,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='WEB_PUSH_NOTIFICATION', index=12, number=12,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='GROUP_UNREAD_SUBSCRIBED_TOPIC_COUNT_UPDATED', index=13, number=13,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='INVITE_COUNT_UPDATED', index=14, number=14,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='MEMBERSHIP_CHANGED', index=15, number=15,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='GROUP_HIDE_CHANGED', index=16, number=16,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='DRIVE_ACL_FIX_PROCESSED', index=17, number=17,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='GROUP_NOTIFICATION_SETTINGS_UPDATED', index=18, number=18,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='RETENTION_SETTINGS_UPDATED', index=19, number=19,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='TOPIC_CREATED', index=20, number=20,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='ON_HOLD_MESSAGE_POSTED', index=21, number=21,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='ON_HOLD_MESSAGE_UPDATED', index=22, number=22,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='ON_HOLD_MESSAGE_PUBLISHED', index=23, number=23,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='MESSAGE_REACTED', index=24, number=24,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='USER_STATUS_UPDATED_EVENT', index=25, number=25,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='GROUP_RETENTION_SETTINGS_UPDATED', index=26, number=26,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='USER_WORKING_HOURS_UPDATED_EVENT', index=27, number=27,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='MESSAGE_SMART_REPLIES', index=28, number=28,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='TYPING_STATE_CHANGED', index=29, number=29,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='GROUP_DELETED', index=30, number=30,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='BLOCK_STATE_CHANGED', index=31, number=31,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='CLEAR_HISTORY', index=32, number=32,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='SESSION_READY', index=33, number=33,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='GROUP_SORT_TIMESTAMP_CHANGED', index=34, number=34,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='GSUITE_INTEGRATION_UPDATED', index=35, number=35,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='READ_RECEIPT_CHANGED', index=36, number=36,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='MARK_AS_UNREAD', index=37, number=37,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='GROUP_NO_OP', index=38, number=38,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='INVALIDATE_GROUP_CACHE', index=39, number=39,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='USER_NO_OP', index=40, number=40,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='INVALIDATE_USER_CACHE', index=41, number=41,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='USER_DENORMALIZED_GROUP_UPDATED', index=42, number=42,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='USER_PRESENCE_SHARED_UPDATED_EVENT', index=43, number=43,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='NOTIFICATIONS_CARD_UPDATED', index=44, number=44,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='USER_HUB_AVAILABILITY_UPDATED_EVENT', index=45, number=45,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='USER_OWNERSHIP_UPDATED', index=46, number=46,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='SHARED_DRIVE_CREATE_SCHEDULED', index=47, number=47,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='SHARED_DRIVE_UPDATED', index=48, number=48,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='MESSAGE_PERSONAL_LABEL_UPDATED', index=49, number=49,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='USER_QUOTA_EXCEEDED', index=50, number=50,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
  ],
  containing_type=None,
  serialized_options=None,
  serialized_start=27159,
  serialized_end=28546,
)
_sym_db.RegisterEnumDescriptor(_EVENT_EVENTTYPE)

_INVITETYPE_TYPE = _descriptor.EnumDescriptor(
  name='Type',
  full_name='InviteType.Type',
  filename=None,
  file=DESCRIPTOR,
  create_key=_descriptor._internal_create_key,
  values=[
    _descriptor.EnumValueDescriptor(
      name='UNKNOWN', index=0, number=0,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='AT_MENTION', index=1, number=1,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='DIRECT_ADD', index=2, number=2,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
  ],
  containing_type=None,
  serialized_options=None,
  serialized_start=28562,
  serialized_end=28613,
)
_sym_db.RegisterEnumDescriptor(_INVITETYPE_TYPE)

_MEETINGSPACE_RECORDINGINFO_RECORDINGEVENT_EVENTTYPE = _descriptor.EnumDescriptor(
  name='EventType',
  full_name='MeetingSpace.RecordingInfo.RecordingEvent.EventType',
  filename=None,
  file=DESCRIPTOR,
  create_key=_descriptor._internal_create_key,
  values=[
    _descriptor.EnumValueDescriptor(
      name='RECORDING_EVENT_UNSPECIFIED', index=0, number=0,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='USER_ACTION', index=1, number=1,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
  ],
  containing_type=None,
  serialized_options=None,
  serialized_start=30350,
  serialized_end=30411,
)
_sym_db.RegisterEnumDescriptor(_MEETINGSPACE_RECORDINGINFO_RECORDINGEVENT_EVENTTYPE)

_MEETINGSPACE_RECORDINGINFO_RECORDINGSTATUS = _descriptor.EnumDescriptor(
  name='RecordingStatus',
  full_name='MeetingSpace.RecordingInfo.RecordingStatus',
  filename=None,
  file=DESCRIPTOR,
  create_key=_descriptor._internal_create_key,
  values=[
    _descriptor.EnumValueDescriptor(
      name='RECORDING_UNSPECIFIED', index=0, number=0,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='RECORDING_INACTIVE', index=1, number=1,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='RECORDING_STARTING', index=2, number=2,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='RECORDING_STARTED', index=3, number=3,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
  ],
  containing_type=None,
  serialized_options=None,
  serialized_start=30413,
  serialized_end=30528,
)
_sym_db.RegisterEnumDescriptor(_MEETINGSPACE_RECORDINGINFO_RECORDINGSTATUS)

_MEETINGSPACE_STREAMINGSESSIONINFO_SESSIONEVENT_EVENTTYPE = _descriptor.EnumDescriptor(
  name='EventType',
  full_name='MeetingSpace.StreamingSessionInfo.SessionEvent.EventType',
  filename=None,
  file=DESCRIPTOR,
  create_key=_descriptor._internal_create_key,
  values=[
    _descriptor.EnumValueDescriptor(
      name='EVENT_UNSPECIFIED', index=0, number=0,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
  ],
  containing_type=None,
  serialized_options=None,
  serialized_start=31059,
  serialized_end=31093,
)
_sym_db.RegisterEnumDescriptor(_MEETINGSPACE_STREAMINGSESSIONINFO_SESSIONEVENT_EVENTTYPE)

_MEETINGSPACE_STREAMINGSESSIONINFO_STATUS = _descriptor.EnumDescriptor(
  name='Status',
  full_name='MeetingSpace.StreamingSessionInfo.Status',
  filename=None,
  file=DESCRIPTOR,
  create_key=_descriptor._internal_create_key,
  values=[
    _descriptor.EnumValueDescriptor(
      name='STATUS_UNSPECIFIED', index=0, number=0,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='INACTIVE', index=1, number=1,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='STARTING', index=2, number=2,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='LIVE', index=3, number=3,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
  ],
  containing_type=None,
  serialized_options=None,
  serialized_start=31148,
  serialized_end=31218,
)
_sym_db.RegisterEnumDescriptor(_MEETINGSPACE_STREAMINGSESSIONINFO_STATUS)

_MEMBERSHIPCHANGEDMETADATA_TYPE = _descriptor.EnumDescriptor(
  name='Type',
  full_name='MembershipChangedMetadata.Type',
  filename=None,
  file=DESCRIPTOR,
  create_key=_descriptor._internal_create_key,
  values=[
    _descriptor.EnumValueDescriptor(
      name='TYPE_UNSPECIFIED', index=0, number=0,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='INVITED', index=1, number=1,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='JOINED', index=2, number=2,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='ADDED', index=3, number=3,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='REMOVED', index=4, number=4,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='LEFT', index=5, number=5,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='BOT_ADDED', index=6, number=6,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='BOT_REMOVED', index=7, number=7,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='KICKED_DUE_TO_OTR_CONFLICT', index=8, number=8,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='ROLE_UPDATED', index=9, number=9,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
  ],
  containing_type=None,
  serialized_options=None,
  serialized_start=33022,
  serialized_end=33191,
)
_sym_db.RegisterEnumDescriptor(_MEMBERSHIPCHANGEDMETADATA_TYPE)

_USERMENTIONMETADATA_TYPE = _descriptor.EnumDescriptor(
  name='Type',
  full_name='UserMentionMetadata.Type',
  filename=None,
  file=DESCRIPTOR,
  create_key=_descriptor._internal_create_key,
  values=[
    _descriptor.EnumValueDescriptor(
      name='TYPE_UNSPECIFIED', index=0, number=0,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='INVITE', index=1, number=1,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='UNINVITE', index=2, number=2,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='MENTION', index=3, number=3,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='MENTION_ALL', index=4, number=4,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='FAILED_TO_ADD', index=5, number=5,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
  ],
  containing_type=None,
  serialized_options=None,
  serialized_start=33353,
  serialized_end=33456,
)
_sym_db.RegisterEnumDescriptor(_USERMENTIONMETADATA_TYPE)

_SLASHCOMMANDMETADATA_TYPE = _descriptor.EnumDescriptor(
  name='Type',
  full_name='SlashCommandMetadata.Type',
  filename=None,
  file=DESCRIPTOR,
  create_key=_descriptor._internal_create_key,
  values=[
    _descriptor.EnumValueDescriptor(
      name='TYPE_UNSPECIFIED', index=0, number=0,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='ADD', index=1, number=1,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='INVOKE', index=2, number=2,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='FAILED_TO_ADD', index=3, number=3,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
  ],
  containing_type=None,
  serialized_options=None,
  serialized_start=33637,
  serialized_end=33705,
)
_sym_db.RegisterEnumDescriptor(_SLASHCOMMANDMETADATA_TYPE)

_GROUPVISIBILITY_VISIBILITYSTATE = _descriptor.EnumDescriptor(
  name='VisibilityState',
  full_name='GroupVisibility.VisibilityState',
  filename=None,
  file=DESCRIPTOR,
  create_key=_descriptor._internal_create_key,
  values=[
    _descriptor.EnumValueDescriptor(
      name='UNKNOWN', index=0, number=0,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='PRIVATE', index=1, number=1,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='PUBLIC', index=2, number=2,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
  ],
  containing_type=None,
  serialized_options=None,
  serialized_start=33726,
  serialized_end=33781,
)
_sym_db.RegisterEnumDescriptor(_GROUPVISIBILITY_VISIBILITYSTATE)

_CATCHUPRESPONSE_RESPONSESTATUS = _descriptor.EnumDescriptor(
  name='ResponseStatus',
  full_name='CatchUpResponse.ResponseStatus',
  filename=None,
  file=DESCRIPTOR,
  create_key=_descriptor._internal_create_key,
  values=[
    _descriptor.EnumValueDescriptor(
      name='UNKNOWN', index=0, number=0,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='COMPLETED', index=1, number=1,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='PAGINATED', index=2, number=2,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='ABORTED_CUTOFF_EXCEEDED', index=3, number=3,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='ABORTED_CACHE_INVALIDATION', index=4, number=4,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='ABORTED_FROM_REVISION_TOO_OLD', index=5, number=5,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
  ],
  containing_type=None,
  serialized_options=None,
  serialized_start=34451,
  serialized_end=34606,
)
_sym_db.RegisterEnumDescriptor(_CATCHUPRESPONSE_RESPONSESTATUS)

_GETGROUPREQUEST_FETCHOPTIONS = _descriptor.EnumDescriptor(
  name='FetchOptions',
  full_name='GetGroupRequest.FetchOptions',
  filename=None,
  file=DESCRIPTOR,
  create_key=_descriptor._internal_create_key,
  values=[
    _descriptor.EnumValueDescriptor(
      name='MEMBERS', index=0, number=0,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='INVITEES', index=1, number=1,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='MEMBER_IDS_ONLY', index=2, number=2,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='PROFILE_IN_READ_RECEIPTS', index=3, number=3,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='INCLUDE_SNIPPET', index=4, number=4,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='INCLUDE_DYNAMIC_GROUP_NAME', index=5, number=5,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='INCLUDE_ACCOUNT_USER_CAPABILITIES', index=6, number=6,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
  ],
  containing_type=None,
  serialized_options=None,
  serialized_start=34878,
  serialized_end=35062,
)
_sym_db.RegisterEnumDescriptor(_GETGROUPREQUEST_FETCHOPTIONS)

_WORLDSECTION_WORLDSECTIONTYPE = _descriptor.EnumDescriptor(
  name='WorldSectionType',
  full_name='WorldSection.WorldSectionType',
  filename=None,
  file=DESCRIPTOR,
  create_key=_descriptor._internal_create_key,
  values=[
    _descriptor.EnumValueDescriptor(
      name='WORLD_SECTION_TYPE_UNSPECIFIED', index=0, number=0,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='STARRED_DIRECT_MESSAGE_PEOPLE', index=1, number=1,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='STARRED_ROOMS', index=2, number=2,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='STARRED_DIRECT_MESSAGE_BOTS', index=3, number=3,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='NON_STARRED_DIRECT_MESSAGE_PEOPLE', index=4, number=4,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='NON_STARRED_ROOMS', index=5, number=5,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='NON_STARRED_DIRECT_MESSAGE_BOTS', index=6, number=6,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='ALL_DIRECT_MESSAGE_PEOPLE', index=7, number=7,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='ALL_ROOMS', index=8, number=8,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='ALL_DIRECT_MESSAGE_BOTS', index=9, number=9,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='INVITED_DM_PEOPLE', index=10, number=10,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='SPAM_INVITED_DM_PEOPLE', index=11, number=11,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='STARRED_DIRECT_MESSAGE_EVERYONE', index=12, number=12,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='NON_STARRED_DIRECT_MESSAGE_EVERYONE', index=13, number=13,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='ALL_DIRECT_MESSAGE_EVERYONE', index=14, number=14,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='STARRED_DMS_AND_STARRED_UNNAMED_ROOMS', index=15, number=15,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='NON_STARRED_DMS_AND_NON_STARRED_UNNAMED_ROOMS', index=16, number=16,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
  ],
  containing_type=None,
  serialized_options=None,
  serialized_start=35488,
  serialized_end=36059,
)
_sym_db.RegisterEnumDescriptor(_WORLDSECTION_WORLDSECTIONTYPE)

_WORLDFILTER_STARREDSTATE = _descriptor.EnumDescriptor(
  name='StarredState',
  full_name='WorldFilter.StarredState',
  filename=None,
  file=DESCRIPTOR,
  create_key=_descriptor._internal_create_key,
  values=[
    _descriptor.EnumValueDescriptor(
      name='STARRED_STATE_UNSPECIFIED', index=0, number=0,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='STARRED', index=1, number=1,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='NON_STARRED', index=2, number=2,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
  ],
  containing_type=None,
  serialized_options=None,
  serialized_start=36495,
  serialized_end=36570,
)
_sym_db.RegisterEnumDescriptor(_WORLDFILTER_STARREDSTATE)

_WORLDFILTER_VISIBILITYSTATE = _descriptor.EnumDescriptor(
  name='VisibilityState',
  full_name='WorldFilter.VisibilityState',
  filename=None,
  file=DESCRIPTOR,
  create_key=_descriptor._internal_create_key,
  values=[
    _descriptor.EnumValueDescriptor(
      name='VISIBILITY_STATE_UNSPECIFIED', index=0, number=0,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='VISIBLE', index=1, number=1,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='HIDDEN', index=2, number=2,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
  ],
  containing_type=None,
  serialized_options=None,
  serialized_start=36572,
  serialized_end=36648,
)
_sym_db.RegisterEnumDescriptor(_WORLDFILTER_VISIBILITYSTATE)

_WORLDFILTER_READSTATE = _descriptor.EnumDescriptor(
  name='ReadState',
  full_name='WorldFilter.ReadState',
  filename=None,
  file=DESCRIPTOR,
  create_key=_descriptor._internal_create_key,
  values=[
    _descriptor.EnumValueDescriptor(
      name='READ_STATE_UNSPECIFIED', index=0, number=0,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='READ', index=1, number=1,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='UNREAD', index=2, number=2,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='UNREAD_SUBSCRIBED_TOPIC', index=3, number=3,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
  ],
  containing_type=None,
  serialized_options=None,
  serialized_start=36650,
  serialized_end=36740,
)
_sym_db.RegisterEnumDescriptor(_WORLDFILTER_READSTATE)

_WORLDFILTER_BLOCKSTATE = _descriptor.EnumDescriptor(
  name='BlockState',
  full_name='WorldFilter.BlockState',
  filename=None,
  file=DESCRIPTOR,
  create_key=_descriptor._internal_create_key,
  values=[
    _descriptor.EnumValueDescriptor(
      name='BLOCK_STATE_UNSPECIFIED', index=0, number=0,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='BLOCKED', index=1, number=1,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='UNBLOCKED', index=2, number=2,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
  ],
  containing_type=None,
  serialized_options=None,
  serialized_start=36742,
  serialized_end=36811,
)
_sym_db.RegisterEnumDescriptor(_WORLDFILTER_BLOCKSTATE)

_WORLDFILTER_NAMEDSTATE = _descriptor.EnumDescriptor(
  name='NamedState',
  full_name='WorldFilter.NamedState',
  filename=None,
  file=DESCRIPTOR,
  create_key=_descriptor._internal_create_key,
  values=[
    _descriptor.EnumValueDescriptor(
      name='NAMED_STATE_UNSPECIFIED', index=0, number=0,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='NAMED', index=1, number=1,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='UNNAMED', index=2, number=2,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
  ],
  containing_type=None,
  serialized_options=None,
  serialized_start=36813,
  serialized_end=36878,
)
_sym_db.RegisterEnumDescriptor(_WORLDFILTER_NAMEDSTATE)

_WORLDFILTER_MEMBERTYPE = _descriptor.EnumDescriptor(
  name='MemberType',
  full_name='WorldFilter.MemberType',
  filename=None,
  file=DESCRIPTOR,
  create_key=_descriptor._internal_create_key,
  values=[
    _descriptor.EnumValueDescriptor(
      name='MEMBER_TYPE_UNSPECIFIED', index=0, number=0,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='HUMAN', index=1, number=1,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='BOT', index=2, number=2,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
  ],
  containing_type=None,
  serialized_options=None,
  serialized_start=36880,
  serialized_end=36941,
)
_sym_db.RegisterEnumDescriptor(_WORLDFILTER_MEMBERTYPE)

_WORLDFILTER_GROUPTYPE = _descriptor.EnumDescriptor(
  name='GroupType',
  full_name='WorldFilter.GroupType',
  filename=None,
  file=DESCRIPTOR,
  create_key=_descriptor._internal_create_key,
  values=[
    _descriptor.EnumValueDescriptor(
      name='GROUP_TYPE_UNSPECIFIED', index=0, number=0,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='DM', index=1, number=1,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='ROOM', index=2, number=2,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
  ],
  containing_type=None,
  serialized_options=None,
  serialized_start=36943,
  serialized_end=37000,
)
_sym_db.RegisterEnumDescriptor(_WORLDFILTER_GROUPTYPE)

_PAGINATEDWORLDREQUEST_FETCHOPTIONS = _descriptor.EnumDescriptor(
  name='FetchOptions',
  full_name='PaginatedWorldRequest.FetchOptions',
  filename=None,
  file=DESCRIPTOR,
  create_key=_descriptor._internal_create_key,
  values=[
    _descriptor.EnumValueDescriptor(
      name='UNKNOWN', index=0, number=0,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='EXCLUDE_GROUP_LITE', index=1, number=1,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='FETCH_BOTS_IN_HUMAN_DM', index=2, number=2,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='FETCH_SPACE_INTEGRATION_PAYLOADS', index=3, number=3,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='FETCH_GROUPS_D3_POLICIES', index=4, number=4,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
  ],
  containing_type=None,
  serialized_options=None,
  serialized_start=39014,
  serialized_end=39161,
)
_sym_db.RegisterEnumDescriptor(_PAGINATEDWORLDREQUEST_FETCHOPTIONS)

_INVITENOTIFICATIONSETTINGS_NOTIFICATIONOPTION = _descriptor.EnumDescriptor(
  name='NotificationOption',
  full_name='InviteNotificationSettings.NotificationOption',
  filename=None,
  file=DESCRIPTOR,
  create_key=_descriptor._internal_create_key,
  values=[
    _descriptor.EnumValueDescriptor(
      name='UNSPECIFIED', index=0, number=0,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='EMAIL_NOTIFICATION', index=1, number=1,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
  ],
  containing_type=None,
  serialized_options=None,
  serialized_start=40120,
  serialized_end=40181,
)
_sym_db.RegisterEnumDescriptor(_INVITENOTIFICATIONSETTINGS_NOTIFICATIONOPTION)

_SETDNDDURATIONREQUEST_STATE = _descriptor.EnumDescriptor(
  name='State',
  full_name='SetDndDurationRequest.State',
  filename=None,
  file=DESCRIPTOR,
  create_key=_descriptor._internal_create_key,
  values=[
    _descriptor.EnumValueDescriptor(
      name='UNKNOWN', index=0, number=0,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='AVAILABLE', index=1, number=1,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='DND', index=2, number=2,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
  ],
  containing_type=None,
  serialized_options=None,
  serialized_start=41287,
  serialized_end=41331,
)
_sym_db.RegisterEnumDescriptor(_SETDNDDURATIONREQUEST_STATE)

_UPDATEGROUPREQUEST_UPDATEMASK = _descriptor.EnumDescriptor(
  name='UpdateMask',
  full_name='UpdateGroupRequest.UpdateMask',
  filename=None,
  file=DESCRIPTOR,
  create_key=_descriptor._internal_create_key,
  values=[
    _descriptor.EnumValueDescriptor(
      name='UPDATE_MASK_UNSPECIFIED', index=0, number=0,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='NAME', index=1, number=1,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='VISIBILITY', index=2, number=2,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='AVATAR_INFO', index=3, number=3,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='SPACE_DETAILS', index=4, number=4,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='SHARED_DRIVE', index=5, number=5,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
  ],
  containing_type=None,
  serialized_options=None,
  serialized_start=41629,
  serialized_end=41750,
)
_sym_db.RegisterEnumDescriptor(_UPDATEGROUPREQUEST_UPDATEMASK)


_USERID = _descriptor.Descriptor(
  name='UserId',
  full_name='UserId',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='id', full_name='UserId.id', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='type', full_name='UserId.type', index=1,
      number=2, type=14, cpp_type=8, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=28,
  serialized_end=73,
)


_USER = _descriptor.Descriptor(
  name='User',
  full_name='User',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='user_id', full_name='User.user_id', index=0,
      number=1, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='name', full_name='User.name', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='avatar_url', full_name='User.avatar_url', index=2,
      number=3, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='email', full_name='User.email', index=3,
      number=4, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='first_name', full_name='User.first_name', index=4,
      number=5, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='last_name', full_name='User.last_name', index=5,
      number=6, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='deleted', full_name='User.deleted', index=6,
      number=7, type=8, cpp_type=7, label=1,
      has_default_value=False, default_value=False,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='is_anonymous', full_name='User.is_anonymous', index=7,
      number=8, type=8, cpp_type=7, label=1,
      has_default_value=False, default_value=False,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='gender', full_name='User.gender', index=8,
      number=10, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='block_relationship', full_name='User.block_relationship', index=9,
      number=13, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=76,
  serialized_end=303,
)


_USERBLOCKRELATIONSHIP = _descriptor.Descriptor(
  name='UserBlockRelationship',
  full_name='UserBlockRelationship',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='is_blocked_by_requester', full_name='UserBlockRelationship.is_blocked_by_requester', index=0,
      number=1, type=8, cpp_type=7, label=1,
      has_default_value=False, default_value=False,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='has_blocked_requester', full_name='UserBlockRelationship.has_blocked_requester', index=1,
      number=2, type=8, cpp_type=7, label=1,
      has_default_value=False, default_value=False,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=305,
  serialized_end=392,
)


_DNDSETTINGS = _descriptor.Descriptor(
  name='DndSettings',
  full_name='DndSettings',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='dnd_state', full_name='DndSettings.dnd_state', index=0,
      number=1, type=14, cpp_type=8, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='dnd_expiry_time_usec', full_name='DndSettings.dnd_expiry_time_usec', index=1,
      number=2, type=3, cpp_type=2, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='state_remaining_duration_usec', full_name='DndSettings.state_remaining_duration_usec', index=2,
      number=3, type=3, cpp_type=2, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
    _DNDSETTINGS_DNDSTATE_STATE,
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=395,
  serialized_end=580,
)


_EMOJI = _descriptor.Descriptor(
  name='Emoji',
  full_name='Emoji',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='unicode', full_name='Emoji.unicode', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=582,
  serialized_end=606,
)


_CUSTOMSTATUS = _descriptor.Descriptor(
  name='CustomStatus',
  full_name='CustomStatus',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='status_text', full_name='CustomStatus.status_text', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='status_emoji', full_name='CustomStatus.status_emoji', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='state_expiry_timestamp_usec', full_name='CustomStatus.state_expiry_timestamp_usec', index=2,
      number=3, type=3, cpp_type=2, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='emoji', full_name='CustomStatus.emoji', index=3,
      number=4, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=608,
  serialized_end=725,
)


_USERSTATUS = _descriptor.Descriptor(
  name='UserStatus',
  full_name='UserStatus',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='user_id', full_name='UserStatus.user_id', index=0,
      number=1, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='dnd_settings', full_name='UserStatus.dnd_settings', index=1,
      number=2, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='update_timestamp_usec', full_name='UserStatus.update_timestamp_usec', index=2,
      number=3, type=3, cpp_type=2, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='dnd_settings_update_timestamp_usec', full_name='UserStatus.dnd_settings_update_timestamp_usec', index=3,
      number=7, type=3, cpp_type=2, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='status_timestamp_usec', full_name='UserStatus.status_timestamp_usec', index=4,
      number=4, type=3, cpp_type=2, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='presence_shared', full_name='UserStatus.presence_shared', index=5,
      number=5, type=8, cpp_type=7, label=1,
      has_default_value=False, default_value=False,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='custom_status', full_name='UserStatus.custom_status', index=6,
      number=6, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='custom_status_update_timestamp_usec', full_name='UserStatus.custom_status_update_timestamp_usec', index=7,
      number=8, type=3, cpp_type=2, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=728,
  serialized_end=1016,
)


_USERSTATUSUPDATEDEVENT = _descriptor.Descriptor(
  name='UserStatusUpdatedEvent',
  full_name='UserStatusUpdatedEvent',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='user_status', full_name='UserStatusUpdatedEvent.user_status', index=0,
      number=1, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=1018,
  serialized_end=1076,
)


_GETUSERSTATUSREQUEST = _descriptor.Descriptor(
  name='GetUserStatusRequest',
  full_name='GetUserStatusRequest',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='user_ids', full_name='GetUserStatusRequest.user_ids', index=0,
      number=1, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='request_header', full_name='GetUserStatusRequest.request_header', index=1,
      number=100, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=1078,
  serialized_end=1167,
)


_GETUSERSTATUSRESPONSE = _descriptor.Descriptor(
  name='GetUserStatusResponse',
  full_name='GetUserStatusResponse',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='user_statuses', full_name='GetUserStatusResponse.user_statuses', index=0,
      number=1, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=1169,
  serialized_end=1228,
)


_GETSELFUSERSTATUSREQUEST = _descriptor.Descriptor(
  name='GetSelfUserStatusRequest',
  full_name='GetSelfUserStatusRequest',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='request_header', full_name='GetSelfUserStatusRequest.request_header', index=0,
      number=100, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=1230,
  serialized_end=1296,
)


_GETSELFUSERSTATUSRESPONSE = _descriptor.Descriptor(
  name='GetSelfUserStatusResponse',
  full_name='GetSelfUserStatusResponse',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='user_status', full_name='GetSelfUserStatusResponse.user_status', index=0,
      number=1, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='user_revision', full_name='GetSelfUserStatusResponse.user_revision', index=1,
      number=2, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=1298,
  serialized_end=1397,
)


_CLIENTFEATURECAPABILITIES = _descriptor.Descriptor(
  name='ClientFeatureCapabilities',
  full_name='ClientFeatureCapabilities',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='enable_all_features', full_name='ClientFeatureCapabilities.enable_all_features', index=0,
      number=1, type=8, cpp_type=7, label=1,
      has_default_value=False, default_value=False,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='spaces_level_for_testing', full_name='ClientFeatureCapabilities.spaces_level_for_testing', index=1,
      number=2, type=14, cpp_type=8, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='dms_level_for_testing', full_name='ClientFeatureCapabilities.dms_level_for_testing', index=2,
      number=3, type=14, cpp_type=8, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='post_rooms_level', full_name='ClientFeatureCapabilities.post_rooms_level', index=3,
      number=4, type=14, cpp_type=8, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='spam_room_invites_level', full_name='ClientFeatureCapabilities.spam_room_invites_level', index=4,
      number=5, type=14, cpp_type=8, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='tombstone_level', full_name='ClientFeatureCapabilities.tombstone_level', index=5,
      number=6, type=14, cpp_type=8, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='rich_text_viewing_level', full_name='ClientFeatureCapabilities.rich_text_viewing_level', index=6,
      number=7, type=14, cpp_type=8, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
    _CLIENTFEATURECAPABILITIES_CAPABILITYLEVEL,
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=1400,
  serialized_end=1979,
)


_REQUESTHEADER = _descriptor.Descriptor(
  name='RequestHeader',
  full_name='RequestHeader',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='trace_id', full_name='RequestHeader.trace_id', index=0,
      number=1, type=3, cpp_type=2, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='client_type', full_name='RequestHeader.client_type', index=1,
      number=2, type=14, cpp_type=8, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='client_version', full_name='RequestHeader.client_version', index=2,
      number=3, type=3, cpp_type=2, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='locale', full_name='RequestHeader.locale', index=3,
      number=4, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='client_feature_capabilities', full_name='RequestHeader.client_feature_capabilities', index=4,
      number=5, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
    _REQUESTHEADER_CLIENTTYPE,
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=1982,
  serialized_end=2302,
)


_MEMBER = _descriptor.Descriptor(
  name='Member',
  full_name='Member',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='user', full_name='Member.user', index=0,
      number=1, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=2304,
  serialized_end=2333,
)


_MEMBERID = _descriptor.Descriptor(
  name='MemberId',
  full_name='MemberId',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='user_id', full_name='MemberId.user_id', index=0,
      number=1, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=2335,
  serialized_end=2371,
)


_MEMBERSHIPID = _descriptor.Descriptor(
  name='MembershipId',
  full_name='MembershipId',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='member_id', full_name='MembershipId.member_id', index=0,
      number=1, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='space_id', full_name='MembershipId.space_id', index=1,
      number=2, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='group_id', full_name='MembershipId.group_id', index=2,
      number=3, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=2373,
  serialized_end=2473,
)


_MEMBERSHIP = _descriptor.Descriptor(
  name='Membership',
  full_name='Membership',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='id', full_name='Membership.id', index=0,
      number=1, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='create_time', full_name='Membership.create_time', index=1,
      number=2, type=3, cpp_type=2, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='membership_state', full_name='Membership.membership_state', index=2,
      number=3, type=14, cpp_type=8, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='invite_category', full_name='Membership.invite_category', index=3,
      number=4, type=14, cpp_type=8, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='membership_role', full_name='Membership.membership_role', index=4,
      number=5, type=14, cpp_type=8, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=2476,
  serialized_end=2664,
)


_MEMBERPROFILE = _descriptor.Descriptor(
  name='MemberProfile',
  full_name='MemberProfile',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='id', full_name='MemberProfile.id', index=0,
      number=1, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='member', full_name='MemberProfile.member', index=1,
      number=2, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=2666,
  serialized_end=2733,
)


_GETMEMBERSREQUEST = _descriptor.Descriptor(
  name='GetMembersRequest',
  full_name='GetMembersRequest',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='request_header', full_name='GetMembersRequest.request_header', index=0,
      number=100, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='member_ids', full_name='GetMembersRequest.member_ids', index=1,
      number=1, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='membership_ids', full_name='GetMembersRequest.membership_ids', index=2,
      number=2, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=2736,
  serialized_end=2865,
)


_GETMEMBERSRESPONSE = _descriptor.Descriptor(
  name='GetMembersResponse',
  full_name='GetMembersResponse',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='members', full_name='GetMembersResponse.members', index=0,
      number=1, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='member_profiles', full_name='GetMembersResponse.member_profiles', index=1,
      number=2, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=2867,
  serialized_end=2954,
)


_USERPRESENCE = _descriptor.Descriptor(
  name='UserPresence',
  full_name='UserPresence',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='user_id', full_name='UserPresence.user_id', index=0,
      number=1, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='presence', full_name='UserPresence.presence', index=1,
      number=2, type=14, cpp_type=8, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='active_until', full_name='UserPresence.active_until', index=2,
      number=4, type=3, cpp_type=2, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='dnd_state', full_name='UserPresence.dnd_state', index=3,
      number=3, type=14, cpp_type=8, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='user_status', full_name='UserPresence.user_status', index=4,
      number=5, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=2957,
  serialized_end=3118,
)


_GETUSERPRESENCEREQUEST = _descriptor.Descriptor(
  name='GetUserPresenceRequest',
  full_name='GetUserPresenceRequest',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='request_header', full_name='GetUserPresenceRequest.request_header', index=0,
      number=100, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='user_ids', full_name='GetUserPresenceRequest.user_ids', index=1,
      number=1, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='include_active_until', full_name='GetUserPresenceRequest.include_active_until', index=2,
      number=2, type=8, cpp_type=7, label=1,
      has_default_value=False, default_value=False,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='include_user_status', full_name='GetUserPresenceRequest.include_user_status', index=3,
      number=3, type=8, cpp_type=7, label=1,
      has_default_value=False, default_value=False,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=3121,
  serialized_end=3271,
)


_GETUSERPRESENCERESPONSE = _descriptor.Descriptor(
  name='GetUserPresenceResponse',
  full_name='GetUserPresenceResponse',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='user_presences', full_name='GetUserPresenceResponse.user_presences', index=0,
      number=1, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=3273,
  serialized_end=3337,
)


_JADDONSIDENTIFIER = _descriptor.Descriptor(
  name='JAddOnsIdentifier',
  full_name='JAddOnsIdentifier',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='id', full_name='JAddOnsIdentifier.id', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=3339,
  serialized_end=3370,
)


_JADDONSFORMATTEDTEXT_FORMATTEDTEXTELEMENT_DATETIME = _descriptor.Descriptor(
  name='DateTime',
  full_name='JAddOnsFormattedText.FormattedTextElement.DateTime',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='time_millis', full_name='JAddOnsFormattedText.FormattedTextElement.DateTime.time_millis', index=0,
      number=1, type=3, cpp_type=2, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='time_zone_offset_minutes', full_name='JAddOnsFormattedText.FormattedTextElement.DateTime.time_zone_offset_minutes', index=1,
      number=2, type=5, cpp_type=1, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='date_only', full_name='JAddOnsFormattedText.FormattedTextElement.DateTime.date_only', index=2,
      number=3, type=8, cpp_type=7, label=1,
      has_default_value=False, default_value=False,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='floating_time', full_name='JAddOnsFormattedText.FormattedTextElement.DateTime.floating_time', index=3,
      number=4, type=8, cpp_type=7, label=1,
      has_default_value=False, default_value=False,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=3756,
  serialized_end=3863,
)

_JADDONSFORMATTEDTEXT_FORMATTEDTEXTELEMENT_STYLEDTEXT = _descriptor.Descriptor(
  name='StyledText',
  full_name='JAddOnsFormattedText.FormattedTextElement.StyledText',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='text', full_name='JAddOnsFormattedText.FormattedTextElement.StyledText.text', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='datetime', full_name='JAddOnsFormattedText.FormattedTextElement.StyledText.datetime', index=1,
      number=4, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='styles', full_name='JAddOnsFormattedText.FormattedTextElement.StyledText.styles', index=2,
      number=2, type=14, cpp_type=8, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='font_weight', full_name='JAddOnsFormattedText.FormattedTextElement.StyledText.font_weight', index=3,
      number=5, type=14, cpp_type=8, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='color', full_name='JAddOnsFormattedText.FormattedTextElement.StyledText.color', index=4,
      number=3, type=5, cpp_type=1, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='theme_colors', full_name='JAddOnsFormattedText.FormattedTextElement.StyledText.theme_colors', index=5,
      number=7, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
    _JADDONSFORMATTEDTEXT_FORMATTEDTEXTELEMENT_STYLEDTEXT_STYLE,
    _JADDONSFORMATTEDTEXT_FORMATTEDTEXTELEMENT_STYLEDTEXT_FONTWEIGHT,
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=3866,
  serialized_end=4354,
)

_JADDONSFORMATTEDTEXT_FORMATTEDTEXTELEMENT_HYPERLINK = _descriptor.Descriptor(
  name='HyperLink',
  full_name='JAddOnsFormattedText.FormattedTextElement.HyperLink',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='link', full_name='JAddOnsFormattedText.FormattedTextElement.HyperLink.link', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='original_link', full_name='JAddOnsFormattedText.FormattedTextElement.HyperLink.original_link', index=1,
      number=3, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='text', full_name='JAddOnsFormattedText.FormattedTextElement.HyperLink.text', index=2,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=4356,
  serialized_end=4418,
)

_JADDONSFORMATTEDTEXT_FORMATTEDTEXTELEMENT = _descriptor.Descriptor(
  name='FormattedTextElement',
  full_name='JAddOnsFormattedText.FormattedTextElement',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='styled_text', full_name='JAddOnsFormattedText.FormattedTextElement.styled_text', index=0,
      number=1, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='hyperlink', full_name='JAddOnsFormattedText.FormattedTextElement.hyperlink', index=1,
      number=2, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[_JADDONSFORMATTEDTEXT_FORMATTEDTEXTELEMENT_DATETIME, _JADDONSFORMATTEDTEXT_FORMATTEDTEXTELEMENT_STYLEDTEXT, _JADDONSFORMATTEDTEXT_FORMATTEDTEXTELEMENT_HYPERLINK, ],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=3583,
  serialized_end=4418,
)

_JADDONSFORMATTEDTEXT = _descriptor.Descriptor(
  name='JAddOnsFormattedText',
  full_name='JAddOnsFormattedText',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='id', full_name='JAddOnsFormattedText.id', index=0,
      number=4, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='original_text', full_name='JAddOnsFormattedText.original_text', index=1,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='formatted_text_elements', full_name='JAddOnsFormattedText.formatted_text_elements', index=2,
      number=2, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='text_align', full_name='JAddOnsFormattedText.text_align', index=3,
      number=3, type=14, cpp_type=8, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[_JADDONSFORMATTEDTEXT_FORMATTEDTEXTELEMENT, ],
  enum_types=[
    _JADDONSFORMATTEDTEXT_TEXTALIGN,
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=3373,
  serialized_end=4464,
)


_JADDONSIMAGECROPSTYLE = _descriptor.Descriptor(
  name='JAddOnsImageCropStyle',
  full_name='JAddOnsImageCropStyle',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='type', full_name='JAddOnsImageCropStyle.type', index=0,
      number=1, type=14, cpp_type=8, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='aspect_ratio', full_name='JAddOnsImageCropStyle.aspect_ratio', index=1,
      number=2, type=1, cpp_type=5, label=1,
      has_default_value=False, default_value=float(0),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
    _JADDONSIMAGECROPSTYLE_IMAGECROPTYPE,
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=4467,
  serialized_end=4669,
)


_JADDONSTHEMECOLORS = _descriptor.Descriptor(
  name='JAddOnsThemeColors',
  full_name='JAddOnsThemeColors',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='light_theme_color', full_name='JAddOnsThemeColors.light_theme_color', index=0,
      number=1, type=5, cpp_type=1, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='dark_theme_color', full_name='JAddOnsThemeColors.dark_theme_color', index=1,
      number=2, type=5, cpp_type=1, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=4671,
  serialized_end=4744,
)


_JADDONSOPENLINK = _descriptor.Descriptor(
  name='JAddOnsOpenLink',
  full_name='JAddOnsOpenLink',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='url', full_name='JAddOnsOpenLink.url', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='original_link', full_name='JAddOnsOpenLink.original_link', index=1,
      number=5, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='open_as', full_name='JAddOnsOpenLink.open_as', index=2,
      number=2, type=14, cpp_type=8, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='on_close', full_name='JAddOnsOpenLink.on_close', index=3,
      number=3, type=14, cpp_type=8, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='load_indicator', full_name='JAddOnsOpenLink.load_indicator', index=4,
      number=4, type=14, cpp_type=8, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
    _JADDONSOPENLINK_OPENAS,
    _JADDONSOPENLINK_ONCLOSE,
    _JADDONSOPENLINK_LOADINDICATOR,
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=4747,
  serialized_end=5063,
)


_JADDONSCARDITEM_CARDITEMHEADER = _descriptor.Descriptor(
  name='CardItemHeader',
  full_name='JAddOnsCardItem.CardItemHeader',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='title', full_name='JAddOnsCardItem.CardItemHeader.title', index=0,
      number=1, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='subtitle', full_name='JAddOnsCardItem.CardItemHeader.subtitle', index=1,
      number=2, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='image_style', full_name='JAddOnsCardItem.CardItemHeader.image_style', index=2,
      number=3, type=14, cpp_type=8, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='image_url', full_name='JAddOnsCardItem.CardItemHeader.image_url', index=3,
      number=4, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='image_alt_text', full_name='JAddOnsCardItem.CardItemHeader.image_alt_text', index=4,
      number=5, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=5547,
  serialized_end=5744,
)

_JADDONSCARDITEM_CARDITEMSECTION = _descriptor.Descriptor(
  name='CardItemSection',
  full_name='JAddOnsCardItem.CardItemSection',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='id', full_name='JAddOnsCardItem.CardItemSection.id', index=0,
      number=6, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='description', full_name='JAddOnsCardItem.CardItemSection.description', index=1,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='header', full_name='JAddOnsCardItem.CardItemSection.header', index=2,
      number=5, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='widgets', full_name='JAddOnsCardItem.CardItemSection.widgets', index=3,
      number=2, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='collapsable', full_name='JAddOnsCardItem.CardItemSection.collapsable', index=4,
      number=3, type=8, cpp_type=7, label=1,
      has_default_value=False, default_value=False,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='num_uncollapsable_widgets', full_name='JAddOnsCardItem.CardItemSection.num_uncollapsable_widgets', index=5,
      number=4, type=5, cpp_type=1, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=5747,
  serialized_end=5945,
)

_JADDONSCARDITEM_CARDITEMACTION = _descriptor.Descriptor(
  name='CardItemAction',
  full_name='JAddOnsCardItem.CardItemAction',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='action_label', full_name='JAddOnsCardItem.CardItemAction.action_label', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='on_click', full_name='JAddOnsCardItem.CardItemAction.on_click', index=1,
      number=2, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=5947,
  serialized_end=6020,
)

_JADDONSCARDITEM_CARDITEMFIXEDFOOTER = _descriptor.Descriptor(
  name='CardItemFixedFooter',
  full_name='JAddOnsCardItem.CardItemFixedFooter',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='buttons', full_name='JAddOnsCardItem.CardItemFixedFooter.buttons', index=0,
      number=1, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='primary_button', full_name='JAddOnsCardItem.CardItemFixedFooter.primary_button', index=1,
      number=2, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='secondary_button', full_name='JAddOnsCardItem.CardItemFixedFooter.secondary_button', index=2,
      number=3, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=6023,
  serialized_end=6188,
)

_JADDONSCARDITEM_CARDITEMREFRESHACTION = _descriptor.Descriptor(
  name='CardItemRefreshAction',
  full_name='JAddOnsCardItem.CardItemRefreshAction',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='method', full_name='JAddOnsCardItem.CardItemRefreshAction.method', index=0,
      number=1, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=6190,
  serialized_end=6249,
)

_JADDONSCARDITEM = _descriptor.Descriptor(
  name='JAddOnsCardItem',
  full_name='JAddOnsCardItem',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='header', full_name='JAddOnsCardItem.header', index=0,
      number=1, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='sections', full_name='JAddOnsCardItem.sections', index=1,
      number=2, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='card_actions', full_name='JAddOnsCardItem.card_actions', index=2,
      number=3, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='name', full_name='JAddOnsCardItem.name', index=3,
      number=4, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='fixed_footer', full_name='JAddOnsCardItem.fixed_footer', index=4,
      number=5, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='refresh_action', full_name='JAddOnsCardItem.refresh_action', index=5,
      number=6, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='display_style', full_name='JAddOnsCardItem.display_style', index=6,
      number=7, type=14, cpp_type=8, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='peek_card_header', full_name='JAddOnsCardItem.peek_card_header', index=7,
      number=8, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='background_theme_colors', full_name='JAddOnsCardItem.background_theme_colors', index=8,
      number=9, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[_JADDONSCARDITEM_CARDITEMHEADER, _JADDONSCARDITEM_CARDITEMSECTION, _JADDONSCARDITEM_CARDITEMACTION, _JADDONSCARDITEM_CARDITEMFIXEDFOOTER, _JADDONSCARDITEM_CARDITEMREFRESHACTION, ],
  enum_types=[
    _JADDONSCARDITEM_DISPLAYSTYLE,
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=5066,
  serialized_end=6319,
)


_JADDONSPUSHCARD = _descriptor.Descriptor(
  name='JAddOnsPushCard',
  full_name='JAddOnsPushCard',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='card', full_name='JAddOnsPushCard.card', index=0,
      number=1, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=6321,
  serialized_end=6370,
)


_JADDONSONCLICK = _descriptor.Descriptor(
  name='JAddOnsOnClick',
  full_name='JAddOnsOnClick',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='link', full_name='JAddOnsOnClick.link', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='action', full_name='JAddOnsOnClick.action', index=1,
      number=2, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='open_link', full_name='JAddOnsOnClick.open_link', index=2,
      number=5, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='open_link_action', full_name='JAddOnsOnClick.open_link_action', index=3,
      number=7, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='push_card', full_name='JAddOnsOnClick.push_card', index=4,
      number=8, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=6373,
  serialized_end=6559,
)


_JADDONSTEXTWIDGET = _descriptor.Descriptor(
  name='JAddOnsTextWidget',
  full_name='JAddOnsTextWidget',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='line', full_name='JAddOnsTextWidget.line', index=0,
      number=1, type=9, cpp_type=9, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=6561,
  serialized_end=6594,
)


_JADDONSIMAGECOMPONENT = _descriptor.Descriptor(
  name='JAddOnsImageComponent',
  full_name='JAddOnsImageComponent',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='image_url', full_name='JAddOnsImageComponent.image_url', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='alt_text', full_name='JAddOnsImageComponent.alt_text', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='crop_style', full_name='JAddOnsImageComponent.crop_style', index=2,
      number=3, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='border_style', full_name='JAddOnsImageComponent.border_style', index=3,
      number=4, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=6597,
  serialized_end=6744,
)


_JADDONSGRID_GRIDITEM = _descriptor.Descriptor(
  name='GridItem',
  full_name='JAddOnsGrid.GridItem',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='identifier', full_name='JAddOnsGrid.GridItem.identifier', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='image', full_name='JAddOnsGrid.GridItem.image', index=1,
      number=2, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='title', full_name='JAddOnsGrid.GridItem.title', index=2,
      number=3, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='subtitle', full_name='JAddOnsGrid.GridItem.subtitle', index=3,
      number=4, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='text_alignment', full_name='JAddOnsGrid.GridItem.text_alignment', index=4,
      number=5, type=14, cpp_type=8, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='layout', full_name='JAddOnsGrid.GridItem.layout', index=5,
      number=9, type=14, cpp_type=8, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='on_click', full_name='JAddOnsGrid.GridItem.on_click', index=6,
      number=10, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
    _JADDONSGRID_GRIDITEM_GRIDITEMLAYOUT,
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=6947,
  serialized_end=7257,
)

_JADDONSGRID = _descriptor.Descriptor(
  name='JAddOnsGrid',
  full_name='JAddOnsGrid',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='id', full_name='JAddOnsGrid.id', index=0,
      number=1, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='title', full_name='JAddOnsGrid.title', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='items', full_name='JAddOnsGrid.items', index=2,
      number=3, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='border_style', full_name='JAddOnsGrid.border_style', index=3,
      number=4, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='num_columns', full_name='JAddOnsGrid.num_columns', index=4,
      number=5, type=5, cpp_type=1, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='on_click', full_name='JAddOnsGrid.on_click', index=5,
      number=6, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[_JADDONSGRID_GRIDITEM, ],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=6747,
  serialized_end=7257,
)


_JADDONSBORDERSTYLE = _descriptor.Descriptor(
  name='JAddOnsBorderStyle',
  full_name='JAddOnsBorderStyle',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='type', full_name='JAddOnsBorderStyle.type', index=0,
      number=1, type=14, cpp_type=8, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='stroke_color', full_name='JAddOnsBorderStyle.stroke_color', index=1,
      number=2, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='corner_radius', full_name='JAddOnsBorderStyle.corner_radius', index=2,
      number=3, type=5, cpp_type=1, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
    _JADDONSBORDERSTYLE_BORDERTYPE,
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=7260,
  serialized_end=7458,
)


_JADDONSLABELCONTENTPAIR = _descriptor.Descriptor(
  name='JAddOnsLabelContentPair',
  full_name='JAddOnsLabelContentPair',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='label', full_name='JAddOnsLabelContentPair.label', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='content', full_name='JAddOnsLabelContentPair.content', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=7460,
  serialized_end=7517,
)


_JADDONSLABELCONTENTPAIRWIDGET = _descriptor.Descriptor(
  name='JAddOnsLabelContentPairWidget',
  full_name='JAddOnsLabelContentPairWidget',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='label_content_pair', full_name='JAddOnsLabelContentPairWidget.label_content_pair', index=0,
      number=1, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=7519,
  serialized_end=7604,
)


_JADDONSWIDGET_TEXTPARAGRAPH = _descriptor.Descriptor(
  name='TextParagraph',
  full_name='JAddOnsWidget.TextParagraph',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='text', full_name='JAddOnsWidget.TextParagraph.text', index=0,
      number=1, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=8340,
  serialized_end=8392,
)

_JADDONSWIDGET_TEXTKEYVALUE = _descriptor.Descriptor(
  name='TextKeyValue',
  full_name='JAddOnsWidget.TextKeyValue',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='key', full_name='JAddOnsWidget.TextKeyValue.key', index=0,
      number=1, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='text', full_name='JAddOnsWidget.TextKeyValue.text', index=1,
      number=2, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='on_click', full_name='JAddOnsWidget.TextKeyValue.on_click', index=2,
      number=3, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=8394,
  serialized_end=8516,
)

_JADDONSWIDGET_IMAGEKEYVALUE = _descriptor.Descriptor(
  name='ImageKeyValue',
  full_name='JAddOnsWidget.ImageKeyValue',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='icon_url', full_name='JAddOnsWidget.ImageKeyValue.icon_url', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='text', full_name='JAddOnsWidget.ImageKeyValue.text', index=1,
      number=2, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='on_click', full_name='JAddOnsWidget.ImageKeyValue.on_click', index=2,
      number=3, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=8518,
  serialized_end=8623,
)

_JADDONSWIDGET_IMAGE = _descriptor.Descriptor(
  name='Image',
  full_name='JAddOnsWidget.Image',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='id', full_name='JAddOnsWidget.Image.id', index=0,
      number=5, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='fife_image_url', full_name='JAddOnsWidget.Image.fife_image_url', index=1,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='on_click', full_name='JAddOnsWidget.Image.on_click', index=2,
      number=2, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='aspect_ratio', full_name='JAddOnsWidget.Image.aspect_ratio', index=3,
      number=3, type=1, cpp_type=5, label=1,
      has_default_value=False, default_value=float(0),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='alt_text', full_name='JAddOnsWidget.Image.alt_text', index=4,
      number=4, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=8626,
  serialized_end=8764,
)

_JADDONSWIDGET_ICON = _descriptor.Descriptor(
  name='Icon',
  full_name='JAddOnsWidget.Icon',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='icon_url', full_name='JAddOnsWidget.Icon.icon_url', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='alt_text', full_name='JAddOnsWidget.Icon.alt_text', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='image_style', full_name='JAddOnsWidget.Icon.image_style', index=2,
      number=3, type=14, cpp_type=8, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=8766,
  serialized_end=8867,
)

_JADDONSWIDGET_KEYVALUE_SWITCHWIDGET = _descriptor.Descriptor(
  name='SwitchWidget',
  full_name='JAddOnsWidget.KeyValue.SwitchWidget',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='id', full_name='JAddOnsWidget.KeyValue.SwitchWidget.id', index=0,
      number=5, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='name', full_name='JAddOnsWidget.KeyValue.SwitchWidget.name', index=1,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='value', full_name='JAddOnsWidget.KeyValue.SwitchWidget.value', index=2,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='selected', full_name='JAddOnsWidget.KeyValue.SwitchWidget.selected', index=3,
      number=3, type=8, cpp_type=7, label=1,
      has_default_value=False, default_value=False,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='on_change', full_name='JAddOnsWidget.KeyValue.SwitchWidget.on_change', index=4,
      number=4, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='control_type', full_name='JAddOnsWidget.KeyValue.SwitchWidget.control_type', index=5,
      number=6, type=14, cpp_type=8, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
    _JADDONSWIDGET_KEYVALUE_SWITCHWIDGET_CONTROLTYPE,
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=9352,
  serialized_end=9614,
)

_JADDONSWIDGET_KEYVALUE = _descriptor.Descriptor(
  name='KeyValue',
  full_name='JAddOnsWidget.KeyValue',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='icon_url', full_name='JAddOnsWidget.KeyValue.icon_url', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='icon_alt_text', full_name='JAddOnsWidget.KeyValue.icon_alt_text', index=1,
      number=9, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='image_style', full_name='JAddOnsWidget.KeyValue.image_style', index=2,
      number=10, type=14, cpp_type=8, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='start_icon', full_name='JAddOnsWidget.KeyValue.start_icon', index=3,
      number=11, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='top_label', full_name='JAddOnsWidget.KeyValue.top_label', index=4,
      number=2, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='content', full_name='JAddOnsWidget.KeyValue.content', index=5,
      number=3, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='content_multiline', full_name='JAddOnsWidget.KeyValue.content_multiline', index=6,
      number=8, type=8, cpp_type=7, label=1,
      has_default_value=False, default_value=False,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='bottom_label', full_name='JAddOnsWidget.KeyValue.bottom_label', index=7,
      number=4, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='on_click', full_name='JAddOnsWidget.KeyValue.on_click', index=8,
      number=5, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='button', full_name='JAddOnsWidget.KeyValue.button', index=9,
      number=6, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='switch_widget', full_name='JAddOnsWidget.KeyValue.switch_widget', index=10,
      number=7, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='end_icon', full_name='JAddOnsWidget.KeyValue.end_icon', index=11,
      number=12, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[_JADDONSWIDGET_KEYVALUE_SWITCHWIDGET, ],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=8870,
  serialized_end=9614,
)

_JADDONSWIDGET_DIVIDER = _descriptor.Descriptor(
  name='Divider',
  full_name='JAddOnsWidget.Divider',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=9616,
  serialized_end=9625,
)

_JADDONSWIDGET_MENU_MENUITEM = _descriptor.Descriptor(
  name='MenuItem',
  full_name='JAddOnsWidget.Menu.MenuItem',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='text', full_name='JAddOnsWidget.Menu.MenuItem.text', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='value', full_name='JAddOnsWidget.Menu.MenuItem.value', index=1,
      number=4, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='selected', full_name='JAddOnsWidget.Menu.MenuItem.selected', index=2,
      number=3, type=8, cpp_type=7, label=1,
      has_default_value=False, default_value=False,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=9749,
  serialized_end=9806,
)

_JADDONSWIDGET_MENU = _descriptor.Descriptor(
  name='Menu',
  full_name='JAddOnsWidget.Menu',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='name', full_name='JAddOnsWidget.Menu.name', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='items', full_name='JAddOnsWidget.Menu.items', index=1,
      number=2, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='on_change', full_name='JAddOnsWidget.Menu.on_change', index=2,
      number=4, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='label', full_name='JAddOnsWidget.Menu.label', index=3,
      number=5, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[_JADDONSWIDGET_MENU_MENUITEM, ],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=9628,
  serialized_end=9806,
)

_JADDONSWIDGET_AUTOCOMPLETE_AUTOCOMPLETEITEM = _descriptor.Descriptor(
  name='AutoCompleteItem',
  full_name='JAddOnsWidget.AutoComplete.AutoCompleteItem',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='text', full_name='JAddOnsWidget.AutoComplete.AutoCompleteItem.text', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=9885,
  serialized_end=9917,
)

_JADDONSWIDGET_AUTOCOMPLETE = _descriptor.Descriptor(
  name='AutoComplete',
  full_name='JAddOnsWidget.AutoComplete',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='items', full_name='JAddOnsWidget.AutoComplete.items', index=0,
      number=1, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[_JADDONSWIDGET_AUTOCOMPLETE_AUTOCOMPLETEITEM, ],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=9808,
  serialized_end=9917,
)

_JADDONSWIDGET_TEXTFIELD = _descriptor.Descriptor(
  name='TextField',
  full_name='JAddOnsWidget.TextField',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='id', full_name='JAddOnsWidget.TextField.id', index=0,
      number=11, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='name', full_name='JAddOnsWidget.TextField.name', index=1,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='max_lines', full_name='JAddOnsWidget.TextField.max_lines', index=2,
      number=2, type=5, cpp_type=1, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='type', full_name='JAddOnsWidget.TextField.type', index=3,
      number=5, type=14, cpp_type=8, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='label', full_name='JAddOnsWidget.TextField.label', index=4,
      number=3, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='hint_text', full_name='JAddOnsWidget.TextField.hint_text', index=5,
      number=4, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='value', full_name='JAddOnsWidget.TextField.value', index=6,
      number=6, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='line_type', full_name='JAddOnsWidget.TextField.line_type', index=7,
      number=7, type=14, cpp_type=8, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='on_change', full_name='JAddOnsWidget.TextField.on_change', index=8,
      number=8, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='auto_complete', full_name='JAddOnsWidget.TextField.auto_complete', index=9,
      number=9, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='auto_complete_callback', full_name='JAddOnsWidget.TextField.auto_complete_callback', index=10,
      number=10, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='auto_complete_multiple_selections', full_name='JAddOnsWidget.TextField.auto_complete_multiple_selections', index=11,
      number=12, type=8, cpp_type=7, label=1,
      has_default_value=False, default_value=False,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
    _JADDONSWIDGET_TEXTFIELD_TEXTFIELDTYPE,
    _JADDONSWIDGET_TEXTFIELD_LINETYPE,
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=9920,
  serialized_end=10421,
)

_JADDONSWIDGET_SELECTIONCONTROL_SELECTIONITEM = _descriptor.Descriptor(
  name='SelectionItem',
  full_name='JAddOnsWidget.SelectionControl.SelectionItem',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='id', full_name='JAddOnsWidget.SelectionControl.SelectionItem.id', index=0,
      number=4, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='text', full_name='JAddOnsWidget.SelectionControl.SelectionItem.text', index=1,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='selected', full_name='JAddOnsWidget.SelectionControl.SelectionItem.selected', index=2,
      number=2, type=8, cpp_type=7, label=1,
      has_default_value=False, default_value=False,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='value', full_name='JAddOnsWidget.SelectionControl.SelectionItem.value', index=3,
      number=3, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=10667,
  serialized_end=10761,
)

_JADDONSWIDGET_SELECTIONCONTROL = _descriptor.Descriptor(
  name='SelectionControl',
  full_name='JAddOnsWidget.SelectionControl',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='id', full_name='JAddOnsWidget.SelectionControl.id', index=0,
      number=6, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='name', full_name='JAddOnsWidget.SelectionControl.name', index=1,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='type', full_name='JAddOnsWidget.SelectionControl.type', index=2,
      number=2, type=14, cpp_type=8, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='items', full_name='JAddOnsWidget.SelectionControl.items', index=3,
      number=3, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='on_change', full_name='JAddOnsWidget.SelectionControl.on_change', index=4,
      number=4, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='label', full_name='JAddOnsWidget.SelectionControl.label', index=5,
      number=5, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[_JADDONSWIDGET_SELECTIONCONTROL_SELECTIONITEM, ],
  enum_types=[
    _JADDONSWIDGET_SELECTIONCONTROL_SELECTIONTYPE,
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=10424,
  serialized_end=10837,
)

_JADDONSWIDGET_DATETIMEPICKER = _descriptor.Descriptor(
  name='DateTimePicker',
  full_name='JAddOnsWidget.DateTimePicker',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='name', full_name='JAddOnsWidget.DateTimePicker.name', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='label', full_name='JAddOnsWidget.DateTimePicker.label', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='type', full_name='JAddOnsWidget.DateTimePicker.type', index=2,
      number=3, type=14, cpp_type=8, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='value_ms_epoch', full_name='JAddOnsWidget.DateTimePicker.value_ms_epoch', index=3,
      number=4, type=3, cpp_type=2, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='timezone_offset_date', full_name='JAddOnsWidget.DateTimePicker.timezone_offset_date', index=4,
      number=5, type=5, cpp_type=1, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='on_change', full_name='JAddOnsWidget.DateTimePicker.on_change', index=5,
      number=6, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='color', full_name='JAddOnsWidget.DateTimePicker.color', index=6,
      number=7, type=5, cpp_type=1, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='theme_colors', full_name='JAddOnsWidget.DateTimePicker.theme_colors', index=7,
      number=8, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
    _JADDONSWIDGET_DATETIMEPICKER_DATETIMEPICKERTYPE,
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=10840,
  serialized_end=11193,
)

_JADDONSWIDGET_TEXTBUTTON = _descriptor.Descriptor(
  name='TextButton',
  full_name='JAddOnsWidget.TextButton',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='id', full_name='JAddOnsWidget.TextButton.id', index=0,
      number=3, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='text', full_name='JAddOnsWidget.TextButton.text', index=1,
      number=1, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='on_click', full_name='JAddOnsWidget.TextButton.on_click', index=2,
      number=2, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='disabled', full_name='JAddOnsWidget.TextButton.disabled', index=3,
      number=4, type=8, cpp_type=7, label=1,
      has_default_value=False, default_value=False,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='style', full_name='JAddOnsWidget.TextButton.style', index=4,
      number=5, type=14, cpp_type=8, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='background_color', full_name='JAddOnsWidget.TextButton.background_color', index=5,
      number=6, type=5, cpp_type=1, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='background_theme_colors', full_name='JAddOnsWidget.TextButton.background_theme_colors', index=6,
      number=7, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='alt_text', full_name='JAddOnsWidget.TextButton.alt_text', index=7,
      number=8, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
    _JADDONSWIDGET_TEXTBUTTON_STYLE,
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=11196,
  serialized_end=11524,
)

_JADDONSWIDGET_IMAGEBUTTON = _descriptor.Descriptor(
  name='ImageButton',
  full_name='JAddOnsWidget.ImageButton',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='id', full_name='JAddOnsWidget.ImageButton.id', index=0,
      number=4, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='icon_url', full_name='JAddOnsWidget.ImageButton.icon_url', index=1,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='on_click', full_name='JAddOnsWidget.ImageButton.on_click', index=2,
      number=2, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='name', full_name='JAddOnsWidget.ImageButton.name', index=3,
      number=3, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=11526,
  serialized_end=11638,
)

_JADDONSWIDGET_BUTTON = _descriptor.Descriptor(
  name='Button',
  full_name='JAddOnsWidget.Button',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='text_button', full_name='JAddOnsWidget.Button.text_button', index=0,
      number=1, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='image_button', full_name='JAddOnsWidget.Button.image_button', index=1,
      number=2, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=11640,
  serialized_end=11746,
)

_JADDONSWIDGET = _descriptor.Descriptor(
  name='JAddOnsWidget',
  full_name='JAddOnsWidget',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='text_widget', full_name='JAddOnsWidget.text_widget', index=0,
      number=1, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='label_content_pair_widget', full_name='JAddOnsWidget.label_content_pair_widget', index=1,
      number=2, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='text_paragraph', full_name='JAddOnsWidget.text_paragraph', index=2,
      number=3, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='text_key_value', full_name='JAddOnsWidget.text_key_value', index=3,
      number=4, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='image_key_value', full_name='JAddOnsWidget.image_key_value', index=4,
      number=5, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='image', full_name='JAddOnsWidget.image', index=5,
      number=9, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='key_value', full_name='JAddOnsWidget.key_value', index=6,
      number=13, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='divider', full_name='JAddOnsWidget.divider', index=7,
      number=16, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='grid', full_name='JAddOnsWidget.grid', index=8,
      number=17, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='menu', full_name='JAddOnsWidget.menu', index=9,
      number=10, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='text_field', full_name='JAddOnsWidget.text_field', index=10,
      number=11, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='selection_control', full_name='JAddOnsWidget.selection_control', index=11,
      number=12, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='date_time_picker', full_name='JAddOnsWidget.date_time_picker', index=12,
      number=14, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='buttons', full_name='JAddOnsWidget.buttons', index=13,
      number=8, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='horizontal_align', full_name='JAddOnsWidget.horizontal_align', index=14,
      number=15, type=14, cpp_type=8, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[_JADDONSWIDGET_TEXTPARAGRAPH, _JADDONSWIDGET_TEXTKEYVALUE, _JADDONSWIDGET_IMAGEKEYVALUE, _JADDONSWIDGET_IMAGE, _JADDONSWIDGET_ICON, _JADDONSWIDGET_KEYVALUE, _JADDONSWIDGET_DIVIDER, _JADDONSWIDGET_MENU, _JADDONSWIDGET_AUTOCOMPLETE, _JADDONSWIDGET_TEXTFIELD, _JADDONSWIDGET_SELECTIONCONTROL, _JADDONSWIDGET_DATETIMEPICKER, _JADDONSWIDGET_TEXTBUTTON, _JADDONSWIDGET_IMAGEBUTTON, _JADDONSWIDGET_BUTTON, ],
  enum_types=[
    _JADDONSWIDGET_HORIZONTALALIGN,
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=7607,
  serialized_end=11810,
)


_JADDONSFORMACTION_ACTIONPARAMETER = _descriptor.Descriptor(
  name='ActionParameter',
  full_name='JAddOnsFormAction.ActionParameter',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='key', full_name='JAddOnsFormAction.ActionParameter.key', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='value', full_name='JAddOnsFormAction.ActionParameter.value', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=12057,
  serialized_end=12102,
)

_JADDONSFORMACTION = _descriptor.Descriptor(
  name='JAddOnsFormAction',
  full_name='JAddOnsFormAction',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='action_method_name', full_name='JAddOnsFormAction.action_method_name', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='parameters', full_name='JAddOnsFormAction.parameters', index=1,
      number=4, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='parameters_deprecated', full_name='JAddOnsFormAction.parameters_deprecated', index=2,
      number=3, type=9, cpp_type=9, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='encrypted_action', full_name='JAddOnsFormAction.encrypted_action', index=3,
      number=7, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='load_indicator', full_name='JAddOnsFormAction.load_indicator', index=4,
      number=5, type=14, cpp_type=8, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='persist_values', full_name='JAddOnsFormAction.persist_values', index=5,
      number=6, type=8, cpp_type=7, label=1,
      has_default_value=False, default_value=False,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[_JADDONSFORMACTION_ACTIONPARAMETER, ],
  enum_types=[
    _JADDONSFORMACTION_LOADINDICATOR,
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=11813,
  serialized_end=12142,
)


_JADDONSCONTEXTUALADDON_TOOLBAR = _descriptor.Descriptor(
  name='Toolbar',
  full_name='JAddOnsContextualAddOn.Toolbar',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='name', full_name='JAddOnsContextualAddOn.Toolbar.name', index=0,
      number=1, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='icon_url', full_name='JAddOnsContextualAddOn.Toolbar.icon_url', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='color', full_name='JAddOnsContextualAddOn.Toolbar.color', index=2,
      number=3, type=5, cpp_type=1, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='theme_colors', full_name='JAddOnsContextualAddOn.Toolbar.theme_colors', index=3,
      number=4, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=12266,
  serialized_end=12388,
)

_JADDONSCONTEXTUALADDON_CARD_CARDHEADER = _descriptor.Descriptor(
  name='CardHeader',
  full_name='JAddOnsContextualAddOn.Card.CardHeader',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='title', full_name='JAddOnsContextualAddOn.Card.CardHeader.title', index=0,
      number=1, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='subtitle', full_name='JAddOnsContextualAddOn.Card.CardHeader.subtitle', index=1,
      number=2, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='image_style', full_name='JAddOnsContextualAddOn.Card.CardHeader.image_style', index=2,
      number=3, type=14, cpp_type=8, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='image_url', full_name='JAddOnsContextualAddOn.Card.CardHeader.image_url', index=3,
      number=4, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='image_alt_text', full_name='JAddOnsContextualAddOn.Card.CardHeader.image_alt_text', index=4,
      number=5, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=12776,
  serialized_end=12969,
)

_JADDONSCONTEXTUALADDON_CARD_SECTION = _descriptor.Descriptor(
  name='Section',
  full_name='JAddOnsContextualAddOn.Card.Section',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='id', full_name='JAddOnsContextualAddOn.Card.Section.id', index=0,
      number=6, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='description', full_name='JAddOnsContextualAddOn.Card.Section.description', index=1,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='header', full_name='JAddOnsContextualAddOn.Card.Section.header', index=2,
      number=5, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='widgets', full_name='JAddOnsContextualAddOn.Card.Section.widgets', index=3,
      number=2, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='collapsable', full_name='JAddOnsContextualAddOn.Card.Section.collapsable', index=4,
      number=3, type=8, cpp_type=7, label=1,
      has_default_value=False, default_value=False,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='num_uncollapsable_widgets', full_name='JAddOnsContextualAddOn.Card.Section.num_uncollapsable_widgets', index=5,
      number=4, type=5, cpp_type=1, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=12972,
  serialized_end=13162,
)

_JADDONSCONTEXTUALADDON_CARD_CARDACTION = _descriptor.Descriptor(
  name='CardAction',
  full_name='JAddOnsContextualAddOn.Card.CardAction',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='action_label', full_name='JAddOnsContextualAddOn.Card.CardAction.action_label', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='on_click', full_name='JAddOnsContextualAddOn.Card.CardAction.on_click', index=1,
      number=2, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=13164,
  serialized_end=13233,
)

_JADDONSCONTEXTUALADDON_CARD_FIXEDFOOTER = _descriptor.Descriptor(
  name='FixedFooter',
  full_name='JAddOnsContextualAddOn.Card.FixedFooter',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='buttons', full_name='JAddOnsContextualAddOn.Card.FixedFooter.buttons', index=0,
      number=1, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='primary_button', full_name='JAddOnsContextualAddOn.Card.FixedFooter.primary_button', index=1,
      number=2, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='secondary_button', full_name='JAddOnsContextualAddOn.Card.FixedFooter.secondary_button', index=2,
      number=3, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=13236,
  serialized_end=13393,
)

_JADDONSCONTEXTUALADDON_CARD_REFRESHACTION = _descriptor.Descriptor(
  name='RefreshAction',
  full_name='JAddOnsContextualAddOn.Card.RefreshAction',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='method', full_name='JAddOnsContextualAddOn.Card.RefreshAction.method', index=0,
      number=1, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=13395,
  serialized_end=13446,
)

_JADDONSCONTEXTUALADDON_CARD = _descriptor.Descriptor(
  name='Card',
  full_name='JAddOnsContextualAddOn.Card',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='header', full_name='JAddOnsContextualAddOn.Card.header', index=0,
      number=1, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='sections', full_name='JAddOnsContextualAddOn.Card.sections', index=1,
      number=2, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='card_actions', full_name='JAddOnsContextualAddOn.Card.card_actions', index=2,
      number=3, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='name', full_name='JAddOnsContextualAddOn.Card.name', index=3,
      number=4, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='fixed_footer', full_name='JAddOnsContextualAddOn.Card.fixed_footer', index=4,
      number=5, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='refresh_action', full_name='JAddOnsContextualAddOn.Card.refresh_action', index=5,
      number=6, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='background_theme_colors', full_name='JAddOnsContextualAddOn.Card.background_theme_colors', index=6,
      number=7, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[_JADDONSCONTEXTUALADDON_CARD_CARDHEADER, _JADDONSCONTEXTUALADDON_CARD_SECTION, _JADDONSCONTEXTUALADDON_CARD_CARDACTION, _JADDONSCONTEXTUALADDON_CARD_FIXEDFOOTER, _JADDONSCONTEXTUALADDON_CARD_REFRESHACTION, ],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=12391,
  serialized_end=13446,
)

_JADDONSCONTEXTUALADDON = _descriptor.Descriptor(
  name='JAddOnsContextualAddOn',
  full_name='JAddOnsContextualAddOn',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='toolbar', full_name='JAddOnsContextualAddOn.toolbar', index=0,
      number=1, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='cards', full_name='JAddOnsContextualAddOn.cards', index=1,
      number=2, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[_JADDONSCONTEXTUALADDON_TOOLBAR, _JADDONSCONTEXTUALADDON_CARD, ],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=12145,
  serialized_end=13446,
)


_HTML = _descriptor.Descriptor(
  name='Html',
  full_name='Html',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='html', full_name='Html.html', index=0,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=13448,
  serialized_end=13468,
)


_HTMLATTACHMENT = _descriptor.Descriptor(
  name='HtmlAttachment',
  full_name='HtmlAttachment',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='html', full_name='HtmlAttachment.html', index=0,
      number=2, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=13470,
  serialized_end=13507,
)


_ATTACHMENT = _descriptor.Descriptor(
  name='Attachment',
  full_name='Attachment',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='html', full_name='Attachment.html', index=0,
      number=1, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='add_on_data', full_name='Attachment.add_on_data', index=1,
      number=2, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='app_id', full_name='Attachment.app_id', index=2,
      number=5, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='app_user', full_name='Attachment.app_user', index=3,
      number=6, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=13510,
  serialized_end=13649,
)


_REACTION = _descriptor.Descriptor(
  name='Reaction',
  full_name='Reaction',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='emoji', full_name='Reaction.emoji', index=0,
      number=1, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='count', full_name='Reaction.count', index=1,
      number=2, type=5, cpp_type=1, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='current_user_participated', full_name='Reaction.current_user_participated', index=2,
      number=3, type=8, cpp_type=7, label=1,
      has_default_value=False, default_value=False,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='create_timestamp', full_name='Reaction.create_timestamp', index=3,
      number=4, type=3, cpp_type=2, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=13651,
  serialized_end=13760,
)


_MESSAGEPARENTID = _descriptor.Descriptor(
  name='MessageParentId',
  full_name='MessageParentId',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='topic_id', full_name='MessageParentId.topic_id', index=0,
      number=4, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=13762,
  serialized_end=13807,
)


_MESSAGEID = _descriptor.Descriptor(
  name='MessageId',
  full_name='MessageId',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='parent_id', full_name='MessageId.parent_id', index=0,
      number=1, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='message_id', full_name='MessageId.message_id', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=13809,
  serialized_end=13877,
)


_MESSAGE = _descriptor.Descriptor(
  name='Message',
  full_name='Message',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='id', full_name='Message.id', index=0,
      number=1, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='message_state', full_name='Message.message_state', index=1,
      number=20, type=14, cpp_type=8, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='creator', full_name='Message.creator', index=2,
      number=2, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='create_time', full_name='Message.create_time', index=3,
      number=3, type=3, cpp_type=2, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='last_update_time', full_name='Message.last_update_time', index=4,
      number=4, type=3, cpp_type=2, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='delete_time', full_name='Message.delete_time', index=5,
      number=8, type=3, cpp_type=2, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='last_edit_time', full_name='Message.last_edit_time', index=6,
      number=17, type=3, cpp_type=2, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='text_body', full_name='Message.text_body', index=7,
      number=10, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='annotations', full_name='Message.annotations', index=8,
      number=11, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='local_id', full_name='Message.local_id', index=9,
      number=14, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='attachments', full_name='Message.attachments', index=10,
      number=15, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='retention_settings', full_name='Message.retention_settings', index=11,
      number=18, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='reactions', full_name='Message.reactions', index=12,
      number=21, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='last_reply_time', full_name='Message.last_reply_time', index=13,
      number=5, type=3, cpp_type=2, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='last_view_time', full_name='Message.last_view_time', index=14,
      number=6, type=3, cpp_type=2, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='last_mention_time', full_name='Message.last_mention_time', index=15,
      number=7, type=3, cpp_type=2, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='num_unread_replies', full_name='Message.num_unread_replies', index=16,
      number=9, type=3, cpp_type=2, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='last_reply', full_name='Message.last_reply', index=17,
      number=12, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='sort_time', full_name='Message.sort_time', index=18,
      number=13, type=3, cpp_type=2, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='message_type', full_name='Message.message_type', index=19,
      number=28, type=14, cpp_type=8, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='creator_membership', full_name='Message.creator_membership', index=20,
      number=30, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
    _MESSAGE_MESSAGESTATE,
    _MESSAGE_MESSAGETYPE,
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=13880,
  serialized_end=14640,
)


_DRIVEMETADATA = _descriptor.Descriptor(
  name='DriveMetadata',
  full_name='DriveMetadata',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='id', full_name='DriveMetadata.id', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='title', full_name='DriveMetadata.title', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='thumbnail_url', full_name='DriveMetadata.thumbnail_url', index=2,
      number=3, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='thumbnail_width', full_name='DriveMetadata.thumbnail_width', index=3,
      number=7, type=5, cpp_type=1, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='thumbnail_height', full_name='DriveMetadata.thumbnail_height', index=4,
      number=8, type=5, cpp_type=1, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='mimetype', full_name='DriveMetadata.mimetype', index=5,
      number=4, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='external_mimetype', full_name='DriveMetadata.external_mimetype', index=6,
      number=14, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='organization_display_name', full_name='DriveMetadata.organization_display_name', index=7,
      number=13, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='encrypted_doc_id', full_name='DriveMetadata.encrypted_doc_id', index=8,
      number=10, type=8, cpp_type=7, label=1,
      has_default_value=False, default_value=False,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='url_fragment', full_name='DriveMetadata.url_fragment', index=9,
      number=16, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='embed_url', full_name='DriveMetadata.embed_url', index=10,
      number=18, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='is_owner', full_name='DriveMetadata.is_owner', index=11,
      number=19, type=8, cpp_type=7, label=1,
      has_default_value=False, default_value=False,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='can_share', full_name='DriveMetadata.can_share', index=12,
      number=5, type=8, cpp_type=7, label=1,
      has_default_value=False, default_value=False,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='can_edit', full_name='DriveMetadata.can_edit', index=13,
      number=15, type=8, cpp_type=7, label=1,
      has_default_value=False, default_value=False,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='can_view', full_name='DriveMetadata.can_view', index=14,
      number=6, type=8, cpp_type=7, label=1,
      has_default_value=False, default_value=False,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='should_not_render', full_name='DriveMetadata.should_not_render', index=15,
      number=9, type=8, cpp_type=7, label=1,
      has_default_value=False, default_value=False,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='is_download_restricted', full_name='DriveMetadata.is_download_restricted', index=16,
      number=17, type=8, cpp_type=7, label=1,
      has_default_value=False, default_value=False,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=14643,
  serialized_end=15059,
)


_URL = _descriptor.Descriptor(
  name='Url',
  full_name='Url',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='url', full_name='Url.url', index=0,
      number=3, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=15061,
  serialized_end=15079,
)


_TRUSTEDRESOURCEURL = _descriptor.Descriptor(
  name='TrustedResourceUrl',
  full_name='TrustedResourceUrl',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='resource_url', full_name='TrustedResourceUrl.resource_url', index=0,
      number=4, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=15081,
  serialized_end=15123,
)


_URLMETADATA = _descriptor.Descriptor(
  name='UrlMetadata',
  full_name='UrlMetadata',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='title', full_name='UrlMetadata.title', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='snippet', full_name='UrlMetadata.snippet', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='image_url', full_name='UrlMetadata.image_url', index=2,
      number=3, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='image_height', full_name='UrlMetadata.image_height', index=3,
      number=4, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='image_width', full_name='UrlMetadata.image_width', index=4,
      number=5, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='url', full_name='UrlMetadata.url', index=5,
      number=7, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='gws_url', full_name='UrlMetadata.gws_url', index=6,
      number=13, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='redirect_url', full_name='UrlMetadata.redirect_url', index=7,
      number=14, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='gws_url_expiration_timestamp', full_name='UrlMetadata.gws_url_expiration_timestamp', index=8,
      number=15, type=3, cpp_type=2, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='domain', full_name='UrlMetadata.domain', index=9,
      number=8, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='should_not_render', full_name='UrlMetadata.should_not_render', index=10,
      number=9, type=8, cpp_type=7, label=1,
      has_default_value=False, default_value=False,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='int_image_height', full_name='UrlMetadata.int_image_height', index=11,
      number=10, type=5, cpp_type=1, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='int_image_width', full_name='UrlMetadata.int_image_width', index=12,
      number=11, type=5, cpp_type=1, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='mime_type', full_name='UrlMetadata.mime_type', index=13,
      number=12, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=15126,
  serialized_end=15454,
)


_UPLOADMETADATA = _descriptor.Descriptor(
  name='UploadMetadata',
  full_name='UploadMetadata',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='attachment_token', full_name='UploadMetadata.attachment_token', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='content_name', full_name='UploadMetadata.content_name', index=1,
      number=3, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='content_type', full_name='UploadMetadata.content_type', index=2,
      number=4, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='local_id', full_name='UploadMetadata.local_id', index=3,
      number=6, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='cloned_drive_id', full_name='UploadMetadata.cloned_drive_id', index=4,
      number=9, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=15457,
  serialized_end=15586,
)


_FORMATMETADATA = _descriptor.Descriptor(
  name='FormatMetadata',
  full_name='FormatMetadata',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='format_type', full_name='FormatMetadata.format_type', index=0,
      number=1, type=14, cpp_type=8, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='font_color', full_name='FormatMetadata.font_color', index=1,
      number=2, type=5, cpp_type=1, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
    _FORMATMETADATA_FORMATTYPE,
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=15589,
  serialized_end=15903,
)


_ANNOTATION = _descriptor.Descriptor(
  name='Annotation',
  full_name='Annotation',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='type', full_name='Annotation.type', index=0,
      number=1, type=14, cpp_type=8, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='start_index', full_name='Annotation.start_index', index=1,
      number=2, type=5, cpp_type=1, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='length', full_name='Annotation.length', index=2,
      number=3, type=5, cpp_type=1, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='local_id', full_name='Annotation.local_id', index=3,
      number=9, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='unique_id', full_name='Annotation.unique_id', index=4,
      number=19, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='chip_render_type', full_name='Annotation.chip_render_type', index=5,
      number=20, type=14, cpp_type=8, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='server_invalidated', full_name='Annotation.server_invalidated', index=6,
      number=13, type=8, cpp_type=7, label=1,
      has_default_value=False, default_value=False,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='user_mention_metadata', full_name='Annotation.user_mention_metadata', index=7,
      number=5, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='format_metadata', full_name='Annotation.format_metadata', index=8,
      number=8, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='slash_command_metadata', full_name='Annotation.slash_command_metadata', index=9,
      number=15, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='drive_metadata', full_name='Annotation.drive_metadata', index=10,
      number=4, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='url_metadata', full_name='Annotation.url_metadata', index=11,
      number=7, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='upload_metadata', full_name='Annotation.upload_metadata', index=12,
      number=10, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='membership_changed', full_name='Annotation.membership_changed', index=13,
      number=11, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='room_updated', full_name='Annotation.room_updated', index=14,
      number=14, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
    _ANNOTATION_CHIPRENDERTYPE,
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=15906,
  serialized_end=16559,
)


_TYPINGCONTEXT = _descriptor.Descriptor(
  name='TypingContext',
  full_name='TypingContext',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='group_id', full_name='TypingContext.group_id', index=0,
      number=1, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='topic_id', full_name='TypingContext.topic_id', index=1,
      number=2, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=16561,
  serialized_end=16632,
)


_SETTYPINGSTATEREQUEST = _descriptor.Descriptor(
  name='SetTypingStateRequest',
  full_name='SetTypingStateRequest',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='request_header', full_name='SetTypingStateRequest.request_header', index=0,
      number=100, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='state', full_name='SetTypingStateRequest.state', index=1,
      number=1, type=14, cpp_type=8, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='context', full_name='SetTypingStateRequest.context', index=2,
      number=2, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=16634,
  serialized_end=16759,
)


_SETTYPINGSTATERESPONSE = _descriptor.Descriptor(
  name='SetTypingStateResponse',
  full_name='SetTypingStateResponse',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='start_timestamp_usec', full_name='SetTypingStateResponse.start_timestamp_usec', index=0,
      number=1, type=3, cpp_type=2, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=16761,
  serialized_end=16815,
)


_DMID = _descriptor.Descriptor(
  name='DmId',
  full_name='DmId',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='dm_id', full_name='DmId.dm_id', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=16817,
  serialized_end=16838,
)


_SPACEID = _descriptor.Descriptor(
  name='SpaceId',
  full_name='SpaceId',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='space_id', full_name='SpaceId.space_id', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=16840,
  serialized_end=16867,
)


_GROUPID = _descriptor.Descriptor(
  name='GroupId',
  full_name='GroupId',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='space_id', full_name='GroupId.space_id', index=0,
      number=1, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='dm_id', full_name='GroupId.dm_id', index=1,
      number=3, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=16869,
  serialized_end=16928,
)


_GROUP_FLATGROUP = _descriptor.Descriptor(
  name='FlatGroup',
  full_name='Group.FlatGroup',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=17747,
  serialized_end=17758,
)

_GROUP_THREADEDGROUP = _descriptor.Descriptor(
  name='ThreadedGroup',
  full_name='Group.ThreadedGroup',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=17760,
  serialized_end=17775,
)

_GROUP = _descriptor.Descriptor(
  name='Group',
  full_name='Group',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='group_id', full_name='Group.group_id', index=0,
      number=1, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='name', full_name='Group.name', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='creator', full_name='Group.creator', index=2,
      number=4, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='create_time', full_name='Group.create_time', index=3,
      number=5, type=3, cpp_type=2, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='last_modified_time', full_name='Group.last_modified_time', index=4,
      number=6, type=3, cpp_type=2, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='sort_time', full_name='Group.sort_time', index=5,
      number=7, type=3, cpp_type=2, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='revision_time', full_name='Group.revision_time', index=6,
      number=13, type=3, cpp_type=2, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='group_read_state', full_name='Group.group_read_state', index=7,
      number=12, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='retention_horizon_time', full_name='Group.retention_horizon_time', index=8,
      number=14, type=3, cpp_type=2, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='retention_settings', full_name='Group.retention_settings', index=9,
      number=16, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='is_flat', full_name='Group.is_flat', index=10,
      number=17, type=8, cpp_type=7, label=1,
      has_default_value=False, default_value=False,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='interop_enabled', full_name='Group.interop_enabled', index=11,
      number=18, type=8, cpp_type=7, label=1,
      has_default_value=False, default_value=False,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='retention_duration_seconds', full_name='Group.retention_duration_seconds', index=12,
      number=19, type=3, cpp_type=2, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='group_type', full_name='Group.group_type', index=13,
      number=22, type=14, cpp_type=8, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='visibility', full_name='Group.visibility', index=14,
      number=23, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='roster_email', full_name='Group.roster_email', index=15,
      number=25, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='flat_group', full_name='Group.flat_group', index=16,
      number=26, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='threaded_group', full_name='Group.threaded_group', index=17,
      number=27, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='avatar_url', full_name='Group.avatar_url', index=18,
      number=28, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='attribute_checker_group_type', full_name='Group.attribute_checker_group_type', index=19,
      number=33, type=14, cpp_type=8, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='name_users', full_name='Group.name_users', index=20,
      number=34, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='group_support_level', full_name='Group.group_support_level', index=21,
      number=38, type=14, cpp_type=8, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='group_unsupported_reason', full_name='Group.group_unsupported_reason', index=22,
      number=39, type=14, cpp_type=8, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='typing_indicators_enabled', full_name='Group.typing_indicators_enabled', index=23,
      number=43, type=8, cpp_type=7, label=1,
      has_default_value=False, default_value=False,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[_GROUP_FLATGROUP, _GROUP_THREADEDGROUP, ],
  enum_types=[
    _GROUP_GROUPTYPE,
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=16931,
  serialized_end=17852,
)


_GROUPREADSTATEID = _descriptor.Descriptor(
  name='GroupReadStateId',
  full_name='GroupReadStateId',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='user_id', full_name='GroupReadStateId.user_id', index=0,
      number=1, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='group_id', full_name='GroupReadStateId.group_id', index=1,
      number=2, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=17854,
  serialized_end=17926,
)


_INVITESTATE = _descriptor.Descriptor(
  name='InviteState',
  full_name='InviteState',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='inviter_user_id', full_name='InviteState.inviter_user_id', index=0,
      number=1, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='invitation_message_id', full_name='InviteState.invitation_message_id', index=1,
      number=2, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='show_welcome_mat', full_name='InviteState.show_welcome_mat', index=2,
      number=3, type=8, cpp_type=7, label=1,
      has_default_value=False, default_value=False,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='invite_type', full_name='InviteState.invite_type', index=3,
      number=5, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=17929,
  serialized_end=18079,
)


_GROUPREADSTATE = _descriptor.Descriptor(
  name='GroupReadState',
  full_name='GroupReadState',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='id', full_name='GroupReadState.id', index=0,
      number=1, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='last_read_time', full_name='GroupReadState.last_read_time', index=1,
      number=2, type=3, cpp_type=2, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='unread_message_count', full_name='GroupReadState.unread_message_count', index=2,
      number=4, type=3, cpp_type=2, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='starred', full_name='GroupReadState.starred', index=3,
      number=5, type=8, cpp_type=7, label=1,
      has_default_value=False, default_value=False,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='update_timestamp', full_name='GroupReadState.update_timestamp', index=4,
      number=6, type=3, cpp_type=2, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='unread_subscribed_topic_count', full_name='GroupReadState.unread_subscribed_topic_count', index=5,
      number=7, type=3, cpp_type=2, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='unread_subscribed_topic_count_consistency_timestamp', full_name='GroupReadState.unread_subscribed_topic_count_consistency_timestamp', index=6,
      number=11, type=3, cpp_type=2, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='unread_subscribed_topics', full_name='GroupReadState.unread_subscribed_topics', index=7,
      number=8, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='hide_timestamp', full_name='GroupReadState.hide_timestamp', index=8,
      number=9, type=3, cpp_type=2, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='clear_history_timestamp', full_name='GroupReadState.clear_history_timestamp', index=9,
      number=14, type=3, cpp_type=2, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='invite_state', full_name='GroupReadState.invite_state', index=10,
      number=10, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='notification_settings', full_name='GroupReadState.notification_settings', index=11,
      number=12, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='retention_settings', full_name='GroupReadState.retention_settings', index=12,
      number=13, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='blocked', full_name='GroupReadState.blocked', index=13,
      number=15, type=8, cpp_type=7, label=1,
      has_default_value=False, default_value=False,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='membership_role', full_name='GroupReadState.membership_role', index=14,
      number=22, type=14, cpp_type=8, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='membership_state', full_name='GroupReadState.membership_state', index=15,
      number=16, type=14, cpp_type=8, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='invite_category', full_name='GroupReadState.invite_category', index=16,
      number=17, type=14, cpp_type=8, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='mark_as_unread_timestamp_usec', full_name='GroupReadState.mark_as_unread_timestamp_usec', index=17,
      number=18, type=3, cpp_type=2, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='visible_in_world_view', full_name='GroupReadState.visible_in_world_view', index=18,
      number=19, type=8, cpp_type=7, label=1,
      has_default_value=False, default_value=False,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='show_notification_card_in_stream', full_name='GroupReadState.show_notification_card_in_stream', index=19,
      number=20, type=8, cpp_type=7, label=1,
      has_default_value=False, default_value=False,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='joined_users', full_name='GroupReadState.joined_users', index=20,
      number=23, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=18082,
  serialized_end=18858,
)


_GROUPNOTIFICATIONSETTINGS = _descriptor.Descriptor(
  name='GroupNotificationSettings',
  full_name='GroupNotificationSettings',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='state', full_name='GroupNotificationSettings.state', index=0,
      number=1, type=14, cpp_type=8, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='room_state', full_name='GroupNotificationSettings.room_state', index=1,
      number=2, type=14, cpp_type=8, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
    _GROUPNOTIFICATIONSETTINGS_GROUPNOTIFICATIONSTATE,
    _GROUPNOTIFICATIONSETTINGS_ROOMNOTIFICATIONSTATE,
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=18861,
  serialized_end=19219,
)


_RETENTIONSETTINGS = _descriptor.Descriptor(
  name='RetentionSettings',
  full_name='RetentionSettings',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='state', full_name='RetentionSettings.state', index=0,
      number=1, type=14, cpp_type=8, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='expiry_timestamp', full_name='RetentionSettings.expiry_timestamp', index=1,
      number=2, type=3, cpp_type=2, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
    _RETENTIONSETTINGS_RETENTIONSTATE,
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=19222,
  serialized_end=19402,
)


_TOPICID = _descriptor.Descriptor(
  name='TopicId',
  full_name='TopicId',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='group_id', full_name='TopicId.group_id', index=0,
      number=3, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='topic_id', full_name='TopicId.topic_id', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=19404,
  serialized_end=19459,
)


_TOPIC = _descriptor.Descriptor(
  name='Topic',
  full_name='Topic',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='id', full_name='Topic.id', index=0,
      number=1, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='sort_time', full_name='Topic.sort_time', index=1,
      number=2, type=3, cpp_type=2, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='create_time_usec', full_name='Topic.create_time_usec', index=2,
      number=15, type=3, cpp_type=2, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='replies', full_name='Topic.replies', index=3,
      number=7, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='is_system_message', full_name='Topic.is_system_message', index=4,
      number=12, type=8, cpp_type=7, label=1,
      has_default_value=False, default_value=False,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='retention_settings', full_name='Topic.retention_settings', index=5,
      number=13, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='contains_more_unread_replies', full_name='Topic.contains_more_unread_replies', index=6,
      number=14, type=8, cpp_type=7, label=1,
      has_default_value=False, default_value=False,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=19462,
  serialized_end=19676,
)


_MESSAGEINFO = _descriptor.Descriptor(
  name='MessageInfo',
  full_name='MessageInfo',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='accept_format_annotations', full_name='MessageInfo.accept_format_annotations', index=0,
      number=1, type=8, cpp_type=7, label=1,
      has_default_value=False, default_value=False,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=19678,
  serialized_end=19726,
)


_CREATETOPICREQUEST = _descriptor.Descriptor(
  name='CreateTopicRequest',
  full_name='CreateTopicRequest',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='request_header', full_name='CreateTopicRequest.request_header', index=0,
      number=100, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='group_id', full_name='CreateTopicRequest.group_id', index=1,
      number=5, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='text_body', full_name='CreateTopicRequest.text_body', index=2,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='annotations', full_name='CreateTopicRequest.annotations', index=3,
      number=3, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='retention_settings', full_name='CreateTopicRequest.retention_settings', index=4,
      number=6, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='local_id', full_name='CreateTopicRequest.local_id', index=5,
      number=4, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='topic_and_message_id', full_name='CreateTopicRequest.topic_and_message_id', index=6,
      number=7, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='history_v2', full_name='CreateTopicRequest.history_v2', index=7,
      number=8, type=8, cpp_type=7, label=1,
      has_default_value=False, default_value=False,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='message_info', full_name='CreateTopicRequest.message_info', index=8,
      number=9, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=19729,
  serialized_end=20022,
)


_CREATETOPICRESPONSE = _descriptor.Descriptor(
  name='CreateTopicResponse',
  full_name='CreateTopicResponse',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='topic', full_name='CreateTopicResponse.topic', index=0,
      number=1, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='group_revision', full_name='CreateTopicResponse.group_revision', index=1,
      number=2, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='current_group_revision', full_name='CreateTopicResponse.current_group_revision', index=2,
      number=3, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=20025,
  serialized_end=20156,
)


_INVITEEINFO = _descriptor.Descriptor(
  name='InviteeInfo',
  full_name='InviteeInfo',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='user_id', full_name='InviteeInfo.user_id', index=0,
      number=1, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='email', full_name='InviteeInfo.email', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=20158,
  serialized_end=20212,
)


_INVITEEMEMBERINFO = _descriptor.Descriptor(
  name='InviteeMemberInfo',
  full_name='InviteeMemberInfo',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='invitee_info', full_name='InviteeMemberInfo.invitee_info', index=0,
      number=1, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=20214,
  serialized_end=20269,
)


_SPACECREATIONINFO_FLATGROUP = _descriptor.Descriptor(
  name='FlatGroup',
  full_name='SpaceCreationInfo.FlatGroup',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=17747,
  serialized_end=17758,
)

_SPACECREATIONINFO_THREADEDGROUP = _descriptor.Descriptor(
  name='ThreadedGroup',
  full_name='SpaceCreationInfo.ThreadedGroup',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=17760,
  serialized_end=17775,
)

_SPACECREATIONINFO_SPACETYPE = _descriptor.Descriptor(
  name='SpaceType',
  full_name='SpaceCreationInfo.SpaceType',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=20717,
  serialized_end=20728,
)

_SPACECREATIONINFO = _descriptor.Descriptor(
  name='SpaceCreationInfo',
  full_name='SpaceCreationInfo',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='name', full_name='SpaceCreationInfo.name', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='visibility', full_name='SpaceCreationInfo.visibility', index=1,
      number=4, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='flat_group', full_name='SpaceCreationInfo.flat_group', index=2,
      number=5, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='threaded_group', full_name='SpaceCreationInfo.threaded_group', index=3,
      number=6, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='has_server_generated_name', full_name='SpaceCreationInfo.has_server_generated_name', index=4,
      number=7, type=8, cpp_type=7, label=1,
      has_default_value=False, default_value=False,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='invitee_member_infos', full_name='SpaceCreationInfo.invitee_member_infos', index=5,
      number=8, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='space_type', full_name='SpaceCreationInfo.space_type', index=6,
      number=11, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='attribute_checker_group_type', full_name='SpaceCreationInfo.attribute_checker_group_type', index=7,
      number=13, type=14, cpp_type=8, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='shared_drive_name', full_name='SpaceCreationInfo.shared_drive_name', index=8,
      number=14, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[_SPACECREATIONINFO_FLATGROUP, _SPACECREATIONINFO_THREADEDGROUP, _SPACECREATIONINFO_SPACETYPE, ],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=20272,
  serialized_end=20728,
)


_CREATEGROUPREQUEST = _descriptor.Descriptor(
  name='CreateGroupRequest',
  full_name='CreateGroupRequest',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='request_header', full_name='CreateGroupRequest.request_header', index=0,
      number=100, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='space', full_name='CreateGroupRequest.space', index=1,
      number=1, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='local_id', full_name='CreateGroupRequest.local_id', index=2,
      number=3, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='should_find_existing_space', full_name='CreateGroupRequest.should_find_existing_space', index=3,
      number=4, type=8, cpp_type=7, label=1,
      has_default_value=False, default_value=False,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=20731,
  serialized_end=20880,
)


_CREATEMEMBERSHIPRESULT = _descriptor.Descriptor(
  name='CreateMembershipResult',
  full_name='CreateMembershipResult',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='membership', full_name='CreateMembershipResult.membership', index=0,
      number=1, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=20882,
  serialized_end=20939,
)


_CREATEGROUPRESPONSE = _descriptor.Descriptor(
  name='CreateGroupResponse',
  full_name='CreateGroupResponse',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='group', full_name='CreateGroupResponse.group', index=0,
      number=1, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='group_revision', full_name='CreateGroupResponse.group_revision', index=1,
      number=2, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='user_revision', full_name='CreateGroupResponse.user_revision', index=2,
      number=3, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='results', full_name='CreateGroupResponse.results', index=3,
      number=4, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=20942,
  serialized_end=21107,
)


_CREATEDMREQUEST = _descriptor.Descriptor(
  name='CreateDmRequest',
  full_name='CreateDmRequest',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='request_header', full_name='CreateDmRequest.request_header', index=0,
      number=100, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='fetch_options', full_name='CreateDmRequest.fetch_options', index=1,
      number=8, type=14, cpp_type=8, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='members', full_name='CreateDmRequest.members', index=2,
      number=1, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='invitees', full_name='CreateDmRequest.invitees', index=3,
      number=7, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='retention_settings', full_name='CreateDmRequest.retention_settings', index=4,
      number=3, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='local_id', full_name='CreateDmRequest.local_id', index=5,
      number=4, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='topic_and_message_id', full_name='CreateDmRequest.topic_and_message_id', index=6,
      number=5, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
    _CREATEDMREQUEST_FETCHOPTIONS,
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=21110,
  serialized_end=21445,
)


_CREATEDMRESPONSE = _descriptor.Descriptor(
  name='CreateDmResponse',
  full_name='CreateDmResponse',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='dm', full_name='CreateDmResponse.dm', index=0,
      number=1, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='topic', full_name='CreateDmResponse.topic', index=1,
      number=2, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='group_revision', full_name='CreateDmResponse.group_revision', index=2,
      number=3, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='memberships', full_name='CreateDmResponse.memberships', index=3,
      number=4, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=21448,
  serialized_end=21583,
)


_LISTTOPICSRESPONSE = _descriptor.Descriptor(
  name='ListTopicsResponse',
  full_name='ListTopicsResponse',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='topics', full_name='ListTopicsResponse.topics', index=0,
      number=1, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='user_revision', full_name='ListTopicsResponse.user_revision', index=1,
      number=2, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='group_revision', full_name='ListTopicsResponse.group_revision', index=2,
      number=3, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='contains_first_topic', full_name='ListTopicsResponse.contains_first_topic', index=3,
      number=4, type=8, cpp_type=7, label=1,
      has_default_value=False, default_value=False,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='contains_last_topic', full_name='ListTopicsResponse.contains_last_topic', index=4,
      number=5, type=8, cpp_type=7, label=1,
      has_default_value=False, default_value=False,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='read_receipt_set', full_name='ListTopicsResponse.read_receipt_set', index=5,
      number=6, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=21586,
  serialized_end=21809,
)


_LISTTOPICSREQUEST = _descriptor.Descriptor(
  name='ListTopicsRequest',
  full_name='ListTopicsRequest',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='request_header', full_name='ListTopicsRequest.request_header', index=0,
      number=100, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='group_id', full_name='ListTopicsRequest.group_id', index=1,
      number=8, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='page_size_for_topics', full_name='ListTopicsRequest.page_size_for_topics', index=2,
      number=2, type=5, cpp_type=1, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='page_size_for_unread_replies', full_name='ListTopicsRequest.page_size_for_unread_replies', index=3,
      number=6, type=5, cpp_type=1, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='page_size_for_read_replies', full_name='ListTopicsRequest.page_size_for_read_replies', index=4,
      number=7, type=5, cpp_type=1, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='page_size_for_replies', full_name='ListTopicsRequest.page_size_for_replies', index=5,
      number=3, type=5, cpp_type=1, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='fetch_options', full_name='ListTopicsRequest.fetch_options', index=6,
      number=5, type=14, cpp_type=8, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='user_not_older_than', full_name='ListTopicsRequest.user_not_older_than', index=7,
      number=9, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='group_not_older_than', full_name='ListTopicsRequest.group_not_older_than', index=8,
      number=10, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
    _LISTTOPICSREQUEST_FETCHOPTIONS,
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=21812,
  serialized_end=22260,
)


_LISTMEMBERSREQUEST = _descriptor.Descriptor(
  name='ListMembersRequest',
  full_name='ListMembersRequest',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='request_header', full_name='ListMembersRequest.request_header', index=0,
      number=100, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='space_id', full_name='ListMembersRequest.space_id', index=1,
      number=1, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='group_id', full_name='ListMembersRequest.group_id', index=2,
      number=5, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='page_size', full_name='ListMembersRequest.page_size', index=3,
      number=3, type=5, cpp_type=1, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='page_token', full_name='ListMembersRequest.page_token', index=4,
      number=4, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='not_older_than', full_name='ListMembersRequest.not_older_than', index=5,
      number=6, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=22263,
  serialized_end=22462,
)


_LISTMEMBERSRESPONSE = _descriptor.Descriptor(
  name='ListMembersResponse',
  full_name='ListMembersResponse',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='memberships', full_name='ListMembersResponse.memberships', index=0,
      number=1, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='members', full_name='ListMembersResponse.members', index=1,
      number=2, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='member_ids', full_name='ListMembersResponse.member_ids', index=2,
      number=5, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='next_page_token', full_name='ListMembersResponse.next_page_token', index=3,
      number=3, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='group_revision', full_name='ListMembersResponse.group_revision', index=4,
      number=4, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='invited_member_ids', full_name='ListMembersResponse.invited_member_ids', index=5,
      number=6, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=22465,
  serialized_end=22680,
)


_READRECEIPT = _descriptor.Descriptor(
  name='ReadReceipt',
  full_name='ReadReceipt',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='read_time_micros', full_name='ReadReceipt.read_time_micros', index=0,
      number=2, type=3, cpp_type=2, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='user', full_name='ReadReceipt.user', index=1,
      number=3, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=22682,
  serialized_end=22742,
)


_READRECEIPTSET = _descriptor.Descriptor(
  name='ReadReceiptSet',
  full_name='ReadReceiptSet',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='enabled', full_name='ReadReceiptSet.enabled', index=0,
      number=1, type=8, cpp_type=7, label=1,
      has_default_value=False, default_value=False,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='read_receipts', full_name='ReadReceiptSet.read_receipts', index=1,
      number=2, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=22744,
  serialized_end=22814,
)


_WEBPUSHNOTIFICATION = _descriptor.Descriptor(
  name='WebPushNotification',
  full_name='WebPushNotification',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='message', full_name='WebPushNotification.message', index=0,
      number=1, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='sender_name', full_name='WebPushNotification.sender_name', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='sender_avatar_url', full_name='WebPushNotification.sender_avatar_url', index=2,
      number=3, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='group_name', full_name='WebPushNotification.group_name', index=3,
      number=4, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='cause', full_name='WebPushNotification.cause', index=4,
      number=5, type=14, cpp_type=8, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=22817,
  serialized_end=22968,
)


_ANDROIDLOCALNOTIFICATION = _descriptor.Descriptor(
  name='AndroidLocalNotification',
  full_name='AndroidLocalNotification',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='thread_id', full_name='AndroidLocalNotification.thread_id', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='type_id', full_name='AndroidLocalNotification.type_id', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='group_name', full_name='AndroidLocalNotification.group_name', index=2,
      number=3, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='text_body', full_name='AndroidLocalNotification.text_body', index=3,
      number=6, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='icon_url', full_name='AndroidLocalNotification.icon_url', index=4,
      number=4, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='title', full_name='AndroidLocalNotification.title', index=5,
      number=7, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='group_id', full_name='AndroidLocalNotification.group_id', index=6,
      number=8, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='sub_text', full_name='AndroidLocalNotification.sub_text', index=7,
      number=9, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=22971,
  serialized_end=23141,
)


_IOSLOCALNOTIFICATION = _descriptor.Descriptor(
  name='IosLocalNotification',
  full_name='IosLocalNotification',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='apns_collapse_id', full_name='IosLocalNotification.apns_collapse_id', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='thread_id', full_name='IosLocalNotification.thread_id', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='title', full_name='IosLocalNotification.title', index=2,
      number=3, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='body', full_name='IosLocalNotification.body', index=3,
      number=4, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='category', full_name='IosLocalNotification.category', index=4,
      number=5, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='group_is_flat', full_name='IosLocalNotification.group_is_flat', index=5,
      number=6, type=8, cpp_type=7, label=1,
      has_default_value=False, default_value=False,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='message_is_off_the_record', full_name='IosLocalNotification.message_is_off_the_record', index=6,
      number=7, type=8, cpp_type=7, label=1,
      has_default_value=False, default_value=False,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='attribute_checker_group_type', full_name='IosLocalNotification.attribute_checker_group_type', index=7,
      number=9, type=14, cpp_type=8, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='navigation', full_name='IosLocalNotification.navigation', index=8,
      number=8, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=23144,
  serialized_end=23408,
)


_MOBILELOCALNOTIFICATION = _descriptor.Descriptor(
  name='MobileLocalNotification',
  full_name='MobileLocalNotification',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='android_local_notification', full_name='MobileLocalNotification.android_local_notification', index=0,
      number=1, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='ios_local_notification', full_name='MobileLocalNotification.ios_local_notification', index=1,
      number=2, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=23411,
  serialized_end=23554,
)


_MESSAGEEVENT = _descriptor.Descriptor(
  name='MessageEvent',
  full_name='MessageEvent',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='message', full_name='MessageEvent.message', index=0,
      number=1, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='last_message_in_topic_time', full_name='MessageEvent.last_message_in_topic_time', index=1,
      number=4, type=3, cpp_type=2, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='prev_revision_time', full_name='MessageEvent.prev_revision_time', index=2,
      number=5, type=3, cpp_type=2, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='is_head_message', full_name='MessageEvent.is_head_message', index=3,
      number=6, type=8, cpp_type=7, label=1,
      has_default_value=False, default_value=False,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=23557,
  serialized_end=23687,
)


_TYPINGSTATECHANGEDEVENT = _descriptor.Descriptor(
  name='TypingStateChangedEvent',
  full_name='TypingStateChangedEvent',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='state', full_name='TypingStateChangedEvent.state', index=0,
      number=1, type=14, cpp_type=8, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='user_id', full_name='TypingStateChangedEvent.user_id', index=1,
      number=2, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='context', full_name='TypingStateChangedEvent.context', index=2,
      number=3, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='start_timestamp_usec', full_name='TypingStateChangedEvent.start_timestamp_usec', index=3,
      number=4, type=3, cpp_type=2, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=23690,
  serialized_end=23833,
)


_MEMBERSHIPCHANGEDEVENT = _descriptor.Descriptor(
  name='MembershipChangedEvent',
  full_name='MembershipChangedEvent',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='new_membership', full_name='MembershipChangedEvent.new_membership', index=0,
      number=1, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='prior_membership_state', full_name='MembershipChangedEvent.prior_membership_state', index=1,
      number=2, type=14, cpp_type=8, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='prior_membership_role', full_name='MembershipChangedEvent.prior_membership_role', index=2,
      number=3, type=14, cpp_type=8, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=23836,
  serialized_end=23995,
)


_READRECEIPTCHANGEDEVENT = _descriptor.Descriptor(
  name='ReadReceiptChangedEvent',
  full_name='ReadReceiptChangedEvent',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='group_id', full_name='ReadReceiptChangedEvent.group_id', index=0,
      number=1, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='read_receipt_set', full_name='ReadReceiptChangedEvent.read_receipt_set', index=1,
      number=2, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=23997,
  serialized_end=24093,
)


_GROUPVIEWEDEVENT = _descriptor.Descriptor(
  name='GroupViewedEvent',
  full_name='GroupViewedEvent',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='group_id', full_name='GroupViewedEvent.group_id', index=0,
      number=1, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='view_time', full_name='GroupViewedEvent.view_time', index=1,
      number=2, type=3, cpp_type=2, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=24095,
  serialized_end=24160,
)


_WEBPUSHNOTIFICATIONEVENT = _descriptor.Descriptor(
  name='WebPushNotificationEvent',
  full_name='WebPushNotificationEvent',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='notification', full_name='WebPushNotificationEvent.notification', index=0,
      number=1, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='dispatch_approach_type', full_name='WebPushNotificationEvent.dispatch_approach_type', index=1,
      number=2, type=14, cpp_type=8, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='mobile_local_notification', full_name='WebPushNotificationEvent.mobile_local_notification', index=2,
      number=3, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='endpoint_types', full_name='WebPushNotificationEvent.endpoint_types', index=3,
      number=4, type=14, cpp_type=8, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
    _WEBPUSHNOTIFICATIONEVENT_DISPATCHAPPROACHTYPE,
    _WEBPUSHNOTIFICATIONEVENT_ENDPOINTTYPE,
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=24163,
  serialized_end=24709,
)


_STREAMEVENTSREQUEST = _descriptor.Descriptor(
  name='StreamEventsRequest',
  full_name='StreamEventsRequest',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='platform', full_name='StreamEventsRequest.platform', index=0,
      number=4, type=14, cpp_type=8, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='client_info', full_name='StreamEventsRequest.client_info', index=1,
      number=5, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='client_session_id', full_name='StreamEventsRequest.client_session_id', index=2,
      number=6, type=3, cpp_type=2, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='sample_id', full_name='StreamEventsRequest.sample_id', index=3,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='sample_ids', full_name='StreamEventsRequest.sample_ids', index=4,
      number=7, type=9, cpp_type=9, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='ping_event', full_name='StreamEventsRequest.ping_event', index=5,
      number=2, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='clock_sync_request', full_name='StreamEventsRequest.clock_sync_request', index=6,
      number=3, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='group_subscription_event', full_name='StreamEventsRequest.group_subscription_event', index=7,
      number=8, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='test_user_gaia_id', full_name='StreamEventsRequest.test_user_gaia_id', index=8,
      number=100, type=3, cpp_type=2, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=24712,
  serialized_end=25027,
)


_CLIENTINFO = _descriptor.Descriptor(
  name='ClientInfo',
  full_name='ClientInfo',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='platform', full_name='ClientInfo.platform', index=0,
      number=1, type=14, cpp_type=8, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='origin', full_name='ClientInfo.origin', index=1,
      number=2, type=14, cpp_type=8, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='source_machine', full_name='ClientInfo.source_machine', index=2,
      number=3, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=25029,
  serialized_end=25124,
)


_CLIENTNOTIFICATIONSSTATE = _descriptor.Descriptor(
  name='ClientNotificationsState',
  full_name='ClientNotificationsState',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='device_setting_state', full_name='ClientNotificationsState.device_setting_state', index=0,
      number=1, type=14, cpp_type=8, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
    _CLIENTNOTIFICATIONSSTATE_DEVICENOTIFICATIONSETTINGSTATE,
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=25127,
  serialized_end=25392,
)


_PINGEVENT = _descriptor.Descriptor(
  name='PingEvent',
  full_name='PingEvent',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='state', full_name='PingEvent.state', index=0,
      number=1, type=14, cpp_type=8, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='application_focus_state', full_name='PingEvent.application_focus_state', index=1,
      number=3, type=14, cpp_type=8, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='last_interactive_time_ms', full_name='PingEvent.last_interactive_time_ms', index=2,
      number=4, type=3, cpp_type=2, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='client_interactive_state', full_name='PingEvent.client_interactive_state', index=3,
      number=5, type=14, cpp_type=8, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='client_notifications_enabled', full_name='PingEvent.client_notifications_enabled', index=4,
      number=6, type=8, cpp_type=7, label=1,
      has_default_value=False, default_value=False,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='notifications_state', full_name='PingEvent.notifications_state', index=5,
      number=7, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='pwa_dedupe_enabled', full_name='PingEvent.pwa_dedupe_enabled', index=6,
      number=8, type=8, cpp_type=7, label=1,
      has_default_value=False, default_value=False,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='device_active_state', full_name='PingEvent.device_active_state', index=7,
      number=9, type=14, cpp_type=8, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
    _PINGEVENT_STATE,
    _PINGEVENT_APPLICATIONFOCUSSTATE,
    _PINGEVENT_CLIENTINTERACTIVESTATE,
    _PINGEVENT_DEVICEACTIVESTATE,
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=25395,
  serialized_end=26158,
)


_CLOCKSYNCREQUEST = _descriptor.Descriptor(
  name='ClockSyncRequest',
  full_name='ClockSyncRequest',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='origin_time_msec', full_name='ClockSyncRequest.origin_time_msec', index=0,
      number=1, type=3, cpp_type=2, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=26160,
  serialized_end=26204,
)


_CLOCKSYNCRESPONSE = _descriptor.Descriptor(
  name='ClockSyncResponse',
  full_name='ClockSyncResponse',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='origin_time_msec', full_name='ClockSyncResponse.origin_time_msec', index=0,
      number=1, type=3, cpp_type=2, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='receive_time_msec', full_name='ClockSyncResponse.receive_time_msec', index=1,
      number=2, type=3, cpp_type=2, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='transmit_time_msec', full_name='ClockSyncResponse.transmit_time_msec', index=2,
      number=3, type=3, cpp_type=2, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=26206,
  serialized_end=26306,
)


_GROUPSUBSCRIPTIONEVENT = _descriptor.Descriptor(
  name='GroupSubscriptionEvent',
  full_name='GroupSubscriptionEvent',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='group_ids', full_name='GroupSubscriptionEvent.group_ids', index=0,
      number=1, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=26308,
  serialized_end=26361,
)


_STREAMEVENTSRESPONSE = _descriptor.Descriptor(
  name='StreamEventsResponse',
  full_name='StreamEventsResponse',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='event', full_name='StreamEventsResponse.event', index=0,
      number=1, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='sample_id', full_name='StreamEventsResponse.sample_id', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='clock_sync_response', full_name='StreamEventsResponse.clock_sync_response', index=2,
      number=3, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=26363,
  serialized_end=26476,
)


_EVENT_EVENTBODY = _descriptor.Descriptor(
  name='EventBody',
  full_name='Event.EventBody',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='group_viewed', full_name='Event.EventBody.group_viewed', index=0,
      number=3, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='message_posted', full_name='Event.EventBody.message_posted', index=1,
      number=6, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='web_push_notification', full_name='Event.EventBody.web_push_notification', index=2,
      number=10, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='membership_changed', full_name='Event.EventBody.membership_changed', index=3,
      number=14, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='user_status_updated_event', full_name='Event.EventBody.user_status_updated_event', index=4,
      number=23, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='typing_state_changed_event', full_name='Event.EventBody.typing_state_changed_event', index=5,
      number=26, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='read_receipt_changed', full_name='Event.EventBody.read_receipt_changed', index=6,
      number=33, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='event_type', full_name='Event.EventBody.event_type', index=7,
      number=12, type=14, cpp_type=8, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='trace_id', full_name='Event.EventBody.trace_id', index=8,
      number=20, type=3, cpp_type=2, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=26720,
  serialized_end=27156,
)

_EVENT = _descriptor.Descriptor(
  name='Event',
  full_name='Event',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='group_id', full_name='Event.group_id', index=0,
      number=1, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='type', full_name='Event.type', index=1,
      number=3, type=14, cpp_type=8, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='body', full_name='Event.body', index=2,
      number=4, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='user_id', full_name='Event.user_id', index=3,
      number=5, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='user_revision', full_name='Event.user_revision', index=4,
      number=6, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='group_revision', full_name='Event.group_revision', index=5,
      number=7, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='bodies', full_name='Event.bodies', index=6,
      number=8, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[_EVENT_EVENTBODY, ],
  enum_types=[
    _EVENT_EVENTTYPE,
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=26479,
  serialized_end=28546,
)


_INVITETYPE = _descriptor.Descriptor(
  name='InviteType',
  full_name='InviteType',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
    _INVITETYPE_TYPE,
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=28548,
  serialized_end=28613,
)


_GROUPDETAILS = _descriptor.Descriptor(
  name='GroupDetails',
  full_name='GroupDetails',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='description', full_name='GroupDetails.description', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='guidelines', full_name='GroupDetails.guidelines', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=28615,
  serialized_end=28670,
)


_ROOMUPDATEDMETADATA_ROOMRENAMEMETADATA = _descriptor.Descriptor(
  name='RoomRenameMetadata',
  full_name='RoomUpdatedMetadata.RoomRenameMetadata',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='new_name', full_name='RoomUpdatedMetadata.RoomRenameMetadata.new_name', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='prev_name', full_name='RoomUpdatedMetadata.RoomRenameMetadata.prev_name', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=28958,
  serialized_end=29015,
)

_ROOMUPDATEDMETADATA_GROUPDETAILSUPDATEDMETADATA = _descriptor.Descriptor(
  name='GroupDetailsUpdatedMetadata',
  full_name='RoomUpdatedMetadata.GroupDetailsUpdatedMetadata',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='new_group_details', full_name='RoomUpdatedMetadata.GroupDetailsUpdatedMetadata.new_group_details', index=0,
      number=1, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='prev_group_details', full_name='RoomUpdatedMetadata.GroupDetailsUpdatedMetadata.prev_group_details', index=1,
      number=2, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=29017,
  serialized_end=29131,
)

_ROOMUPDATEDMETADATA = _descriptor.Descriptor(
  name='RoomUpdatedMetadata',
  full_name='RoomUpdatedMetadata',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='name', full_name='RoomUpdatedMetadata.name', index=0,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='visibility', full_name='RoomUpdatedMetadata.visibility', index=1,
      number=3, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='group_link_sharing_enabled', full_name='RoomUpdatedMetadata.group_link_sharing_enabled', index=2,
      number=5, type=8, cpp_type=7, label=1,
      has_default_value=False, default_value=False,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='rename_metadata', full_name='RoomUpdatedMetadata.rename_metadata', index=3,
      number=6, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='group_details_metadata', full_name='RoomUpdatedMetadata.group_details_metadata', index=4,
      number=7, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='initiator', full_name='RoomUpdatedMetadata.initiator', index=5,
      number=4, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[_ROOMUPDATEDMETADATA_ROOMRENAMEMETADATA, _ROOMUPDATEDMETADATA_GROUPDETAILSUPDATEDMETADATA, ],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=28673,
  serialized_end=29131,
)


_MEETINGSPACE_PHONEACCESS = _descriptor.Descriptor(
  name='PhoneAccess',
  full_name='MeetingSpace.PhoneAccess',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='phone_number', full_name='MeetingSpace.PhoneAccess.phone_number', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='formatted_phone_number', full_name='MeetingSpace.PhoneAccess.formatted_phone_number', index=1,
      number=5, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='pin', full_name='MeetingSpace.PhoneAccess.pin', index=2,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='region_code', full_name='MeetingSpace.PhoneAccess.region_code', index=3,
      number=3, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='language_code', full_name='MeetingSpace.PhoneAccess.language_code', index=4,
      number=4, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=29688,
  serialized_end=29812,
)

_MEETINGSPACE_UNIVERSALPHONEACCESS = _descriptor.Descriptor(
  name='UniversalPhoneAccess',
  full_name='MeetingSpace.UniversalPhoneAccess',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='pin', full_name='MeetingSpace.UniversalPhoneAccess.pin', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='pstn_info_url', full_name='MeetingSpace.UniversalPhoneAccess.pstn_info_url', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=29814,
  serialized_end=29872,
)

_MEETINGSPACE_PRESENTER = _descriptor.Descriptor(
  name='Presenter',
  full_name='MeetingSpace.Presenter',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='presenter_device_id', full_name='MeetingSpace.Presenter.presenter_device_id', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='by_device_id', full_name='MeetingSpace.Presenter.by_device_id', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=29874,
  serialized_end=29936,
)

_MEETINGSPACE_RECORDINGINFO_RECORDINGEVENT = _descriptor.Descriptor(
  name='RecordingEvent',
  full_name='MeetingSpace.RecordingInfo.RecordingEvent',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='device_id', full_name='MeetingSpace.RecordingInfo.RecordingEvent.device_id', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='type', full_name='MeetingSpace.RecordingInfo.RecordingEvent.type', index=1,
      number=2, type=14, cpp_type=8, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
    _MEETINGSPACE_RECORDINGINFO_RECORDINGEVENT_EVENTTYPE,
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=30245,
  serialized_end=30411,
)

_MEETINGSPACE_RECORDINGINFO = _descriptor.Descriptor(
  name='RecordingInfo',
  full_name='MeetingSpace.RecordingInfo',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='recording_status', full_name='MeetingSpace.RecordingInfo.recording_status', index=0,
      number=1, type=14, cpp_type=8, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='recording_id', full_name='MeetingSpace.RecordingInfo.recording_id', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='producer_device_id', full_name='MeetingSpace.RecordingInfo.producer_device_id', index=2,
      number=3, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='latest_recording_event', full_name='MeetingSpace.RecordingInfo.latest_recording_event', index=3,
      number=4, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='owner_display_name', full_name='MeetingSpace.RecordingInfo.owner_display_name', index=4,
      number=5, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='recording_application_type', full_name='MeetingSpace.RecordingInfo.recording_application_type', index=5,
      number=6, type=14, cpp_type=8, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[_MEETINGSPACE_RECORDINGINFO_RECORDINGEVENT, ],
  enum_types=[
    _MEETINGSPACE_RECORDINGINFO_RECORDINGSTATUS,
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=29939,
  serialized_end=30528,
)

_MEETINGSPACE_STREAMINGSESSIONINFO_SESSIONEVENT = _descriptor.Descriptor(
  name='SessionEvent',
  full_name='MeetingSpace.StreamingSessionInfo.SessionEvent',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='device_id', full_name='MeetingSpace.StreamingSessionInfo.SessionEvent.device_id', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='type', full_name='MeetingSpace.StreamingSessionInfo.SessionEvent.type', index=1,
      number=2, type=14, cpp_type=8, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
    _MEETINGSPACE_STREAMINGSESSIONINFO_SESSIONEVENT_EVENTTYPE,
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=30951,
  serialized_end=31093,
)

_MEETINGSPACE_STREAMINGSESSIONINFO_STREAMVIEWERSTATS = _descriptor.Descriptor(
  name='StreamViewerStats',
  full_name='MeetingSpace.StreamingSessionInfo.StreamViewerStats',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='estimated_viewer_count', full_name='MeetingSpace.StreamingSessionInfo.StreamViewerStats.estimated_viewer_count', index=0,
      number=1, type=3, cpp_type=2, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=31095,
  serialized_end=31146,
)

_MEETINGSPACE_STREAMINGSESSIONINFO = _descriptor.Descriptor(
  name='StreamingSessionInfo',
  full_name='MeetingSpace.StreamingSessionInfo',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='status', full_name='MeetingSpace.StreamingSessionInfo.status', index=0,
      number=1, type=14, cpp_type=8, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='session_id', full_name='MeetingSpace.StreamingSessionInfo.session_id', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='application_type', full_name='MeetingSpace.StreamingSessionInfo.application_type', index=2,
      number=3, type=14, cpp_type=8, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='latest_session_event', full_name='MeetingSpace.StreamingSessionInfo.latest_session_event', index=3,
      number=4, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='owner_display_name', full_name='MeetingSpace.StreamingSessionInfo.owner_display_name', index=4,
      number=5, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='viewer_access_policy', full_name='MeetingSpace.StreamingSessionInfo.viewer_access_policy', index=5,
      number=6, type=14, cpp_type=8, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='training_enabled', full_name='MeetingSpace.StreamingSessionInfo.training_enabled', index=6,
      number=7, type=8, cpp_type=7, label=1,
      has_default_value=False, default_value=False,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='viewer_stats', full_name='MeetingSpace.StreamingSessionInfo.viewer_stats', index=7,
      number=8, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[_MEETINGSPACE_STREAMINGSESSIONINFO_SESSIONEVENT, _MEETINGSPACE_STREAMINGSESSIONINFO_STREAMVIEWERSTATS, ],
  enum_types=[
    _MEETINGSPACE_STREAMINGSESSIONINFO_STATUS,
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=30531,
  serialized_end=31218,
)

_MEETINGSPACE_CALLINFO_CALLSETTINGS = _descriptor.Descriptor(
  name='CallSettings',
  full_name='MeetingSpace.CallInfo.CallSettings',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='access_lock', full_name='MeetingSpace.CallInfo.CallSettings.access_lock', index=0,
      number=1, type=8, cpp_type=7, label=1,
      has_default_value=False, default_value=False,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='chat_lock', full_name='MeetingSpace.CallInfo.CallSettings.chat_lock', index=1,
      number=2, type=8, cpp_type=7, label=1,
      has_default_value=False, default_value=False,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='present_lock', full_name='MeetingSpace.CallInfo.CallSettings.present_lock', index=2,
      number=3, type=8, cpp_type=7, label=1,
      has_default_value=False, default_value=False,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='attendance_report_enabled', full_name='MeetingSpace.CallInfo.CallSettings.attendance_report_enabled', index=3,
      number=4, type=8, cpp_type=7, label=1,
      has_default_value=False, default_value=False,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='audio_lock', full_name='MeetingSpace.CallInfo.CallSettings.audio_lock', index=4,
      number=5, type=8, cpp_type=7, label=1,
      has_default_value=False, default_value=False,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='video_lock', full_name='MeetingSpace.CallInfo.CallSettings.video_lock', index=5,
      number=6, type=8, cpp_type=7, label=1,
      has_default_value=False, default_value=False,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='moderation_enabled', full_name='MeetingSpace.CallInfo.CallSettings.moderation_enabled', index=6,
      number=7, type=8, cpp_type=7, label=1,
      has_default_value=False, default_value=False,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='cse_enabled', full_name='MeetingSpace.CallInfo.CallSettings.cse_enabled', index=7,
      number=8, type=8, cpp_type=7, label=1,
      has_default_value=False, default_value=False,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=31705,
  serialized_end=31905,
)

_MEETINGSPACE_CALLINFO_PAYGATEINFO = _descriptor.Descriptor(
  name='PaygateInfo',
  full_name='MeetingSpace.CallInfo.PaygateInfo',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='show_upgrade_promos', full_name='MeetingSpace.CallInfo.PaygateInfo.show_upgrade_promos', index=0,
      number=1, type=8, cpp_type=7, label=1,
      has_default_value=False, default_value=False,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='call_ending_soon_warning_time', full_name='MeetingSpace.CallInfo.PaygateInfo.call_ending_soon_warning_time', index=1,
      number=2, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='call_ending_time', full_name='MeetingSpace.CallInfo.PaygateInfo.call_ending_time', index=2,
      number=3, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=31908,
  serialized_end=32073,
)

_MEETINGSPACE_CALLINFO_CSEINFO = _descriptor.Descriptor(
  name='CseInfo',
  full_name='MeetingSpace.CallInfo.CseInfo',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='wrapped_key', full_name='MeetingSpace.CallInfo.CseInfo.wrapped_key', index=0,
      number=1, type=12, cpp_type=9, label=1,
      has_default_value=False, default_value=b"",
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=32075,
  serialized_end=32105,
)

_MEETINGSPACE_CALLINFO = _descriptor.Descriptor(
  name='CallInfo',
  full_name='MeetingSpace.CallInfo',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='presenter', full_name='MeetingSpace.CallInfo.presenter', index=0,
      number=1, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='recording_info', full_name='MeetingSpace.CallInfo.recording_info', index=1,
      number=2, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='calendar_event_id', full_name='MeetingSpace.CallInfo.calendar_event_id', index=2,
      number=3, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='organization_name', full_name='MeetingSpace.CallInfo.organization_name', index=3,
      number=4, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='max_joined_devices', full_name='MeetingSpace.CallInfo.max_joined_devices', index=4,
      number=6, type=5, cpp_type=1, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='media_backend_info', full_name='MeetingSpace.CallInfo.media_backend_info', index=5,
      number=8, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='streaming_sessions', full_name='MeetingSpace.CallInfo.streaming_sessions', index=6,
      number=9, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='settings', full_name='MeetingSpace.CallInfo.settings', index=7,
      number=11, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='paygate_info', full_name='MeetingSpace.CallInfo.paygate_info', index=8,
      number=13, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='supported_caption_languages', full_name='MeetingSpace.CallInfo.supported_caption_languages', index=9,
      number=14, type=9, cpp_type=9, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='cse_info', full_name='MeetingSpace.CallInfo.cse_info', index=10,
      number=15, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[_MEETINGSPACE_CALLINFO_CALLSETTINGS, _MEETINGSPACE_CALLINFO_PAYGATEINFO, _MEETINGSPACE_CALLINFO_CSEINFO, ],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=31221,
  serialized_end=32105,
)

_MEETINGSPACE_GATEWAYACCESS = _descriptor.Descriptor(
  name='GatewayAccess',
  full_name='MeetingSpace.GatewayAccess',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='enabled', full_name='MeetingSpace.GatewayAccess.enabled', index=0,
      number=1, type=8, cpp_type=7, label=1,
      has_default_value=False, default_value=False,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=32107,
  serialized_end=32139,
)

_MEETINGSPACE_GATEWAYSIPACCESS = _descriptor.Descriptor(
  name='GatewaySipAccess',
  full_name='MeetingSpace.GatewaySipAccess',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='uri', full_name='MeetingSpace.GatewaySipAccess.uri', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='sip_access_code', full_name='MeetingSpace.GatewaySipAccess.sip_access_code', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=32141,
  serialized_end=32197,
)

_MEETINGSPACE_BROADCASTACCESS = _descriptor.Descriptor(
  name='BroadcastAccess',
  full_name='MeetingSpace.BroadcastAccess',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='view_url', full_name='MeetingSpace.BroadcastAccess.view_url', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=32199,
  serialized_end=32234,
)

_MEETINGSPACE_SETTINGS = _descriptor.Descriptor(
  name='Settings',
  full_name='MeetingSpace.Settings',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='access_lock', full_name='MeetingSpace.Settings.access_lock', index=0,
      number=1, type=8, cpp_type=7, label=1,
      has_default_value=False, default_value=False,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='attendance_report_enabled', full_name='MeetingSpace.Settings.attendance_report_enabled', index=1,
      number=2, type=8, cpp_type=7, label=1,
      has_default_value=False, default_value=False,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='chat_lock', full_name='MeetingSpace.Settings.chat_lock', index=2,
      number=3, type=8, cpp_type=7, label=1,
      has_default_value=False, default_value=False,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='present_lock', full_name='MeetingSpace.Settings.present_lock', index=3,
      number=4, type=8, cpp_type=7, label=1,
      has_default_value=False, default_value=False,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='moderation_enabled', full_name='MeetingSpace.Settings.moderation_enabled', index=4,
      number=5, type=8, cpp_type=7, label=1,
      has_default_value=False, default_value=False,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='cse_enabled', full_name='MeetingSpace.Settings.cse_enabled', index=5,
      number=6, type=8, cpp_type=7, label=1,
      has_default_value=False, default_value=False,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=32237,
  serialized_end=32393,
)

_MEETINGSPACE = _descriptor.Descriptor(
  name='MeetingSpace',
  full_name='MeetingSpace',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='meeting_space_id', full_name='MeetingSpace.meeting_space_id', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='meeting_code', full_name='MeetingSpace.meeting_code', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='meeting_url', full_name='MeetingSpace.meeting_url', index=2,
      number=3, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='phone_access', full_name='MeetingSpace.phone_access', index=3,
      number=5, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='universal_phone_access', full_name='MeetingSpace.universal_phone_access', index=4,
      number=8, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='call_info', full_name='MeetingSpace.call_info', index=5,
      number=6, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='meeting_alias', full_name='MeetingSpace.meeting_alias', index=6,
      number=7, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='gateway_access', full_name='MeetingSpace.gateway_access', index=7,
      number=9, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='more_join_url', full_name='MeetingSpace.more_join_url', index=8,
      number=10, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='accepted_number_class', full_name='MeetingSpace.accepted_number_class', index=9,
      number=11, type=14, cpp_type=8, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='gateway_sip_access', full_name='MeetingSpace.gateway_sip_access', index=10,
      number=13, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='broadcast_access', full_name='MeetingSpace.broadcast_access', index=11,
      number=14, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='settings', full_name='MeetingSpace.settings', index=12,
      number=15, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[_MEETINGSPACE_PHONEACCESS, _MEETINGSPACE_UNIVERSALPHONEACCESS, _MEETINGSPACE_PRESENTER, _MEETINGSPACE_RECORDINGINFO, _MEETINGSPACE_STREAMINGSESSIONINFO, _MEETINGSPACE_CALLINFO, _MEETINGSPACE_GATEWAYACCESS, _MEETINGSPACE_GATEWAYSIPACCESS, _MEETINGSPACE_BROADCASTACCESS, _MEETINGSPACE_SETTINGS, ],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=29134,
  serialized_end=32393,
)


_VIDEOCALLMETADATA = _descriptor.Descriptor(
  name='VideoCallMetadata',
  full_name='VideoCallMetadata',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='meeting_space', full_name='VideoCallMetadata.meeting_space', index=0,
      number=1, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='was_created_in_current_group', full_name='VideoCallMetadata.was_created_in_current_group', index=1,
      number=2, type=8, cpp_type=7, label=1,
      has_default_value=False, default_value=False,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='should_not_render', full_name='VideoCallMetadata.should_not_render', index=2,
      number=3, type=8, cpp_type=7, label=1,
      has_default_value=False, default_value=False,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=32395,
  serialized_end=32517,
)


_MEMBERSHIPCHANGEDMETADATA_AFFECTEDMEMBERSHIP = _descriptor.Descriptor(
  name='AffectedMembership',
  full_name='MembershipChangedMetadata.AffectedMembership',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='affected_member', full_name='MembershipChangedMetadata.AffectedMembership.affected_member', index=0,
      number=1, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='prior_membership_state', full_name='MembershipChangedMetadata.AffectedMembership.prior_membership_state', index=1,
      number=2, type=14, cpp_type=8, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='prior_membership_role', full_name='MembershipChangedMetadata.AffectedMembership.prior_membership_role', index=2,
      number=3, type=14, cpp_type=8, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='target_membership_role', full_name='MembershipChangedMetadata.AffectedMembership.target_membership_role', index=3,
      number=4, type=14, cpp_type=8, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=32816,
  serialized_end=33019,
)

_MEMBERSHIPCHANGEDMETADATA = _descriptor.Descriptor(
  name='MembershipChangedMetadata',
  full_name='MembershipChangedMetadata',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='type', full_name='MembershipChangedMetadata.type', index=0,
      number=1, type=14, cpp_type=8, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='affected_memberships', full_name='MembershipChangedMetadata.affected_memberships', index=1,
      number=6, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='initiator', full_name='MembershipChangedMetadata.initiator', index=2,
      number=2, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='affected_members', full_name='MembershipChangedMetadata.affected_members', index=3,
      number=3, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='initiator_profile', full_name='MembershipChangedMetadata.initiator_profile', index=4,
      number=4, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='affected_member_profiles', full_name='MembershipChangedMetadata.affected_member_profiles', index=5,
      number=5, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[_MEMBERSHIPCHANGEDMETADATA_AFFECTEDMEMBERSHIP, ],
  enum_types=[
    _MEMBERSHIPCHANGEDMETADATA_TYPE,
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=32520,
  serialized_end=33191,
)


_USERMENTIONMETADATA = _descriptor.Descriptor(
  name='UserMentionMetadata',
  full_name='UserMentionMetadata',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='id', full_name='UserMentionMetadata.id', index=0,
      number=1, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='invitee_info', full_name='UserMentionMetadata.invitee_info', index=1,
      number=3, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='type', full_name='UserMentionMetadata.type', index=2,
      number=2, type=14, cpp_type=8, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='display_name', full_name='UserMentionMetadata.display_name', index=3,
      number=4, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='gender', full_name='UserMentionMetadata.gender', index=4,
      number=5, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
    _USERMENTIONMETADATA_TYPE,
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=33194,
  serialized_end=33456,
)


_SLASHCOMMANDMETADATA = _descriptor.Descriptor(
  name='SlashCommandMetadata',
  full_name='SlashCommandMetadata',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='id', full_name='SlashCommandMetadata.id', index=0,
      number=1, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='type', full_name='SlashCommandMetadata.type', index=1,
      number=2, type=14, cpp_type=8, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='command_name', full_name='SlashCommandMetadata.command_name', index=2,
      number=3, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='command_id', full_name='SlashCommandMetadata.command_id', index=3,
      number=4, type=3, cpp_type=2, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='arguments_hint', full_name='SlashCommandMetadata.arguments_hint', index=4,
      number=5, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='triggers_dialog', full_name='SlashCommandMetadata.triggers_dialog', index=5,
      number=6, type=8, cpp_type=7, label=1,
      has_default_value=False, default_value=False,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
    _SLASHCOMMANDMETADATA_TYPE,
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=33459,
  serialized_end=33705,
)


_GROUPVISIBILITY = _descriptor.Descriptor(
  name='GroupVisibility',
  full_name='GroupVisibility',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
    _GROUPVISIBILITY_VISIBILITYSTATE,
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=33707,
  serialized_end=33781,
)


_COMGOOGLEPROTOBUFTIMESTAMP = _descriptor.Descriptor(
  name='ComGoogleProtobufTimestamp',
  full_name='ComGoogleProtobufTimestamp',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='seconds', full_name='ComGoogleProtobufTimestamp.seconds', index=0,
      number=1, type=3, cpp_type=2, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='nanos', full_name='ComGoogleProtobufTimestamp.nanos', index=1,
      number=2, type=5, cpp_type=1, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=33783,
  serialized_end=33843,
)


_GETSERVERTIMEREQUEST = _descriptor.Descriptor(
  name='GetServerTimeRequest',
  full_name='GetServerTimeRequest',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='request_header', full_name='GetServerTimeRequest.request_header', index=0,
      number=100, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=33845,
  serialized_end=33907,
)


_GETSERVERTIMERESPONSE = _descriptor.Descriptor(
  name='GetServerTimeResponse',
  full_name='GetServerTimeResponse',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='timestamp', full_name='GetServerTimeResponse.timestamp', index=0,
      number=1, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=33909,
  serialized_end=33980,
)


_CATCHUPRANGE = _descriptor.Descriptor(
  name='CatchUpRange',
  full_name='CatchUpRange',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='from_revision_timestamp', full_name='CatchUpRange.from_revision_timestamp', index=0,
      number=1, type=3, cpp_type=2, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='to_revision_timestamp', full_name='CatchUpRange.to_revision_timestamp', index=1,
      number=2, type=3, cpp_type=2, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=33982,
  serialized_end=34060,
)


_CATCHUPGROUPREQUEST = _descriptor.Descriptor(
  name='CatchUpGroupRequest',
  full_name='CatchUpGroupRequest',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='request_header', full_name='CatchUpGroupRequest.request_header', index=0,
      number=100, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='group_id', full_name='CatchUpGroupRequest.group_id', index=1,
      number=1, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='range', full_name='CatchUpGroupRequest.range', index=2,
      number=2, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='page_size', full_name='CatchUpGroupRequest.page_size', index=3,
      number=3, type=5, cpp_type=1, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='cutoff_size', full_name='CatchUpGroupRequest.cutoff_size', index=4,
      number=4, type=5, cpp_type=1, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=34063,
  serialized_end=34222,
)


_CATCHUPUSERREQUEST = _descriptor.Descriptor(
  name='CatchUpUserRequest',
  full_name='CatchUpUserRequest',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='request_header', full_name='CatchUpUserRequest.request_header', index=0,
      number=100, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='range', full_name='CatchUpUserRequest.range', index=1,
      number=1, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='page_size', full_name='CatchUpUserRequest.page_size', index=2,
      number=2, type=5, cpp_type=1, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='cutoff_size', full_name='CatchUpUserRequest.cutoff_size', index=3,
      number=3, type=5, cpp_type=1, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=34225,
  serialized_end=34355,
)


_CATCHUPRESPONSE = _descriptor.Descriptor(
  name='CatchUpResponse',
  full_name='CatchUpResponse',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='events', full_name='CatchUpResponse.events', index=0,
      number=1, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='status', full_name='CatchUpResponse.status', index=1,
      number=2, type=14, cpp_type=8, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
    _CATCHUPRESPONSE_RESPONSESTATUS,
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=34358,
  serialized_end=34606,
)


_GETGROUPREQUEST = _descriptor.Descriptor(
  name='GetGroupRequest',
  full_name='GetGroupRequest',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='request_header', full_name='GetGroupRequest.request_header', index=0,
      number=100, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='group_id', full_name='GetGroupRequest.group_id', index=1,
      number=1, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='fetch_options', full_name='GetGroupRequest.fetch_options', index=2,
      number=4, type=14, cpp_type=8, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='user_not_older_than', full_name='GetGroupRequest.user_not_older_than', index=3,
      number=2, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='group_not_older_than', full_name='GetGroupRequest.group_not_older_than', index=4,
      number=3, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='include_invite_dms', full_name='GetGroupRequest.include_invite_dms', index=5,
      number=5, type=8, cpp_type=7, label=1,
      has_default_value=False, default_value=False,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
    _GETGROUPREQUEST_FETCHOPTIONS,
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=34609,
  serialized_end=35062,
)


_GETGROUPRESPONSE = _descriptor.Descriptor(
  name='GetGroupResponse',
  full_name='GetGroupResponse',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='group', full_name='GetGroupResponse.group', index=0,
      number=1, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='memberships', full_name='GetGroupResponse.memberships', index=1,
      number=4, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='user_revision', full_name='GetGroupResponse.user_revision', index=2,
      number=2, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='group_revision', full_name='GetGroupResponse.group_revision', index=3,
      number=3, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='membership_state', full_name='GetGroupResponse.membership_state', index=4,
      number=6, type=14, cpp_type=8, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='joined_member_ids', full_name='GetGroupResponse.joined_member_ids', index=5,
      number=7, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='invited_member_ids', full_name='GetGroupResponse.invited_member_ids', index=6,
      number=8, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='read_receipt_set', full_name='GetGroupResponse.read_receipt_set', index=7,
      number=9, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='snippet', full_name='GetGroupResponse.snippet', index=8,
      number=10, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=35065,
  serialized_end=35408,
)


_WORLDSECTION = _descriptor.Descriptor(
  name='WorldSection',
  full_name='WorldSection',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='world_section_type', full_name='WorldSection.world_section_type', index=0,
      number=1, type=14, cpp_type=8, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
    _WORLDSECTION_WORLDSECTIONTYPE,
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=35411,
  serialized_end=36059,
)


_WORLDFILTER = _descriptor.Descriptor(
  name='WorldFilter',
  full_name='WorldFilter',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='starred_state', full_name='WorldFilter.starred_state', index=0,
      number=1, type=14, cpp_type=8, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='visibility_state', full_name='WorldFilter.visibility_state', index=1,
      number=2, type=14, cpp_type=8, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='read_state', full_name='WorldFilter.read_state', index=2,
      number=7, type=14, cpp_type=8, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='block_state', full_name='WorldFilter.block_state', index=3,
      number=8, type=14, cpp_type=8, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='named_state', full_name='WorldFilter.named_state', index=4,
      number=9, type=14, cpp_type=8, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='membership_state', full_name='WorldFilter.membership_state', index=5,
      number=3, type=14, cpp_type=8, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='invite_category', full_name='WorldFilter.invite_category', index=6,
      number=4, type=14, cpp_type=8, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='member_type', full_name='WorldFilter.member_type', index=7,
      number=5, type=14, cpp_type=8, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='group_type', full_name='WorldFilter.group_type', index=8,
      number=6, type=14, cpp_type=8, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
    _WORLDFILTER_STARREDSTATE,
    _WORLDFILTER_VISIBILITYSTATE,
    _WORLDFILTER_READSTATE,
    _WORLDFILTER_BLOCKSTATE,
    _WORLDFILTER_NAMEDSTATE,
    _WORLDFILTER_MEMBERTYPE,
    _WORLDFILTER_GROUPTYPE,
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=36062,
  serialized_end=37000,
)


_NAMEUSERS = _descriptor.Descriptor(
  name='NameUsers',
  full_name='NameUsers',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='name_user_ids', full_name='NameUsers.name_user_ids', index=0,
      number=1, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='has_more_name_users', full_name='NameUsers.has_more_name_users', index=1,
      number=2, type=8, cpp_type=7, label=1,
      has_default_value=False, default_value=False,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='group_name', full_name='NameUsers.group_name', index=2,
      number=3, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=37002,
  serialized_end=37094,
)


_WORLDITEMLITE_MEMBERSHIPLITE = _descriptor.Descriptor(
  name='MembershipLite',
  full_name='WorldItemLite.MembershipLite',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='user_id', full_name='WorldItemLite.MembershipLite.user_id', index=0,
      number=1, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='membership_state', full_name='WorldItemLite.MembershipLite.membership_state', index=1,
      number=2, type=14, cpp_type=8, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=37815,
  serialized_end=37901,
)

_WORLDITEMLITE_DMMEMBERS = _descriptor.Descriptor(
  name='DmMembers',
  full_name='WorldItemLite.DmMembers',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='members', full_name='WorldItemLite.DmMembers.members', index=0,
      number=1, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='memberships', full_name='WorldItemLite.DmMembers.memberships', index=1,
      number=2, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='has_all_memberships', full_name='WorldItemLite.DmMembers.has_all_memberships', index=2,
      number=3, type=8, cpp_type=7, label=1,
      has_default_value=False, default_value=False,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=37903,
  serialized_end=38021,
)

_WORLDITEMLITE_GROUPLITE = _descriptor.Descriptor(
  name='GroupLite',
  full_name='WorldItemLite.GroupLite',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='creator_id', full_name='WorldItemLite.GroupLite.creator_id', index=0,
      number=1, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='create_time', full_name='WorldItemLite.GroupLite.create_time', index=1,
      number=2, type=3, cpp_type=2, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='is_flat', full_name='WorldItemLite.GroupLite.is_flat', index=2,
      number=3, type=8, cpp_type=7, label=1,
      has_default_value=False, default_value=False,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='retention_settings', full_name='WorldItemLite.GroupLite.retention_settings', index=3,
      number=4, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='interop_enabled', full_name='WorldItemLite.GroupLite.interop_enabled', index=4,
      number=5, type=8, cpp_type=7, label=1,
      has_default_value=False, default_value=False,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='roster_email', full_name='WorldItemLite.GroupLite.roster_email', index=5,
      number=7, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=38024,
  serialized_end=38197,
)

_WORLDITEMLITE_FLATGROUP = _descriptor.Descriptor(
  name='FlatGroup',
  full_name='WorldItemLite.FlatGroup',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=17747,
  serialized_end=17758,
)

_WORLDITEMLITE_THREADEDGROUP = _descriptor.Descriptor(
  name='ThreadedGroup',
  full_name='WorldItemLite.ThreadedGroup',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=17760,
  serialized_end=17775,
)

_WORLDITEMLITE = _descriptor.Descriptor(
  name='WorldItemLite',
  full_name='WorldItemLite',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='group_id', full_name='WorldItemLite.group_id', index=0,
      number=1, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='group_revision', full_name='WorldItemLite.group_revision', index=1,
      number=2, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='sort_timestamp', full_name='WorldItemLite.sort_timestamp', index=2,
      number=3, type=3, cpp_type=2, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='read_state', full_name='WorldItemLite.read_state', index=3,
      number=4, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='room_name', full_name='WorldItemLite.room_name', index=4,
      number=5, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='dm_members', full_name='WorldItemLite.dm_members', index=5,
      number=6, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='name_users', full_name='WorldItemLite.name_users', index=6,
      number=20, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='retention_horizon_time_micros', full_name='WorldItemLite.retention_horizon_time_micros', index=7,
      number=8, type=3, cpp_type=2, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='retention_duration_seconds', full_name='WorldItemLite.retention_duration_seconds', index=8,
      number=9, type=3, cpp_type=2, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='group_lite', full_name='WorldItemLite.group_lite', index=9,
      number=7, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='message', full_name='WorldItemLite.message', index=10,
      number=13, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='is_message_blocked', full_name='WorldItemLite.is_message_blocked', index=11,
      number=25, type=8, cpp_type=7, label=1,
      has_default_value=False, default_value=False,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='flat_group', full_name='WorldItemLite.flat_group', index=12,
      number=14, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='threaded_group', full_name='WorldItemLite.threaded_group', index=13,
      number=15, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='avatar_url', full_name='WorldItemLite.avatar_url', index=14,
      number=16, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='attribute_checker_group_type', full_name='WorldItemLite.attribute_checker_group_type', index=15,
      number=19, type=14, cpp_type=8, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='group_support_level', full_name='WorldItemLite.group_support_level', index=16,
      number=23, type=14, cpp_type=8, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='group_unsupported_reason', full_name='WorldItemLite.group_unsupported_reason', index=17,
      number=24, type=14, cpp_type=8, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[_WORLDITEMLITE_MEMBERSHIPLITE, _WORLDITEMLITE_DMMEMBERS, _WORLDITEMLITE_GROUPLITE, _WORLDITEMLITE_FLATGROUP, _WORLDITEMLITE_THREADEDGROUP, ],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=37097,
  serialized_end=38227,
)


_WORLDSECTIONREQUEST = _descriptor.Descriptor(
  name='WorldSectionRequest',
  full_name='WorldSectionRequest',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='page_size', full_name='WorldSectionRequest.page_size', index=0,
      number=1, type=5, cpp_type=1, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='world_section', full_name='WorldSectionRequest.world_section', index=1,
      number=2, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='world_filter', full_name='WorldSectionRequest.world_filter', index=2,
      number=4, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='num_world_items_with_snippet', full_name='WorldSectionRequest.num_world_items_with_snippet', index=3,
      number=5, type=5, cpp_type=1, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='anchor_sort_timestamp_micros', full_name='WorldSectionRequest.anchor_sort_timestamp_micros', index=4,
      number=3, type=3, cpp_type=2, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='pagination_token', full_name='WorldSectionRequest.pagination_token', index=5,
      number=6, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=38230,
  serialized_end=38446,
)


_WORLDSECTIONRESPONSE = _descriptor.Descriptor(
  name='WorldSectionResponse',
  full_name='WorldSectionResponse',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='world_section', full_name='WorldSectionResponse.world_section', index=0,
      number=1, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='world_filter', full_name='WorldSectionResponse.world_filter', index=1,
      number=4, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='world_items', full_name='WorldSectionResponse.world_items', index=2,
      number=2, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='next_page_anchor_sort_timestamp_micros', full_name='WorldSectionResponse.next_page_anchor_sort_timestamp_micros', index=3,
      number=3, type=3, cpp_type=2, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='has_more_items', full_name='WorldSectionResponse.has_more_items', index=4,
      number=5, type=8, cpp_type=7, label=1,
      has_default_value=False, default_value=False,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='pagination_token', full_name='WorldSectionResponse.pagination_token', index=5,
      number=6, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=38449,
  serialized_end=38680,
)


_PAGINATEDWORLDREQUEST = _descriptor.Descriptor(
  name='PaginatedWorldRequest',
  full_name='PaginatedWorldRequest',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='request_header', full_name='PaginatedWorldRequest.request_header', index=0,
      number=1, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='world_section_requests', full_name='PaginatedWorldRequest.world_section_requests', index=1,
      number=2, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='world_consistency_token', full_name='PaginatedWorldRequest.world_consistency_token', index=2,
      number=3, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='fetch_options', full_name='PaginatedWorldRequest.fetch_options', index=3,
      number=4, type=14, cpp_type=8, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='fetch_from_user_spaces', full_name='PaginatedWorldRequest.fetch_from_user_spaces', index=4,
      number=5, type=8, cpp_type=7, label=1,
      has_default_value=False, default_value=False,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='receive_world_update_notifications', full_name='PaginatedWorldRequest.receive_world_update_notifications', index=5,
      number=6, type=8, cpp_type=7, label=1,
      has_default_value=False, default_value=False,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='fetch_snippets_for_unnamed_rooms', full_name='PaginatedWorldRequest.fetch_snippets_for_unnamed_rooms', index=6,
      number=7, type=8, cpp_type=7, label=1,
      has_default_value=False, default_value=False,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
    _PAGINATEDWORLDREQUEST_FETCHOPTIONS,
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=38683,
  serialized_end=39161,
)


_PAGINATEDWORLDRESPONSE = _descriptor.Descriptor(
  name='PaginatedWorldResponse',
  full_name='PaginatedWorldResponse',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='world_section_responses', full_name='PaginatedWorldResponse.world_section_responses', index=0,
      number=1, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='world_consistency_token', full_name='PaginatedWorldResponse.world_consistency_token', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='user_revision', full_name='PaginatedWorldResponse.user_revision', index=2,
      number=3, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='world_items', full_name='PaginatedWorldResponse.world_items', index=3,
      number=4, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=39164,
  serialized_end=39352,
)


_REMOVEMEMBERSHIPSREQUEST = _descriptor.Descriptor(
  name='RemoveMembershipsRequest',
  full_name='RemoveMembershipsRequest',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='request_header', full_name='RemoveMembershipsRequest.request_header', index=0,
      number=100, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='member_ids', full_name='RemoveMembershipsRequest.member_ids', index=1,
      number=1, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='group_id', full_name='RemoveMembershipsRequest.group_id', index=2,
      number=2, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='membership_state', full_name='RemoveMembershipsRequest.membership_state', index=3,
      number=3, type=14, cpp_type=8, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=39355,
  serialized_end=39524,
)


_REMOVEMEMBERSHIPRESULT = _descriptor.Descriptor(
  name='RemoveMembershipResult',
  full_name='RemoveMembershipResult',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='request_header', full_name='RemoveMembershipResult.request_header', index=0,
      number=100, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='member_id', full_name='RemoveMembershipResult.member_id', index=1,
      number=1, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='attached_roster_names', full_name='RemoveMembershipResult.attached_roster_names', index=2,
      number=4, type=9, cpp_type=9, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=39526,
  serialized_end=39651,
)


_REMOVEMEMBERSHIPSRESPONSE = _descriptor.Descriptor(
  name='RemoveMembershipsResponse',
  full_name='RemoveMembershipsResponse',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='results', full_name='RemoveMembershipsResponse.results', index=0,
      number=1, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='retention_settings', full_name='RemoveMembershipsResponse.retention_settings', index=1,
      number=3, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='retention_settings_group_revision', full_name='RemoveMembershipsResponse.retention_settings_group_revision', index=2,
      number=4, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=39654,
  serialized_end=39830,
)


_HIDEGROUPREQUEST = _descriptor.Descriptor(
  name='HideGroupRequest',
  full_name='HideGroupRequest',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='request_header', full_name='HideGroupRequest.request_header', index=0,
      number=100, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='id', full_name='HideGroupRequest.id', index=1,
      number=1, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='hide', full_name='HideGroupRequest.hide', index=2,
      number=2, type=8, cpp_type=7, label=1,
      has_default_value=False, default_value=False,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=39832,
  serialized_end=39926,
)


_HIDEGROUPRESPONSE = _descriptor.Descriptor(
  name='HideGroupResponse',
  full_name='HideGroupResponse',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='read_state', full_name='HideGroupResponse.read_state', index=0,
      number=1, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='user_revision', full_name='HideGroupResponse.user_revision', index=1,
      number=3, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=39928,
  serialized_end=40023,
)


_INVITENOTIFICATIONSETTINGS = _descriptor.Descriptor(
  name='InviteNotificationSettings',
  full_name='InviteNotificationSettings',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='option', full_name='InviteNotificationSettings.option', index=0,
      number=1, type=14, cpp_type=8, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
    _INVITENOTIFICATIONSETTINGS_NOTIFICATIONOPTION,
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=40026,
  serialized_end=40181,
)


_CREATEMEMBERSHIPREQUEST = _descriptor.Descriptor(
  name='CreateMembershipRequest',
  full_name='CreateMembershipRequest',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='request_header', full_name='CreateMembershipRequest.request_header', index=0,
      number=100, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='member_ids', full_name='CreateMembershipRequest.member_ids', index=1,
      number=2, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='invitee_member_infos', full_name='CreateMembershipRequest.invitee_member_infos', index=2,
      number=8, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='membership_state', full_name='CreateMembershipRequest.membership_state', index=3,
      number=3, type=14, cpp_type=8, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='group_id', full_name='CreateMembershipRequest.group_id', index=4,
      number=4, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='notification_settings', full_name='CreateMembershipRequest.notification_settings', index=5,
      number=5, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=40184,
  serialized_end=40462,
)


_CREATEMEMBERSHIPRESPONSE = _descriptor.Descriptor(
  name='CreateMembershipResponse',
  full_name='CreateMembershipResponse',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='results', full_name='CreateMembershipResponse.results', index=0,
      number=2, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='group_revision', full_name='CreateMembershipResponse.group_revision', index=1,
      number=4, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='retention_settings', full_name='CreateMembershipResponse.retention_settings', index=2,
      number=5, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='retention_settings_group_revision', full_name='CreateMembershipResponse.retention_settings_group_revision', index=3,
      number=6, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=40465,
  serialized_end=40680,
)


_MARKGROUPREADSTATEREQUEST = _descriptor.Descriptor(
  name='MarkGroupReadstateRequest',
  full_name='MarkGroupReadstateRequest',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='request_header', full_name='MarkGroupReadstateRequest.request_header', index=0,
      number=100, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='id', full_name='MarkGroupReadstateRequest.id', index=1,
      number=1, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='last_read_time', full_name='MarkGroupReadstateRequest.last_read_time', index=2,
      number=2, type=3, cpp_type=2, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=40682,
  serialized_end=40795,
)


_MARKGROUPREADSTATERESPONSE = _descriptor.Descriptor(
  name='MarkGroupReadstateResponse',
  full_name='MarkGroupReadstateResponse',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='read_state', full_name='MarkGroupReadstateResponse.read_state', index=0,
      number=1, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='user_revision', full_name='MarkGroupReadstateResponse.user_revision', index=1,
      number=2, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=40797,
  serialized_end=40901,
)


_SETPRESENCESHAREDREQUEST = _descriptor.Descriptor(
  name='SetPresenceSharedRequest',
  full_name='SetPresenceSharedRequest',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='request_header', full_name='SetPresenceSharedRequest.request_header', index=0,
      number=100, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='presence_shared', full_name='SetPresenceSharedRequest.presence_shared', index=1,
      number=1, type=8, cpp_type=7, label=1,
      has_default_value=False, default_value=False,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=40903,
  serialized_end=40994,
)


_SETPRESENCESHAREDRESPONSE = _descriptor.Descriptor(
  name='SetPresenceSharedResponse',
  full_name='SetPresenceSharedResponse',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='user_status', full_name='SetPresenceSharedResponse.user_status', index=0,
      number=1, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='user_revision', full_name='SetPresenceSharedResponse.user_revision', index=1,
      number=2, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=40996,
  serialized_end=41096,
)


_SETDNDDURATIONREQUEST = _descriptor.Descriptor(
  name='SetDndDurationRequest',
  full_name='SetDndDurationRequest',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='request_header', full_name='SetDndDurationRequest.request_header', index=0,
      number=100, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='current_dnd_state', full_name='SetDndDurationRequest.current_dnd_state', index=1,
      number=2, type=14, cpp_type=8, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='new_dnd_duration_usec', full_name='SetDndDurationRequest.new_dnd_duration_usec', index=2,
      number=1, type=3, cpp_type=2, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='dnd_expiry_timestamp_usec', full_name='SetDndDurationRequest.dnd_expiry_timestamp_usec', index=3,
      number=3, type=3, cpp_type=2, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
    _SETDNDDURATIONREQUEST_STATE,
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=41099,
  serialized_end=41331,
)


_SETDNDDURATIONRESPONSE = _descriptor.Descriptor(
  name='SetDndDurationResponse',
  full_name='SetDndDurationResponse',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='user_status', full_name='SetDndDurationResponse.user_status', index=0,
      number=1, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='user_revision', full_name='SetDndDurationResponse.user_revision', index=1,
      number=2, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=41333,
  serialized_end=41430,
)


_UPDATEGROUPREQUEST = _descriptor.Descriptor(
  name='UpdateGroupRequest',
  full_name='UpdateGroupRequest',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='request_header', full_name='UpdateGroupRequest.request_header', index=0,
      number=100, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='space_id', full_name='UpdateGroupRequest.space_id', index=1,
      number=1, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='update_masks', full_name='UpdateGroupRequest.update_masks', index=2,
      number=4, type=14, cpp_type=8, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='name', full_name='UpdateGroupRequest.name', index=3,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='visibility', full_name='UpdateGroupRequest.visibility', index=4,
      number=5, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
    _UPDATEGROUPREQUEST_UPDATEMASK,
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=41433,
  serialized_end=41750,
)


_UPDATEGROUPRESPONSE = _descriptor.Descriptor(
  name='UpdateGroupResponse',
  full_name='UpdateGroupResponse',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='group', full_name='UpdateGroupResponse.group', index=0,
      number=1, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='group_revision', full_name='UpdateGroupResponse.group_revision', index=1,
      number=2, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=41752,
  serialized_end=41836,
)


_BLOCKENTITYREQUEST = _descriptor.Descriptor(
  name='BlockEntityRequest',
  full_name='BlockEntityRequest',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='request_header', full_name='BlockEntityRequest.request_header', index=0,
      number=100, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='user_id', full_name='BlockEntityRequest.user_id', index=1,
      number=1, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='group_id', full_name='BlockEntityRequest.group_id', index=2,
      number=2, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='blocked', full_name='BlockEntityRequest.blocked', index=3,
      number=3, type=8, cpp_type=7, label=1,
      has_default_value=False, default_value=False,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='reported', full_name='BlockEntityRequest.reported', index=4,
      number=4, type=8, cpp_type=7, label=1,
      has_default_value=False, default_value=False,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=41839,
  serialized_end=41988,
)


_BLOCKENTITYRESPONSE = _descriptor.Descriptor(
  name='BlockEntityResponse',
  full_name='BlockEntityResponse',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='read_state', full_name='BlockEntityResponse.read_state', index=0,
      number=1, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='user_revision', full_name='BlockEntityResponse.user_revision', index=1,
      number=2, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=41990,
  serialized_end=42087,
)


_SETCUSTOMSTATUSREQUEST = _descriptor.Descriptor(
  name='SetCustomStatusRequest',
  full_name='SetCustomStatusRequest',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='request_header', full_name='SetCustomStatusRequest.request_header', index=0,
      number=100, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='custom_status', full_name='SetCustomStatusRequest.custom_status', index=1,
      number=1, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='custom_status_expiry_timestamp_usec', full_name='SetCustomStatusRequest.custom_status_expiry_timestamp_usec', index=2,
      number=2, type=3, cpp_type=2, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='custom_status_remaining_duration_usec', full_name='SetCustomStatusRequest.custom_status_remaining_duration_usec', index=3,
      number=3, type=3, cpp_type=2, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=42090,
  serialized_end=42284,
)


_SETCUSTOMSTATUSRESPONSE = _descriptor.Descriptor(
  name='SetCustomStatusResponse',
  full_name='SetCustomStatusResponse',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='user_status', full_name='SetCustomStatusResponse.user_status', index=0,
      number=1, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='user_revision', full_name='SetCustomStatusResponse.user_revision', index=1,
      number=2, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=42286,
  serialized_end=42384,
)


_WRITEREVISION = _descriptor.Descriptor(
  name='WriteRevision',
  full_name='WriteRevision',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='timestamp', full_name='WriteRevision.timestamp', index=0,
      number=1, type=3, cpp_type=2, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='prev_timestamp', full_name='WriteRevision.prev_timestamp', index=1,
      number=2, type=3, cpp_type=2, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=42386,
  serialized_end=42444,
)


_READREVISION = _descriptor.Descriptor(
  name='ReadRevision',
  full_name='ReadRevision',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='timestamp', full_name='ReadRevision.timestamp', index=0,
      number=1, type=3, cpp_type=2, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=42446,
  serialized_end=42479,
)


_REFERENCEREVISION = _descriptor.Descriptor(
  name='ReferenceRevision',
  full_name='ReferenceRevision',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='timestamp', full_name='ReferenceRevision.timestamp', index=0,
      number=1, type=3, cpp_type=2, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=42481,
  serialized_end=42519,
)

_USERID.fields_by_name['type'].enum_type = _USERTYPE
_USER.fields_by_name['user_id'].message_type = _USERID
_USER.fields_by_name['block_relationship'].message_type = _USERBLOCKRELATIONSHIP
_DNDSETTINGS.fields_by_name['dnd_state'].enum_type = _DNDSETTINGS_DNDSTATE_STATE
_DNDSETTINGS_DNDSTATE_STATE.containing_type = _DNDSETTINGS
_CUSTOMSTATUS.fields_by_name['emoji'].message_type = _EMOJI
_USERSTATUS.fields_by_name['user_id'].message_type = _USERID
_USERSTATUS.fields_by_name['dnd_settings'].message_type = _DNDSETTINGS
_USERSTATUS.fields_by_name['custom_status'].message_type = _CUSTOMSTATUS
_USERSTATUSUPDATEDEVENT.fields_by_name['user_status'].message_type = _USERSTATUS
_GETUSERSTATUSREQUEST.fields_by_name['user_ids'].message_type = _USERID
_GETUSERSTATUSREQUEST.fields_by_name['request_header'].message_type = _REQUESTHEADER
_GETUSERSTATUSRESPONSE.fields_by_name['user_statuses'].message_type = _USERSTATUS
_GETSELFUSERSTATUSREQUEST.fields_by_name['request_header'].message_type = _REQUESTHEADER
_GETSELFUSERSTATUSRESPONSE.fields_by_name['user_status'].message_type = _USERSTATUS
_GETSELFUSERSTATUSRESPONSE.fields_by_name['user_revision'].message_type = _READREVISION
_CLIENTFEATURECAPABILITIES.fields_by_name['spaces_level_for_testing'].enum_type = _CLIENTFEATURECAPABILITIES_CAPABILITYLEVEL
_CLIENTFEATURECAPABILITIES.fields_by_name['dms_level_for_testing'].enum_type = _CLIENTFEATURECAPABILITIES_CAPABILITYLEVEL
_CLIENTFEATURECAPABILITIES.fields_by_name['post_rooms_level'].enum_type = _CLIENTFEATURECAPABILITIES_CAPABILITYLEVEL
_CLIENTFEATURECAPABILITIES.fields_by_name['spam_room_invites_level'].enum_type = _CLIENTFEATURECAPABILITIES_CAPABILITYLEVEL
_CLIENTFEATURECAPABILITIES.fields_by_name['tombstone_level'].enum_type = _CLIENTFEATURECAPABILITIES_CAPABILITYLEVEL
_CLIENTFEATURECAPABILITIES.fields_by_name['rich_text_viewing_level'].enum_type = _CLIENTFEATURECAPABILITIES_CAPABILITYLEVEL
_CLIENTFEATURECAPABILITIES_CAPABILITYLEVEL.containing_type = _CLIENTFEATURECAPABILITIES
_REQUESTHEADER.fields_by_name['client_type'].enum_type = _REQUESTHEADER_CLIENTTYPE
_REQUESTHEADER.fields_by_name['client_feature_capabilities'].message_type = _CLIENTFEATURECAPABILITIES
_REQUESTHEADER_CLIENTTYPE.containing_type = _REQUESTHEADER
_MEMBER.fields_by_name['user'].message_type = _USER
_MEMBERID.fields_by_name['user_id'].message_type = _USERID
_MEMBERSHIPID.fields_by_name['member_id'].message_type = _MEMBERID
_MEMBERSHIPID.fields_by_name['space_id'].message_type = _SPACEID
_MEMBERSHIPID.fields_by_name['group_id'].message_type = _GROUPID
_MEMBERSHIP.fields_by_name['id'].message_type = _MEMBERSHIPID
_MEMBERSHIP.fields_by_name['membership_state'].enum_type = _MEMBERSHIPSTATE
_MEMBERSHIP.fields_by_name['invite_category'].enum_type = _INVITECATEGORY
_MEMBERSHIP.fields_by_name['membership_role'].enum_type = _MEMBERSHIPROLE
_MEMBERPROFILE.fields_by_name['id'].message_type = _MEMBERSHIPID
_MEMBERPROFILE.fields_by_name['member'].message_type = _MEMBER
_GETMEMBERSREQUEST.fields_by_name['request_header'].message_type = _REQUESTHEADER
_GETMEMBERSREQUEST.fields_by_name['member_ids'].message_type = _MEMBERID
_GETMEMBERSREQUEST.fields_by_name['membership_ids'].message_type = _MEMBERSHIPID
_GETMEMBERSRESPONSE.fields_by_name['members'].message_type = _MEMBER
_GETMEMBERSRESPONSE.fields_by_name['member_profiles'].message_type = _MEMBERPROFILE
_USERPRESENCE.fields_by_name['user_id'].message_type = _USERID
_USERPRESENCE.fields_by_name['presence'].enum_type = _PRESENCE
_USERPRESENCE.fields_by_name['dnd_state'].enum_type = _DNDSTATE_STATE
_USERPRESENCE.fields_by_name['user_status'].message_type = _USERSTATUS
_GETUSERPRESENCEREQUEST.fields_by_name['request_header'].message_type = _REQUESTHEADER
_GETUSERPRESENCEREQUEST.fields_by_name['user_ids'].message_type = _USERID
_GETUSERPRESENCERESPONSE.fields_by_name['user_presences'].message_type = _USERPRESENCE
_JADDONSFORMATTEDTEXT_FORMATTEDTEXTELEMENT_DATETIME.containing_type = _JADDONSFORMATTEDTEXT_FORMATTEDTEXTELEMENT
_JADDONSFORMATTEDTEXT_FORMATTEDTEXTELEMENT_STYLEDTEXT.fields_by_name['datetime'].message_type = _JADDONSFORMATTEDTEXT_FORMATTEDTEXTELEMENT_DATETIME
_JADDONSFORMATTEDTEXT_FORMATTEDTEXTELEMENT_STYLEDTEXT.fields_by_name['styles'].enum_type = _JADDONSFORMATTEDTEXT_FORMATTEDTEXTELEMENT_STYLEDTEXT_STYLE
_JADDONSFORMATTEDTEXT_FORMATTEDTEXTELEMENT_STYLEDTEXT.fields_by_name['font_weight'].enum_type = _JADDONSFORMATTEDTEXT_FORMATTEDTEXTELEMENT_STYLEDTEXT_FONTWEIGHT
_JADDONSFORMATTEDTEXT_FORMATTEDTEXTELEMENT_STYLEDTEXT.fields_by_name['theme_colors'].message_type = _JADDONSTHEMECOLORS
_JADDONSFORMATTEDTEXT_FORMATTEDTEXTELEMENT_STYLEDTEXT.containing_type = _JADDONSFORMATTEDTEXT_FORMATTEDTEXTELEMENT
_JADDONSFORMATTEDTEXT_FORMATTEDTEXTELEMENT_STYLEDTEXT_STYLE.containing_type = _JADDONSFORMATTEDTEXT_FORMATTEDTEXTELEMENT_STYLEDTEXT
_JADDONSFORMATTEDTEXT_FORMATTEDTEXTELEMENT_STYLEDTEXT_FONTWEIGHT.containing_type = _JADDONSFORMATTEDTEXT_FORMATTEDTEXTELEMENT_STYLEDTEXT
_JADDONSFORMATTEDTEXT_FORMATTEDTEXTELEMENT_HYPERLINK.containing_type = _JADDONSFORMATTEDTEXT_FORMATTEDTEXTELEMENT
_JADDONSFORMATTEDTEXT_FORMATTEDTEXTELEMENT.fields_by_name['styled_text'].message_type = _JADDONSFORMATTEDTEXT_FORMATTEDTEXTELEMENT_STYLEDTEXT
_JADDONSFORMATTEDTEXT_FORMATTEDTEXTELEMENT.fields_by_name['hyperlink'].message_type = _JADDONSFORMATTEDTEXT_FORMATTEDTEXTELEMENT_HYPERLINK
_JADDONSFORMATTEDTEXT_FORMATTEDTEXTELEMENT.containing_type = _JADDONSFORMATTEDTEXT
_JADDONSFORMATTEDTEXT.fields_by_name['id'].message_type = _JADDONSIDENTIFIER
_JADDONSFORMATTEDTEXT.fields_by_name['formatted_text_elements'].message_type = _JADDONSFORMATTEDTEXT_FORMATTEDTEXTELEMENT
_JADDONSFORMATTEDTEXT.fields_by_name['text_align'].enum_type = _JADDONSFORMATTEDTEXT_TEXTALIGN
_JADDONSFORMATTEDTEXT_TEXTALIGN.containing_type = _JADDONSFORMATTEDTEXT
_JADDONSIMAGECROPSTYLE.fields_by_name['type'].enum_type = _JADDONSIMAGECROPSTYLE_IMAGECROPTYPE
_JADDONSIMAGECROPSTYLE_IMAGECROPTYPE.containing_type = _JADDONSIMAGECROPSTYLE
_JADDONSOPENLINK.fields_by_name['open_as'].enum_type = _JADDONSOPENLINK_OPENAS
_JADDONSOPENLINK.fields_by_name['on_close'].enum_type = _JADDONSOPENLINK_ONCLOSE
_JADDONSOPENLINK.fields_by_name['load_indicator'].enum_type = _JADDONSOPENLINK_LOADINDICATOR
_JADDONSOPENLINK_OPENAS.containing_type = _JADDONSOPENLINK
_JADDONSOPENLINK_ONCLOSE.containing_type = _JADDONSOPENLINK
_JADDONSOPENLINK_LOADINDICATOR.containing_type = _JADDONSOPENLINK
_JADDONSCARDITEM_CARDITEMHEADER.fields_by_name['title'].message_type = _JADDONSFORMATTEDTEXT
_JADDONSCARDITEM_CARDITEMHEADER.fields_by_name['subtitle'].message_type = _JADDONSFORMATTEDTEXT
_JADDONSCARDITEM_CARDITEMHEADER.fields_by_name['image_style'].enum_type = _JADDONSIMAGECROPSTYLE_IMAGECROPTYPE
_JADDONSCARDITEM_CARDITEMHEADER.containing_type = _JADDONSCARDITEM
_JADDONSCARDITEM_CARDITEMSECTION.fields_by_name['id'].message_type = _JADDONSIDENTIFIER
_JADDONSCARDITEM_CARDITEMSECTION.fields_by_name['header'].message_type = _JADDONSFORMATTEDTEXT
_JADDONSCARDITEM_CARDITEMSECTION.fields_by_name['widgets'].message_type = _JADDONSWIDGET
_JADDONSCARDITEM_CARDITEMSECTION.containing_type = _JADDONSCARDITEM
_JADDONSCARDITEM_CARDITEMACTION.fields_by_name['on_click'].message_type = _JADDONSONCLICK
_JADDONSCARDITEM_CARDITEMACTION.containing_type = _JADDONSCARDITEM
_JADDONSCARDITEM_CARDITEMFIXEDFOOTER.fields_by_name['buttons'].message_type = _JADDONSWIDGET_BUTTON
_JADDONSCARDITEM_CARDITEMFIXEDFOOTER.fields_by_name['primary_button'].message_type = _JADDONSWIDGET_TEXTBUTTON
_JADDONSCARDITEM_CARDITEMFIXEDFOOTER.fields_by_name['secondary_button'].message_type = _JADDONSWIDGET_TEXTBUTTON
_JADDONSCARDITEM_CARDITEMFIXEDFOOTER.containing_type = _JADDONSCARDITEM
_JADDONSCARDITEM_CARDITEMREFRESHACTION.fields_by_name['method'].message_type = _JADDONSFORMACTION
_JADDONSCARDITEM_CARDITEMREFRESHACTION.containing_type = _JADDONSCARDITEM
_JADDONSCARDITEM.fields_by_name['header'].message_type = _JADDONSCARDITEM_CARDITEMHEADER
_JADDONSCARDITEM.fields_by_name['sections'].message_type = _JADDONSCARDITEM_CARDITEMSECTION
_JADDONSCARDITEM.fields_by_name['card_actions'].message_type = _JADDONSCARDITEM_CARDITEMACTION
_JADDONSCARDITEM.fields_by_name['fixed_footer'].message_type = _JADDONSCARDITEM_CARDITEMFIXEDFOOTER
_JADDONSCARDITEM.fields_by_name['refresh_action'].message_type = _JADDONSCARDITEM_CARDITEMREFRESHACTION
_JADDONSCARDITEM.fields_by_name['display_style'].enum_type = _JADDONSCARDITEM_DISPLAYSTYLE
_JADDONSCARDITEM.fields_by_name['peek_card_header'].message_type = _JADDONSCARDITEM_CARDITEMHEADER
_JADDONSCARDITEM.fields_by_name['background_theme_colors'].message_type = _JADDONSTHEMECOLORS
_JADDONSCARDITEM_DISPLAYSTYLE.containing_type = _JADDONSCARDITEM
_JADDONSPUSHCARD.fields_by_name['card'].message_type = _JADDONSCARDITEM
_JADDONSONCLICK.fields_by_name['action'].message_type = _JADDONSFORMACTION
_JADDONSONCLICK.fields_by_name['open_link'].message_type = _JADDONSOPENLINK
_JADDONSONCLICK.fields_by_name['open_link_action'].message_type = _JADDONSFORMACTION
_JADDONSONCLICK.fields_by_name['push_card'].message_type = _JADDONSPUSHCARD
_JADDONSIMAGECOMPONENT.fields_by_name['crop_style'].message_type = _JADDONSIMAGECROPSTYLE
_JADDONSIMAGECOMPONENT.fields_by_name['border_style'].message_type = _JADDONSBORDERSTYLE
_JADDONSGRID_GRIDITEM.fields_by_name['image'].message_type = _JADDONSIMAGECOMPONENT
_JADDONSGRID_GRIDITEM.fields_by_name['text_alignment'].enum_type = _JADDONSWIDGET_HORIZONTALALIGN
_JADDONSGRID_GRIDITEM.fields_by_name['layout'].enum_type = _JADDONSGRID_GRIDITEM_GRIDITEMLAYOUT
_JADDONSGRID_GRIDITEM.fields_by_name['on_click'].message_type = _JADDONSONCLICK
_JADDONSGRID_GRIDITEM.containing_type = _JADDONSGRID
_JADDONSGRID_GRIDITEM_GRIDITEMLAYOUT.containing_type = _JADDONSGRID_GRIDITEM
_JADDONSGRID.fields_by_name['id'].message_type = _JADDONSIDENTIFIER
_JADDONSGRID.fields_by_name['items'].message_type = _JADDONSGRID_GRIDITEM
_JADDONSGRID.fields_by_name['border_style'].message_type = _JADDONSBORDERSTYLE
_JADDONSGRID.fields_by_name['on_click'].message_type = _JADDONSONCLICK
_JADDONSBORDERSTYLE.fields_by_name['type'].enum_type = _JADDONSBORDERSTYLE_BORDERTYPE
_JADDONSBORDERSTYLE.fields_by_name['stroke_color'].message_type = _JADDONSTHEMECOLORS
_JADDONSBORDERSTYLE_BORDERTYPE.containing_type = _JADDONSBORDERSTYLE
_JADDONSLABELCONTENTPAIRWIDGET.fields_by_name['label_content_pair'].message_type = _JADDONSLABELCONTENTPAIR
_JADDONSWIDGET_TEXTPARAGRAPH.fields_by_name['text'].message_type = _JADDONSFORMATTEDTEXT
_JADDONSWIDGET_TEXTPARAGRAPH.containing_type = _JADDONSWIDGET
_JADDONSWIDGET_TEXTKEYVALUE.fields_by_name['key'].message_type = _JADDONSFORMATTEDTEXT
_JADDONSWIDGET_TEXTKEYVALUE.fields_by_name['text'].message_type = _JADDONSFORMATTEDTEXT
_JADDONSWIDGET_TEXTKEYVALUE.fields_by_name['on_click'].message_type = _JADDONSONCLICK
_JADDONSWIDGET_TEXTKEYVALUE.containing_type = _JADDONSWIDGET
_JADDONSWIDGET_IMAGEKEYVALUE.fields_by_name['text'].message_type = _JADDONSFORMATTEDTEXT
_JADDONSWIDGET_IMAGEKEYVALUE.fields_by_name['on_click'].message_type = _JADDONSONCLICK
_JADDONSWIDGET_IMAGEKEYVALUE.containing_type = _JADDONSWIDGET
_JADDONSWIDGET_IMAGE.fields_by_name['id'].message_type = _JADDONSIDENTIFIER
_JADDONSWIDGET_IMAGE.fields_by_name['on_click'].message_type = _JADDONSONCLICK
_JADDONSWIDGET_IMAGE.containing_type = _JADDONSWIDGET
_JADDONSWIDGET_ICON.fields_by_name['image_style'].enum_type = _JADDONSIMAGECROPSTYLE_IMAGECROPTYPE
_JADDONSWIDGET_ICON.containing_type = _JADDONSWIDGET
_JADDONSWIDGET_KEYVALUE_SWITCHWIDGET.fields_by_name['id'].message_type = _JADDONSIDENTIFIER
_JADDONSWIDGET_KEYVALUE_SWITCHWIDGET.fields_by_name['on_change'].message_type = _JADDONSFORMACTION
_JADDONSWIDGET_KEYVALUE_SWITCHWIDGET.fields_by_name['control_type'].enum_type = _JADDONSWIDGET_KEYVALUE_SWITCHWIDGET_CONTROLTYPE
_JADDONSWIDGET_KEYVALUE_SWITCHWIDGET.containing_type = _JADDONSWIDGET_KEYVALUE
_JADDONSWIDGET_KEYVALUE_SWITCHWIDGET_CONTROLTYPE.containing_type = _JADDONSWIDGET_KEYVALUE_SWITCHWIDGET
_JADDONSWIDGET_KEYVALUE.fields_by_name['image_style'].enum_type = _JADDONSIMAGECROPSTYLE_IMAGECROPTYPE
_JADDONSWIDGET_KEYVALUE.fields_by_name['start_icon'].message_type = _JADDONSWIDGET_ICON
_JADDONSWIDGET_KEYVALUE.fields_by_name['top_label'].message_type = _JADDONSFORMATTEDTEXT
_JADDONSWIDGET_KEYVALUE.fields_by_name['content'].message_type = _JADDONSFORMATTEDTEXT
_JADDONSWIDGET_KEYVALUE.fields_by_name['bottom_label'].message_type = _JADDONSFORMATTEDTEXT
_JADDONSWIDGET_KEYVALUE.fields_by_name['on_click'].message_type = _JADDONSONCLICK
_JADDONSWIDGET_KEYVALUE.fields_by_name['button'].message_type = _JADDONSWIDGET_BUTTON
_JADDONSWIDGET_KEYVALUE.fields_by_name['switch_widget'].message_type = _JADDONSWIDGET_KEYVALUE_SWITCHWIDGET
_JADDONSWIDGET_KEYVALUE.fields_by_name['end_icon'].message_type = _JADDONSWIDGET_ICON
_JADDONSWIDGET_KEYVALUE.containing_type = _JADDONSWIDGET
_JADDONSWIDGET_DIVIDER.containing_type = _JADDONSWIDGET
_JADDONSWIDGET_MENU_MENUITEM.containing_type = _JADDONSWIDGET_MENU
_JADDONSWIDGET_MENU.fields_by_name['items'].message_type = _JADDONSWIDGET_MENU_MENUITEM
_JADDONSWIDGET_MENU.fields_by_name['on_change'].message_type = _JADDONSFORMACTION
_JADDONSWIDGET_MENU.containing_type = _JADDONSWIDGET
_JADDONSWIDGET_AUTOCOMPLETE_AUTOCOMPLETEITEM.containing_type = _JADDONSWIDGET_AUTOCOMPLETE
_JADDONSWIDGET_AUTOCOMPLETE.fields_by_name['items'].message_type = _JADDONSWIDGET_AUTOCOMPLETE_AUTOCOMPLETEITEM
_JADDONSWIDGET_AUTOCOMPLETE.containing_type = _JADDONSWIDGET
_JADDONSWIDGET_TEXTFIELD.fields_by_name['id'].message_type = _JADDONSIDENTIFIER
_JADDONSWIDGET_TEXTFIELD.fields_by_name['type'].enum_type = _JADDONSWIDGET_TEXTFIELD_TEXTFIELDTYPE
_JADDONSWIDGET_TEXTFIELD.fields_by_name['line_type'].enum_type = _JADDONSWIDGET_TEXTFIELD_LINETYPE
_JADDONSWIDGET_TEXTFIELD.fields_by_name['on_change'].message_type = _JADDONSFORMACTION
_JADDONSWIDGET_TEXTFIELD.fields_by_name['auto_complete'].message_type = _JADDONSWIDGET_AUTOCOMPLETE
_JADDONSWIDGET_TEXTFIELD.fields_by_name['auto_complete_callback'].message_type = _JADDONSFORMACTION
_JADDONSWIDGET_TEXTFIELD.containing_type = _JADDONSWIDGET
_JADDONSWIDGET_TEXTFIELD_TEXTFIELDTYPE.containing_type = _JADDONSWIDGET_TEXTFIELD
_JADDONSWIDGET_TEXTFIELD_LINETYPE.containing_type = _JADDONSWIDGET_TEXTFIELD
_JADDONSWIDGET_SELECTIONCONTROL_SELECTIONITEM.fields_by_name['id'].message_type = _JADDONSIDENTIFIER
_JADDONSWIDGET_SELECTIONCONTROL_SELECTIONITEM.containing_type = _JADDONSWIDGET_SELECTIONCONTROL
_JADDONSWIDGET_SELECTIONCONTROL.fields_by_name['id'].message_type = _JADDONSIDENTIFIER
_JADDONSWIDGET_SELECTIONCONTROL.fields_by_name['type'].enum_type = _JADDONSWIDGET_SELECTIONCONTROL_SELECTIONTYPE
_JADDONSWIDGET_SELECTIONCONTROL.fields_by_name['items'].message_type = _JADDONSWIDGET_SELECTIONCONTROL_SELECTIONITEM
_JADDONSWIDGET_SELECTIONCONTROL.fields_by_name['on_change'].message_type = _JADDONSFORMACTION
_JADDONSWIDGET_SELECTIONCONTROL.containing_type = _JADDONSWIDGET
_JADDONSWIDGET_SELECTIONCONTROL_SELECTIONTYPE.containing_type = _JADDONSWIDGET_SELECTIONCONTROL
_JADDONSWIDGET_DATETIMEPICKER.fields_by_name['type'].enum_type = _JADDONSWIDGET_DATETIMEPICKER_DATETIMEPICKERTYPE
_JADDONSWIDGET_DATETIMEPICKER.fields_by_name['on_change'].message_type = _JADDONSFORMACTION
_JADDONSWIDGET_DATETIMEPICKER.fields_by_name['theme_colors'].message_type = _JADDONSTHEMECOLORS
_JADDONSWIDGET_DATETIMEPICKER.containing_type = _JADDONSWIDGET
_JADDONSWIDGET_DATETIMEPICKER_DATETIMEPICKERTYPE.containing_type = _JADDONSWIDGET_DATETIMEPICKER
_JADDONSWIDGET_TEXTBUTTON.fields_by_name['id'].message_type = _JADDONSIDENTIFIER
_JADDONSWIDGET_TEXTBUTTON.fields_by_name['text'].message_type = _JADDONSFORMATTEDTEXT
_JADDONSWIDGET_TEXTBUTTON.fields_by_name['on_click'].message_type = _JADDONSONCLICK
_JADDONSWIDGET_TEXTBUTTON.fields_by_name['style'].enum_type = _JADDONSWIDGET_TEXTBUTTON_STYLE
_JADDONSWIDGET_TEXTBUTTON.fields_by_name['background_theme_colors'].message_type = _JADDONSTHEMECOLORS
_JADDONSWIDGET_TEXTBUTTON.containing_type = _JADDONSWIDGET
_JADDONSWIDGET_TEXTBUTTON_STYLE.containing_type = _JADDONSWIDGET_TEXTBUTTON
_JADDONSWIDGET_IMAGEBUTTON.fields_by_name['id'].message_type = _JADDONSIDENTIFIER
_JADDONSWIDGET_IMAGEBUTTON.fields_by_name['on_click'].message_type = _JADDONSONCLICK
_JADDONSWIDGET_IMAGEBUTTON.containing_type = _JADDONSWIDGET
_JADDONSWIDGET_BUTTON.fields_by_name['text_button'].message_type = _JADDONSWIDGET_TEXTBUTTON
_JADDONSWIDGET_BUTTON.fields_by_name['image_button'].message_type = _JADDONSWIDGET_IMAGEBUTTON
_JADDONSWIDGET_BUTTON.containing_type = _JADDONSWIDGET
_JADDONSWIDGET.fields_by_name['text_widget'].message_type = _JADDONSTEXTWIDGET
_JADDONSWIDGET.fields_by_name['label_content_pair_widget'].message_type = _JADDONSLABELCONTENTPAIRWIDGET
_JADDONSWIDGET.fields_by_name['text_paragraph'].message_type = _JADDONSWIDGET_TEXTPARAGRAPH
_JADDONSWIDGET.fields_by_name['text_key_value'].message_type = _JADDONSWIDGET_TEXTKEYVALUE
_JADDONSWIDGET.fields_by_name['image_key_value'].message_type = _JADDONSWIDGET_IMAGEKEYVALUE
_JADDONSWIDGET.fields_by_name['image'].message_type = _JADDONSWIDGET_IMAGE
_JADDONSWIDGET.fields_by_name['key_value'].message_type = _JADDONSWIDGET_KEYVALUE
_JADDONSWIDGET.fields_by_name['divider'].message_type = _JADDONSWIDGET_DIVIDER
_JADDONSWIDGET.fields_by_name['grid'].message_type = _JADDONSGRID
_JADDONSWIDGET.fields_by_name['menu'].message_type = _JADDONSWIDGET_MENU
_JADDONSWIDGET.fields_by_name['text_field'].message_type = _JADDONSWIDGET_TEXTFIELD
_JADDONSWIDGET.fields_by_name['selection_control'].message_type = _JADDONSWIDGET_SELECTIONCONTROL
_JADDONSWIDGET.fields_by_name['date_time_picker'].message_type = _JADDONSWIDGET_DATETIMEPICKER
_JADDONSWIDGET.fields_by_name['buttons'].message_type = _JADDONSWIDGET_BUTTON
_JADDONSWIDGET.fields_by_name['horizontal_align'].enum_type = _JADDONSWIDGET_HORIZONTALALIGN
_JADDONSWIDGET_HORIZONTALALIGN.containing_type = _JADDONSWIDGET
_JADDONSFORMACTION_ACTIONPARAMETER.containing_type = _JADDONSFORMACTION
_JADDONSFORMACTION.fields_by_name['parameters'].message_type = _JADDONSFORMACTION_ACTIONPARAMETER
_JADDONSFORMACTION.fields_by_name['load_indicator'].enum_type = _JADDONSFORMACTION_LOADINDICATOR
_JADDONSFORMACTION_LOADINDICATOR.containing_type = _JADDONSFORMACTION
_JADDONSCONTEXTUALADDON_TOOLBAR.fields_by_name['name'].message_type = _JADDONSFORMATTEDTEXT
_JADDONSCONTEXTUALADDON_TOOLBAR.fields_by_name['theme_colors'].message_type = _JADDONSTHEMECOLORS
_JADDONSCONTEXTUALADDON_TOOLBAR.containing_type = _JADDONSCONTEXTUALADDON
_JADDONSCONTEXTUALADDON_CARD_CARDHEADER.fields_by_name['title'].message_type = _JADDONSFORMATTEDTEXT
_JADDONSCONTEXTUALADDON_CARD_CARDHEADER.fields_by_name['subtitle'].message_type = _JADDONSFORMATTEDTEXT
_JADDONSCONTEXTUALADDON_CARD_CARDHEADER.fields_by_name['image_style'].enum_type = _JADDONSIMAGECROPSTYLE_IMAGECROPTYPE
_JADDONSCONTEXTUALADDON_CARD_CARDHEADER.containing_type = _JADDONSCONTEXTUALADDON_CARD
_JADDONSCONTEXTUALADDON_CARD_SECTION.fields_by_name['id'].message_type = _JADDONSIDENTIFIER
_JADDONSCONTEXTUALADDON_CARD_SECTION.fields_by_name['header'].message_type = _JADDONSFORMATTEDTEXT
_JADDONSCONTEXTUALADDON_CARD_SECTION.fields_by_name['widgets'].message_type = _JADDONSWIDGET
_JADDONSCONTEXTUALADDON_CARD_SECTION.containing_type = _JADDONSCONTEXTUALADDON_CARD
_JADDONSCONTEXTUALADDON_CARD_CARDACTION.fields_by_name['on_click'].message_type = _JADDONSONCLICK
_JADDONSCONTEXTUALADDON_CARD_CARDACTION.containing_type = _JADDONSCONTEXTUALADDON_CARD
_JADDONSCONTEXTUALADDON_CARD_FIXEDFOOTER.fields_by_name['buttons'].message_type = _JADDONSWIDGET_BUTTON
_JADDONSCONTEXTUALADDON_CARD_FIXEDFOOTER.fields_by_name['primary_button'].message_type = _JADDONSWIDGET_TEXTBUTTON
_JADDONSCONTEXTUALADDON_CARD_FIXEDFOOTER.fields_by_name['secondary_button'].message_type = _JADDONSWIDGET_TEXTBUTTON
_JADDONSCONTEXTUALADDON_CARD_FIXEDFOOTER.containing_type = _JADDONSCONTEXTUALADDON_CARD
_JADDONSCONTEXTUALADDON_CARD_REFRESHACTION.fields_by_name['method'].message_type = _JADDONSFORMACTION
_JADDONSCONTEXTUALADDON_CARD_REFRESHACTION.containing_type = _JADDONSCONTEXTUALADDON_CARD
_JADDONSCONTEXTUALADDON_CARD.fields_by_name['header'].message_type = _JADDONSCONTEXTUALADDON_CARD_CARDHEADER
_JADDONSCONTEXTUALADDON_CARD.fields_by_name['sections'].message_type = _JADDONSCONTEXTUALADDON_CARD_SECTION
_JADDONSCONTEXTUALADDON_CARD.fields_by_name['card_actions'].message_type = _JADDONSCONTEXTUALADDON_CARD_CARDACTION
_JADDONSCONTEXTUALADDON_CARD.fields_by_name['fixed_footer'].message_type = _JADDONSCONTEXTUALADDON_CARD_FIXEDFOOTER
_JADDONSCONTEXTUALADDON_CARD.fields_by_name['refresh_action'].message_type = _JADDONSCONTEXTUALADDON_CARD_REFRESHACTION
_JADDONSCONTEXTUALADDON_CARD.fields_by_name['background_theme_colors'].message_type = _JADDONSTHEMECOLORS
_JADDONSCONTEXTUALADDON_CARD.containing_type = _JADDONSCONTEXTUALADDON
_JADDONSCONTEXTUALADDON.fields_by_name['toolbar'].message_type = _JADDONSCONTEXTUALADDON_TOOLBAR
_JADDONSCONTEXTUALADDON.fields_by_name['cards'].message_type = _JADDONSCONTEXTUALADDON_CARD
_HTMLATTACHMENT.fields_by_name['html'].message_type = _HTML
_ATTACHMENT.fields_by_name['html'].message_type = _HTMLATTACHMENT
_ATTACHMENT.fields_by_name['add_on_data'].message_type = _JADDONSCONTEXTUALADDON
_ATTACHMENT.fields_by_name['app_id'].message_type = _USERID
_ATTACHMENT.fields_by_name['app_user'].message_type = _USER
_REACTION.fields_by_name['emoji'].message_type = _EMOJI
_MESSAGEPARENTID.fields_by_name['topic_id'].message_type = _TOPICID
_MESSAGEID.fields_by_name['parent_id'].message_type = _MESSAGEPARENTID
_MESSAGE.fields_by_name['id'].message_type = _MESSAGEID
_MESSAGE.fields_by_name['message_state'].enum_type = _MESSAGE_MESSAGESTATE
_MESSAGE.fields_by_name['creator'].message_type = _USER
_MESSAGE.fields_by_name['annotations'].message_type = _ANNOTATION
_MESSAGE.fields_by_name['attachments'].message_type = _ATTACHMENT
_MESSAGE.fields_by_name['retention_settings'].message_type = _RETENTIONSETTINGS
_MESSAGE.fields_by_name['reactions'].message_type = _REACTION
_MESSAGE.fields_by_name['last_reply'].message_type = _MESSAGE
_MESSAGE.fields_by_name['message_type'].enum_type = _MESSAGE_MESSAGETYPE
_MESSAGE.fields_by_name['creator_membership'].message_type = _MEMBERSHIP
_MESSAGE_MESSAGESTATE.containing_type = _MESSAGE
_MESSAGE_MESSAGETYPE.containing_type = _MESSAGE
_DRIVEMETADATA.fields_by_name['embed_url'].message_type = _TRUSTEDRESOURCEURL
_URLMETADATA.fields_by_name['url'].message_type = _URL
_URLMETADATA.fields_by_name['gws_url'].message_type = _URL
_URLMETADATA.fields_by_name['redirect_url'].message_type = _URL
_FORMATMETADATA.fields_by_name['format_type'].enum_type = _FORMATMETADATA_FORMATTYPE
_FORMATMETADATA_FORMATTYPE.containing_type = _FORMATMETADATA
_ANNOTATION.fields_by_name['type'].enum_type = _ANNOTATIONTYPE
_ANNOTATION.fields_by_name['chip_render_type'].enum_type = _ANNOTATION_CHIPRENDERTYPE
_ANNOTATION.fields_by_name['user_mention_metadata'].message_type = _USERMENTIONMETADATA
_ANNOTATION.fields_by_name['format_metadata'].message_type = _FORMATMETADATA
_ANNOTATION.fields_by_name['slash_command_metadata'].message_type = _SLASHCOMMANDMETADATA
_ANNOTATION.fields_by_name['drive_metadata'].message_type = _DRIVEMETADATA
_ANNOTATION.fields_by_name['url_metadata'].message_type = _URLMETADATA
_ANNOTATION.fields_by_name['upload_metadata'].message_type = _UPLOADMETADATA
_ANNOTATION.fields_by_name['membership_changed'].message_type = _MEMBERSHIPCHANGEDMETADATA
_ANNOTATION.fields_by_name['room_updated'].message_type = _ROOMUPDATEDMETADATA
_ANNOTATION_CHIPRENDERTYPE.containing_type = _ANNOTATION
_TYPINGCONTEXT.fields_by_name['group_id'].message_type = _GROUPID
_TYPINGCONTEXT.fields_by_name['topic_id'].message_type = _TOPICID
_SETTYPINGSTATEREQUEST.fields_by_name['request_header'].message_type = _REQUESTHEADER
_SETTYPINGSTATEREQUEST.fields_by_name['state'].enum_type = _TYPINGSTATE
_SETTYPINGSTATEREQUEST.fields_by_name['context'].message_type = _TYPINGCONTEXT
_GROUPID.fields_by_name['space_id'].message_type = _SPACEID
_GROUPID.fields_by_name['dm_id'].message_type = _DMID
_GROUP_FLATGROUP.containing_type = _GROUP
_GROUP_THREADEDGROUP.containing_type = _GROUP
_GROUP.fields_by_name['group_id'].message_type = _GROUPID
_GROUP.fields_by_name['creator'].message_type = _USER
_GROUP.fields_by_name['group_read_state'].message_type = _GROUPREADSTATE
_GROUP.fields_by_name['retention_settings'].message_type = _RETENTIONSETTINGS
_GROUP.fields_by_name['group_type'].enum_type = _GROUP_GROUPTYPE
_GROUP.fields_by_name['visibility'].message_type = _GROUPVISIBILITY
_GROUP.fields_by_name['flat_group'].message_type = _GROUP_FLATGROUP
_GROUP.fields_by_name['threaded_group'].message_type = _GROUP_THREADEDGROUP
_GROUP.fields_by_name['attribute_checker_group_type'].enum_type = _SHAREDATTRIBUTECHECKERGROUPTYPE
_GROUP.fields_by_name['name_users'].message_type = _NAMEUSERS
_GROUP.fields_by_name['group_support_level'].enum_type = _GROUPSUPPORTLEVEL
_GROUP.fields_by_name['group_unsupported_reason'].enum_type = _GROUPUNSUPPORTEDREASON
_GROUP_GROUPTYPE.containing_type = _GROUP
_GROUPREADSTATEID.fields_by_name['user_id'].message_type = _USERID
_GROUPREADSTATEID.fields_by_name['group_id'].message_type = _GROUPID
_INVITESTATE.fields_by_name['inviter_user_id'].message_type = _USERID
_INVITESTATE.fields_by_name['invitation_message_id'].message_type = _MESSAGEID
_INVITESTATE.fields_by_name['invite_type'].message_type = _INVITETYPE
_GROUPREADSTATE.fields_by_name['id'].message_type = _GROUPREADSTATEID
_GROUPREADSTATE.fields_by_name['unread_subscribed_topics'].message_type = _TOPICID
_GROUPREADSTATE.fields_by_name['invite_state'].message_type = _INVITESTATE
_GROUPREADSTATE.fields_by_name['notification_settings'].message_type = _GROUPNOTIFICATIONSETTINGS
_GROUPREADSTATE.fields_by_name['retention_settings'].message_type = _RETENTIONSETTINGS
_GROUPREADSTATE.fields_by_name['membership_role'].enum_type = _MEMBERSHIPROLE
_GROUPREADSTATE.fields_by_name['membership_state'].enum_type = _MEMBERSHIPSTATE
_GROUPREADSTATE.fields_by_name['invite_category'].enum_type = _INVITECATEGORY
_GROUPREADSTATE.fields_by_name['joined_users'].message_type = _USERID
_GROUPNOTIFICATIONSETTINGS.fields_by_name['state'].enum_type = _GROUPNOTIFICATIONSETTINGS_GROUPNOTIFICATIONSTATE
_GROUPNOTIFICATIONSETTINGS.fields_by_name['room_state'].enum_type = _GROUPNOTIFICATIONSETTINGS_ROOMNOTIFICATIONSTATE
_GROUPNOTIFICATIONSETTINGS_GROUPNOTIFICATIONSTATE.containing_type = _GROUPNOTIFICATIONSETTINGS
_GROUPNOTIFICATIONSETTINGS_ROOMNOTIFICATIONSTATE.containing_type = _GROUPNOTIFICATIONSETTINGS
_RETENTIONSETTINGS.fields_by_name['state'].enum_type = _RETENTIONSETTINGS_RETENTIONSTATE
_RETENTIONSETTINGS_RETENTIONSTATE.containing_type = _RETENTIONSETTINGS
_TOPICID.fields_by_name['group_id'].message_type = _GROUPID
_TOPIC.fields_by_name['id'].message_type = _TOPICID
_TOPIC.fields_by_name['replies'].message_type = _MESSAGE
_TOPIC.fields_by_name['retention_settings'].message_type = _RETENTIONSETTINGS
_CREATETOPICREQUEST.fields_by_name['request_header'].message_type = _REQUESTHEADER
_CREATETOPICREQUEST.fields_by_name['group_id'].message_type = _GROUPID
_CREATETOPICREQUEST.fields_by_name['annotations'].message_type = _ANNOTATION
_CREATETOPICREQUEST.fields_by_name['retention_settings'].message_type = _RETENTIONSETTINGS
_CREATETOPICREQUEST.fields_by_name['message_info'].message_type = _MESSAGEINFO
_CREATETOPICRESPONSE.fields_by_name['topic'].message_type = _TOPIC
_CREATETOPICRESPONSE.fields_by_name['group_revision'].message_type = _WRITEREVISION
_CREATETOPICRESPONSE.fields_by_name['current_group_revision'].message_type = _READREVISION
_INVITEEINFO.fields_by_name['user_id'].message_type = _USERID
_INVITEEMEMBERINFO.fields_by_name['invitee_info'].message_type = _INVITEEINFO
_SPACECREATIONINFO_FLATGROUP.containing_type = _SPACECREATIONINFO
_SPACECREATIONINFO_THREADEDGROUP.containing_type = _SPACECREATIONINFO
_SPACECREATIONINFO_SPACETYPE.containing_type = _SPACECREATIONINFO
_SPACECREATIONINFO.fields_by_name['visibility'].message_type = _GROUPVISIBILITY
_SPACECREATIONINFO.fields_by_name['flat_group'].message_type = _SPACECREATIONINFO_FLATGROUP
_SPACECREATIONINFO.fields_by_name['threaded_group'].message_type = _SPACECREATIONINFO_THREADEDGROUP
_SPACECREATIONINFO.fields_by_name['invitee_member_infos'].message_type = _INVITEEMEMBERINFO
_SPACECREATIONINFO.fields_by_name['space_type'].message_type = _SPACECREATIONINFO_SPACETYPE
_SPACECREATIONINFO.fields_by_name['attribute_checker_group_type'].enum_type = _SHAREDATTRIBUTECHECKERGROUPTYPE
_CREATEGROUPREQUEST.fields_by_name['request_header'].message_type = _REQUESTHEADER
_CREATEGROUPREQUEST.fields_by_name['space'].message_type = _SPACECREATIONINFO
_CREATEMEMBERSHIPRESULT.fields_by_name['membership'].message_type = _MEMBERSHIP
_CREATEGROUPRESPONSE.fields_by_name['group'].message_type = _GROUP
_CREATEGROUPRESPONSE.fields_by_name['group_revision'].message_type = _WRITEREVISION
_CREATEGROUPRESPONSE.fields_by_name['user_revision'].message_type = _WRITEREVISION
_CREATEGROUPRESPONSE.fields_by_name['results'].message_type = _CREATEMEMBERSHIPRESULT
_CREATEDMREQUEST.fields_by_name['request_header'].message_type = _REQUESTHEADER
_CREATEDMREQUEST.fields_by_name['fetch_options'].enum_type = _CREATEDMREQUEST_FETCHOPTIONS
_CREATEDMREQUEST.fields_by_name['members'].message_type = _USERID
_CREATEDMREQUEST.fields_by_name['invitees'].message_type = _INVITEEINFO
_CREATEDMREQUEST.fields_by_name['retention_settings'].message_type = _RETENTIONSETTINGS
_CREATEDMREQUEST_FETCHOPTIONS.containing_type = _CREATEDMREQUEST
_CREATEDMRESPONSE.fields_by_name['dm'].message_type = _GROUP
_CREATEDMRESPONSE.fields_by_name['topic'].message_type = _TOPIC
_CREATEDMRESPONSE.fields_by_name['group_revision'].message_type = _WRITEREVISION
_CREATEDMRESPONSE.fields_by_name['memberships'].message_type = _MEMBERSHIP
_LISTTOPICSRESPONSE.fields_by_name['topics'].message_type = _TOPIC
_LISTTOPICSRESPONSE.fields_by_name['user_revision'].message_type = _READREVISION
_LISTTOPICSRESPONSE.fields_by_name['group_revision'].message_type = _READREVISION
_LISTTOPICSRESPONSE.fields_by_name['read_receipt_set'].message_type = _READRECEIPTSET
_LISTTOPICSREQUEST.fields_by_name['request_header'].message_type = _REQUESTHEADER
_LISTTOPICSREQUEST.fields_by_name['group_id'].message_type = _GROUPID
_LISTTOPICSREQUEST.fields_by_name['fetch_options'].enum_type = _LISTTOPICSREQUEST_FETCHOPTIONS
_LISTTOPICSREQUEST.fields_by_name['user_not_older_than'].message_type = _REFERENCEREVISION
_LISTTOPICSREQUEST.fields_by_name['group_not_older_than'].message_type = _REFERENCEREVISION
_LISTTOPICSREQUEST_FETCHOPTIONS.containing_type = _LISTTOPICSREQUEST
_LISTMEMBERSREQUEST.fields_by_name['request_header'].message_type = _REQUESTHEADER
_LISTMEMBERSREQUEST.fields_by_name['space_id'].message_type = _SPACEID
_LISTMEMBERSREQUEST.fields_by_name['group_id'].message_type = _GROUPID
_LISTMEMBERSREQUEST.fields_by_name['not_older_than'].message_type = _REFERENCEREVISION
_LISTMEMBERSRESPONSE.fields_by_name['memberships'].message_type = _MEMBERSHIP
_LISTMEMBERSRESPONSE.fields_by_name['members'].message_type = _MEMBER
_LISTMEMBERSRESPONSE.fields_by_name['member_ids'].message_type = _MEMBERID
_LISTMEMBERSRESPONSE.fields_by_name['group_revision'].message_type = _READREVISION
_LISTMEMBERSRESPONSE.fields_by_name['invited_member_ids'].message_type = _MEMBERID
_READRECEIPT.fields_by_name['user'].message_type = _USER
_READRECEIPTSET.fields_by_name['read_receipts'].message_type = _READRECEIPT
_WEBPUSHNOTIFICATION.fields_by_name['message'].message_type = _MESSAGE
_WEBPUSHNOTIFICATION.fields_by_name['cause'].enum_type = _NOTIFICATIONCAUSE
_IOSLOCALNOTIFICATION.fields_by_name['attribute_checker_group_type'].enum_type = _SHAREDATTRIBUTECHECKERGROUPTYPE
_MOBILELOCALNOTIFICATION.fields_by_name['android_local_notification'].message_type = _ANDROIDLOCALNOTIFICATION
_MOBILELOCALNOTIFICATION.fields_by_name['ios_local_notification'].message_type = _IOSLOCALNOTIFICATION
_MESSAGEEVENT.fields_by_name['message'].message_type = _MESSAGE
_TYPINGSTATECHANGEDEVENT.fields_by_name['state'].enum_type = _TYPINGSTATE
_TYPINGSTATECHANGEDEVENT.fields_by_name['user_id'].message_type = _USERID
_TYPINGSTATECHANGEDEVENT.fields_by_name['context'].message_type = _TYPINGCONTEXT
_MEMBERSHIPCHANGEDEVENT.fields_by_name['new_membership'].message_type = _MEMBERSHIP
_MEMBERSHIPCHANGEDEVENT.fields_by_name['prior_membership_state'].enum_type = _MEMBERSHIPSTATE
_MEMBERSHIPCHANGEDEVENT.fields_by_name['prior_membership_role'].enum_type = _MEMBERSHIPROLE
_READRECEIPTCHANGEDEVENT.fields_by_name['group_id'].message_type = _GROUPID
_READRECEIPTCHANGEDEVENT.fields_by_name['read_receipt_set'].message_type = _READRECEIPTSET
_GROUPVIEWEDEVENT.fields_by_name['group_id'].message_type = _GROUPID
_WEBPUSHNOTIFICATIONEVENT.fields_by_name['notification'].message_type = _WEBPUSHNOTIFICATION
_WEBPUSHNOTIFICATIONEVENT.fields_by_name['dispatch_approach_type'].enum_type = _WEBPUSHNOTIFICATIONEVENT_DISPATCHAPPROACHTYPE
_WEBPUSHNOTIFICATIONEVENT.fields_by_name['mobile_local_notification'].message_type = _MOBILELOCALNOTIFICATION
_WEBPUSHNOTIFICATIONEVENT.fields_by_name['endpoint_types'].enum_type = _WEBPUSHNOTIFICATIONEVENT_ENDPOINTTYPE
_WEBPUSHNOTIFICATIONEVENT_DISPATCHAPPROACHTYPE.containing_type = _WEBPUSHNOTIFICATIONEVENT
_WEBPUSHNOTIFICATIONEVENT_ENDPOINTTYPE.containing_type = _WEBPUSHNOTIFICATIONEVENT
_STREAMEVENTSREQUEST.fields_by_name['platform'].enum_type = _PLATFORM
_STREAMEVENTSREQUEST.fields_by_name['client_info'].message_type = _CLIENTINFO
_STREAMEVENTSREQUEST.fields_by_name['ping_event'].message_type = _PINGEVENT
_STREAMEVENTSREQUEST.fields_by_name['clock_sync_request'].message_type = _CLOCKSYNCREQUEST
_STREAMEVENTSREQUEST.fields_by_name['group_subscription_event'].message_type = _GROUPSUBSCRIPTIONEVENT
_CLIENTINFO.fields_by_name['platform'].enum_type = _PLATFORM
_CLIENTINFO.fields_by_name['origin'].enum_type = _EVENTORIGIN
_CLIENTNOTIFICATIONSSTATE.fields_by_name['device_setting_state'].enum_type = _CLIENTNOTIFICATIONSSTATE_DEVICENOTIFICATIONSETTINGSTATE
_CLIENTNOTIFICATIONSSTATE_DEVICENOTIFICATIONSETTINGSTATE.containing_type = _CLIENTNOTIFICATIONSSTATE
_PINGEVENT.fields_by_name['state'].enum_type = _PINGEVENT_STATE
_PINGEVENT.fields_by_name['application_focus_state'].enum_type = _PINGEVENT_APPLICATIONFOCUSSTATE
_PINGEVENT.fields_by_name['client_interactive_state'].enum_type = _PINGEVENT_CLIENTINTERACTIVESTATE
_PINGEVENT.fields_by_name['notifications_state'].message_type = _CLIENTNOTIFICATIONSSTATE
_PINGEVENT.fields_by_name['device_active_state'].enum_type = _PINGEVENT_DEVICEACTIVESTATE
_PINGEVENT_STATE.containing_type = _PINGEVENT
_PINGEVENT_APPLICATIONFOCUSSTATE.containing_type = _PINGEVENT
_PINGEVENT_CLIENTINTERACTIVESTATE.containing_type = _PINGEVENT
_PINGEVENT_DEVICEACTIVESTATE.containing_type = _PINGEVENT
_GROUPSUBSCRIPTIONEVENT.fields_by_name['group_ids'].message_type = _GROUPID
_STREAMEVENTSRESPONSE.fields_by_name['event'].message_type = _EVENT
_STREAMEVENTSRESPONSE.fields_by_name['clock_sync_response'].message_type = _CLOCKSYNCRESPONSE
_EVENT_EVENTBODY.fields_by_name['group_viewed'].message_type = _GROUPVIEWEDEVENT
_EVENT_EVENTBODY.fields_by_name['message_posted'].message_type = _MESSAGEEVENT
_EVENT_EVENTBODY.fields_by_name['web_push_notification'].message_type = _WEBPUSHNOTIFICATIONEVENT
_EVENT_EVENTBODY.fields_by_name['membership_changed'].message_type = _MEMBERSHIPCHANGEDEVENT
_EVENT_EVENTBODY.fields_by_name['user_status_updated_event'].message_type = _USERSTATUSUPDATEDEVENT
_EVENT_EVENTBODY.fields_by_name['typing_state_changed_event'].message_type = _TYPINGSTATECHANGEDEVENT
_EVENT_EVENTBODY.fields_by_name['read_receipt_changed'].message_type = _READRECEIPTCHANGEDEVENT
_EVENT_EVENTBODY.fields_by_name['event_type'].enum_type = _EVENT_EVENTTYPE
_EVENT_EVENTBODY.containing_type = _EVENT
_EVENT.fields_by_name['group_id'].message_type = _GROUPID
_EVENT.fields_by_name['type'].enum_type = _EVENT_EVENTTYPE
_EVENT.fields_by_name['body'].message_type = _EVENT_EVENTBODY
_EVENT.fields_by_name['user_id'].message_type = _USERID
_EVENT.fields_by_name['user_revision'].message_type = _WRITEREVISION
_EVENT.fields_by_name['group_revision'].message_type = _WRITEREVISION
_EVENT.fields_by_name['bodies'].message_type = _EVENT_EVENTBODY
_EVENT_EVENTTYPE.containing_type = _EVENT
_INVITETYPE_TYPE.containing_type = _INVITETYPE
_ROOMUPDATEDMETADATA_ROOMRENAMEMETADATA.containing_type = _ROOMUPDATEDMETADATA
_ROOMUPDATEDMETADATA_GROUPDETAILSUPDATEDMETADATA.fields_by_name['new_group_details'].message_type = _GROUPDETAILS
_ROOMUPDATEDMETADATA_GROUPDETAILSUPDATEDMETADATA.fields_by_name['prev_group_details'].message_type = _GROUPDETAILS
_ROOMUPDATEDMETADATA_GROUPDETAILSUPDATEDMETADATA.containing_type = _ROOMUPDATEDMETADATA
_ROOMUPDATEDMETADATA.fields_by_name['visibility'].message_type = _GROUPVISIBILITY
_ROOMUPDATEDMETADATA.fields_by_name['rename_metadata'].message_type = _ROOMUPDATEDMETADATA_ROOMRENAMEMETADATA
_ROOMUPDATEDMETADATA.fields_by_name['group_details_metadata'].message_type = _ROOMUPDATEDMETADATA_GROUPDETAILSUPDATEDMETADATA
_ROOMUPDATEDMETADATA.fields_by_name['initiator'].message_type = _USER
_MEETINGSPACE_PHONEACCESS.containing_type = _MEETINGSPACE
_MEETINGSPACE_UNIVERSALPHONEACCESS.containing_type = _MEETINGSPACE
_MEETINGSPACE_PRESENTER.containing_type = _MEETINGSPACE
_MEETINGSPACE_RECORDINGINFO_RECORDINGEVENT.fields_by_name['type'].enum_type = _MEETINGSPACE_RECORDINGINFO_RECORDINGEVENT_EVENTTYPE
_MEETINGSPACE_RECORDINGINFO_RECORDINGEVENT.containing_type = _MEETINGSPACE_RECORDINGINFO
_MEETINGSPACE_RECORDINGINFO_RECORDINGEVENT_EVENTTYPE.containing_type = _MEETINGSPACE_RECORDINGINFO_RECORDINGEVENT
_MEETINGSPACE_RECORDINGINFO.fields_by_name['recording_status'].enum_type = _MEETINGSPACE_RECORDINGINFO_RECORDINGSTATUS
_MEETINGSPACE_RECORDINGINFO.fields_by_name['latest_recording_event'].message_type = _MEETINGSPACE_RECORDINGINFO_RECORDINGEVENT
_MEETINGSPACE_RECORDINGINFO.fields_by_name['recording_application_type'].enum_type = _RECORDINGAPPLICATIONTYPE
_MEETINGSPACE_RECORDINGINFO.containing_type = _MEETINGSPACE
_MEETINGSPACE_RECORDINGINFO_RECORDINGSTATUS.containing_type = _MEETINGSPACE_RECORDINGINFO
_MEETINGSPACE_STREAMINGSESSIONINFO_SESSIONEVENT.fields_by_name['type'].enum_type = _MEETINGSPACE_STREAMINGSESSIONINFO_SESSIONEVENT_EVENTTYPE
_MEETINGSPACE_STREAMINGSESSIONINFO_SESSIONEVENT.containing_type = _MEETINGSPACE_STREAMINGSESSIONINFO
_MEETINGSPACE_STREAMINGSESSIONINFO_SESSIONEVENT_EVENTTYPE.containing_type = _MEETINGSPACE_STREAMINGSESSIONINFO_SESSIONEVENT
_MEETINGSPACE_STREAMINGSESSIONINFO_STREAMVIEWERSTATS.containing_type = _MEETINGSPACE_STREAMINGSESSIONINFO
_MEETINGSPACE_STREAMINGSESSIONINFO.fields_by_name['status'].enum_type = _MEETINGSPACE_STREAMINGSESSIONINFO_STATUS
_MEETINGSPACE_STREAMINGSESSIONINFO.fields_by_name['application_type'].enum_type = _RECORDINGAPPLICATIONTYPE
_MEETINGSPACE_STREAMINGSESSIONINFO.fields_by_name['latest_session_event'].message_type = _MEETINGSPACE_STREAMINGSESSIONINFO_SESSIONEVENT
_MEETINGSPACE_STREAMINGSESSIONINFO.fields_by_name['viewer_access_policy'].enum_type = _BROADCASTACCESSPOLICY
_MEETINGSPACE_STREAMINGSESSIONINFO.fields_by_name['viewer_stats'].message_type = _MEETINGSPACE_STREAMINGSESSIONINFO_STREAMVIEWERSTATS
_MEETINGSPACE_STREAMINGSESSIONINFO.containing_type = _MEETINGSPACE
_MEETINGSPACE_STREAMINGSESSIONINFO_STATUS.containing_type = _MEETINGSPACE_STREAMINGSESSIONINFO
_MEETINGSPACE_CALLINFO_CALLSETTINGS.containing_type = _MEETINGSPACE_CALLINFO
_MEETINGSPACE_CALLINFO_PAYGATEINFO.fields_by_name['call_ending_soon_warning_time'].message_type = _COMGOOGLEPROTOBUFTIMESTAMP
_MEETINGSPACE_CALLINFO_PAYGATEINFO.fields_by_name['call_ending_time'].message_type = _COMGOOGLEPROTOBUFTIMESTAMP
_MEETINGSPACE_CALLINFO_PAYGATEINFO.containing_type = _MEETINGSPACE_CALLINFO
_MEETINGSPACE_CALLINFO_CSEINFO.containing_type = _MEETINGSPACE_CALLINFO
_MEETINGSPACE_CALLINFO.fields_by_name['presenter'].message_type = _MEETINGSPACE_PRESENTER
_MEETINGSPACE_CALLINFO.fields_by_name['recording_info'].message_type = _MEETINGSPACE_RECORDINGINFO
_MEETINGSPACE_CALLINFO.fields_by_name['streaming_sessions'].message_type = _MEETINGSPACE_STREAMINGSESSIONINFO
_MEETINGSPACE_CALLINFO.fields_by_name['settings'].message_type = _MEETINGSPACE_CALLINFO_CALLSETTINGS
_MEETINGSPACE_CALLINFO.fields_by_name['paygate_info'].message_type = _MEETINGSPACE_CALLINFO_PAYGATEINFO
_MEETINGSPACE_CALLINFO.fields_by_name['cse_info'].message_type = _MEETINGSPACE_CALLINFO_CSEINFO
_MEETINGSPACE_CALLINFO.containing_type = _MEETINGSPACE
_MEETINGSPACE_GATEWAYACCESS.containing_type = _MEETINGSPACE
_MEETINGSPACE_GATEWAYSIPACCESS.containing_type = _MEETINGSPACE
_MEETINGSPACE_BROADCASTACCESS.containing_type = _MEETINGSPACE
_MEETINGSPACE_SETTINGS.containing_type = _MEETINGSPACE
_MEETINGSPACE.fields_by_name['phone_access'].message_type = _MEETINGSPACE_PHONEACCESS
_MEETINGSPACE.fields_by_name['universal_phone_access'].message_type = _MEETINGSPACE_UNIVERSALPHONEACCESS
_MEETINGSPACE.fields_by_name['call_info'].message_type = _MEETINGSPACE_CALLINFO
_MEETINGSPACE.fields_by_name['gateway_access'].message_type = _MEETINGSPACE_GATEWAYACCESS
_MEETINGSPACE.fields_by_name['accepted_number_class'].enum_type = _DIALINNUMBERCLASS
_MEETINGSPACE.fields_by_name['gateway_sip_access'].message_type = _MEETINGSPACE_GATEWAYSIPACCESS
_MEETINGSPACE.fields_by_name['broadcast_access'].message_type = _MEETINGSPACE_BROADCASTACCESS
_MEETINGSPACE.fields_by_name['settings'].message_type = _MEETINGSPACE_SETTINGS
_VIDEOCALLMETADATA.fields_by_name['meeting_space'].message_type = _MEETINGSPACE
_MEMBERSHIPCHANGEDMETADATA_AFFECTEDMEMBERSHIP.fields_by_name['affected_member'].message_type = _MEMBERID
_MEMBERSHIPCHANGEDMETADATA_AFFECTEDMEMBERSHIP.fields_by_name['prior_membership_state'].enum_type = _MEMBERSHIPSTATE
_MEMBERSHIPCHANGEDMETADATA_AFFECTEDMEMBERSHIP.fields_by_name['prior_membership_role'].enum_type = _MEMBERSHIPROLE
_MEMBERSHIPCHANGEDMETADATA_AFFECTEDMEMBERSHIP.fields_by_name['target_membership_role'].enum_type = _MEMBERSHIPROLE
_MEMBERSHIPCHANGEDMETADATA_AFFECTEDMEMBERSHIP.containing_type = _MEMBERSHIPCHANGEDMETADATA
_MEMBERSHIPCHANGEDMETADATA.fields_by_name['type'].enum_type = _MEMBERSHIPCHANGEDMETADATA_TYPE
_MEMBERSHIPCHANGEDMETADATA.fields_by_name['affected_memberships'].message_type = _MEMBERSHIPCHANGEDMETADATA_AFFECTEDMEMBERSHIP
_MEMBERSHIPCHANGEDMETADATA.fields_by_name['initiator'].message_type = _USERID
_MEMBERSHIPCHANGEDMETADATA.fields_by_name['affected_members'].message_type = _MEMBERID
_MEMBERSHIPCHANGEDMETADATA.fields_by_name['initiator_profile'].message_type = _USER
_MEMBERSHIPCHANGEDMETADATA.fields_by_name['affected_member_profiles'].message_type = _MEMBER
_MEMBERSHIPCHANGEDMETADATA_TYPE.containing_type = _MEMBERSHIPCHANGEDMETADATA
_USERMENTIONMETADATA.fields_by_name['id'].message_type = _USERID
_USERMENTIONMETADATA.fields_by_name['invitee_info'].message_type = _INVITEEINFO
_USERMENTIONMETADATA.fields_by_name['type'].enum_type = _USERMENTIONMETADATA_TYPE
_USERMENTIONMETADATA_TYPE.containing_type = _USERMENTIONMETADATA
_SLASHCOMMANDMETADATA.fields_by_name['id'].message_type = _USERID
_SLASHCOMMANDMETADATA.fields_by_name['type'].enum_type = _SLASHCOMMANDMETADATA_TYPE
_SLASHCOMMANDMETADATA_TYPE.containing_type = _SLASHCOMMANDMETADATA
_GROUPVISIBILITY_VISIBILITYSTATE.containing_type = _GROUPVISIBILITY
_GETSERVERTIMEREQUEST.fields_by_name['request_header'].message_type = _REQUESTHEADER
_GETSERVERTIMERESPONSE.fields_by_name['timestamp'].message_type = _COMGOOGLEPROTOBUFTIMESTAMP
_CATCHUPGROUPREQUEST.fields_by_name['request_header'].message_type = _REQUESTHEADER
_CATCHUPGROUPREQUEST.fields_by_name['group_id'].message_type = _GROUPID
_CATCHUPGROUPREQUEST.fields_by_name['range'].message_type = _CATCHUPRANGE
_CATCHUPUSERREQUEST.fields_by_name['request_header'].message_type = _REQUESTHEADER
_CATCHUPUSERREQUEST.fields_by_name['range'].message_type = _CATCHUPRANGE
_CATCHUPRESPONSE.fields_by_name['events'].message_type = _EVENT
_CATCHUPRESPONSE.fields_by_name['status'].enum_type = _CATCHUPRESPONSE_RESPONSESTATUS
_CATCHUPRESPONSE_RESPONSESTATUS.containing_type = _CATCHUPRESPONSE
_GETGROUPREQUEST.fields_by_name['request_header'].message_type = _REQUESTHEADER
_GETGROUPREQUEST.fields_by_name['group_id'].message_type = _GROUPID
_GETGROUPREQUEST.fields_by_name['fetch_options'].enum_type = _GETGROUPREQUEST_FETCHOPTIONS
_GETGROUPREQUEST.fields_by_name['user_not_older_than'].message_type = _REFERENCEREVISION
_GETGROUPREQUEST.fields_by_name['group_not_older_than'].message_type = _REFERENCEREVISION
_GETGROUPREQUEST_FETCHOPTIONS.containing_type = _GETGROUPREQUEST
_GETGROUPRESPONSE.fields_by_name['group'].message_type = _GROUP
_GETGROUPRESPONSE.fields_by_name['memberships'].message_type = _MEMBERSHIP
_GETGROUPRESPONSE.fields_by_name['user_revision'].message_type = _READREVISION
_GETGROUPRESPONSE.fields_by_name['group_revision'].message_type = _READREVISION
_GETGROUPRESPONSE.fields_by_name['membership_state'].enum_type = _MEMBERSHIPSTATE
_GETGROUPRESPONSE.fields_by_name['joined_member_ids'].message_type = _MEMBERID
_GETGROUPRESPONSE.fields_by_name['invited_member_ids'].message_type = _MEMBERID
_GETGROUPRESPONSE.fields_by_name['read_receipt_set'].message_type = _READRECEIPTSET
_GETGROUPRESPONSE.fields_by_name['snippet'].message_type = _MESSAGE
_WORLDSECTION.fields_by_name['world_section_type'].enum_type = _WORLDSECTION_WORLDSECTIONTYPE
_WORLDSECTION_WORLDSECTIONTYPE.containing_type = _WORLDSECTION
_WORLDFILTER.fields_by_name['starred_state'].enum_type = _WORLDFILTER_STARREDSTATE
_WORLDFILTER.fields_by_name['visibility_state'].enum_type = _WORLDFILTER_VISIBILITYSTATE
_WORLDFILTER.fields_by_name['read_state'].enum_type = _WORLDFILTER_READSTATE
_WORLDFILTER.fields_by_name['block_state'].enum_type = _WORLDFILTER_BLOCKSTATE
_WORLDFILTER.fields_by_name['named_state'].enum_type = _WORLDFILTER_NAMEDSTATE
_WORLDFILTER.fields_by_name['membership_state'].enum_type = _MEMBERSHIPSTATE
_WORLDFILTER.fields_by_name['invite_category'].enum_type = _INVITECATEGORY
_WORLDFILTER.fields_by_name['member_type'].enum_type = _WORLDFILTER_MEMBERTYPE
_WORLDFILTER.fields_by_name['group_type'].enum_type = _WORLDFILTER_GROUPTYPE
_WORLDFILTER_STARREDSTATE.containing_type = _WORLDFILTER
_WORLDFILTER_VISIBILITYSTATE.containing_type = _WORLDFILTER
_WORLDFILTER_READSTATE.containing_type = _WORLDFILTER
_WORLDFILTER_BLOCKSTATE.containing_type = _WORLDFILTER
_WORLDFILTER_NAMEDSTATE.containing_type = _WORLDFILTER
_WORLDFILTER_MEMBERTYPE.containing_type = _WORLDFILTER
_WORLDFILTER_GROUPTYPE.containing_type = _WORLDFILTER
_NAMEUSERS.fields_by_name['name_user_ids'].message_type = _USERID
_WORLDITEMLITE_MEMBERSHIPLITE.fields_by_name['user_id'].message_type = _USERID
_WORLDITEMLITE_MEMBERSHIPLITE.fields_by_name['membership_state'].enum_type = _MEMBERSHIPSTATE
_WORLDITEMLITE_MEMBERSHIPLITE.containing_type = _WORLDITEMLITE
_WORLDITEMLITE_DMMEMBERS.fields_by_name['members'].message_type = _USERID
_WORLDITEMLITE_DMMEMBERS.fields_by_name['memberships'].message_type = _WORLDITEMLITE_MEMBERSHIPLITE
_WORLDITEMLITE_DMMEMBERS.containing_type = _WORLDITEMLITE
_WORLDITEMLITE_GROUPLITE.fields_by_name['creator_id'].message_type = _USERID
_WORLDITEMLITE_GROUPLITE.fields_by_name['retention_settings'].message_type = _RETENTIONSETTINGS
_WORLDITEMLITE_GROUPLITE.containing_type = _WORLDITEMLITE
_WORLDITEMLITE_FLATGROUP.containing_type = _WORLDITEMLITE
_WORLDITEMLITE_THREADEDGROUP.containing_type = _WORLDITEMLITE
_WORLDITEMLITE.fields_by_name['group_id'].message_type = _GROUPID
_WORLDITEMLITE.fields_by_name['group_revision'].message_type = _READREVISION
_WORLDITEMLITE.fields_by_name['read_state'].message_type = _GROUPREADSTATE
_WORLDITEMLITE.fields_by_name['dm_members'].message_type = _WORLDITEMLITE_DMMEMBERS
_WORLDITEMLITE.fields_by_name['name_users'].message_type = _NAMEUSERS
_WORLDITEMLITE.fields_by_name['group_lite'].message_type = _WORLDITEMLITE_GROUPLITE
_WORLDITEMLITE.fields_by_name['message'].message_type = _MESSAGE
_WORLDITEMLITE.fields_by_name['flat_group'].message_type = _WORLDITEMLITE_FLATGROUP
_WORLDITEMLITE.fields_by_name['threaded_group'].message_type = _WORLDITEMLITE_THREADEDGROUP
_WORLDITEMLITE.fields_by_name['attribute_checker_group_type'].enum_type = _SHAREDATTRIBUTECHECKERGROUPTYPE
_WORLDITEMLITE.fields_by_name['group_support_level'].enum_type = _GROUPSUPPORTLEVEL
_WORLDITEMLITE.fields_by_name['group_unsupported_reason'].enum_type = _GROUPUNSUPPORTEDREASON
_WORLDSECTIONREQUEST.fields_by_name['world_section'].message_type = _WORLDSECTION
_WORLDSECTIONREQUEST.fields_by_name['world_filter'].message_type = _WORLDFILTER
_WORLDSECTIONRESPONSE.fields_by_name['world_section'].message_type = _WORLDSECTION
_WORLDSECTIONRESPONSE.fields_by_name['world_filter'].message_type = _WORLDFILTER
_WORLDSECTIONRESPONSE.fields_by_name['world_items'].message_type = _WORLDITEMLITE
_PAGINATEDWORLDREQUEST.fields_by_name['request_header'].message_type = _REQUESTHEADER
_PAGINATEDWORLDREQUEST.fields_by_name['world_section_requests'].message_type = _WORLDSECTIONREQUEST
_PAGINATEDWORLDREQUEST.fields_by_name['fetch_options'].enum_type = _PAGINATEDWORLDREQUEST_FETCHOPTIONS
_PAGINATEDWORLDREQUEST_FETCHOPTIONS.containing_type = _PAGINATEDWORLDREQUEST
_PAGINATEDWORLDRESPONSE.fields_by_name['world_section_responses'].message_type = _WORLDSECTIONRESPONSE
_PAGINATEDWORLDRESPONSE.fields_by_name['user_revision'].message_type = _READREVISION
_PAGINATEDWORLDRESPONSE.fields_by_name['world_items'].message_type = _WORLDITEMLITE
_REMOVEMEMBERSHIPSREQUEST.fields_by_name['request_header'].message_type = _REQUESTHEADER
_REMOVEMEMBERSHIPSREQUEST.fields_by_name['member_ids'].message_type = _MEMBERID
_REMOVEMEMBERSHIPSREQUEST.fields_by_name['group_id'].message_type = _GROUPID
_REMOVEMEMBERSHIPSREQUEST.fields_by_name['membership_state'].enum_type = _MEMBERSHIPSTATE
_REMOVEMEMBERSHIPRESULT.fields_by_name['request_header'].message_type = _REQUESTHEADER
_REMOVEMEMBERSHIPRESULT.fields_by_name['member_id'].message_type = _MEMBERID
_REMOVEMEMBERSHIPSRESPONSE.fields_by_name['results'].message_type = _REMOVEMEMBERSHIPRESULT
_REMOVEMEMBERSHIPSRESPONSE.fields_by_name['retention_settings'].message_type = _RETENTIONSETTINGS
_REMOVEMEMBERSHIPSRESPONSE.fields_by_name['retention_settings_group_revision'].message_type = _WRITEREVISION
_HIDEGROUPREQUEST.fields_by_name['request_header'].message_type = _REQUESTHEADER
_HIDEGROUPREQUEST.fields_by_name['id'].message_type = _GROUPID
_HIDEGROUPRESPONSE.fields_by_name['read_state'].message_type = _GROUPREADSTATE
_HIDEGROUPRESPONSE.fields_by_name['user_revision'].message_type = _WRITEREVISION
_INVITENOTIFICATIONSETTINGS.fields_by_name['option'].enum_type = _INVITENOTIFICATIONSETTINGS_NOTIFICATIONOPTION
_INVITENOTIFICATIONSETTINGS_NOTIFICATIONOPTION.containing_type = _INVITENOTIFICATIONSETTINGS
_CREATEMEMBERSHIPREQUEST.fields_by_name['request_header'].message_type = _REQUESTHEADER
_CREATEMEMBERSHIPREQUEST.fields_by_name['member_ids'].message_type = _MEMBERID
_CREATEMEMBERSHIPREQUEST.fields_by_name['invitee_member_infos'].message_type = _INVITEEMEMBERINFO
_CREATEMEMBERSHIPREQUEST.fields_by_name['membership_state'].enum_type = _MEMBERSHIPSTATE
_CREATEMEMBERSHIPREQUEST.fields_by_name['group_id'].message_type = _GROUPID
_CREATEMEMBERSHIPREQUEST.fields_by_name['notification_settings'].message_type = _INVITENOTIFICATIONSETTINGS
_CREATEMEMBERSHIPRESPONSE.fields_by_name['results'].message_type = _CREATEMEMBERSHIPRESULT
_CREATEMEMBERSHIPRESPONSE.fields_by_name['group_revision'].message_type = _WRITEREVISION
_CREATEMEMBERSHIPRESPONSE.fields_by_name['retention_settings'].message_type = _RETENTIONSETTINGS
_CREATEMEMBERSHIPRESPONSE.fields_by_name['retention_settings_group_revision'].message_type = _WRITEREVISION
_MARKGROUPREADSTATEREQUEST.fields_by_name['request_header'].message_type = _REQUESTHEADER
_MARKGROUPREADSTATEREQUEST.fields_by_name['id'].message_type = _GROUPID
_MARKGROUPREADSTATERESPONSE.fields_by_name['read_state'].message_type = _GROUPREADSTATE
_MARKGROUPREADSTATERESPONSE.fields_by_name['user_revision'].message_type = _WRITEREVISION
_SETPRESENCESHAREDREQUEST.fields_by_name['request_header'].message_type = _REQUESTHEADER
_SETPRESENCESHAREDRESPONSE.fields_by_name['user_status'].message_type = _USERSTATUS
_SETPRESENCESHAREDRESPONSE.fields_by_name['user_revision'].message_type = _WRITEREVISION
_SETDNDDURATIONREQUEST.fields_by_name['request_header'].message_type = _REQUESTHEADER
_SETDNDDURATIONREQUEST.fields_by_name['current_dnd_state'].enum_type = _SETDNDDURATIONREQUEST_STATE
_SETDNDDURATIONREQUEST_STATE.containing_type = _SETDNDDURATIONREQUEST
_SETDNDDURATIONRESPONSE.fields_by_name['user_status'].message_type = _USERSTATUS
_SETDNDDURATIONRESPONSE.fields_by_name['user_revision'].message_type = _WRITEREVISION
_UPDATEGROUPREQUEST.fields_by_name['request_header'].message_type = _REQUESTHEADER
_UPDATEGROUPREQUEST.fields_by_name['space_id'].message_type = _SPACEID
_UPDATEGROUPREQUEST.fields_by_name['update_masks'].enum_type = _UPDATEGROUPREQUEST_UPDATEMASK
_UPDATEGROUPREQUEST.fields_by_name['visibility'].message_type = _GROUPVISIBILITY
_UPDATEGROUPREQUEST_UPDATEMASK.containing_type = _UPDATEGROUPREQUEST
_UPDATEGROUPRESPONSE.fields_by_name['group'].message_type = _GROUP
_UPDATEGROUPRESPONSE.fields_by_name['group_revision'].message_type = _WRITEREVISION
_BLOCKENTITYREQUEST.fields_by_name['request_header'].message_type = _REQUESTHEADER
_BLOCKENTITYREQUEST.fields_by_name['user_id'].message_type = _USERID
_BLOCKENTITYREQUEST.fields_by_name['group_id'].message_type = _GROUPID
_BLOCKENTITYRESPONSE.fields_by_name['read_state'].message_type = _GROUPREADSTATE
_BLOCKENTITYRESPONSE.fields_by_name['user_revision'].message_type = _WRITEREVISION
_SETCUSTOMSTATUSREQUEST.fields_by_name['request_header'].message_type = _REQUESTHEADER
_SETCUSTOMSTATUSREQUEST.fields_by_name['custom_status'].message_type = _CUSTOMSTATUS
_SETCUSTOMSTATUSRESPONSE.fields_by_name['user_status'].message_type = _USERSTATUS
_SETCUSTOMSTATUSRESPONSE.fields_by_name['user_revision'].message_type = _WRITEREVISION
DESCRIPTOR.message_types_by_name['UserId'] = _USERID
DESCRIPTOR.message_types_by_name['User'] = _USER
DESCRIPTOR.message_types_by_name['UserBlockRelationship'] = _USERBLOCKRELATIONSHIP
DESCRIPTOR.message_types_by_name['DndSettings'] = _DNDSETTINGS
DESCRIPTOR.message_types_by_name['Emoji'] = _EMOJI
DESCRIPTOR.message_types_by_name['CustomStatus'] = _CUSTOMSTATUS
DESCRIPTOR.message_types_by_name['UserStatus'] = _USERSTATUS
DESCRIPTOR.message_types_by_name['UserStatusUpdatedEvent'] = _USERSTATUSUPDATEDEVENT
DESCRIPTOR.message_types_by_name['GetUserStatusRequest'] = _GETUSERSTATUSREQUEST
DESCRIPTOR.message_types_by_name['GetUserStatusResponse'] = _GETUSERSTATUSRESPONSE
DESCRIPTOR.message_types_by_name['GetSelfUserStatusRequest'] = _GETSELFUSERSTATUSREQUEST
DESCRIPTOR.message_types_by_name['GetSelfUserStatusResponse'] = _GETSELFUSERSTATUSRESPONSE
DESCRIPTOR.message_types_by_name['ClientFeatureCapabilities'] = _CLIENTFEATURECAPABILITIES
DESCRIPTOR.message_types_by_name['RequestHeader'] = _REQUESTHEADER
DESCRIPTOR.message_types_by_name['Member'] = _MEMBER
DESCRIPTOR.message_types_by_name['MemberId'] = _MEMBERID
DESCRIPTOR.message_types_by_name['MembershipId'] = _MEMBERSHIPID
DESCRIPTOR.message_types_by_name['Membership'] = _MEMBERSHIP
DESCRIPTOR.message_types_by_name['MemberProfile'] = _MEMBERPROFILE
DESCRIPTOR.message_types_by_name['GetMembersRequest'] = _GETMEMBERSREQUEST
DESCRIPTOR.message_types_by_name['GetMembersResponse'] = _GETMEMBERSRESPONSE
DESCRIPTOR.message_types_by_name['UserPresence'] = _USERPRESENCE
DESCRIPTOR.message_types_by_name['GetUserPresenceRequest'] = _GETUSERPRESENCEREQUEST
DESCRIPTOR.message_types_by_name['GetUserPresenceResponse'] = _GETUSERPRESENCERESPONSE
DESCRIPTOR.message_types_by_name['JAddOnsIdentifier'] = _JADDONSIDENTIFIER
DESCRIPTOR.message_types_by_name['JAddOnsFormattedText'] = _JADDONSFORMATTEDTEXT
DESCRIPTOR.message_types_by_name['JAddOnsImageCropStyle'] = _JADDONSIMAGECROPSTYLE
DESCRIPTOR.message_types_by_name['JAddOnsThemeColors'] = _JADDONSTHEMECOLORS
DESCRIPTOR.message_types_by_name['JAddOnsOpenLink'] = _JADDONSOPENLINK
DESCRIPTOR.message_types_by_name['JAddOnsCardItem'] = _JADDONSCARDITEM
DESCRIPTOR.message_types_by_name['JAddOnsPushCard'] = _JADDONSPUSHCARD
DESCRIPTOR.message_types_by_name['JAddOnsOnClick'] = _JADDONSONCLICK
DESCRIPTOR.message_types_by_name['JAddOnsTextWidget'] = _JADDONSTEXTWIDGET
DESCRIPTOR.message_types_by_name['JAddOnsImageComponent'] = _JADDONSIMAGECOMPONENT
DESCRIPTOR.message_types_by_name['JAddOnsGrid'] = _JADDONSGRID
DESCRIPTOR.message_types_by_name['JAddOnsBorderStyle'] = _JADDONSBORDERSTYLE
DESCRIPTOR.message_types_by_name['JAddOnsLabelContentPair'] = _JADDONSLABELCONTENTPAIR
DESCRIPTOR.message_types_by_name['JAddOnsLabelContentPairWidget'] = _JADDONSLABELCONTENTPAIRWIDGET
DESCRIPTOR.message_types_by_name['JAddOnsWidget'] = _JADDONSWIDGET
DESCRIPTOR.message_types_by_name['JAddOnsFormAction'] = _JADDONSFORMACTION
DESCRIPTOR.message_types_by_name['JAddOnsContextualAddOn'] = _JADDONSCONTEXTUALADDON
DESCRIPTOR.message_types_by_name['Html'] = _HTML
DESCRIPTOR.message_types_by_name['HtmlAttachment'] = _HTMLATTACHMENT
DESCRIPTOR.message_types_by_name['Attachment'] = _ATTACHMENT
DESCRIPTOR.message_types_by_name['Reaction'] = _REACTION
DESCRIPTOR.message_types_by_name['MessageParentId'] = _MESSAGEPARENTID
DESCRIPTOR.message_types_by_name['MessageId'] = _MESSAGEID
DESCRIPTOR.message_types_by_name['Message'] = _MESSAGE
DESCRIPTOR.message_types_by_name['DriveMetadata'] = _DRIVEMETADATA
DESCRIPTOR.message_types_by_name['Url'] = _URL
DESCRIPTOR.message_types_by_name['TrustedResourceUrl'] = _TRUSTEDRESOURCEURL
DESCRIPTOR.message_types_by_name['UrlMetadata'] = _URLMETADATA
DESCRIPTOR.message_types_by_name['UploadMetadata'] = _UPLOADMETADATA
DESCRIPTOR.message_types_by_name['FormatMetadata'] = _FORMATMETADATA
DESCRIPTOR.message_types_by_name['Annotation'] = _ANNOTATION
DESCRIPTOR.message_types_by_name['TypingContext'] = _TYPINGCONTEXT
DESCRIPTOR.message_types_by_name['SetTypingStateRequest'] = _SETTYPINGSTATEREQUEST
DESCRIPTOR.message_types_by_name['SetTypingStateResponse'] = _SETTYPINGSTATERESPONSE
DESCRIPTOR.message_types_by_name['DmId'] = _DMID
DESCRIPTOR.message_types_by_name['SpaceId'] = _SPACEID
DESCRIPTOR.message_types_by_name['GroupId'] = _GROUPID
DESCRIPTOR.message_types_by_name['Group'] = _GROUP
DESCRIPTOR.message_types_by_name['GroupReadStateId'] = _GROUPREADSTATEID
DESCRIPTOR.message_types_by_name['InviteState'] = _INVITESTATE
DESCRIPTOR.message_types_by_name['GroupReadState'] = _GROUPREADSTATE
DESCRIPTOR.message_types_by_name['GroupNotificationSettings'] = _GROUPNOTIFICATIONSETTINGS
DESCRIPTOR.message_types_by_name['RetentionSettings'] = _RETENTIONSETTINGS
DESCRIPTOR.message_types_by_name['TopicId'] = _TOPICID
DESCRIPTOR.message_types_by_name['Topic'] = _TOPIC
DESCRIPTOR.message_types_by_name['MessageInfo'] = _MESSAGEINFO
DESCRIPTOR.message_types_by_name['CreateTopicRequest'] = _CREATETOPICREQUEST
DESCRIPTOR.message_types_by_name['CreateTopicResponse'] = _CREATETOPICRESPONSE
DESCRIPTOR.message_types_by_name['InviteeInfo'] = _INVITEEINFO
DESCRIPTOR.message_types_by_name['InviteeMemberInfo'] = _INVITEEMEMBERINFO
DESCRIPTOR.message_types_by_name['SpaceCreationInfo'] = _SPACECREATIONINFO
DESCRIPTOR.message_types_by_name['CreateGroupRequest'] = _CREATEGROUPREQUEST
DESCRIPTOR.message_types_by_name['CreateMembershipResult'] = _CREATEMEMBERSHIPRESULT
DESCRIPTOR.message_types_by_name['CreateGroupResponse'] = _CREATEGROUPRESPONSE
DESCRIPTOR.message_types_by_name['CreateDmRequest'] = _CREATEDMREQUEST
DESCRIPTOR.message_types_by_name['CreateDmResponse'] = _CREATEDMRESPONSE
DESCRIPTOR.message_types_by_name['ListTopicsResponse'] = _LISTTOPICSRESPONSE
DESCRIPTOR.message_types_by_name['ListTopicsRequest'] = _LISTTOPICSREQUEST
DESCRIPTOR.message_types_by_name['ListMembersRequest'] = _LISTMEMBERSREQUEST
DESCRIPTOR.message_types_by_name['ListMembersResponse'] = _LISTMEMBERSRESPONSE
DESCRIPTOR.message_types_by_name['ReadReceipt'] = _READRECEIPT
DESCRIPTOR.message_types_by_name['ReadReceiptSet'] = _READRECEIPTSET
DESCRIPTOR.message_types_by_name['WebPushNotification'] = _WEBPUSHNOTIFICATION
DESCRIPTOR.message_types_by_name['AndroidLocalNotification'] = _ANDROIDLOCALNOTIFICATION
DESCRIPTOR.message_types_by_name['IosLocalNotification'] = _IOSLOCALNOTIFICATION
DESCRIPTOR.message_types_by_name['MobileLocalNotification'] = _MOBILELOCALNOTIFICATION
DESCRIPTOR.message_types_by_name['MessageEvent'] = _MESSAGEEVENT
DESCRIPTOR.message_types_by_name['TypingStateChangedEvent'] = _TYPINGSTATECHANGEDEVENT
DESCRIPTOR.message_types_by_name['MembershipChangedEvent'] = _MEMBERSHIPCHANGEDEVENT
DESCRIPTOR.message_types_by_name['ReadReceiptChangedEvent'] = _READRECEIPTCHANGEDEVENT
DESCRIPTOR.message_types_by_name['GroupViewedEvent'] = _GROUPVIEWEDEVENT
DESCRIPTOR.message_types_by_name['WebPushNotificationEvent'] = _WEBPUSHNOTIFICATIONEVENT
DESCRIPTOR.message_types_by_name['StreamEventsRequest'] = _STREAMEVENTSREQUEST
DESCRIPTOR.message_types_by_name['ClientInfo'] = _CLIENTINFO
DESCRIPTOR.message_types_by_name['ClientNotificationsState'] = _CLIENTNOTIFICATIONSSTATE
DESCRIPTOR.message_types_by_name['PingEvent'] = _PINGEVENT
DESCRIPTOR.message_types_by_name['ClockSyncRequest'] = _CLOCKSYNCREQUEST
DESCRIPTOR.message_types_by_name['ClockSyncResponse'] = _CLOCKSYNCRESPONSE
DESCRIPTOR.message_types_by_name['GroupSubscriptionEvent'] = _GROUPSUBSCRIPTIONEVENT
DESCRIPTOR.message_types_by_name['StreamEventsResponse'] = _STREAMEVENTSRESPONSE
DESCRIPTOR.message_types_by_name['Event'] = _EVENT
DESCRIPTOR.message_types_by_name['InviteType'] = _INVITETYPE
DESCRIPTOR.message_types_by_name['GroupDetails'] = _GROUPDETAILS
DESCRIPTOR.message_types_by_name['RoomUpdatedMetadata'] = _ROOMUPDATEDMETADATA
DESCRIPTOR.message_types_by_name['MeetingSpace'] = _MEETINGSPACE
DESCRIPTOR.message_types_by_name['VideoCallMetadata'] = _VIDEOCALLMETADATA
DESCRIPTOR.message_types_by_name['MembershipChangedMetadata'] = _MEMBERSHIPCHANGEDMETADATA
DESCRIPTOR.message_types_by_name['UserMentionMetadata'] = _USERMENTIONMETADATA
DESCRIPTOR.message_types_by_name['SlashCommandMetadata'] = _SLASHCOMMANDMETADATA
DESCRIPTOR.message_types_by_name['GroupVisibility'] = _GROUPVISIBILITY
DESCRIPTOR.message_types_by_name['ComGoogleProtobufTimestamp'] = _COMGOOGLEPROTOBUFTIMESTAMP
DESCRIPTOR.message_types_by_name['GetServerTimeRequest'] = _GETSERVERTIMEREQUEST
DESCRIPTOR.message_types_by_name['GetServerTimeResponse'] = _GETSERVERTIMERESPONSE
DESCRIPTOR.message_types_by_name['CatchUpRange'] = _CATCHUPRANGE
DESCRIPTOR.message_types_by_name['CatchUpGroupRequest'] = _CATCHUPGROUPREQUEST
DESCRIPTOR.message_types_by_name['CatchUpUserRequest'] = _CATCHUPUSERREQUEST
DESCRIPTOR.message_types_by_name['CatchUpResponse'] = _CATCHUPRESPONSE
DESCRIPTOR.message_types_by_name['GetGroupRequest'] = _GETGROUPREQUEST
DESCRIPTOR.message_types_by_name['GetGroupResponse'] = _GETGROUPRESPONSE
DESCRIPTOR.message_types_by_name['WorldSection'] = _WORLDSECTION
DESCRIPTOR.message_types_by_name['WorldFilter'] = _WORLDFILTER
DESCRIPTOR.message_types_by_name['NameUsers'] = _NAMEUSERS
DESCRIPTOR.message_types_by_name['WorldItemLite'] = _WORLDITEMLITE
DESCRIPTOR.message_types_by_name['WorldSectionRequest'] = _WORLDSECTIONREQUEST
DESCRIPTOR.message_types_by_name['WorldSectionResponse'] = _WORLDSECTIONRESPONSE
DESCRIPTOR.message_types_by_name['PaginatedWorldRequest'] = _PAGINATEDWORLDREQUEST
DESCRIPTOR.message_types_by_name['PaginatedWorldResponse'] = _PAGINATEDWORLDRESPONSE
DESCRIPTOR.message_types_by_name['RemoveMembershipsRequest'] = _REMOVEMEMBERSHIPSREQUEST
DESCRIPTOR.message_types_by_name['RemoveMembershipResult'] = _REMOVEMEMBERSHIPRESULT
DESCRIPTOR.message_types_by_name['RemoveMembershipsResponse'] = _REMOVEMEMBERSHIPSRESPONSE
DESCRIPTOR.message_types_by_name['HideGroupRequest'] = _HIDEGROUPREQUEST
DESCRIPTOR.message_types_by_name['HideGroupResponse'] = _HIDEGROUPRESPONSE
DESCRIPTOR.message_types_by_name['InviteNotificationSettings'] = _INVITENOTIFICATIONSETTINGS
DESCRIPTOR.message_types_by_name['CreateMembershipRequest'] = _CREATEMEMBERSHIPREQUEST
DESCRIPTOR.message_types_by_name['CreateMembershipResponse'] = _CREATEMEMBERSHIPRESPONSE
DESCRIPTOR.message_types_by_name['MarkGroupReadstateRequest'] = _MARKGROUPREADSTATEREQUEST
DESCRIPTOR.message_types_by_name['MarkGroupReadstateResponse'] = _MARKGROUPREADSTATERESPONSE
DESCRIPTOR.message_types_by_name['SetPresenceSharedRequest'] = _SETPRESENCESHAREDREQUEST
DESCRIPTOR.message_types_by_name['SetPresenceSharedResponse'] = _SETPRESENCESHAREDRESPONSE
DESCRIPTOR.message_types_by_name['SetDndDurationRequest'] = _SETDNDDURATIONREQUEST
DESCRIPTOR.message_types_by_name['SetDndDurationResponse'] = _SETDNDDURATIONRESPONSE
DESCRIPTOR.message_types_by_name['UpdateGroupRequest'] = _UPDATEGROUPREQUEST
DESCRIPTOR.message_types_by_name['UpdateGroupResponse'] = _UPDATEGROUPRESPONSE
DESCRIPTOR.message_types_by_name['BlockEntityRequest'] = _BLOCKENTITYREQUEST
DESCRIPTOR.message_types_by_name['BlockEntityResponse'] = _BLOCKENTITYRESPONSE
DESCRIPTOR.message_types_by_name['SetCustomStatusRequest'] = _SETCUSTOMSTATUSREQUEST
DESCRIPTOR.message_types_by_name['SetCustomStatusResponse'] = _SETCUSTOMSTATUSRESPONSE
DESCRIPTOR.message_types_by_name['WriteRevision'] = _WRITEREVISION
DESCRIPTOR.message_types_by_name['ReadRevision'] = _READREVISION
DESCRIPTOR.message_types_by_name['ReferenceRevision'] = _REFERENCEREVISION
DESCRIPTOR.enum_types_by_name['UserType'] = _USERTYPE
DESCRIPTOR.enum_types_by_name['InviteCategory'] = _INVITECATEGORY
DESCRIPTOR.enum_types_by_name['Presence'] = _PRESENCE
DESCRIPTOR.enum_types_by_name['DndState_State'] = _DNDSTATE_STATE
DESCRIPTOR.enum_types_by_name['TypingState'] = _TYPINGSTATE
DESCRIPTOR.enum_types_by_name['GroupUnsupportedReason'] = _GROUPUNSUPPORTEDREASON
DESCRIPTOR.enum_types_by_name['GroupSupportLevel'] = _GROUPSUPPORTLEVEL
DESCRIPTOR.enum_types_by_name['NotificationCause'] = _NOTIFICATIONCAUSE
DESCRIPTOR.enum_types_by_name['EventOrigin'] = _EVENTORIGIN
DESCRIPTOR.enum_types_by_name['Platform'] = _PLATFORM
DESCRIPTOR.enum_types_by_name['AnnotationType'] = _ANNOTATIONTYPE
DESCRIPTOR.enum_types_by_name['DialInNumberClass'] = _DIALINNUMBERCLASS
DESCRIPTOR.enum_types_by_name['RecordingApplicationType'] = _RECORDINGAPPLICATIONTYPE
DESCRIPTOR.enum_types_by_name['BroadcastAccessPolicy'] = _BROADCASTACCESSPOLICY
DESCRIPTOR.enum_types_by_name['CallStatus'] = _CALLSTATUS
DESCRIPTOR.enum_types_by_name['RingStatus'] = _RINGSTATUS
DESCRIPTOR.enum_types_by_name['UserAccountState'] = _USERACCOUNTSTATE
DESCRIPTOR.enum_types_by_name['MembershipRole'] = _MEMBERSHIPROLE
DESCRIPTOR.enum_types_by_name['AppType'] = _APPTYPE
DESCRIPTOR.enum_types_by_name['MembershipState'] = _MEMBERSHIPSTATE
DESCRIPTOR.enum_types_by_name['SharedAttributeCheckerGroupType'] = _SHAREDATTRIBUTECHECKERGROUPTYPE
_sym_db.RegisterFileDescriptor(DESCRIPTOR)

UserId = _reflection.GeneratedProtocolMessageType('UserId', (_message.Message,), {
  'DESCRIPTOR' : _USERID,
  '__module__' : 'hangups.googlechat_pb2'
  # @@protoc_insertion_point(class_scope:UserId)
  })
_sym_db.RegisterMessage(UserId)

User = _reflection.GeneratedProtocolMessageType('User', (_message.Message,), {
  'DESCRIPTOR' : _USER,
  '__module__' : 'hangups.googlechat_pb2'
  # @@protoc_insertion_point(class_scope:User)
  })
_sym_db.RegisterMessage(User)

UserBlockRelationship = _reflection.GeneratedProtocolMessageType('UserBlockRelationship', (_message.Message,), {
  'DESCRIPTOR' : _USERBLOCKRELATIONSHIP,
  '__module__' : 'hangups.googlechat_pb2'
  # @@protoc_insertion_point(class_scope:UserBlockRelationship)
  })
_sym_db.RegisterMessage(UserBlockRelationship)

DndSettings = _reflection.GeneratedProtocolMessageType('DndSettings', (_message.Message,), {
  'DESCRIPTOR' : _DNDSETTINGS,
  '__module__' : 'hangups.googlechat_pb2'
  # @@protoc_insertion_point(class_scope:DndSettings)
  })
_sym_db.RegisterMessage(DndSettings)

Emoji = _reflection.GeneratedProtocolMessageType('Emoji', (_message.Message,), {
  'DESCRIPTOR' : _EMOJI,
  '__module__' : 'hangups.googlechat_pb2'
  # @@protoc_insertion_point(class_scope:Emoji)
  })
_sym_db.RegisterMessage(Emoji)

CustomStatus = _reflection.GeneratedProtocolMessageType('CustomStatus', (_message.Message,), {
  'DESCRIPTOR' : _CUSTOMSTATUS,
  '__module__' : 'hangups.googlechat_pb2'
  # @@protoc_insertion_point(class_scope:CustomStatus)
  })
_sym_db.RegisterMessage(CustomStatus)

UserStatus = _reflection.GeneratedProtocolMessageType('UserStatus', (_message.Message,), {
  'DESCRIPTOR' : _USERSTATUS,
  '__module__' : 'hangups.googlechat_pb2'
  # @@protoc_insertion_point(class_scope:UserStatus)
  })
_sym_db.RegisterMessage(UserStatus)

UserStatusUpdatedEvent = _reflection.GeneratedProtocolMessageType('UserStatusUpdatedEvent', (_message.Message,), {
  'DESCRIPTOR' : _USERSTATUSUPDATEDEVENT,
  '__module__' : 'hangups.googlechat_pb2'
  # @@protoc_insertion_point(class_scope:UserStatusUpdatedEvent)
  })
_sym_db.RegisterMessage(UserStatusUpdatedEvent)

GetUserStatusRequest = _reflection.GeneratedProtocolMessageType('GetUserStatusRequest', (_message.Message,), {
  'DESCRIPTOR' : _GETUSERSTATUSREQUEST,
  '__module__' : 'hangups.googlechat_pb2'
  # @@protoc_insertion_point(class_scope:GetUserStatusRequest)
  })
_sym_db.RegisterMessage(GetUserStatusRequest)

GetUserStatusResponse = _reflection.GeneratedProtocolMessageType('GetUserStatusResponse', (_message.Message,), {
  'DESCRIPTOR' : _GETUSERSTATUSRESPONSE,
  '__module__' : 'hangups.googlechat_pb2'
  # @@protoc_insertion_point(class_scope:GetUserStatusResponse)
  })
_sym_db.RegisterMessage(GetUserStatusResponse)

GetSelfUserStatusRequest = _reflection.GeneratedProtocolMessageType('GetSelfUserStatusRequest', (_message.Message,), {
  'DESCRIPTOR' : _GETSELFUSERSTATUSREQUEST,
  '__module__' : 'hangups.googlechat_pb2'
  # @@protoc_insertion_point(class_scope:GetSelfUserStatusRequest)
  })
_sym_db.RegisterMessage(GetSelfUserStatusRequest)

GetSelfUserStatusResponse = _reflection.GeneratedProtocolMessageType('GetSelfUserStatusResponse', (_message.Message,), {
  'DESCRIPTOR' : _GETSELFUSERSTATUSRESPONSE,
  '__module__' : 'hangups.googlechat_pb2'
  # @@protoc_insertion_point(class_scope:GetSelfUserStatusResponse)
  })
_sym_db.RegisterMessage(GetSelfUserStatusResponse)

ClientFeatureCapabilities = _reflection.GeneratedProtocolMessageType('ClientFeatureCapabilities', (_message.Message,), {
  'DESCRIPTOR' : _CLIENTFEATURECAPABILITIES,
  '__module__' : 'hangups.googlechat_pb2'
  # @@protoc_insertion_point(class_scope:ClientFeatureCapabilities)
  })
_sym_db.RegisterMessage(ClientFeatureCapabilities)

RequestHeader = _reflection.GeneratedProtocolMessageType('RequestHeader', (_message.Message,), {
  'DESCRIPTOR' : _REQUESTHEADER,
  '__module__' : 'hangups.googlechat_pb2'
  # @@protoc_insertion_point(class_scope:RequestHeader)
  })
_sym_db.RegisterMessage(RequestHeader)

Member = _reflection.GeneratedProtocolMessageType('Member', (_message.Message,), {
  'DESCRIPTOR' : _MEMBER,
  '__module__' : 'hangups.googlechat_pb2'
  # @@protoc_insertion_point(class_scope:Member)
  })
_sym_db.RegisterMessage(Member)

MemberId = _reflection.GeneratedProtocolMessageType('MemberId', (_message.Message,), {
  'DESCRIPTOR' : _MEMBERID,
  '__module__' : 'hangups.googlechat_pb2'
  # @@protoc_insertion_point(class_scope:MemberId)
  })
_sym_db.RegisterMessage(MemberId)

MembershipId = _reflection.GeneratedProtocolMessageType('MembershipId', (_message.Message,), {
  'DESCRIPTOR' : _MEMBERSHIPID,
  '__module__' : 'hangups.googlechat_pb2'
  # @@protoc_insertion_point(class_scope:MembershipId)
  })
_sym_db.RegisterMessage(MembershipId)

Membership = _reflection.GeneratedProtocolMessageType('Membership', (_message.Message,), {
  'DESCRIPTOR' : _MEMBERSHIP,
  '__module__' : 'hangups.googlechat_pb2'
  # @@protoc_insertion_point(class_scope:Membership)
  })
_sym_db.RegisterMessage(Membership)

MemberProfile = _reflection.GeneratedProtocolMessageType('MemberProfile', (_message.Message,), {
  'DESCRIPTOR' : _MEMBERPROFILE,
  '__module__' : 'hangups.googlechat_pb2'
  # @@protoc_insertion_point(class_scope:MemberProfile)
  })
_sym_db.RegisterMessage(MemberProfile)

GetMembersRequest = _reflection.GeneratedProtocolMessageType('GetMembersRequest', (_message.Message,), {
  'DESCRIPTOR' : _GETMEMBERSREQUEST,
  '__module__' : 'hangups.googlechat_pb2'
  # @@protoc_insertion_point(class_scope:GetMembersRequest)
  })
_sym_db.RegisterMessage(GetMembersRequest)

GetMembersResponse = _reflection.GeneratedProtocolMessageType('GetMembersResponse', (_message.Message,), {
  'DESCRIPTOR' : _GETMEMBERSRESPONSE,
  '__module__' : 'hangups.googlechat_pb2'
  # @@protoc_insertion_point(class_scope:GetMembersResponse)
  })
_sym_db.RegisterMessage(GetMembersResponse)

UserPresence = _reflection.GeneratedProtocolMessageType('UserPresence', (_message.Message,), {
  'DESCRIPTOR' : _USERPRESENCE,
  '__module__' : 'hangups.googlechat_pb2'
  # @@protoc_insertion_point(class_scope:UserPresence)
  })
_sym_db.RegisterMessage(UserPresence)

GetUserPresenceRequest = _reflection.GeneratedProtocolMessageType('GetUserPresenceRequest', (_message.Message,), {
  'DESCRIPTOR' : _GETUSERPRESENCEREQUEST,
  '__module__' : 'hangups.googlechat_pb2'
  # @@protoc_insertion_point(class_scope:GetUserPresenceRequest)
  })
_sym_db.RegisterMessage(GetUserPresenceRequest)

GetUserPresenceResponse = _reflection.GeneratedProtocolMessageType('GetUserPresenceResponse', (_message.Message,), {
  'DESCRIPTOR' : _GETUSERPRESENCERESPONSE,
  '__module__' : 'hangups.googlechat_pb2'
  # @@protoc_insertion_point(class_scope:GetUserPresenceResponse)
  })
_sym_db.RegisterMessage(GetUserPresenceResponse)

JAddOnsIdentifier = _reflection.GeneratedProtocolMessageType('JAddOnsIdentifier', (_message.Message,), {
  'DESCRIPTOR' : _JADDONSIDENTIFIER,
  '__module__' : 'hangups.googlechat_pb2'
  # @@protoc_insertion_point(class_scope:JAddOnsIdentifier)
  })
_sym_db.RegisterMessage(JAddOnsIdentifier)

JAddOnsFormattedText = _reflection.GeneratedProtocolMessageType('JAddOnsFormattedText', (_message.Message,), {

  'FormattedTextElement' : _reflection.GeneratedProtocolMessageType('FormattedTextElement', (_message.Message,), {

    'DateTime' : _reflection.GeneratedProtocolMessageType('DateTime', (_message.Message,), {
      'DESCRIPTOR' : _JADDONSFORMATTEDTEXT_FORMATTEDTEXTELEMENT_DATETIME,
      '__module__' : 'hangups.googlechat_pb2'
      # @@protoc_insertion_point(class_scope:JAddOnsFormattedText.FormattedTextElement.DateTime)
      })
    ,

    'StyledText' : _reflection.GeneratedProtocolMessageType('StyledText', (_message.Message,), {
      'DESCRIPTOR' : _JADDONSFORMATTEDTEXT_FORMATTEDTEXTELEMENT_STYLEDTEXT,
      '__module__' : 'hangups.googlechat_pb2'
      # @@protoc_insertion_point(class_scope:JAddOnsFormattedText.FormattedTextElement.StyledText)
      })
    ,

    'HyperLink' : _reflection.GeneratedProtocolMessageType('HyperLink', (_message.Message,), {
      'DESCRIPTOR' : _JADDONSFORMATTEDTEXT_FORMATTEDTEXTELEMENT_HYPERLINK,
      '__module__' : 'hangups.googlechat_pb2'
      # @@protoc_insertion_point(class_scope:JAddOnsFormattedText.FormattedTextElement.HyperLink)
      })
    ,
    'DESCRIPTOR' : _JADDONSFORMATTEDTEXT_FORMATTEDTEXTELEMENT,
    '__module__' : 'hangups.googlechat_pb2'
    # @@protoc_insertion_point(class_scope:JAddOnsFormattedText.FormattedTextElement)
    })
  ,
  'DESCRIPTOR' : _JADDONSFORMATTEDTEXT,
  '__module__' : 'hangups.googlechat_pb2'
  # @@protoc_insertion_point(class_scope:JAddOnsFormattedText)
  })
_sym_db.RegisterMessage(JAddOnsFormattedText)
_sym_db.RegisterMessage(JAddOnsFormattedText.FormattedTextElement)
_sym_db.RegisterMessage(JAddOnsFormattedText.FormattedTextElement.DateTime)
_sym_db.RegisterMessage(JAddOnsFormattedText.FormattedTextElement.StyledText)
_sym_db.RegisterMessage(JAddOnsFormattedText.FormattedTextElement.HyperLink)

JAddOnsImageCropStyle = _reflection.GeneratedProtocolMessageType('JAddOnsImageCropStyle', (_message.Message,), {
  'DESCRIPTOR' : _JADDONSIMAGECROPSTYLE,
  '__module__' : 'hangups.googlechat_pb2'
  # @@protoc_insertion_point(class_scope:JAddOnsImageCropStyle)
  })
_sym_db.RegisterMessage(JAddOnsImageCropStyle)

JAddOnsThemeColors = _reflection.GeneratedProtocolMessageType('JAddOnsThemeColors', (_message.Message,), {
  'DESCRIPTOR' : _JADDONSTHEMECOLORS,
  '__module__' : 'hangups.googlechat_pb2'
  # @@protoc_insertion_point(class_scope:JAddOnsThemeColors)
  })
_sym_db.RegisterMessage(JAddOnsThemeColors)

JAddOnsOpenLink = _reflection.GeneratedProtocolMessageType('JAddOnsOpenLink', (_message.Message,), {
  'DESCRIPTOR' : _JADDONSOPENLINK,
  '__module__' : 'hangups.googlechat_pb2'
  # @@protoc_insertion_point(class_scope:JAddOnsOpenLink)
  })
_sym_db.RegisterMessage(JAddOnsOpenLink)

JAddOnsCardItem = _reflection.GeneratedProtocolMessageType('JAddOnsCardItem', (_message.Message,), {

  'CardItemHeader' : _reflection.GeneratedProtocolMessageType('CardItemHeader', (_message.Message,), {
    'DESCRIPTOR' : _JADDONSCARDITEM_CARDITEMHEADER,
    '__module__' : 'hangups.googlechat_pb2'
    # @@protoc_insertion_point(class_scope:JAddOnsCardItem.CardItemHeader)
    })
  ,

  'CardItemSection' : _reflection.GeneratedProtocolMessageType('CardItemSection', (_message.Message,), {
    'DESCRIPTOR' : _JADDONSCARDITEM_CARDITEMSECTION,
    '__module__' : 'hangups.googlechat_pb2'
    # @@protoc_insertion_point(class_scope:JAddOnsCardItem.CardItemSection)
    })
  ,

  'CardItemAction' : _reflection.GeneratedProtocolMessageType('CardItemAction', (_message.Message,), {
    'DESCRIPTOR' : _JADDONSCARDITEM_CARDITEMACTION,
    '__module__' : 'hangups.googlechat_pb2'
    # @@protoc_insertion_point(class_scope:JAddOnsCardItem.CardItemAction)
    })
  ,

  'CardItemFixedFooter' : _reflection.GeneratedProtocolMessageType('CardItemFixedFooter', (_message.Message,), {
    'DESCRIPTOR' : _JADDONSCARDITEM_CARDITEMFIXEDFOOTER,
    '__module__' : 'hangups.googlechat_pb2'
    # @@protoc_insertion_point(class_scope:JAddOnsCardItem.CardItemFixedFooter)
    })
  ,

  'CardItemRefreshAction' : _reflection.GeneratedProtocolMessageType('CardItemRefreshAction', (_message.Message,), {
    'DESCRIPTOR' : _JADDONSCARDITEM_CARDITEMREFRESHACTION,
    '__module__' : 'hangups.googlechat_pb2'
    # @@protoc_insertion_point(class_scope:JAddOnsCardItem.CardItemRefreshAction)
    })
  ,
  'DESCRIPTOR' : _JADDONSCARDITEM,
  '__module__' : 'hangups.googlechat_pb2'
  # @@protoc_insertion_point(class_scope:JAddOnsCardItem)
  })
_sym_db.RegisterMessage(JAddOnsCardItem)
_sym_db.RegisterMessage(JAddOnsCardItem.CardItemHeader)
_sym_db.RegisterMessage(JAddOnsCardItem.CardItemSection)
_sym_db.RegisterMessage(JAddOnsCardItem.CardItemAction)
_sym_db.RegisterMessage(JAddOnsCardItem.CardItemFixedFooter)
_sym_db.RegisterMessage(JAddOnsCardItem.CardItemRefreshAction)

JAddOnsPushCard = _reflection.GeneratedProtocolMessageType('JAddOnsPushCard', (_message.Message,), {
  'DESCRIPTOR' : _JADDONSPUSHCARD,
  '__module__' : 'hangups.googlechat_pb2'
  # @@protoc_insertion_point(class_scope:JAddOnsPushCard)
  })
_sym_db.RegisterMessage(JAddOnsPushCard)

JAddOnsOnClick = _reflection.GeneratedProtocolMessageType('JAddOnsOnClick', (_message.Message,), {
  'DESCRIPTOR' : _JADDONSONCLICK,
  '__module__' : 'hangups.googlechat_pb2'
  # @@protoc_insertion_point(class_scope:JAddOnsOnClick)
  })
_sym_db.RegisterMessage(JAddOnsOnClick)

JAddOnsTextWidget = _reflection.GeneratedProtocolMessageType('JAddOnsTextWidget', (_message.Message,), {
  'DESCRIPTOR' : _JADDONSTEXTWIDGET,
  '__module__' : 'hangups.googlechat_pb2'
  # @@protoc_insertion_point(class_scope:JAddOnsTextWidget)
  })
_sym_db.RegisterMessage(JAddOnsTextWidget)

JAddOnsImageComponent = _reflection.GeneratedProtocolMessageType('JAddOnsImageComponent', (_message.Message,), {
  'DESCRIPTOR' : _JADDONSIMAGECOMPONENT,
  '__module__' : 'hangups.googlechat_pb2'
  # @@protoc_insertion_point(class_scope:JAddOnsImageComponent)
  })
_sym_db.RegisterMessage(JAddOnsImageComponent)

JAddOnsGrid = _reflection.GeneratedProtocolMessageType('JAddOnsGrid', (_message.Message,), {

  'GridItem' : _reflection.GeneratedProtocolMessageType('GridItem', (_message.Message,), {
    'DESCRIPTOR' : _JADDONSGRID_GRIDITEM,
    '__module__' : 'hangups.googlechat_pb2'
    # @@protoc_insertion_point(class_scope:JAddOnsGrid.GridItem)
    })
  ,
  'DESCRIPTOR' : _JADDONSGRID,
  '__module__' : 'hangups.googlechat_pb2'
  # @@protoc_insertion_point(class_scope:JAddOnsGrid)
  })
_sym_db.RegisterMessage(JAddOnsGrid)
_sym_db.RegisterMessage(JAddOnsGrid.GridItem)

JAddOnsBorderStyle = _reflection.GeneratedProtocolMessageType('JAddOnsBorderStyle', (_message.Message,), {
  'DESCRIPTOR' : _JADDONSBORDERSTYLE,
  '__module__' : 'hangups.googlechat_pb2'
  # @@protoc_insertion_point(class_scope:JAddOnsBorderStyle)
  })
_sym_db.RegisterMessage(JAddOnsBorderStyle)

JAddOnsLabelContentPair = _reflection.GeneratedProtocolMessageType('JAddOnsLabelContentPair', (_message.Message,), {
  'DESCRIPTOR' : _JADDONSLABELCONTENTPAIR,
  '__module__' : 'hangups.googlechat_pb2'
  # @@protoc_insertion_point(class_scope:JAddOnsLabelContentPair)
  })
_sym_db.RegisterMessage(JAddOnsLabelContentPair)

JAddOnsLabelContentPairWidget = _reflection.GeneratedProtocolMessageType('JAddOnsLabelContentPairWidget', (_message.Message,), {
  'DESCRIPTOR' : _JADDONSLABELCONTENTPAIRWIDGET,
  '__module__' : 'hangups.googlechat_pb2'
  # @@protoc_insertion_point(class_scope:JAddOnsLabelContentPairWidget)
  })
_sym_db.RegisterMessage(JAddOnsLabelContentPairWidget)

JAddOnsWidget = _reflection.GeneratedProtocolMessageType('JAddOnsWidget', (_message.Message,), {

  'TextParagraph' : _reflection.GeneratedProtocolMessageType('TextParagraph', (_message.Message,), {
    'DESCRIPTOR' : _JADDONSWIDGET_TEXTPARAGRAPH,
    '__module__' : 'hangups.googlechat_pb2'
    # @@protoc_insertion_point(class_scope:JAddOnsWidget.TextParagraph)
    })
  ,

  'TextKeyValue' : _reflection.GeneratedProtocolMessageType('TextKeyValue', (_message.Message,), {
    'DESCRIPTOR' : _JADDONSWIDGET_TEXTKEYVALUE,
    '__module__' : 'hangups.googlechat_pb2'
    # @@protoc_insertion_point(class_scope:JAddOnsWidget.TextKeyValue)
    })
  ,

  'ImageKeyValue' : _reflection.GeneratedProtocolMessageType('ImageKeyValue', (_message.Message,), {
    'DESCRIPTOR' : _JADDONSWIDGET_IMAGEKEYVALUE,
    '__module__' : 'hangups.googlechat_pb2'
    # @@protoc_insertion_point(class_scope:JAddOnsWidget.ImageKeyValue)
    })
  ,

  'Image' : _reflection.GeneratedProtocolMessageType('Image', (_message.Message,), {
    'DESCRIPTOR' : _JADDONSWIDGET_IMAGE,
    '__module__' : 'hangups.googlechat_pb2'
    # @@protoc_insertion_point(class_scope:JAddOnsWidget.Image)
    })
  ,

  'Icon' : _reflection.GeneratedProtocolMessageType('Icon', (_message.Message,), {
    'DESCRIPTOR' : _JADDONSWIDGET_ICON,
    '__module__' : 'hangups.googlechat_pb2'
    # @@protoc_insertion_point(class_scope:JAddOnsWidget.Icon)
    })
  ,

  'KeyValue' : _reflection.GeneratedProtocolMessageType('KeyValue', (_message.Message,), {

    'SwitchWidget' : _reflection.GeneratedProtocolMessageType('SwitchWidget', (_message.Message,), {
      'DESCRIPTOR' : _JADDONSWIDGET_KEYVALUE_SWITCHWIDGET,
      '__module__' : 'hangups.googlechat_pb2'
      # @@protoc_insertion_point(class_scope:JAddOnsWidget.KeyValue.SwitchWidget)
      })
    ,
    'DESCRIPTOR' : _JADDONSWIDGET_KEYVALUE,
    '__module__' : 'hangups.googlechat_pb2'
    # @@protoc_insertion_point(class_scope:JAddOnsWidget.KeyValue)
    })
  ,

  'Divider' : _reflection.GeneratedProtocolMessageType('Divider', (_message.Message,), {
    'DESCRIPTOR' : _JADDONSWIDGET_DIVIDER,
    '__module__' : 'hangups.googlechat_pb2'
    # @@protoc_insertion_point(class_scope:JAddOnsWidget.Divider)
    })
  ,

  'Menu' : _reflection.GeneratedProtocolMessageType('Menu', (_message.Message,), {

    'MenuItem' : _reflection.GeneratedProtocolMessageType('MenuItem', (_message.Message,), {
      'DESCRIPTOR' : _JADDONSWIDGET_MENU_MENUITEM,
      '__module__' : 'hangups.googlechat_pb2'
      # @@protoc_insertion_point(class_scope:JAddOnsWidget.Menu.MenuItem)
      })
    ,
    'DESCRIPTOR' : _JADDONSWIDGET_MENU,
    '__module__' : 'hangups.googlechat_pb2'
    # @@protoc_insertion_point(class_scope:JAddOnsWidget.Menu)
    })
  ,

  'AutoComplete' : _reflection.GeneratedProtocolMessageType('AutoComplete', (_message.Message,), {

    'AutoCompleteItem' : _reflection.GeneratedProtocolMessageType('AutoCompleteItem', (_message.Message,), {
      'DESCRIPTOR' : _JADDONSWIDGET_AUTOCOMPLETE_AUTOCOMPLETEITEM,
      '__module__' : 'hangups.googlechat_pb2'
      # @@protoc_insertion_point(class_scope:JAddOnsWidget.AutoComplete.AutoCompleteItem)
      })
    ,
    'DESCRIPTOR' : _JADDONSWIDGET_AUTOCOMPLETE,
    '__module__' : 'hangups.googlechat_pb2'
    # @@protoc_insertion_point(class_scope:JAddOnsWidget.AutoComplete)
    })
  ,

  'TextField' : _reflection.GeneratedProtocolMessageType('TextField', (_message.Message,), {
    'DESCRIPTOR' : _JADDONSWIDGET_TEXTFIELD,
    '__module__' : 'hangups.googlechat_pb2'
    # @@protoc_insertion_point(class_scope:JAddOnsWidget.TextField)
    })
  ,

  'SelectionControl' : _reflection.GeneratedProtocolMessageType('SelectionControl', (_message.Message,), {

    'SelectionItem' : _reflection.GeneratedProtocolMessageType('SelectionItem', (_message.Message,), {
      'DESCRIPTOR' : _JADDONSWIDGET_SELECTIONCONTROL_SELECTIONITEM,
      '__module__' : 'hangups.googlechat_pb2'
      # @@protoc_insertion_point(class_scope:JAddOnsWidget.SelectionControl.SelectionItem)
      })
    ,
    'DESCRIPTOR' : _JADDONSWIDGET_SELECTIONCONTROL,
    '__module__' : 'hangups.googlechat_pb2'
    # @@protoc_insertion_point(class_scope:JAddOnsWidget.SelectionControl)
    })
  ,

  'DateTimePicker' : _reflection.GeneratedProtocolMessageType('DateTimePicker', (_message.Message,), {
    'DESCRIPTOR' : _JADDONSWIDGET_DATETIMEPICKER,
    '__module__' : 'hangups.googlechat_pb2'
    # @@protoc_insertion_point(class_scope:JAddOnsWidget.DateTimePicker)
    })
  ,

  'TextButton' : _reflection.GeneratedProtocolMessageType('TextButton', (_message.Message,), {
    'DESCRIPTOR' : _JADDONSWIDGET_TEXTBUTTON,
    '__module__' : 'hangups.googlechat_pb2'
    # @@protoc_insertion_point(class_scope:JAddOnsWidget.TextButton)
    })
  ,

  'ImageButton' : _reflection.GeneratedProtocolMessageType('ImageButton', (_message.Message,), {
    'DESCRIPTOR' : _JADDONSWIDGET_IMAGEBUTTON,
    '__module__' : 'hangups.googlechat_pb2'
    # @@protoc_insertion_point(class_scope:JAddOnsWidget.ImageButton)
    })
  ,

  'Button' : _reflection.GeneratedProtocolMessageType('Button', (_message.Message,), {
    'DESCRIPTOR' : _JADDONSWIDGET_BUTTON,
    '__module__' : 'hangups.googlechat_pb2'
    # @@protoc_insertion_point(class_scope:JAddOnsWidget.Button)
    })
  ,
  'DESCRIPTOR' : _JADDONSWIDGET,
  '__module__' : 'hangups.googlechat_pb2'
  # @@protoc_insertion_point(class_scope:JAddOnsWidget)
  })
_sym_db.RegisterMessage(JAddOnsWidget)
_sym_db.RegisterMessage(JAddOnsWidget.TextParagraph)
_sym_db.RegisterMessage(JAddOnsWidget.TextKeyValue)
_sym_db.RegisterMessage(JAddOnsWidget.ImageKeyValue)
_sym_db.RegisterMessage(JAddOnsWidget.Image)
_sym_db.RegisterMessage(JAddOnsWidget.Icon)
_sym_db.RegisterMessage(JAddOnsWidget.KeyValue)
_sym_db.RegisterMessage(JAddOnsWidget.KeyValue.SwitchWidget)
_sym_db.RegisterMessage(JAddOnsWidget.Divider)
_sym_db.RegisterMessage(JAddOnsWidget.Menu)
_sym_db.RegisterMessage(JAddOnsWidget.Menu.MenuItem)
_sym_db.RegisterMessage(JAddOnsWidget.AutoComplete)
_sym_db.RegisterMessage(JAddOnsWidget.AutoComplete.AutoCompleteItem)
_sym_db.RegisterMessage(JAddOnsWidget.TextField)
_sym_db.RegisterMessage(JAddOnsWidget.SelectionControl)
_sym_db.RegisterMessage(JAddOnsWidget.SelectionControl.SelectionItem)
_sym_db.RegisterMessage(JAddOnsWidget.DateTimePicker)
_sym_db.RegisterMessage(JAddOnsWidget.TextButton)
_sym_db.RegisterMessage(JAddOnsWidget.ImageButton)
_sym_db.RegisterMessage(JAddOnsWidget.Button)

JAddOnsFormAction = _reflection.GeneratedProtocolMessageType('JAddOnsFormAction', (_message.Message,), {

  'ActionParameter' : _reflection.GeneratedProtocolMessageType('ActionParameter', (_message.Message,), {
    'DESCRIPTOR' : _JADDONSFORMACTION_ACTIONPARAMETER,
    '__module__' : 'hangups.googlechat_pb2'
    # @@protoc_insertion_point(class_scope:JAddOnsFormAction.ActionParameter)
    })
  ,
  'DESCRIPTOR' : _JADDONSFORMACTION,
  '__module__' : 'hangups.googlechat_pb2'
  # @@protoc_insertion_point(class_scope:JAddOnsFormAction)
  })
_sym_db.RegisterMessage(JAddOnsFormAction)
_sym_db.RegisterMessage(JAddOnsFormAction.ActionParameter)

JAddOnsContextualAddOn = _reflection.GeneratedProtocolMessageType('JAddOnsContextualAddOn', (_message.Message,), {

  'Toolbar' : _reflection.GeneratedProtocolMessageType('Toolbar', (_message.Message,), {
    'DESCRIPTOR' : _JADDONSCONTEXTUALADDON_TOOLBAR,
    '__module__' : 'hangups.googlechat_pb2'
    # @@protoc_insertion_point(class_scope:JAddOnsContextualAddOn.Toolbar)
    })
  ,

  'Card' : _reflection.GeneratedProtocolMessageType('Card', (_message.Message,), {

    'CardHeader' : _reflection.GeneratedProtocolMessageType('CardHeader', (_message.Message,), {
      'DESCRIPTOR' : _JADDONSCONTEXTUALADDON_CARD_CARDHEADER,
      '__module__' : 'hangups.googlechat_pb2'
      # @@protoc_insertion_point(class_scope:JAddOnsContextualAddOn.Card.CardHeader)
      })
    ,

    'Section' : _reflection.GeneratedProtocolMessageType('Section', (_message.Message,), {
      'DESCRIPTOR' : _JADDONSCONTEXTUALADDON_CARD_SECTION,
      '__module__' : 'hangups.googlechat_pb2'
      # @@protoc_insertion_point(class_scope:JAddOnsContextualAddOn.Card.Section)
      })
    ,

    'CardAction' : _reflection.GeneratedProtocolMessageType('CardAction', (_message.Message,), {
      'DESCRIPTOR' : _JADDONSCONTEXTUALADDON_CARD_CARDACTION,
      '__module__' : 'hangups.googlechat_pb2'
      # @@protoc_insertion_point(class_scope:JAddOnsContextualAddOn.Card.CardAction)
      })
    ,

    'FixedFooter' : _reflection.GeneratedProtocolMessageType('FixedFooter', (_message.Message,), {
      'DESCRIPTOR' : _JADDONSCONTEXTUALADDON_CARD_FIXEDFOOTER,
      '__module__' : 'hangups.googlechat_pb2'
      # @@protoc_insertion_point(class_scope:JAddOnsContextualAddOn.Card.FixedFooter)
      })
    ,

    'RefreshAction' : _reflection.GeneratedProtocolMessageType('RefreshAction', (_message.Message,), {
      'DESCRIPTOR' : _JADDONSCONTEXTUALADDON_CARD_REFRESHACTION,
      '__module__' : 'hangups.googlechat_pb2'
      # @@protoc_insertion_point(class_scope:JAddOnsContextualAddOn.Card.RefreshAction)
      })
    ,
    'DESCRIPTOR' : _JADDONSCONTEXTUALADDON_CARD,
    '__module__' : 'hangups.googlechat_pb2'
    # @@protoc_insertion_point(class_scope:JAddOnsContextualAddOn.Card)
    })
  ,
  'DESCRIPTOR' : _JADDONSCONTEXTUALADDON,
  '__module__' : 'hangups.googlechat_pb2'
  # @@protoc_insertion_point(class_scope:JAddOnsContextualAddOn)
  })
_sym_db.RegisterMessage(JAddOnsContextualAddOn)
_sym_db.RegisterMessage(JAddOnsContextualAddOn.Toolbar)
_sym_db.RegisterMessage(JAddOnsContextualAddOn.Card)
_sym_db.RegisterMessage(JAddOnsContextualAddOn.Card.CardHeader)
_sym_db.RegisterMessage(JAddOnsContextualAddOn.Card.Section)
_sym_db.RegisterMessage(JAddOnsContextualAddOn.Card.CardAction)
_sym_db.RegisterMessage(JAddOnsContextualAddOn.Card.FixedFooter)
_sym_db.RegisterMessage(JAddOnsContextualAddOn.Card.RefreshAction)

Html = _reflection.GeneratedProtocolMessageType('Html', (_message.Message,), {
  'DESCRIPTOR' : _HTML,
  '__module__' : 'hangups.googlechat_pb2'
  # @@protoc_insertion_point(class_scope:Html)
  })
_sym_db.RegisterMessage(Html)

HtmlAttachment = _reflection.GeneratedProtocolMessageType('HtmlAttachment', (_message.Message,), {
  'DESCRIPTOR' : _HTMLATTACHMENT,
  '__module__' : 'hangups.googlechat_pb2'
  # @@protoc_insertion_point(class_scope:HtmlAttachment)
  })
_sym_db.RegisterMessage(HtmlAttachment)

Attachment = _reflection.GeneratedProtocolMessageType('Attachment', (_message.Message,), {
  'DESCRIPTOR' : _ATTACHMENT,
  '__module__' : 'hangups.googlechat_pb2'
  # @@protoc_insertion_point(class_scope:Attachment)
  })
_sym_db.RegisterMessage(Attachment)

Reaction = _reflection.GeneratedProtocolMessageType('Reaction', (_message.Message,), {
  'DESCRIPTOR' : _REACTION,
  '__module__' : 'hangups.googlechat_pb2'
  # @@protoc_insertion_point(class_scope:Reaction)
  })
_sym_db.RegisterMessage(Reaction)

MessageParentId = _reflection.GeneratedProtocolMessageType('MessageParentId', (_message.Message,), {
  'DESCRIPTOR' : _MESSAGEPARENTID,
  '__module__' : 'hangups.googlechat_pb2'
  # @@protoc_insertion_point(class_scope:MessageParentId)
  })
_sym_db.RegisterMessage(MessageParentId)

MessageId = _reflection.GeneratedProtocolMessageType('MessageId', (_message.Message,), {
  'DESCRIPTOR' : _MESSAGEID,
  '__module__' : 'hangups.googlechat_pb2'
  # @@protoc_insertion_point(class_scope:MessageId)
  })
_sym_db.RegisterMessage(MessageId)

Message = _reflection.GeneratedProtocolMessageType('Message', (_message.Message,), {
  'DESCRIPTOR' : _MESSAGE,
  '__module__' : 'hangups.googlechat_pb2'
  # @@protoc_insertion_point(class_scope:Message)
  })
_sym_db.RegisterMessage(Message)

DriveMetadata = _reflection.GeneratedProtocolMessageType('DriveMetadata', (_message.Message,), {
  'DESCRIPTOR' : _DRIVEMETADATA,
  '__module__' : 'hangups.googlechat_pb2'
  # @@protoc_insertion_point(class_scope:DriveMetadata)
  })
_sym_db.RegisterMessage(DriveMetadata)

Url = _reflection.GeneratedProtocolMessageType('Url', (_message.Message,), {
  'DESCRIPTOR' : _URL,
  '__module__' : 'hangups.googlechat_pb2'
  # @@protoc_insertion_point(class_scope:Url)
  })
_sym_db.RegisterMessage(Url)

TrustedResourceUrl = _reflection.GeneratedProtocolMessageType('TrustedResourceUrl', (_message.Message,), {
  'DESCRIPTOR' : _TRUSTEDRESOURCEURL,
  '__module__' : 'hangups.googlechat_pb2'
  # @@protoc_insertion_point(class_scope:TrustedResourceUrl)
  })
_sym_db.RegisterMessage(TrustedResourceUrl)

UrlMetadata = _reflection.GeneratedProtocolMessageType('UrlMetadata', (_message.Message,), {
  'DESCRIPTOR' : _URLMETADATA,
  '__module__' : 'hangups.googlechat_pb2'
  # @@protoc_insertion_point(class_scope:UrlMetadata)
  })
_sym_db.RegisterMessage(UrlMetadata)

UploadMetadata = _reflection.GeneratedProtocolMessageType('UploadMetadata', (_message.Message,), {
  'DESCRIPTOR' : _UPLOADMETADATA,
  '__module__' : 'hangups.googlechat_pb2'
  # @@protoc_insertion_point(class_scope:UploadMetadata)
  })
_sym_db.RegisterMessage(UploadMetadata)

FormatMetadata = _reflection.GeneratedProtocolMessageType('FormatMetadata', (_message.Message,), {
  'DESCRIPTOR' : _FORMATMETADATA,
  '__module__' : 'hangups.googlechat_pb2'
  # @@protoc_insertion_point(class_scope:FormatMetadata)
  })
_sym_db.RegisterMessage(FormatMetadata)

Annotation = _reflection.GeneratedProtocolMessageType('Annotation', (_message.Message,), {
  'DESCRIPTOR' : _ANNOTATION,
  '__module__' : 'hangups.googlechat_pb2'
  # @@protoc_insertion_point(class_scope:Annotation)
  })
_sym_db.RegisterMessage(Annotation)

TypingContext = _reflection.GeneratedProtocolMessageType('TypingContext', (_message.Message,), {
  'DESCRIPTOR' : _TYPINGCONTEXT,
  '__module__' : 'hangups.googlechat_pb2'
  # @@protoc_insertion_point(class_scope:TypingContext)
  })
_sym_db.RegisterMessage(TypingContext)

SetTypingStateRequest = _reflection.GeneratedProtocolMessageType('SetTypingStateRequest', (_message.Message,), {
  'DESCRIPTOR' : _SETTYPINGSTATEREQUEST,
  '__module__' : 'hangups.googlechat_pb2'
  # @@protoc_insertion_point(class_scope:SetTypingStateRequest)
  })
_sym_db.RegisterMessage(SetTypingStateRequest)

SetTypingStateResponse = _reflection.GeneratedProtocolMessageType('SetTypingStateResponse', (_message.Message,), {
  'DESCRIPTOR' : _SETTYPINGSTATERESPONSE,
  '__module__' : 'hangups.googlechat_pb2'
  # @@protoc_insertion_point(class_scope:SetTypingStateResponse)
  })
_sym_db.RegisterMessage(SetTypingStateResponse)

DmId = _reflection.GeneratedProtocolMessageType('DmId', (_message.Message,), {
  'DESCRIPTOR' : _DMID,
  '__module__' : 'hangups.googlechat_pb2'
  # @@protoc_insertion_point(class_scope:DmId)
  })
_sym_db.RegisterMessage(DmId)

SpaceId = _reflection.GeneratedProtocolMessageType('SpaceId', (_message.Message,), {
  'DESCRIPTOR' : _SPACEID,
  '__module__' : 'hangups.googlechat_pb2'
  # @@protoc_insertion_point(class_scope:SpaceId)
  })
_sym_db.RegisterMessage(SpaceId)

GroupId = _reflection.GeneratedProtocolMessageType('GroupId', (_message.Message,), {
  'DESCRIPTOR' : _GROUPID,
  '__module__' : 'hangups.googlechat_pb2'
  # @@protoc_insertion_point(class_scope:GroupId)
  })
_sym_db.RegisterMessage(GroupId)

Group = _reflection.GeneratedProtocolMessageType('Group', (_message.Message,), {

  'FlatGroup' : _reflection.GeneratedProtocolMessageType('FlatGroup', (_message.Message,), {
    'DESCRIPTOR' : _GROUP_FLATGROUP,
    '__module__' : 'hangups.googlechat_pb2'
    # @@protoc_insertion_point(class_scope:Group.FlatGroup)
    })
  ,

  'ThreadedGroup' : _reflection.GeneratedProtocolMessageType('ThreadedGroup', (_message.Message,), {
    'DESCRIPTOR' : _GROUP_THREADEDGROUP,
    '__module__' : 'hangups.googlechat_pb2'
    # @@protoc_insertion_point(class_scope:Group.ThreadedGroup)
    })
  ,
  'DESCRIPTOR' : _GROUP,
  '__module__' : 'hangups.googlechat_pb2'
  # @@protoc_insertion_point(class_scope:Group)
  })
_sym_db.RegisterMessage(Group)
_sym_db.RegisterMessage(Group.FlatGroup)
_sym_db.RegisterMessage(Group.ThreadedGroup)

GroupReadStateId = _reflection.GeneratedProtocolMessageType('GroupReadStateId', (_message.Message,), {
  'DESCRIPTOR' : _GROUPREADSTATEID,
  '__module__' : 'hangups.googlechat_pb2'
  # @@protoc_insertion_point(class_scope:GroupReadStateId)
  })
_sym_db.RegisterMessage(GroupReadStateId)

InviteState = _reflection.GeneratedProtocolMessageType('InviteState', (_message.Message,), {
  'DESCRIPTOR' : _INVITESTATE,
  '__module__' : 'hangups.googlechat_pb2'
  # @@protoc_insertion_point(class_scope:InviteState)
  })
_sym_db.RegisterMessage(InviteState)

GroupReadState = _reflection.GeneratedProtocolMessageType('GroupReadState', (_message.Message,), {
  'DESCRIPTOR' : _GROUPREADSTATE,
  '__module__' : 'hangups.googlechat_pb2'
  # @@protoc_insertion_point(class_scope:GroupReadState)
  })
_sym_db.RegisterMessage(GroupReadState)

GroupNotificationSettings = _reflection.GeneratedProtocolMessageType('GroupNotificationSettings', (_message.Message,), {
  'DESCRIPTOR' : _GROUPNOTIFICATIONSETTINGS,
  '__module__' : 'hangups.googlechat_pb2'
  # @@protoc_insertion_point(class_scope:GroupNotificationSettings)
  })
_sym_db.RegisterMessage(GroupNotificationSettings)

RetentionSettings = _reflection.GeneratedProtocolMessageType('RetentionSettings', (_message.Message,), {
  'DESCRIPTOR' : _RETENTIONSETTINGS,
  '__module__' : 'hangups.googlechat_pb2'
  # @@protoc_insertion_point(class_scope:RetentionSettings)
  })
_sym_db.RegisterMessage(RetentionSettings)

TopicId = _reflection.GeneratedProtocolMessageType('TopicId', (_message.Message,), {
  'DESCRIPTOR' : _TOPICID,
  '__module__' : 'hangups.googlechat_pb2'
  # @@protoc_insertion_point(class_scope:TopicId)
  })
_sym_db.RegisterMessage(TopicId)

Topic = _reflection.GeneratedProtocolMessageType('Topic', (_message.Message,), {
  'DESCRIPTOR' : _TOPIC,
  '__module__' : 'hangups.googlechat_pb2'
  # @@protoc_insertion_point(class_scope:Topic)
  })
_sym_db.RegisterMessage(Topic)

MessageInfo = _reflection.GeneratedProtocolMessageType('MessageInfo', (_message.Message,), {
  'DESCRIPTOR' : _MESSAGEINFO,
  '__module__' : 'hangups.googlechat_pb2'
  # @@protoc_insertion_point(class_scope:MessageInfo)
  })
_sym_db.RegisterMessage(MessageInfo)

CreateTopicRequest = _reflection.GeneratedProtocolMessageType('CreateTopicRequest', (_message.Message,), {
  'DESCRIPTOR' : _CREATETOPICREQUEST,
  '__module__' : 'hangups.googlechat_pb2'
  # @@protoc_insertion_point(class_scope:CreateTopicRequest)
  })
_sym_db.RegisterMessage(CreateTopicRequest)

CreateTopicResponse = _reflection.GeneratedProtocolMessageType('CreateTopicResponse', (_message.Message,), {
  'DESCRIPTOR' : _CREATETOPICRESPONSE,
  '__module__' : 'hangups.googlechat_pb2'
  # @@protoc_insertion_point(class_scope:CreateTopicResponse)
  })
_sym_db.RegisterMessage(CreateTopicResponse)

InviteeInfo = _reflection.GeneratedProtocolMessageType('InviteeInfo', (_message.Message,), {
  'DESCRIPTOR' : _INVITEEINFO,
  '__module__' : 'hangups.googlechat_pb2'
  # @@protoc_insertion_point(class_scope:InviteeInfo)
  })
_sym_db.RegisterMessage(InviteeInfo)

InviteeMemberInfo = _reflection.GeneratedProtocolMessageType('InviteeMemberInfo', (_message.Message,), {
  'DESCRIPTOR' : _INVITEEMEMBERINFO,
  '__module__' : 'hangups.googlechat_pb2'
  # @@protoc_insertion_point(class_scope:InviteeMemberInfo)
  })
_sym_db.RegisterMessage(InviteeMemberInfo)

SpaceCreationInfo = _reflection.GeneratedProtocolMessageType('SpaceCreationInfo', (_message.Message,), {

  'FlatGroup' : _reflection.GeneratedProtocolMessageType('FlatGroup', (_message.Message,), {
    'DESCRIPTOR' : _SPACECREATIONINFO_FLATGROUP,
    '__module__' : 'hangups.googlechat_pb2'
    # @@protoc_insertion_point(class_scope:SpaceCreationInfo.FlatGroup)
    })
  ,

  'ThreadedGroup' : _reflection.GeneratedProtocolMessageType('ThreadedGroup', (_message.Message,), {
    'DESCRIPTOR' : _SPACECREATIONINFO_THREADEDGROUP,
    '__module__' : 'hangups.googlechat_pb2'
    # @@protoc_insertion_point(class_scope:SpaceCreationInfo.ThreadedGroup)
    })
  ,

  'SpaceType' : _reflection.GeneratedProtocolMessageType('SpaceType', (_message.Message,), {
    'DESCRIPTOR' : _SPACECREATIONINFO_SPACETYPE,
    '__module__' : 'hangups.googlechat_pb2'
    # @@protoc_insertion_point(class_scope:SpaceCreationInfo.SpaceType)
    })
  ,
  'DESCRIPTOR' : _SPACECREATIONINFO,
  '__module__' : 'hangups.googlechat_pb2'
  # @@protoc_insertion_point(class_scope:SpaceCreationInfo)
  })
_sym_db.RegisterMessage(SpaceCreationInfo)
_sym_db.RegisterMessage(SpaceCreationInfo.FlatGroup)
_sym_db.RegisterMessage(SpaceCreationInfo.ThreadedGroup)
_sym_db.RegisterMessage(SpaceCreationInfo.SpaceType)

CreateGroupRequest = _reflection.GeneratedProtocolMessageType('CreateGroupRequest', (_message.Message,), {
  'DESCRIPTOR' : _CREATEGROUPREQUEST,
  '__module__' : 'hangups.googlechat_pb2'
  # @@protoc_insertion_point(class_scope:CreateGroupRequest)
  })
_sym_db.RegisterMessage(CreateGroupRequest)

CreateMembershipResult = _reflection.GeneratedProtocolMessageType('CreateMembershipResult', (_message.Message,), {
  'DESCRIPTOR' : _CREATEMEMBERSHIPRESULT,
  '__module__' : 'hangups.googlechat_pb2'
  # @@protoc_insertion_point(class_scope:CreateMembershipResult)
  })
_sym_db.RegisterMessage(CreateMembershipResult)

CreateGroupResponse = _reflection.GeneratedProtocolMessageType('CreateGroupResponse', (_message.Message,), {
  'DESCRIPTOR' : _CREATEGROUPRESPONSE,
  '__module__' : 'hangups.googlechat_pb2'
  # @@protoc_insertion_point(class_scope:CreateGroupResponse)
  })
_sym_db.RegisterMessage(CreateGroupResponse)

CreateDmRequest = _reflection.GeneratedProtocolMessageType('CreateDmRequest', (_message.Message,), {
  'DESCRIPTOR' : _CREATEDMREQUEST,
  '__module__' : 'hangups.googlechat_pb2'
  # @@protoc_insertion_point(class_scope:CreateDmRequest)
  })
_sym_db.RegisterMessage(CreateDmRequest)

CreateDmResponse = _reflection.GeneratedProtocolMessageType('CreateDmResponse', (_message.Message,), {
  'DESCRIPTOR' : _CREATEDMRESPONSE,
  '__module__' : 'hangups.googlechat_pb2'
  # @@protoc_insertion_point(class_scope:CreateDmResponse)
  })
_sym_db.RegisterMessage(CreateDmResponse)

ListTopicsResponse = _reflection.GeneratedProtocolMessageType('ListTopicsResponse', (_message.Message,), {
  'DESCRIPTOR' : _LISTTOPICSRESPONSE,
  '__module__' : 'hangups.googlechat_pb2'
  # @@protoc_insertion_point(class_scope:ListTopicsResponse)
  })
_sym_db.RegisterMessage(ListTopicsResponse)

ListTopicsRequest = _reflection.GeneratedProtocolMessageType('ListTopicsRequest', (_message.Message,), {
  'DESCRIPTOR' : _LISTTOPICSREQUEST,
  '__module__' : 'hangups.googlechat_pb2'
  # @@protoc_insertion_point(class_scope:ListTopicsRequest)
  })
_sym_db.RegisterMessage(ListTopicsRequest)

ListMembersRequest = _reflection.GeneratedProtocolMessageType('ListMembersRequest', (_message.Message,), {
  'DESCRIPTOR' : _LISTMEMBERSREQUEST,
  '__module__' : 'hangups.googlechat_pb2'
  # @@protoc_insertion_point(class_scope:ListMembersRequest)
  })
_sym_db.RegisterMessage(ListMembersRequest)

ListMembersResponse = _reflection.GeneratedProtocolMessageType('ListMembersResponse', (_message.Message,), {
  'DESCRIPTOR' : _LISTMEMBERSRESPONSE,
  '__module__' : 'hangups.googlechat_pb2'
  # @@protoc_insertion_point(class_scope:ListMembersResponse)
  })
_sym_db.RegisterMessage(ListMembersResponse)

ReadReceipt = _reflection.GeneratedProtocolMessageType('ReadReceipt', (_message.Message,), {
  'DESCRIPTOR' : _READRECEIPT,
  '__module__' : 'hangups.googlechat_pb2'
  # @@protoc_insertion_point(class_scope:ReadReceipt)
  })
_sym_db.RegisterMessage(ReadReceipt)

ReadReceiptSet = _reflection.GeneratedProtocolMessageType('ReadReceiptSet', (_message.Message,), {
  'DESCRIPTOR' : _READRECEIPTSET,
  '__module__' : 'hangups.googlechat_pb2'
  # @@protoc_insertion_point(class_scope:ReadReceiptSet)
  })
_sym_db.RegisterMessage(ReadReceiptSet)

WebPushNotification = _reflection.GeneratedProtocolMessageType('WebPushNotification', (_message.Message,), {
  'DESCRIPTOR' : _WEBPUSHNOTIFICATION,
  '__module__' : 'hangups.googlechat_pb2'
  # @@protoc_insertion_point(class_scope:WebPushNotification)
  })
_sym_db.RegisterMessage(WebPushNotification)

AndroidLocalNotification = _reflection.GeneratedProtocolMessageType('AndroidLocalNotification', (_message.Message,), {
  'DESCRIPTOR' : _ANDROIDLOCALNOTIFICATION,
  '__module__' : 'hangups.googlechat_pb2'
  # @@protoc_insertion_point(class_scope:AndroidLocalNotification)
  })
_sym_db.RegisterMessage(AndroidLocalNotification)

IosLocalNotification = _reflection.GeneratedProtocolMessageType('IosLocalNotification', (_message.Message,), {
  'DESCRIPTOR' : _IOSLOCALNOTIFICATION,
  '__module__' : 'hangups.googlechat_pb2'
  # @@protoc_insertion_point(class_scope:IosLocalNotification)
  })
_sym_db.RegisterMessage(IosLocalNotification)

MobileLocalNotification = _reflection.GeneratedProtocolMessageType('MobileLocalNotification', (_message.Message,), {
  'DESCRIPTOR' : _MOBILELOCALNOTIFICATION,
  '__module__' : 'hangups.googlechat_pb2'
  # @@protoc_insertion_point(class_scope:MobileLocalNotification)
  })
_sym_db.RegisterMessage(MobileLocalNotification)

MessageEvent = _reflection.GeneratedProtocolMessageType('MessageEvent', (_message.Message,), {
  'DESCRIPTOR' : _MESSAGEEVENT,
  '__module__' : 'hangups.googlechat_pb2'
  # @@protoc_insertion_point(class_scope:MessageEvent)
  })
_sym_db.RegisterMessage(MessageEvent)

TypingStateChangedEvent = _reflection.GeneratedProtocolMessageType('TypingStateChangedEvent', (_message.Message,), {
  'DESCRIPTOR' : _TYPINGSTATECHANGEDEVENT,
  '__module__' : 'hangups.googlechat_pb2'
  # @@protoc_insertion_point(class_scope:TypingStateChangedEvent)
  })
_sym_db.RegisterMessage(TypingStateChangedEvent)

MembershipChangedEvent = _reflection.GeneratedProtocolMessageType('MembershipChangedEvent', (_message.Message,), {
  'DESCRIPTOR' : _MEMBERSHIPCHANGEDEVENT,
  '__module__' : 'hangups.googlechat_pb2'
  # @@protoc_insertion_point(class_scope:MembershipChangedEvent)
  })
_sym_db.RegisterMessage(MembershipChangedEvent)

ReadReceiptChangedEvent = _reflection.GeneratedProtocolMessageType('ReadReceiptChangedEvent', (_message.Message,), {
  'DESCRIPTOR' : _READRECEIPTCHANGEDEVENT,
  '__module__' : 'hangups.googlechat_pb2'
  # @@protoc_insertion_point(class_scope:ReadReceiptChangedEvent)
  })
_sym_db.RegisterMessage(ReadReceiptChangedEvent)

GroupViewedEvent = _reflection.GeneratedProtocolMessageType('GroupViewedEvent', (_message.Message,), {
  'DESCRIPTOR' : _GROUPVIEWEDEVENT,
  '__module__' : 'hangups.googlechat_pb2'
  # @@protoc_insertion_point(class_scope:GroupViewedEvent)
  })
_sym_db.RegisterMessage(GroupViewedEvent)

WebPushNotificationEvent = _reflection.GeneratedProtocolMessageType('WebPushNotificationEvent', (_message.Message,), {
  'DESCRIPTOR' : _WEBPUSHNOTIFICATIONEVENT,
  '__module__' : 'hangups.googlechat_pb2'
  # @@protoc_insertion_point(class_scope:WebPushNotificationEvent)
  })
_sym_db.RegisterMessage(WebPushNotificationEvent)

StreamEventsRequest = _reflection.GeneratedProtocolMessageType('StreamEventsRequest', (_message.Message,), {
  'DESCRIPTOR' : _STREAMEVENTSREQUEST,
  '__module__' : 'hangups.googlechat_pb2'
  # @@protoc_insertion_point(class_scope:StreamEventsRequest)
  })
_sym_db.RegisterMessage(StreamEventsRequest)

ClientInfo = _reflection.GeneratedProtocolMessageType('ClientInfo', (_message.Message,), {
  'DESCRIPTOR' : _CLIENTINFO,
  '__module__' : 'hangups.googlechat_pb2'
  # @@protoc_insertion_point(class_scope:ClientInfo)
  })
_sym_db.RegisterMessage(ClientInfo)

ClientNotificationsState = _reflection.GeneratedProtocolMessageType('ClientNotificationsState', (_message.Message,), {
  'DESCRIPTOR' : _CLIENTNOTIFICATIONSSTATE,
  '__module__' : 'hangups.googlechat_pb2'
  # @@protoc_insertion_point(class_scope:ClientNotificationsState)
  })
_sym_db.RegisterMessage(ClientNotificationsState)

PingEvent = _reflection.GeneratedProtocolMessageType('PingEvent', (_message.Message,), {
  'DESCRIPTOR' : _PINGEVENT,
  '__module__' : 'hangups.googlechat_pb2'
  # @@protoc_insertion_point(class_scope:PingEvent)
  })
_sym_db.RegisterMessage(PingEvent)

ClockSyncRequest = _reflection.GeneratedProtocolMessageType('ClockSyncRequest', (_message.Message,), {
  'DESCRIPTOR' : _CLOCKSYNCREQUEST,
  '__module__' : 'hangups.googlechat_pb2'
  # @@protoc_insertion_point(class_scope:ClockSyncRequest)
  })
_sym_db.RegisterMessage(ClockSyncRequest)

ClockSyncResponse = _reflection.GeneratedProtocolMessageType('ClockSyncResponse', (_message.Message,), {
  'DESCRIPTOR' : _CLOCKSYNCRESPONSE,
  '__module__' : 'hangups.googlechat_pb2'
  # @@protoc_insertion_point(class_scope:ClockSyncResponse)
  })
_sym_db.RegisterMessage(ClockSyncResponse)

GroupSubscriptionEvent = _reflection.GeneratedProtocolMessageType('GroupSubscriptionEvent', (_message.Message,), {
  'DESCRIPTOR' : _GROUPSUBSCRIPTIONEVENT,
  '__module__' : 'hangups.googlechat_pb2'
  # @@protoc_insertion_point(class_scope:GroupSubscriptionEvent)
  })
_sym_db.RegisterMessage(GroupSubscriptionEvent)

StreamEventsResponse = _reflection.GeneratedProtocolMessageType('StreamEventsResponse', (_message.Message,), {
  'DESCRIPTOR' : _STREAMEVENTSRESPONSE,
  '__module__' : 'hangups.googlechat_pb2'
  # @@protoc_insertion_point(class_scope:StreamEventsResponse)
  })
_sym_db.RegisterMessage(StreamEventsResponse)

Event = _reflection.GeneratedProtocolMessageType('Event', (_message.Message,), {

  'EventBody' : _reflection.GeneratedProtocolMessageType('EventBody', (_message.Message,), {
    'DESCRIPTOR' : _EVENT_EVENTBODY,
    '__module__' : 'hangups.googlechat_pb2'
    # @@protoc_insertion_point(class_scope:Event.EventBody)
    })
  ,
  'DESCRIPTOR' : _EVENT,
  '__module__' : 'hangups.googlechat_pb2'
  # @@protoc_insertion_point(class_scope:Event)
  })
_sym_db.RegisterMessage(Event)
_sym_db.RegisterMessage(Event.EventBody)

InviteType = _reflection.GeneratedProtocolMessageType('InviteType', (_message.Message,), {
  'DESCRIPTOR' : _INVITETYPE,
  '__module__' : 'hangups.googlechat_pb2'
  # @@protoc_insertion_point(class_scope:InviteType)
  })
_sym_db.RegisterMessage(InviteType)

GroupDetails = _reflection.GeneratedProtocolMessageType('GroupDetails', (_message.Message,), {
  'DESCRIPTOR' : _GROUPDETAILS,
  '__module__' : 'hangups.googlechat_pb2'
  # @@protoc_insertion_point(class_scope:GroupDetails)
  })
_sym_db.RegisterMessage(GroupDetails)

RoomUpdatedMetadata = _reflection.GeneratedProtocolMessageType('RoomUpdatedMetadata', (_message.Message,), {

  'RoomRenameMetadata' : _reflection.GeneratedProtocolMessageType('RoomRenameMetadata', (_message.Message,), {
    'DESCRIPTOR' : _ROOMUPDATEDMETADATA_ROOMRENAMEMETADATA,
    '__module__' : 'hangups.googlechat_pb2'
    # @@protoc_insertion_point(class_scope:RoomUpdatedMetadata.RoomRenameMetadata)
    })
  ,

  'GroupDetailsUpdatedMetadata' : _reflection.GeneratedProtocolMessageType('GroupDetailsUpdatedMetadata', (_message.Message,), {
    'DESCRIPTOR' : _ROOMUPDATEDMETADATA_GROUPDETAILSUPDATEDMETADATA,
    '__module__' : 'hangups.googlechat_pb2'
    # @@protoc_insertion_point(class_scope:RoomUpdatedMetadata.GroupDetailsUpdatedMetadata)
    })
  ,
  'DESCRIPTOR' : _ROOMUPDATEDMETADATA,
  '__module__' : 'hangups.googlechat_pb2'
  # @@protoc_insertion_point(class_scope:RoomUpdatedMetadata)
  })
_sym_db.RegisterMessage(RoomUpdatedMetadata)
_sym_db.RegisterMessage(RoomUpdatedMetadata.RoomRenameMetadata)
_sym_db.RegisterMessage(RoomUpdatedMetadata.GroupDetailsUpdatedMetadata)

MeetingSpace = _reflection.GeneratedProtocolMessageType('MeetingSpace', (_message.Message,), {

  'PhoneAccess' : _reflection.GeneratedProtocolMessageType('PhoneAccess', (_message.Message,), {
    'DESCRIPTOR' : _MEETINGSPACE_PHONEACCESS,
    '__module__' : 'hangups.googlechat_pb2'
    # @@protoc_insertion_point(class_scope:MeetingSpace.PhoneAccess)
    })
  ,

  'UniversalPhoneAccess' : _reflection.GeneratedProtocolMessageType('UniversalPhoneAccess', (_message.Message,), {
    'DESCRIPTOR' : _MEETINGSPACE_UNIVERSALPHONEACCESS,
    '__module__' : 'hangups.googlechat_pb2'
    # @@protoc_insertion_point(class_scope:MeetingSpace.UniversalPhoneAccess)
    })
  ,

  'Presenter' : _reflection.GeneratedProtocolMessageType('Presenter', (_message.Message,), {
    'DESCRIPTOR' : _MEETINGSPACE_PRESENTER,
    '__module__' : 'hangups.googlechat_pb2'
    # @@protoc_insertion_point(class_scope:MeetingSpace.Presenter)
    })
  ,

  'RecordingInfo' : _reflection.GeneratedProtocolMessageType('RecordingInfo', (_message.Message,), {

    'RecordingEvent' : _reflection.GeneratedProtocolMessageType('RecordingEvent', (_message.Message,), {
      'DESCRIPTOR' : _MEETINGSPACE_RECORDINGINFO_RECORDINGEVENT,
      '__module__' : 'hangups.googlechat_pb2'
      # @@protoc_insertion_point(class_scope:MeetingSpace.RecordingInfo.RecordingEvent)
      })
    ,
    'DESCRIPTOR' : _MEETINGSPACE_RECORDINGINFO,
    '__module__' : 'hangups.googlechat_pb2'
    # @@protoc_insertion_point(class_scope:MeetingSpace.RecordingInfo)
    })
  ,

  'StreamingSessionInfo' : _reflection.GeneratedProtocolMessageType('StreamingSessionInfo', (_message.Message,), {

    'SessionEvent' : _reflection.GeneratedProtocolMessageType('SessionEvent', (_message.Message,), {
      'DESCRIPTOR' : _MEETINGSPACE_STREAMINGSESSIONINFO_SESSIONEVENT,
      '__module__' : 'hangups.googlechat_pb2'
      # @@protoc_insertion_point(class_scope:MeetingSpace.StreamingSessionInfo.SessionEvent)
      })
    ,

    'StreamViewerStats' : _reflection.GeneratedProtocolMessageType('StreamViewerStats', (_message.Message,), {
      'DESCRIPTOR' : _MEETINGSPACE_STREAMINGSESSIONINFO_STREAMVIEWERSTATS,
      '__module__' : 'hangups.googlechat_pb2'
      # @@protoc_insertion_point(class_scope:MeetingSpace.StreamingSessionInfo.StreamViewerStats)
      })
    ,
    'DESCRIPTOR' : _MEETINGSPACE_STREAMINGSESSIONINFO,
    '__module__' : 'hangups.googlechat_pb2'
    # @@protoc_insertion_point(class_scope:MeetingSpace.StreamingSessionInfo)
    })
  ,

  'CallInfo' : _reflection.GeneratedProtocolMessageType('CallInfo', (_message.Message,), {

    'CallSettings' : _reflection.GeneratedProtocolMessageType('CallSettings', (_message.Message,), {
      'DESCRIPTOR' : _MEETINGSPACE_CALLINFO_CALLSETTINGS,
      '__module__' : 'hangups.googlechat_pb2'
      # @@protoc_insertion_point(class_scope:MeetingSpace.CallInfo.CallSettings)
      })
    ,

    'PaygateInfo' : _reflection.GeneratedProtocolMessageType('PaygateInfo', (_message.Message,), {
      'DESCRIPTOR' : _MEETINGSPACE_CALLINFO_PAYGATEINFO,
      '__module__' : 'hangups.googlechat_pb2'
      # @@protoc_insertion_point(class_scope:MeetingSpace.CallInfo.PaygateInfo)
      })
    ,

    'CseInfo' : _reflection.GeneratedProtocolMessageType('CseInfo', (_message.Message,), {
      'DESCRIPTOR' : _MEETINGSPACE_CALLINFO_CSEINFO,
      '__module__' : 'hangups.googlechat_pb2'
      # @@protoc_insertion_point(class_scope:MeetingSpace.CallInfo.CseInfo)
      })
    ,
    'DESCRIPTOR' : _MEETINGSPACE_CALLINFO,
    '__module__' : 'hangups.googlechat_pb2'
    # @@protoc_insertion_point(class_scope:MeetingSpace.CallInfo)
    })
  ,

  'GatewayAccess' : _reflection.GeneratedProtocolMessageType('GatewayAccess', (_message.Message,), {
    'DESCRIPTOR' : _MEETINGSPACE_GATEWAYACCESS,
    '__module__' : 'hangups.googlechat_pb2'
    # @@protoc_insertion_point(class_scope:MeetingSpace.GatewayAccess)
    })
  ,

  'GatewaySipAccess' : _reflection.GeneratedProtocolMessageType('GatewaySipAccess', (_message.Message,), {
    'DESCRIPTOR' : _MEETINGSPACE_GATEWAYSIPACCESS,
    '__module__' : 'hangups.googlechat_pb2'
    # @@protoc_insertion_point(class_scope:MeetingSpace.GatewaySipAccess)
    })
  ,

  'BroadcastAccess' : _reflection.GeneratedProtocolMessageType('BroadcastAccess', (_message.Message,), {
    'DESCRIPTOR' : _MEETINGSPACE_BROADCASTACCESS,
    '__module__' : 'hangups.googlechat_pb2'
    # @@protoc_insertion_point(class_scope:MeetingSpace.BroadcastAccess)
    })
  ,

  'Settings' : _reflection.GeneratedProtocolMessageType('Settings', (_message.Message,), {
    'DESCRIPTOR' : _MEETINGSPACE_SETTINGS,
    '__module__' : 'hangups.googlechat_pb2'
    # @@protoc_insertion_point(class_scope:MeetingSpace.Settings)
    })
  ,
  'DESCRIPTOR' : _MEETINGSPACE,
  '__module__' : 'hangups.googlechat_pb2'
  # @@protoc_insertion_point(class_scope:MeetingSpace)
  })
_sym_db.RegisterMessage(MeetingSpace)
_sym_db.RegisterMessage(MeetingSpace.PhoneAccess)
_sym_db.RegisterMessage(MeetingSpace.UniversalPhoneAccess)
_sym_db.RegisterMessage(MeetingSpace.Presenter)
_sym_db.RegisterMessage(MeetingSpace.RecordingInfo)
_sym_db.RegisterMessage(MeetingSpace.RecordingInfo.RecordingEvent)
_sym_db.RegisterMessage(MeetingSpace.StreamingSessionInfo)
_sym_db.RegisterMessage(MeetingSpace.StreamingSessionInfo.SessionEvent)
_sym_db.RegisterMessage(MeetingSpace.StreamingSessionInfo.StreamViewerStats)
_sym_db.RegisterMessage(MeetingSpace.CallInfo)
_sym_db.RegisterMessage(MeetingSpace.CallInfo.CallSettings)
_sym_db.RegisterMessage(MeetingSpace.CallInfo.PaygateInfo)
_sym_db.RegisterMessage(MeetingSpace.CallInfo.CseInfo)
_sym_db.RegisterMessage(MeetingSpace.GatewayAccess)
_sym_db.RegisterMessage(MeetingSpace.GatewaySipAccess)
_sym_db.RegisterMessage(MeetingSpace.BroadcastAccess)
_sym_db.RegisterMessage(MeetingSpace.Settings)

VideoCallMetadata = _reflection.GeneratedProtocolMessageType('VideoCallMetadata', (_message.Message,), {
  'DESCRIPTOR' : _VIDEOCALLMETADATA,
  '__module__' : 'hangups.googlechat_pb2'
  # @@protoc_insertion_point(class_scope:VideoCallMetadata)
  })
_sym_db.RegisterMessage(VideoCallMetadata)

MembershipChangedMetadata = _reflection.GeneratedProtocolMessageType('MembershipChangedMetadata', (_message.Message,), {

  'AffectedMembership' : _reflection.GeneratedProtocolMessageType('AffectedMembership', (_message.Message,), {
    'DESCRIPTOR' : _MEMBERSHIPCHANGEDMETADATA_AFFECTEDMEMBERSHIP,
    '__module__' : 'hangups.googlechat_pb2'
    # @@protoc_insertion_point(class_scope:MembershipChangedMetadata.AffectedMembership)
    })
  ,
  'DESCRIPTOR' : _MEMBERSHIPCHANGEDMETADATA,
  '__module__' : 'hangups.googlechat_pb2'
  # @@protoc_insertion_point(class_scope:MembershipChangedMetadata)
  })
_sym_db.RegisterMessage(MembershipChangedMetadata)
_sym_db.RegisterMessage(MembershipChangedMetadata.AffectedMembership)

UserMentionMetadata = _reflection.GeneratedProtocolMessageType('UserMentionMetadata', (_message.Message,), {
  'DESCRIPTOR' : _USERMENTIONMETADATA,
  '__module__' : 'hangups.googlechat_pb2'
  # @@protoc_insertion_point(class_scope:UserMentionMetadata)
  })
_sym_db.RegisterMessage(UserMentionMetadata)

SlashCommandMetadata = _reflection.GeneratedProtocolMessageType('SlashCommandMetadata', (_message.Message,), {
  'DESCRIPTOR' : _SLASHCOMMANDMETADATA,
  '__module__' : 'hangups.googlechat_pb2'
  # @@protoc_insertion_point(class_scope:SlashCommandMetadata)
  })
_sym_db.RegisterMessage(SlashCommandMetadata)

GroupVisibility = _reflection.GeneratedProtocolMessageType('GroupVisibility', (_message.Message,), {
  'DESCRIPTOR' : _GROUPVISIBILITY,
  '__module__' : 'hangups.googlechat_pb2'
  # @@protoc_insertion_point(class_scope:GroupVisibility)
  })
_sym_db.RegisterMessage(GroupVisibility)

ComGoogleProtobufTimestamp = _reflection.GeneratedProtocolMessageType('ComGoogleProtobufTimestamp', (_message.Message,), {
  'DESCRIPTOR' : _COMGOOGLEPROTOBUFTIMESTAMP,
  '__module__' : 'hangups.googlechat_pb2'
  # @@protoc_insertion_point(class_scope:ComGoogleProtobufTimestamp)
  })
_sym_db.RegisterMessage(ComGoogleProtobufTimestamp)

GetServerTimeRequest = _reflection.GeneratedProtocolMessageType('GetServerTimeRequest', (_message.Message,), {
  'DESCRIPTOR' : _GETSERVERTIMEREQUEST,
  '__module__' : 'hangups.googlechat_pb2'
  # @@protoc_insertion_point(class_scope:GetServerTimeRequest)
  })
_sym_db.RegisterMessage(GetServerTimeRequest)

GetServerTimeResponse = _reflection.GeneratedProtocolMessageType('GetServerTimeResponse', (_message.Message,), {
  'DESCRIPTOR' : _GETSERVERTIMERESPONSE,
  '__module__' : 'hangups.googlechat_pb2'
  # @@protoc_insertion_point(class_scope:GetServerTimeResponse)
  })
_sym_db.RegisterMessage(GetServerTimeResponse)

CatchUpRange = _reflection.GeneratedProtocolMessageType('CatchUpRange', (_message.Message,), {
  'DESCRIPTOR' : _CATCHUPRANGE,
  '__module__' : 'hangups.googlechat_pb2'
  # @@protoc_insertion_point(class_scope:CatchUpRange)
  })
_sym_db.RegisterMessage(CatchUpRange)

CatchUpGroupRequest = _reflection.GeneratedProtocolMessageType('CatchUpGroupRequest', (_message.Message,), {
  'DESCRIPTOR' : _CATCHUPGROUPREQUEST,
  '__module__' : 'hangups.googlechat_pb2'
  # @@protoc_insertion_point(class_scope:CatchUpGroupRequest)
  })
_sym_db.RegisterMessage(CatchUpGroupRequest)

CatchUpUserRequest = _reflection.GeneratedProtocolMessageType('CatchUpUserRequest', (_message.Message,), {
  'DESCRIPTOR' : _CATCHUPUSERREQUEST,
  '__module__' : 'hangups.googlechat_pb2'
  # @@protoc_insertion_point(class_scope:CatchUpUserRequest)
  })
_sym_db.RegisterMessage(CatchUpUserRequest)

CatchUpResponse = _reflection.GeneratedProtocolMessageType('CatchUpResponse', (_message.Message,), {
  'DESCRIPTOR' : _CATCHUPRESPONSE,
  '__module__' : 'hangups.googlechat_pb2'
  # @@protoc_insertion_point(class_scope:CatchUpResponse)
  })
_sym_db.RegisterMessage(CatchUpResponse)

GetGroupRequest = _reflection.GeneratedProtocolMessageType('GetGroupRequest', (_message.Message,), {
  'DESCRIPTOR' : _GETGROUPREQUEST,
  '__module__' : 'hangups.googlechat_pb2'
  # @@protoc_insertion_point(class_scope:GetGroupRequest)
  })
_sym_db.RegisterMessage(GetGroupRequest)

GetGroupResponse = _reflection.GeneratedProtocolMessageType('GetGroupResponse', (_message.Message,), {
  'DESCRIPTOR' : _GETGROUPRESPONSE,
  '__module__' : 'hangups.googlechat_pb2'
  # @@protoc_insertion_point(class_scope:GetGroupResponse)
  })
_sym_db.RegisterMessage(GetGroupResponse)

WorldSection = _reflection.GeneratedProtocolMessageType('WorldSection', (_message.Message,), {
  'DESCRIPTOR' : _WORLDSECTION,
  '__module__' : 'hangups.googlechat_pb2'
  # @@protoc_insertion_point(class_scope:WorldSection)
  })
_sym_db.RegisterMessage(WorldSection)

WorldFilter = _reflection.GeneratedProtocolMessageType('WorldFilter', (_message.Message,), {
  'DESCRIPTOR' : _WORLDFILTER,
  '__module__' : 'hangups.googlechat_pb2'
  # @@protoc_insertion_point(class_scope:WorldFilter)
  })
_sym_db.RegisterMessage(WorldFilter)

NameUsers = _reflection.GeneratedProtocolMessageType('NameUsers', (_message.Message,), {
  'DESCRIPTOR' : _NAMEUSERS,
  '__module__' : 'hangups.googlechat_pb2'
  # @@protoc_insertion_point(class_scope:NameUsers)
  })
_sym_db.RegisterMessage(NameUsers)

WorldItemLite = _reflection.GeneratedProtocolMessageType('WorldItemLite', (_message.Message,), {

  'MembershipLite' : _reflection.GeneratedProtocolMessageType('MembershipLite', (_message.Message,), {
    'DESCRIPTOR' : _WORLDITEMLITE_MEMBERSHIPLITE,
    '__module__' : 'hangups.googlechat_pb2'
    # @@protoc_insertion_point(class_scope:WorldItemLite.MembershipLite)
    })
  ,

  'DmMembers' : _reflection.GeneratedProtocolMessageType('DmMembers', (_message.Message,), {
    'DESCRIPTOR' : _WORLDITEMLITE_DMMEMBERS,
    '__module__' : 'hangups.googlechat_pb2'
    # @@protoc_insertion_point(class_scope:WorldItemLite.DmMembers)
    })
  ,

  'GroupLite' : _reflection.GeneratedProtocolMessageType('GroupLite', (_message.Message,), {
    'DESCRIPTOR' : _WORLDITEMLITE_GROUPLITE,
    '__module__' : 'hangups.googlechat_pb2'
    # @@protoc_insertion_point(class_scope:WorldItemLite.GroupLite)
    })
  ,

  'FlatGroup' : _reflection.GeneratedProtocolMessageType('FlatGroup', (_message.Message,), {
    'DESCRIPTOR' : _WORLDITEMLITE_FLATGROUP,
    '__module__' : 'hangups.googlechat_pb2'
    # @@protoc_insertion_point(class_scope:WorldItemLite.FlatGroup)
    })
  ,

  'ThreadedGroup' : _reflection.GeneratedProtocolMessageType('ThreadedGroup', (_message.Message,), {
    'DESCRIPTOR' : _WORLDITEMLITE_THREADEDGROUP,
    '__module__' : 'hangups.googlechat_pb2'
    # @@protoc_insertion_point(class_scope:WorldItemLite.ThreadedGroup)
    })
  ,
  'DESCRIPTOR' : _WORLDITEMLITE,
  '__module__' : 'hangups.googlechat_pb2'
  # @@protoc_insertion_point(class_scope:WorldItemLite)
  })
_sym_db.RegisterMessage(WorldItemLite)
_sym_db.RegisterMessage(WorldItemLite.MembershipLite)
_sym_db.RegisterMessage(WorldItemLite.DmMembers)
_sym_db.RegisterMessage(WorldItemLite.GroupLite)
_sym_db.RegisterMessage(WorldItemLite.FlatGroup)
_sym_db.RegisterMessage(WorldItemLite.ThreadedGroup)

WorldSectionRequest = _reflection.GeneratedProtocolMessageType('WorldSectionRequest', (_message.Message,), {
  'DESCRIPTOR' : _WORLDSECTIONREQUEST,
  '__module__' : 'hangups.googlechat_pb2'
  # @@protoc_insertion_point(class_scope:WorldSectionRequest)
  })
_sym_db.RegisterMessage(WorldSectionRequest)

WorldSectionResponse = _reflection.GeneratedProtocolMessageType('WorldSectionResponse', (_message.Message,), {
  'DESCRIPTOR' : _WORLDSECTIONRESPONSE,
  '__module__' : 'hangups.googlechat_pb2'
  # @@protoc_insertion_point(class_scope:WorldSectionResponse)
  })
_sym_db.RegisterMessage(WorldSectionResponse)

PaginatedWorldRequest = _reflection.GeneratedProtocolMessageType('PaginatedWorldRequest', (_message.Message,), {
  'DESCRIPTOR' : _PAGINATEDWORLDREQUEST,
  '__module__' : 'hangups.googlechat_pb2'
  # @@protoc_insertion_point(class_scope:PaginatedWorldRequest)
  })
_sym_db.RegisterMessage(PaginatedWorldRequest)

PaginatedWorldResponse = _reflection.GeneratedProtocolMessageType('PaginatedWorldResponse', (_message.Message,), {
  'DESCRIPTOR' : _PAGINATEDWORLDRESPONSE,
  '__module__' : 'hangups.googlechat_pb2'
  # @@protoc_insertion_point(class_scope:PaginatedWorldResponse)
  })
_sym_db.RegisterMessage(PaginatedWorldResponse)

RemoveMembershipsRequest = _reflection.GeneratedProtocolMessageType('RemoveMembershipsRequest', (_message.Message,), {
  'DESCRIPTOR' : _REMOVEMEMBERSHIPSREQUEST,
  '__module__' : 'hangups.googlechat_pb2'
  # @@protoc_insertion_point(class_scope:RemoveMembershipsRequest)
  })
_sym_db.RegisterMessage(RemoveMembershipsRequest)

RemoveMembershipResult = _reflection.GeneratedProtocolMessageType('RemoveMembershipResult', (_message.Message,), {
  'DESCRIPTOR' : _REMOVEMEMBERSHIPRESULT,
  '__module__' : 'hangups.googlechat_pb2'
  # @@protoc_insertion_point(class_scope:RemoveMembershipResult)
  })
_sym_db.RegisterMessage(RemoveMembershipResult)

RemoveMembershipsResponse = _reflection.GeneratedProtocolMessageType('RemoveMembershipsResponse', (_message.Message,), {
  'DESCRIPTOR' : _REMOVEMEMBERSHIPSRESPONSE,
  '__module__' : 'hangups.googlechat_pb2'
  # @@protoc_insertion_point(class_scope:RemoveMembershipsResponse)
  })
_sym_db.RegisterMessage(RemoveMembershipsResponse)

HideGroupRequest = _reflection.GeneratedProtocolMessageType('HideGroupRequest', (_message.Message,), {
  'DESCRIPTOR' : _HIDEGROUPREQUEST,
  '__module__' : 'hangups.googlechat_pb2'
  # @@protoc_insertion_point(class_scope:HideGroupRequest)
  })
_sym_db.RegisterMessage(HideGroupRequest)

HideGroupResponse = _reflection.GeneratedProtocolMessageType('HideGroupResponse', (_message.Message,), {
  'DESCRIPTOR' : _HIDEGROUPRESPONSE,
  '__module__' : 'hangups.googlechat_pb2'
  # @@protoc_insertion_point(class_scope:HideGroupResponse)
  })
_sym_db.RegisterMessage(HideGroupResponse)

InviteNotificationSettings = _reflection.GeneratedProtocolMessageType('InviteNotificationSettings', (_message.Message,), {
  'DESCRIPTOR' : _INVITENOTIFICATIONSETTINGS,
  '__module__' : 'hangups.googlechat_pb2'
  # @@protoc_insertion_point(class_scope:InviteNotificationSettings)
  })
_sym_db.RegisterMessage(InviteNotificationSettings)

CreateMembershipRequest = _reflection.GeneratedProtocolMessageType('CreateMembershipRequest', (_message.Message,), {
  'DESCRIPTOR' : _CREATEMEMBERSHIPREQUEST,
  '__module__' : 'hangups.googlechat_pb2'
  # @@protoc_insertion_point(class_scope:CreateMembershipRequest)
  })
_sym_db.RegisterMessage(CreateMembershipRequest)

CreateMembershipResponse = _reflection.GeneratedProtocolMessageType('CreateMembershipResponse', (_message.Message,), {
  'DESCRIPTOR' : _CREATEMEMBERSHIPRESPONSE,
  '__module__' : 'hangups.googlechat_pb2'
  # @@protoc_insertion_point(class_scope:CreateMembershipResponse)
  })
_sym_db.RegisterMessage(CreateMembershipResponse)

MarkGroupReadstateRequest = _reflection.GeneratedProtocolMessageType('MarkGroupReadstateRequest', (_message.Message,), {
  'DESCRIPTOR' : _MARKGROUPREADSTATEREQUEST,
  '__module__' : 'hangups.googlechat_pb2'
  # @@protoc_insertion_point(class_scope:MarkGroupReadstateRequest)
  })
_sym_db.RegisterMessage(MarkGroupReadstateRequest)

MarkGroupReadstateResponse = _reflection.GeneratedProtocolMessageType('MarkGroupReadstateResponse', (_message.Message,), {
  'DESCRIPTOR' : _MARKGROUPREADSTATERESPONSE,
  '__module__' : 'hangups.googlechat_pb2'
  # @@protoc_insertion_point(class_scope:MarkGroupReadstateResponse)
  })
_sym_db.RegisterMessage(MarkGroupReadstateResponse)

SetPresenceSharedRequest = _reflection.GeneratedProtocolMessageType('SetPresenceSharedRequest', (_message.Message,), {
  'DESCRIPTOR' : _SETPRESENCESHAREDREQUEST,
  '__module__' : 'hangups.googlechat_pb2'
  # @@protoc_insertion_point(class_scope:SetPresenceSharedRequest)
  })
_sym_db.RegisterMessage(SetPresenceSharedRequest)

SetPresenceSharedResponse = _reflection.GeneratedProtocolMessageType('SetPresenceSharedResponse', (_message.Message,), {
  'DESCRIPTOR' : _SETPRESENCESHAREDRESPONSE,
  '__module__' : 'hangups.googlechat_pb2'
  # @@protoc_insertion_point(class_scope:SetPresenceSharedResponse)
  })
_sym_db.RegisterMessage(SetPresenceSharedResponse)

SetDndDurationRequest = _reflection.GeneratedProtocolMessageType('SetDndDurationRequest', (_message.Message,), {
  'DESCRIPTOR' : _SETDNDDURATIONREQUEST,
  '__module__' : 'hangups.googlechat_pb2'
  # @@protoc_insertion_point(class_scope:SetDndDurationRequest)
  })
_sym_db.RegisterMessage(SetDndDurationRequest)

SetDndDurationResponse = _reflection.GeneratedProtocolMessageType('SetDndDurationResponse', (_message.Message,), {
  'DESCRIPTOR' : _SETDNDDURATIONRESPONSE,
  '__module__' : 'hangups.googlechat_pb2'
  # @@protoc_insertion_point(class_scope:SetDndDurationResponse)
  })
_sym_db.RegisterMessage(SetDndDurationResponse)

UpdateGroupRequest = _reflection.GeneratedProtocolMessageType('UpdateGroupRequest', (_message.Message,), {
  'DESCRIPTOR' : _UPDATEGROUPREQUEST,
  '__module__' : 'hangups.googlechat_pb2'
  # @@protoc_insertion_point(class_scope:UpdateGroupRequest)
  })
_sym_db.RegisterMessage(UpdateGroupRequest)

UpdateGroupResponse = _reflection.GeneratedProtocolMessageType('UpdateGroupResponse', (_message.Message,), {
  'DESCRIPTOR' : _UPDATEGROUPRESPONSE,
  '__module__' : 'hangups.googlechat_pb2'
  # @@protoc_insertion_point(class_scope:UpdateGroupResponse)
  })
_sym_db.RegisterMessage(UpdateGroupResponse)

BlockEntityRequest = _reflection.GeneratedProtocolMessageType('BlockEntityRequest', (_message.Message,), {
  'DESCRIPTOR' : _BLOCKENTITYREQUEST,
  '__module__' : 'hangups.googlechat_pb2'
  # @@protoc_insertion_point(class_scope:BlockEntityRequest)
  })
_sym_db.RegisterMessage(BlockEntityRequest)

BlockEntityResponse = _reflection.GeneratedProtocolMessageType('BlockEntityResponse', (_message.Message,), {
  'DESCRIPTOR' : _BLOCKENTITYRESPONSE,
  '__module__' : 'hangups.googlechat_pb2'
  # @@protoc_insertion_point(class_scope:BlockEntityResponse)
  })
_sym_db.RegisterMessage(BlockEntityResponse)

SetCustomStatusRequest = _reflection.GeneratedProtocolMessageType('SetCustomStatusRequest', (_message.Message,), {
  'DESCRIPTOR' : _SETCUSTOMSTATUSREQUEST,
  '__module__' : 'hangups.googlechat_pb2'
  # @@protoc_insertion_point(class_scope:SetCustomStatusRequest)
  })
_sym_db.RegisterMessage(SetCustomStatusRequest)

SetCustomStatusResponse = _reflection.GeneratedProtocolMessageType('SetCustomStatusResponse', (_message.Message,), {
  'DESCRIPTOR' : _SETCUSTOMSTATUSRESPONSE,
  '__module__' : 'hangups.googlechat_pb2'
  # @@protoc_insertion_point(class_scope:SetCustomStatusResponse)
  })
_sym_db.RegisterMessage(SetCustomStatusResponse)

WriteRevision = _reflection.GeneratedProtocolMessageType('WriteRevision', (_message.Message,), {
  'DESCRIPTOR' : _WRITEREVISION,
  '__module__' : 'hangups.googlechat_pb2'
  # @@protoc_insertion_point(class_scope:WriteRevision)
  })
_sym_db.RegisterMessage(WriteRevision)

ReadRevision = _reflection.GeneratedProtocolMessageType('ReadRevision', (_message.Message,), {
  'DESCRIPTOR' : _READREVISION,
  '__module__' : 'hangups.googlechat_pb2'
  # @@protoc_insertion_point(class_scope:ReadRevision)
  })
_sym_db.RegisterMessage(ReadRevision)

ReferenceRevision = _reflection.GeneratedProtocolMessageType('ReferenceRevision', (_message.Message,), {
  'DESCRIPTOR' : _REFERENCEREVISION,
  '__module__' : 'hangups.googlechat_pb2'
  # @@protoc_insertion_point(class_scope:ReferenceRevision)
  })
_sym_db.RegisterMessage(ReferenceRevision)


# @@protoc_insertion_point(module_scope)
